/**
 * Redistribution and use of this software and associated
 * documentation ("Software"), with or without modification, are
 * permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright statements
 *    and notices.  Redistributions must also contain a copy of this
 *    document.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Intalio Inc.  For written permission, please
 *    contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Intalio Inc. Exolab is a registered trademark of
 *    Intalio Inc.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY INTALIO AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL INTALIO OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Copyright 2000 (C) Intalio Inc. All Rights Reserved.
 *
 * $Id: Delegate.java,v 1.16 2001/08/13 03:47:32 wood Exp $
 *
 * Date         Author  Changes
 */


/** Modified/extended by Diana Szentivanyi in June 2003
* contact diasz@ida.liu.se
*/

package org.openorb.CORBA;

import org.omg.CORBA.* ;
import org.openorb.net.*;

import java.util.Set;
import java.util.TreeSet;
import java.util.HashSet;
import java.util.Arrays;
import java.util.ArrayList;
import java.util.Iterator;

import org.omg.CORBA.CompletionStatus;

import org.openorb.CORBA.MinorCodes;
import org.omg.CORBA.CompletionStatus;

import org.openorb.policy.ForceMarshalPolicy;
import org.openorb.policy.FORCE_MARSHAL_POLICY_ID;

import java.io.ByteArrayOutputStream;
import org.openorb.io.HexPrintStream;
import java.io.PrintStream;
import org.openorb.util.RepoIDHelper;

/**
 * This class constitutes the top layer in the object request sequence. It
 * manages selecting between alternative paths for invoking requests and
 * manages the invocation sequence, in concert with the server stubs.
 *
 * @author <a href="mailto:wood@intalio.com">Chris Wood &lt;wood@intalio.com&gt;</a>
 * @version $Revision: 1.16 $ $Date: 2001/08/13 03:47:32 $
 */
public class Delegate extends org.omg.CORBA_2_3.portable.Delegate
{
    /**
     * Construct new delegate from orb and IOR.
	 *
	 * @param orb the client orb.
	 * @param ior the target IOR.
   */
    public Delegate(org.omg.CORBA.ORB orb, org.omg.IOP.IOR ior) {
	if(ior == null || orb == null)
	    throw new NullPointerException();
	_ior = ior;
	_effectiveIor = ior;
	_orb = orb;
    }

    // Internal state

  /**
   * Reference to the controlling orb.
   */
    private org.omg.CORBA.ORB _orb;


    private int checkfield=0;

    /**
   * IOR of the object. This is used in comparisons and hashing.
   */
    private org.omg.IOP.IOR _ior;

    /**
   * IOR as written to input/output stream. This can change as a
   * result of a perminent redirect.
   */
    private org.omg.IOP.IOR _effectiveIor;

    /**
   * hash code. The object is always hashed on the original IOR.
   */
    private long _hash = -1;

    /**
   * Policy override set. If no overrides are set this is null
   */
    private org.omg.CORBA.PolicyManagerOperations _policyOver = null;

    /**
   * Reference to the policy Current.
   */
    private org.omg.CORBA.PolicyCurrent _policyCurr = null;

    /**
   * Reference to policy reconciler.
   */
    private org.openorb.policy.PolicyReconciler _orbReconciler = null;

    /**
   * Reference to the orb policy manager.
   */
    private org.omg.CORBA.PolicyManager _orbPolicyManager = null;

    /**
   * Reference to the policy set manager.
   */
    private org.openorb.policy.PolicySetManager _policySetManager = null;

    /**
   * Reference to the client manager.
   */
    private ClientManager _clientManager = null;

    /**
   * Reference to the portable interceptor current implementation
   * note that out PI current has some extra operations.
   */
    private org.openorb.PI.CurrentImpl _piCurrent = null;

    /**
   * Synchronization object for the bindings.
   */
    private java.lang.Object _bindingSync = new java.lang.Object();

    /**
   * Available bindings. Set of ClientBinding. This may grow over time if
   * redirections occour. Must synchronize on _bindingSync to access.
   */
    private Set _bindings = new HashSet();

    /**
   * Binding list version. Each time the binding list is updated due to a
   * redirection this value is incremented. Must synchronize on _bindingSync
   * to access.
   */
    private int _bindingVersion = -1;

    /**
   * The 'selected binding'. This is the binding that new invocations will
   * use. It is set when an invocation returns successfully and unset when
   * the channel closes or an invocation on it fails.
   */
    private ClientBinding _selectedBinding = null;

    // Public functions from org.omg.CORBA.portable.Delegate

  /**
   * Return the object interface definition.
   * @deprecated Deprecated by CORBA 2.3
   */
    public org.omg.CORBA.InterfaceDef get_interface(org.omg.CORBA.Object self)
    {
	return org.omg.CORBA.InterfaceDefHelper.narrow( get_interface_def( self ) );
    }

    /**
   * Return the object interface definition.
   */
    public org.omg.CORBA.Object get_interface_def(org.omg.CORBA.Object self)
    {
	while(true) {
	    RequestState state = begin_invocation(self, false);
	    if(state.binding.getObjectAdapter() == null) {
		org.omg.CORBA.portable.InputStream _input = null;
		try
		    {
			org.omg.CORBA.portable.OutputStream _output = request(self, "_interface",true);
			_input = invoke(self, _output);
			return _input.read_Object();
		    }
		catch( org.omg.CORBA.portable.RemarshalException _exception )
		    {
			continue;
		    }
		catch( org.omg.CORBA.portable.ApplicationException _exception )
		    {
			java.lang.String _exception_id = _exception.getId();
			throw new org.omg.CORBA.UNKNOWN("Unexcepected User Exception: "+ _exception_id,
							org.omg.CORBA.OMGVMCID.value | 1 , CompletionStatus.COMPLETED_YES);
		    }
		finally
		    {
			releaseReply(self, _input);
		    }
	    }
	    else
		{
		    boolean failed = true;
		    state.local_level++;
		    try {
			org.omg.CORBA.Object ret = state.binding.get_interface_def();
			failed = false;
			return ret;
		    }
		    catch(org.omg.CORBA.SystemException ex) {
			receive_sysexception(self, state, ex);
			failed = false;
			return null;
		    }
		    catch(org.omg.PortableInterceptor.ForwardRequest ex) {
			org.omg.IOP.IOR fwd = ((Delegate)((org.omg.CORBA.portable.ObjectImpl)ex.forward)._get_delegate()).ior();
			receive_redirect(self, state, fwd, ex.permanent);
			failed = false;
			return null;
		    }
		    finally {
			complete_invocation(state, !failed);
		    }
		}
	}
    }

    /**
   * Return the object's component. This is currently unused.
   */
    public org.omg.CORBA.Object get_component(org.omg.CORBA.Object self)
    {
	while(true) {
	    RequestState state = begin_invocation(self, false);
	    if(state.binding.getObjectAdapter() == null) {
		org.omg.CORBA.portable.InputStream _input = null;
		try
		    {
			org.omg.CORBA.portable.OutputStream _output = request(self, "_component",true);
			_input = invoke(self, _output);
			return _input.read_Object();
		    }
		catch( org.omg.CORBA.portable.RemarshalException _exception )
		    {
			continue;
		    }
		catch( org.omg.CORBA.portable.ApplicationException _exception )
		    {
			java.lang.String _exception_id = _exception.getId();
			throw new org.omg.CORBA.UNKNOWN("Unexcepected User Exception: "+ _exception_id,
							org.omg.CORBA.OMGVMCID.value | 1 , CompletionStatus.COMPLETED_YES);
		    }
		finally
		    {
			releaseReply(self, _input);
		    }
	    }
	    else
		{
		    boolean failed = true;
		    state.local_level++;
		    try {
			org.omg.CORBA.Object ret = state.binding.get_component();
			failed = false;
			return ret;
		    }
		    catch(org.omg.CORBA.SystemException ex) {
			receive_sysexception(self, state, ex);
			failed = false;
			return null;
		    }
		    catch(org.omg.PortableInterceptor.ForwardRequest ex) {
			org.omg.IOP.IOR fwd = ((Delegate)((org.omg.CORBA.portable.ObjectImpl)ex.forward)._get_delegate()).ior();
			receive_redirect(self, state, fwd, ex.permanent);
			failed = false;
			return null;
		    }
		    finally {
			complete_invocation(state, !failed);
		    }
		}
	}
    }

    /**
   * This operation is used to test if an object implements an interface. This
	 * may result in invoking a remote operation.
   */
    public boolean is_a(org.omg.CORBA.Object self, String id )
    {
	java.lang.Object test = RepoIDHelper.createIsATest(id);

	if ( test.equals("IDL:omg.org/CORBA/Object:1.0") )
	    return true;

	// use the original ior
	if ( test.equals( _ior.type_id ) )
	    return true;

	String [] ids = (( org.omg.CORBA.portable.ObjectImpl )self)._ids();

	for ( int i=0; i<ids.length; i++ )
	    if ( test.equals(ids[i]) )
		return true;

	while(true) {
	    RequestState state = begin_invocation(self, false);
	    if(state.binding.getObjectAdapter() == null) {
		org.omg.CORBA.portable.InputStream _input = null;
		try
		    {
			//	System.out.println("Delegate is_a try");
			org.omg.CORBA.portable.OutputStream _output = request(self, "_is_a",true);
			_output.write_string(id);
			_input = invoke(self, _output);
			return _input.read_boolean();
		    }
		catch( org.omg.CORBA.portable.RemarshalException _exception )
		    {
			continue;
		    }
		catch( org.omg.CORBA.portable.ApplicationException _exception )
		    {
			java.lang.String _exception_id = _exception.getId();
			throw new org.omg.CORBA.UNKNOWN("Unexcepected User Exception: "+ _exception_id,
							org.omg.CORBA.OMGVMCID.value | 1 , CompletionStatus.COMPLETED_YES);
		    }
		finally
		    {
			releaseReply(self, _input);
		    }
	    }
	    else
		{
		    boolean failed = true;
		    state.local_level++;
		    try {
			boolean ret = state.binding.is_a(id);
			failed = false;
			return ret;
		    }
		    catch(org.omg.CORBA.SystemException ex) {
			receive_sysexception(self, state, ex);
			failed = false;
			return false;
		    }
		    catch(org.omg.PortableInterceptor.ForwardRequest ex) {
			org.omg.IOP.IOR fwd = ((Delegate)((org.omg.CORBA.portable.ObjectImpl)ex.forward)._get_delegate()).ior();
			receive_redirect(self, state, fwd, ex.permanent);
			failed = false;
			return false;
		    }
		    finally {
			complete_invocation(state, !failed);
		    }
		}
	}
    }

    /**
   * Find the domain manager list associated to this object.
   *
   * @return the list of domain managers.
   */
    public org.omg.CORBA.DomainManager [] get_domain_managers( org.omg.CORBA.Object self)
    {
	return get_domain_managers_and_policies(self, null, null);
    }

    /**
   * This retrieves the effective profile policy when the invocation succeeds.
   */
    private org.omg.CORBA.DomainManager []
	get_domain_managers_and_policies( org.omg.CORBA.Object self , int [] policy_types,
					  org.omg.CORBA.PolicyListHolder policies)
    {
	while(true) {
	    RequestState state = begin_invocation(self, false);
	    if(state.binding.getObjectAdapter() == null) {
		org.omg.CORBA.portable.InputStream _input = null;
		try
		    {
			org.omg.CORBA.portable.OutputStream _output = request(self, "_get_domain_managers",true);
			_input = invoke(self, _output);
			org.omg.CORBA.DomainManager [] ret = org.omg.CORBA.DomainManagersListHelper.read(_input);
			if(policies != null)
			    policies.value = state.binding.getAddress().get_target_policies(policy_types);
			return ret;
		    }
		catch( org.omg.CORBA.portable.RemarshalException _exception )
		    {
			continue;
		    }
		catch( org.omg.CORBA.portable.ApplicationException _exception )
		    {
			java.lang.String _exception_id = _exception.getId();
			throw new org.omg.CORBA.UNKNOWN("Unexcepected User Exception: "+ _exception_id,
							org.omg.CORBA.OMGVMCID.value | 1 , CompletionStatus.COMPLETED_YES);
		    }
		finally
		    {
			releaseReply(self, _input);
		    }
	    }
	    else
		{
		    boolean failed = true;
		    state.local_level++;
		    try {
			org.omg.CORBA.DomainManager [] ret = state.binding.get_domain_managers();
			failed = false;
			return ret;
		    }
		    catch(org.omg.CORBA.SystemException ex) {
			receive_sysexception(self, state, ex);
			failed = false;
			return null;
		    }
		    catch(org.omg.PortableInterceptor.ForwardRequest ex) {
			org.omg.IOP.IOR fwd = ((Delegate)((org.omg.CORBA.portable.ObjectImpl)ex.forward)._get_delegate()).ior();
			receive_redirect(self, state, fwd, ex.permanent);
			failed = false;
			return null;
		    }
		    finally {
			complete_invocation(state, !failed);
		    }
		}
	}
    }

    /**
   * Check if an object exists
   */
    public boolean non_existent( org.omg.CORBA.Object self )
    {
	RequestState state = begin_invocation(self, false);
	return locate_and_get_policies(self, state, null);
    }

    /**
   * Dulicate a reference. This simply returns the same delegate.
   */
    public org.omg.CORBA.Object duplicate(org.omg.CORBA.Object self)
    {
	return self;
    }

    /**
   * Remove a reference. This is an empty operation.
   */
    public void release(org.omg.CORBA.Object self) {}

    /**
   * Check if two object refereces are equivalent. Two references are considered
	 * equivalent if their IORs are identical.
   */
    public boolean is_equivalent(org.omg.CORBA.Object obj1,org.omg.CORBA.Object obj2)
    {
	if ( obj1 == obj2 )
	    return true;

	org.omg.CORBA.portable.Delegate o_dlg2
	    = ((org.omg.CORBA.portable.ObjectImpl)obj2)._get_delegate();

	if(!(o_dlg2 instanceof Delegate))
	    return false;

	Delegate dlg2 = (Delegate)o_dlg2;

	if(this == dlg2)
	    return true;

	// check the IORs for equivalance.
	org.omg.IOP.IOR ior1 = _ior;
	org.omg.IOP.IOR ior2 = dlg2._ior;

	if(ior1.profiles.length != ior2.profiles.length)
	    return false;

	for(int i = 0; i < _ior.profiles.length; ++i) {
	    org.omg.IOP.TaggedProfile pro1 = ior1.profiles[i];
	    org.omg.IOP.TaggedProfile pro2 = ior2.profiles[i];
	    if(pro1.tag != pro2.tag ||
	       !Arrays.equals(pro1.profile_data, pro2.profile_data))
		return false;
	}

	return true;
    }

    /**
	 * Use is_equivalent definition of equals so objects can be stored in
	 * hash tables.
	 */
    public boolean equals(org.omg.CORBA.Object self, java.lang.Object obj) {
	if(!(obj instanceof org.omg.CORBA.Object))
	    return false;

	return is_equivalent(self, (org.omg.CORBA.Object)obj);
    }

    /**
   * Return an hashcode for an object ( with a max value )
   * Note that this hash code will not change even if the object is
   * redirected.
   */
    public int hash(org.omg.CORBA.Object self,int max)
    {
	if(_hash == -1) {
	    // always use the original ior for doing a hash.
	    _hash = 0;
	    for(int i = 0; i < _ior.profiles.length; ++i) {
		for(int j = 0; j < _ior.profiles[i].profile_data.length; ++j)
		    _hash = 31l*_hash + (long)(_ior.profiles[i].profile_data[j]);
		_hash = 31l*_hash + (long)(_ior.profiles[i].tag);
	    }
	}

	return (int)( _hash % (long)( max + 1 ) );
    }

    /**
	 * Use the IOR hash code so that objects can be used in hash tables.
	 */
    public int hashCode(org.omg.CORBA.Object self) {
	return hash(self, 32);
    }

    /**
   * Create a dynamic request
   */
    public org.omg.CORBA.Request create_request(org.omg.CORBA.Object self,
						org.omg.CORBA.Context ctx,
						String operation,
						org.omg.CORBA.NVList arg_list,
						org.omg.CORBA.NamedValue result)
    {
	return new org.openorb.CORBA.dii.Request(self, operation, arg_list,result,new org.openorb.CORBA.dii.Environment(),new org.openorb.CORBA.dii.ExceptionList(),new org.openorb.CORBA.dii.ContextList(), _orb);
    }

    /**
   * Create a dynamic request
   */
    public org.omg.CORBA.Request create_request(org.omg.CORBA.Object self,
						org.omg.CORBA.Context ctx,
						String operation,
						org.omg.CORBA.NVList arg_list,
						org.omg.CORBA.NamedValue result,
						org.omg.CORBA.ExceptionList excepts,
						org.omg.CORBA.ContextList contexts)
    {
	return new org.openorb.CORBA.dii.Request(self, operation,arg_list,result,new org.openorb.CORBA.dii.Environment(),excepts,contexts,_orb) ;
    }

    /**
   * Create a dynamic request
   */
    public org.omg.CORBA.Request request(org.omg.CORBA.Object self,String operation)
    {
	return new org.openorb.CORBA.dii.Request(self, operation, _orb) ;
    }

    /**
   * Return true if this object is local. This fuction also sets up a request
   * sequence.
   */
    public boolean is_local(org.omg.CORBA.Object self) {
	// test for presence of client policy.
	ForceMarshalPolicy pol = (ForceMarshalPolicy)client_policy(self, FORCE_MARSHAL_POLICY_ID.value);

	// TODO: change this so the default policy is to force marshal.
	if(pol != null && pol.forceMarshal())
	    return false;

	// bind, and return local state.
	RequestState state = begin_invocation(self, false);

	return state.binding.local_invoke();
    }

    public org.omg.CORBA.portable.OutputStream request( org.omg.CORBA.Object self,String operation,boolean responseExpected )
    {
	RequestState state = begin_invocation(self, false);
	//System.out.println("After begin invocation");
	org.omg.IOP.IOR local_ior;
	int local_v;
	org.omg.CORBA.portable.OutputStream ret;
	while(true) {
	    try {
		state.request = state.binding.create_request(self, operation, responseExpected);
	    }
	    catch(org.omg.CORBA.SystemException ex) {
		//System.out.println("REceive exc");
		receive_sysexception(self, state, ex);
		continue;
	    }
	    //System.out.println("Begin marshal in request");
	    ret = state.request.begin_marshal();
	    //System.out.println("After begin marshal");
	    if(ret != null)
		return ret;

	    // a client side interceptor has rejected the request.
	    switch(state.request.reply_status()) {
	    case org.omg.PortableInterceptor.SYSTEM_EXCEPTION.value:
		//System.out.println("receive exc after");
		receive_sysexception(self, state, state.request.received_system_exception());
		break;
	    case org.omg.PortableInterceptor.LOCATION_FORWARD.value:
		local_ior=state.request.forward_reference_ior();
		local_v=org.omg.FT.Utils.extract_group_version(_orb,local_ior);
		if(local_v!=-1){
		    if(local_v!=org.omg.FT.Utils.extract_group_version(_orb,_ior)){
			_ior=state.request.forward_reference_ior();
			_bindings.clear();
			_bindingVersion=-1;
			receive_redirect(self, state, _ior, true/*false*/);
		    }
		    else
			failover_fatal(self,state,new org.omg.CORBA.COMM_FAILURE());
		}
		else
		    receive_redirect(self, state, state.request.forward_reference_ior(), false);
		break;
	    case org.omg.PortableInterceptor.LOCATION_FORWARD_PERMANENT.value:
		/* Modified by Diana on Feb. 6 2002*/
			local_ior=state.request.forward_reference_ior();
		local_v=org.omg.FT.Utils.extract_group_version(_orb,local_ior);
		if(local_v!=-1){
		    if(local_v!=org.omg.FT.Utils.extract_group_version(_orb,_ior)){
			_ior=state.request.forward_reference_ior();
			_bindings.clear();
			_bindingVersion=-1;
			receive_redirect(self, state, _ior, true/*false*/);
		    }
		    else
			failover_fatal(self,state,new org.omg.CORBA.COMM_FAILURE());
		}
		else
		    receive_redirect(self, state, state.request.forward_reference_ior(), false);

		break;
	    default:
		org.openorb.util.Trace.assertTrue(false);
	    }
	}
    }

    /**
   * Create a request based on stream stub
   *
   * @param self	the object to apply this operation
   * @param Operation	the operation name
   * @param responseExpected True if a response is expected
   * @return an Outputstream to marshal data
   */
    public org.omg.CORBA.portable.OutputStream request( org.omg.CORBA.Object self, int retention_id,String client_id,String operation, boolean responseExpected )
    {
	/*	if(operation.equals("m1")){
		System.out.println("checkfield is "+checkfield);
		checkfield++;
	    }*/
		org.omg.IOP.IOR local_ior;
	int local_v;
	RequestState state = begin_invocation(self, false);
	//System.out.println("After begin invocation");

	org.omg.CORBA.portable.OutputStream ret;
	while(true) {
	    try {
		state.request = state.binding.create_request(self, retention_id,client_id,operation, responseExpected);
	    }
	    catch(org.omg.CORBA.SystemException ex) {
		receive_sysexception(self, state, ex);
		continue;
	    }

	    //System.out.println("Begin marshal in request");
	    ret = state.request.begin_marshal();
	    //System.out.println("After begin marshal");
	    if(ret != null)
		return ret;

	    // a client side interceptor has rejected the request.
	    switch(state.request.reply_status()) {
	    case org.omg.PortableInterceptor.SYSTEM_EXCEPTION.value:
		receive_sysexception(self, state, state.request.received_system_exception());
		break;
	    case org.omg.PortableInterceptor.LOCATION_FORWARD.value:
		//_ior=state.request.forward_reference_ior();
		/*  org.omg.FT.Utils.print_IOR(_orb,_ior);*/
		local_ior=state.request.forward_reference_ior();
		local_v=org.omg.FT.Utils.extract_group_version(_orb,local_ior);
		if(local_v!=-1){
		    if(local_v!=org.omg.FT.Utils.extract_group_version(_orb,_ior)){
			_ior=state.request.forward_reference_ior();
			_bindings.clear();
			_bindingVersion=-1;
			receive_redirect(self, state, _ior, true/*false*/);
		    }
		    else
			failover_fatal(self,state,new org.omg.CORBA.COMM_FAILURE());
		}
		else
		    receive_redirect(self, state, state.request.forward_reference_ior(), false);

		break;
	    case org.omg.PortableInterceptor.LOCATION_FORWARD_PERMANENT.value:
		//_ior=state.request.forward_reference_ior();
		/*  org.omg.FT.Utils.print_IOR(_orb,_ior);*/
		local_ior=state.request.forward_reference_ior();
		local_v=org.omg.FT.Utils.extract_group_version(_orb,local_ior);
		if(local_v!=-1){
		    if(local_v!=org.omg.FT.Utils.extract_group_version(_orb,_ior)){
			_ior=state.request.forward_reference_ior();
			_bindings.clear();
			_bindingVersion=-1;
			receive_redirect(self, state, _ior, true/*false*/);
		    }
		    else
			failover_fatal(self,state,new org.omg.CORBA.COMM_FAILURE());
		}
		else
		    receive_redirect(self, state, state.request.forward_reference_ior(), false);

		break;
	    default:
		org.openorb.util.Trace.assertTrue(false);
	    }
	}
    }



    /* Added by Diana on feb. 2 2002*/

      public org.omg.CORBA.portable.OutputStream request( org.omg.CORBA.Object self, int retention_id,String client_id,org.omg.IOP.ServiceContext one_sc,String operation, boolean responseExpected )
    {
	/*	if(operation.equals("m1")){
		System.out.println("checkfield is "+checkfield);
		checkfield++;
	    }*/
		org.omg.IOP.IOR local_ior;
	int local_v;
	RequestState state = begin_invocation(self, false);
	//System.out.println("After begin invocation");

	org.omg.CORBA.portable.OutputStream ret;
	while(true) {
	    try {
		state.request = state.binding.create_request(self, retention_id,client_id,one_sc,operation, responseExpected);
	    }
	    catch(org.omg.CORBA.SystemException ex) {
		receive_sysexception(self, state, ex);
		continue;
	    }

	    //System.out.println("Begin marshal in request");
	    ret = state.request.begin_marshal();
	    //System.out.println("After begin marshal");
	    if(ret != null)
		return ret;

	    // a client side interceptor has rejected the request.
	    switch(state.request.reply_status()) {
	    case org.omg.PortableInterceptor.SYSTEM_EXCEPTION.value:
		receive_sysexception(self, state, state.request.received_system_exception());
		break;
	    case org.omg.PortableInterceptor.LOCATION_FORWARD.value:
		//_ior=state.request.forward_reference_ior();
	//	org.omg.FT.Utils.print_IOR(_orb,_ior);
				local_ior=state.request.forward_reference_ior();
		local_v=org.omg.FT.Utils.extract_group_version(_orb,local_ior);
		if(local_v!=-1){
		    if(local_v!=org.omg.FT.Utils.extract_group_version(_orb,_ior)){
			_ior=state.request.forward_reference_ior();
			_bindings.clear();
			_bindingVersion=-1;
			receive_redirect(self, state, _ior, true/*false*/);
		    }
		    else
			failover_fatal(self,state,new org.omg.CORBA.COMM_FAILURE());
		}
		else
		    receive_redirect(self, state, state.request.forward_reference_ior(), false);


		break;
	    case org.omg.PortableInterceptor.LOCATION_FORWARD_PERMANENT.value:
		//_ior=state.request.forward_reference_ior();
		/*  org.omg.FT.Utils.print_IOR(_orb,_ior);*/
				local_ior=state.request.forward_reference_ior();
		local_v=org.omg.FT.Utils.extract_group_version(_orb,local_ior);
		if(local_v!=-1){
		    if(local_v!=org.omg.FT.Utils.extract_group_version(_orb,_ior)){
			_ior=state.request.forward_reference_ior();
			_bindings.clear();
			_bindingVersion=-1;
			receive_redirect(self, state, _ior, true/*false*/);
		    }
		    else
			failover_fatal(self,state,new org.omg.CORBA.COMM_FAILURE());
		}
		else
		    receive_redirect(self, state, state.request.forward_reference_ior(), false);
		break;
	    default:
		org.openorb.util.Trace.assertTrue(false);
	    }
	}
    }

    /* Modif. on Oct. 28 2001*/

     public org.omg.CORBA.portable.OutputStream request( org.omg.CORBA.Object self, int retention_id,String client_id,long request_start_time,String operation, boolean responseExpected )
    {
	/*	if(operation.equals("m1")){
		System.out.println("checkfield is "+checkfield);
		checkfield++;
	    }*/


	org.omg.IOP.IOR local_ior;
	int local_v;
	RequestState state = begin_invocation(self, false);


	org.omg.CORBA.portable.OutputStream ret;
	while(true) {

	    try {
		state.request = state.binding.create_request(self, retention_id,client_id,request_start_time,operation, responseExpected);
	    }
	    catch(org.omg.CORBA.SystemException ex) {
		receive_sysexception(self, state, ex);

		continue;
	    }
	    ret = state.request.begin_marshal();
	    if(ret != null){
		return ret;
	    }

	    // a client side interceptor has rejected the request.
	    switch(state.request.reply_status()) {
	    case org.omg.PortableInterceptor.SYSTEM_EXCEPTION.value:
		receive_sysexception(self, state, state.request.received_system_exception());
		break;
	    case org.omg.PortableInterceptor.LOCATION_FORWARD.value:
		//_ior=state.request.forward_reference_ior();
		/*  org.omg.FT.Utils.print_IOR(_orb,_ior);*/
				local_ior=state.request.forward_reference_ior();
		local_v=org.omg.FT.Utils.extract_group_version(_orb,local_ior);
		if(local_v!=-1){
		    if(local_v!=org.omg.FT.Utils.extract_group_version(_orb,_ior)){
			_ior=state.request.forward_reference_ior();
			_bindings.clear();
			_bindingVersion=-1;
			receive_redirect(self, state, _ior, true/*false*/);
		    }
		    else
			failover_fatal(self,state,new org.omg.CORBA.COMM_FAILURE());
		}
		else
		    receive_redirect(self, state, state.request.forward_reference_ior(), false);


		break;
	    case org.omg.PortableInterceptor.LOCATION_FORWARD_PERMANENT.value:
		//_ior=state.request.forward_reference_ior();
		/*  org.omg.FT.Utils.print_IOR(_orb,_ior);*/
				local_ior=state.request.forward_reference_ior();
		local_v=org.omg.FT.Utils.extract_group_version(_orb,local_ior);
		if(local_v!=-1){
		    if(local_v!=org.omg.FT.Utils.extract_group_version(_orb,_ior)){
			_ior=state.request.forward_reference_ior();
			_bindings.clear();
			_bindingVersion=-1;
			receive_redirect(self, state, _ior, true/*false*/);
		    }
		    else
			failover_fatal(self,state,new org.omg.CORBA.COMM_FAILURE());
		}
		else
		    receive_redirect(self, state, state.request.forward_reference_ior(), false);
		break;
	    default:
		org.openorb.util.Trace.assertTrue(false);
	    }
	}
    }

    /* Added by diana on feb. 2 2002*/
      public org.omg.CORBA.portable.OutputStream request( org.omg.CORBA.Object self, int retention_id,String client_id,long request_start_time,org.omg.IOP.ServiceContext one_sc,String operation, boolean responseExpected )
    {
	/*	if(operation.equals("m1")){
		System.out.println("checkfield is "+checkfield);
		checkfield++;
	    }*/


	org.omg.IOP.IOR local_ior;
	int local_v;
	RequestState state = begin_invocation(self, false);


	org.omg.CORBA.portable.OutputStream ret;
	while(true) {

	    try {
		state.request = state.binding.create_request(self, retention_id,client_id,request_start_time,one_sc,operation, responseExpected);
	    }
	    catch(org.omg.CORBA.SystemException ex) {
		receive_sysexception(self, state, ex);

		continue;
	    }
	    ret = state.request.begin_marshal();
	    if(ret != null){
		return ret;
	    }

	    // a client side interceptor has rejected the request.
	    switch(state.request.reply_status()) {
	    case org.omg.PortableInterceptor.SYSTEM_EXCEPTION.value:
		receive_sysexception(self, state, state.request.received_system_exception());
		break;
	    case org.omg.PortableInterceptor.LOCATION_FORWARD.value:
		//_ior=state.request.forward_reference_ior();
		/*  org.omg.FT.Utils.print_IOR(_orb,_ior);*/
				local_ior=state.request.forward_reference_ior();
		local_v=org.omg.FT.Utils.extract_group_version(_orb,local_ior);
		if(local_v!=-1){
		    if(local_v!=org.omg.FT.Utils.extract_group_version(_orb,_ior)){
			_ior=state.request.forward_reference_ior();
			_bindings.clear();
			_bindingVersion=-1;
			receive_redirect(self, state, _ior, true/*false*/);
		    }
		    else
			failover_fatal(self,state,new org.omg.CORBA.COMM_FAILURE());
		}
		else
		    receive_redirect(self, state, state.request.forward_reference_ior(), false);
		break;
	    case org.omg.PortableInterceptor.LOCATION_FORWARD_PERMANENT.value:
	       //_ior=state.request.forward_reference_ior();
	       /*  org.omg.FT.Utils.print_IOR(_orb,_ior);*/
				local_ior=state.request.forward_reference_ior();
		local_v=org.omg.FT.Utils.extract_group_version(_orb,local_ior);
		if(local_v!=-1){
		    if(local_v!=org.omg.FT.Utils.extract_group_version(_orb,_ior)){
			_ior=state.request.forward_reference_ior();
			_bindings.clear();
			_bindingVersion=-1;
			receive_redirect(self, state, _ior, true/*false*/);
		    }
		    else
			failover_fatal(self,state,new org.omg.CORBA.COMM_FAILURE());
		}
		else
		    receive_redirect(self, state, state.request.forward_reference_ior(), false);
		break;
	    default:
		org.openorb.util.Trace.assertTrue(false);
	    }
	}
    }



    /* Modified by Diana on Dec. 16 2001*/

    /*  public org.omg.CORBA.portable.OutputStream[] request_multiple_target( org.omg.CORBA.Object self, int retention_id,String client_id,long request_start_time,String operation, boolean responseExpected )
	{


	RequestState state = begin_invocation(self, false);


	org.omg.CORBA.portable.OutputStream[] ret;
	while(true) {

	try {
	state.request = state.binding.create_request(self, retention_id,client_id,request_start_time,operation, responseExpected);
	}
	catch(org.omg.CORBA.SystemException ex) {
	receive_sysexception(self, state, ex);

	continue;
	}


	ret = state.request.begin_marshal_multiple_target();

	if(ret != null){

	return ret;
	}


	switch(state.request.reply_status()) {
	case org.omg.PortableInterceptor.SYSTEM_EXCEPTION.value:
	receive_sysexception(self, state, state.request.received_system_exception());
	break;
	case org.omg.PortableInterceptor.LOCATION_FORWARD.value:
	receive_redirect(self, state, state.request.forward_reference_ior(), false);
	break;
	case org.omg.PortableInterceptor.LOCATION_FORWARD_PERMANENT.value:
	receive_redirect(self, state, state.request.forward_reference_ior(), true);
	break;
	default:
	org.openorb.util.Trace.assertTrue(false);
	}
	}
	}*/

       /**
   * Create a request based on stream stub
   *
   * @param self	the object to apply this operation
   * @param Operation	the operation name
   * @param responseExpected True if a response is expected
   * @return an Outputstream to marshal data
   */
    public org.omg.CORBA.portable.OutputStream request( org.omg.CORBA.Object self, int retention_id,String client_id,org.omg.Dynamic.Parameter[] arguments,String operation, boolean responseExpected )
    {
	/*	if(operation.equals("m1")){
		System.out.println("checkfield is "+checkfield);
		checkfield++;
	    }*/
		org.omg.IOP.IOR local_ior;
	int local_v;
	RequestState state = begin_invocation(self, false);
	//System.out.println("After begin invocation");

	org.omg.CORBA.portable.OutputStream ret;
	while(true) {
	    try {
		state.request = state.binding.create_request(self, retention_id,client_id,arguments,operation, responseExpected);
	    }
	    catch(org.omg.CORBA.SystemException ex) {
		receive_sysexception(self, state, ex);
		continue;
	    }

	    //System.out.println("Begin marshal in request");
	    ret = state.request.begin_marshal();

	    if(ret != null)
		return ret;

	    // a client side interceptor has rejected the request.
	    switch(state.request.reply_status()) {
	    case org.omg.PortableInterceptor.SYSTEM_EXCEPTION.value:
		receive_sysexception(self, state, state.request.received_system_exception());
		break;
	    case org.omg.PortableInterceptor.LOCATION_FORWARD.value:
		//_ior=state.request.forward_reference_ior();
		/*  org.omg.FT.Utils.print_IOR(_orb,_ior);*/
				local_ior=state.request.forward_reference_ior();
		local_v=org.omg.FT.Utils.extract_group_version(_orb,local_ior);
		if(local_v!=-1){
		    if(local_v!=org.omg.FT.Utils.extract_group_version(_orb,_ior)){
			_ior=state.request.forward_reference_ior();
			_bindings.clear();
			_bindingVersion=-1;
			receive_redirect(self, state, _ior, true/*false*/);
		    }
		    else
			failover_fatal(self,state,new org.omg.CORBA.COMM_FAILURE());
		}
		else
		    receive_redirect(self, state, state.request.forward_reference_ior(), false);
		break;
	    case org.omg.PortableInterceptor.LOCATION_FORWARD_PERMANENT.value:
		//_ior=state.request.forward_reference_ior();
		/*  org.omg.FT.Utils.print_IOR(_orb,_ior);*/
				local_ior=state.request.forward_reference_ior();
		local_v=org.omg.FT.Utils.extract_group_version(_orb,local_ior);
		if(local_v!=-1){
		    if(local_v!=org.omg.FT.Utils.extract_group_version(_orb,_ior)){
			_ior=state.request.forward_reference_ior();
			_bindings.clear();
			_bindingVersion=-1;
			receive_redirect(self, state, _ior, true/*false*/);
		    }
		    else
			failover_fatal(self,state,new org.omg.CORBA.COMM_FAILURE());
		}
		else
		    receive_redirect(self, state, state.request.forward_reference_ior(), false);
		break;
	    default:
		org.openorb.util.Trace.assertTrue(false);
	    }
	}
    }

    /* new version by Diana on oct. 10 2001*/

      public org.omg.CORBA.portable.OutputStream request( org.omg.CORBA.Object self, int retention_id,String client_id,org.omg.Dynamic.Parameter[] arguments,org.omg.CORBA.TypeCode[] exceptions,org.omg.CORBA.Any result,String operation, boolean responseExpected )
    {
	/*if(operation.equals("m1")){
		System.out.println("checkfield is "+checkfield);
		checkfield++;
	    }*/
	/*	if(operation.equals("get_state"))
		System.out.println("oper "+operation+" before begin invo");*/
	RequestState state = begin_invocation(self, false);
	/*	if(operation.equals("get_state"))
		System.out.println("After begin invocation");*/
	org.omg.IOP.IOR local_ior;
	int local_v;
	org.omg.CORBA.portable.OutputStream ret;
	while(true) {
	    try {
		/*	if(operation.equals("get_state"))
			System.out.println("Create request for oper get_state");*/
		state.request = state.binding.create_request(self, retention_id,client_id,arguments,exceptions,result,operation, responseExpected);
	    }
	    catch(org.omg.CORBA.SystemException ex) {
		/*	if(operation.equals("get_state"))
			System.out.println("Receive Exception for oper get_state");*/
		receive_sysexception(self, state, ex);
		continue;
	    }
	    /*  if(operation.equals("get_state"))
		System.out.println("Begin marshal in request");*/
	    ret = state.request.begin_marshal();
	  /*  if(operation.equals("get_state"))
	       System.out.println("After begin marshal");*/
	    if(ret != null)
		return ret;

	    // a client side interceptor has rejected the request.
	    switch(state.request.reply_status()) {
	    case org.omg.PortableInterceptor.SYSTEM_EXCEPTION.value:
		receive_sysexception(self, state, state.request.received_system_exception());
		break;
	    case org.omg.PortableInterceptor.LOCATION_FORWARD.value:
		//_ior=state.request.forward_reference_ior();
		/*  org.omg.FT.Utils.print_IOR(_orb,_ior);*/
		local_ior=state.request.forward_reference_ior();
		local_v=org.omg.FT.Utils.extract_group_version(_orb,local_ior);
		if(local_v!=-1){
		    if(local_v!=org.omg.FT.Utils.extract_group_version(_orb,_ior)){
			_ior=state.request.forward_reference_ior();
			_bindings.clear();
			_bindingVersion=-1;
			receive_redirect(self, state, _ior, true/*false*/);
		    }
		    else
			failover_fatal(self,state,new org.omg.CORBA.COMM_FAILURE());
		}
		else
		    receive_redirect(self, state, state.request.forward_reference_ior(), false);
		break;
	    case org.omg.PortableInterceptor.LOCATION_FORWARD_PERMANENT.value:
		//_ior=state.request.forward_reference_ior();
		/*  org.omg.FT.Utils.print_IOR(_orb,_ior);*/
				local_ior=state.request.forward_reference_ior();
		local_v=org.omg.FT.Utils.extract_group_version(_orb,local_ior);
		if(local_v!=-1){
		    if(local_v!=org.omg.FT.Utils.extract_group_version(_orb,_ior)){
			_ior=state.request.forward_reference_ior();
			_bindings.clear();
			_bindingVersion=-1;
			receive_redirect(self, state, _ior, true/*false*/);
		    }
		    else
			failover_fatal(self,state,new org.omg.CORBA.COMM_FAILURE());
		}
		else
		    receive_redirect(self, state, state.request.forward_reference_ior(), false);
		break;
	    default:
		org.openorb.util.Trace.assertTrue(false);
	    }
	}
    }

    /* New version on Oct. 28 2001 by Diana*/
      public org.omg.CORBA.portable.OutputStream request( org.omg.CORBA.Object self, int retention_id,String client_id,long request_start_time,org.omg.Dynamic.Parameter[] arguments,org.omg.CORBA.TypeCode[] exceptions,org.omg.CORBA.Any result,String operation, boolean responseExpected )
    {
	RequestState state = begin_invocation(self, false);

	org.omg.IOP.IOR local_ior;
	int local_v;
	org.omg.CORBA.portable.OutputStream ret;
	while(true) {
	    try {

		state.request = state.binding.create_request(self, retention_id,client_id,request_start_time,arguments,exceptions,result,operation, responseExpected);
	    }
	    catch(org.omg.CORBA.SystemException ex) {

		receive_sysexception(self, state, ex);
		continue;
	    }
	    ret = state.request.begin_marshal();
	    //System.out.println("In request for "+operation+" after begin marshal");
	    if(ret != null)
		return ret;

	    // a client side interceptor has rejected the request.
	    switch(state.request.reply_status()) {
	    case org.omg.PortableInterceptor.SYSTEM_EXCEPTION.value:
		//System.out.println("Receive exc after");
		receive_sysexception(self, state, state.request.received_system_exception());
		break;
	    case org.omg.PortableInterceptor.LOCATION_FORWARD.value:
		//_ior=state.request.forward_reference_ior();
		/*  org.omg.FT.Utils.print_IOR(_orb,_ior);*/
				local_ior=state.request.forward_reference_ior();
		local_v=org.omg.FT.Utils.extract_group_version(_orb,local_ior);
		if(local_v!=-1){
		    if(local_v!=org.omg.FT.Utils.extract_group_version(_orb,_ior)){
			_ior=state.request.forward_reference_ior();
			_bindings.clear();
			_bindingVersion=-1;
			receive_redirect(self, state, _ior, true/*false*/);
		    }
		    else
			failover_fatal(self,state,new org.omg.CORBA.COMM_FAILURE());
		}
		else
		    receive_redirect(self, state, state.request.forward_reference_ior(), false);
		break;
	    case org.omg.PortableInterceptor.LOCATION_FORWARD_PERMANENT.value:
		//_ior=state.request.forward_reference_ior();
		/*  org.omg.FT.Utils.print_IOR(_orb,_ior);*/
				local_ior=state.request.forward_reference_ior();
		local_v=org.omg.FT.Utils.extract_group_version(_orb,local_ior);
		if(local_v!=-1){
		    if(local_v!=org.omg.FT.Utils.extract_group_version(_orb,_ior)){
			_ior=state.request.forward_reference_ior();
			_bindings.clear();
			_bindingVersion=-1;
			receive_redirect(self, state, _ior, true/*false*/);
		    }
		    else
			failover_fatal(self,state,new org.omg.CORBA.COMM_FAILURE());
		}
		else
		    receive_redirect(self, state, state.request.forward_reference_ior(), false);
		break;
	    default:
		org.openorb.util.Trace.assertTrue(false);
	    }
	}
    }
    /**
   * Invoke the request as a deferred request.
   *
   * @param self	the object to apply this operation
   * @param os the output stream returned from the request operation.
   */
    public void invoke_deferred( org.omg.CORBA.Object self,
				 org.omg.CORBA.portable.OutputStream os )
    {
	if(self == null || os == null)
	    throw new org.omg.CORBA.BAD_PARAM();

	RequestState state = locate_state();
	if(state == null || state.request == null)
	    throw new org.omg.CORBA.BAD_INV_ORDER(MinorCodes.BAD_INV_ORDER_DELEGATE, CompletionStatus.COMPLETED_NO);

	state.request.send_request();

	_piCurrent.store_invocation_ctx(os);
	_piCurrent.set_invocation_ctx(state.parentState);
	state.parentState = null;
    }

    /**
   * Poll for a response for a defered request. Note that this operation
   * can be called from a different thread to the original request, providing
   * that the invocation state is migrated.
   *
   * @param self	the object to apply this operation
   * @param os the output stream returned from the request operation.
   * @returns true if the invoke operation would return immediatly without
   *         waiting.
   */
    public boolean poll_response( org.omg.CORBA.Object self,
				  org.omg.CORBA.portable.OutputStream os )
    {
	if(self == null || os == null)
	    throw new org.omg.CORBA.BAD_PARAM();

	RequestState state = (RequestState)_piCurrent.retrieve_invocation_ctx(os, false);

	if(state == null)
	    throw new org.omg.CORBA.BAD_INV_ORDER("Operation not sent deferred",
						  org.omg.CORBA.OMGVMCID.value | 13, CompletionStatus.COMPLETED_MAYBE);

	// return true if a response has arrived or the request timeout has expired.
	return state.request.poll_response() ||
	    (state.timeout > 0 && state.timeout < System.currentTimeMillis());
    }

    /**
   * Invoke a remote operation for stream based stub
   *
   * @param self	the object to apply this operation
   * @param os the output stream returned from the request operation.
   * @return the marshalled data return from remote object
   */
    public org.omg.CORBA.portable.InputStream invoke( org.omg.CORBA.Object self,
						      org.omg.CORBA.portable.OutputStream os )
	throws org.omg.CORBA.portable.ApplicationException, org.omg.CORBA.portable.RemarshalException
    {
	if(self == null || os == null)
	    throw new org.omg.CORBA.BAD_PARAM();
    	org.omg.IOP.IOR local_ior;
	int local_v;
	int request_state;
	RequestState state = (RequestState)_piCurrent.retrieve_invocation_ctx(os, true);
	if(state != null) {
	    // TODO: we must check here to ensure this thread is in the same
	    // transaction as the thread which began the invocation.
	    //System.out.println("State is not null");
	    state.parentState = locate_state();
	    _piCurrent.set_invocation_ctx(state);

	    request_state = state.request.state();
	} else {
	    //System.out.println("State is null");
	    state = locate_state();

	    if(state == null || state.request == null)
		throw new org.omg.CORBA.BAD_INV_ORDER(MinorCodes.BAD_INV_ORDER_DELEGATE, CompletionStatus.COMPLETED_NO);

	    request_state = state.request.send_request();

	    /*  switch(request_state){
		case ClientRequest.STATE_WAITING:
		System.out.println("Waiting");
		break;
		case ClientRequest.STATE_COMPLETE:
		System.out.println("Complete");
		break;
		case ClientRequest.STATE_MARSHAL:
		System.out.println("Marshal");
		break;
		case ClientRequest.STATE_UNMARSHAL:
		System.out.println("Unmarshal");
		break;
		default:System.out.println("laskj"+request_state);
		}*/
	}

	switch(request_state) {
	case ClientRequest.STATE_WAITING:
	    //System.out.println("STATE_WAITING");
	    if(state.timeout < 0) {
		//System.out.println("wait for response in invoke "+state.timeout);
		request_state = state.request.wait_for_response(0);
	    } else {
		long waittime = state.timeout - System.currentTimeMillis();
		//		System.out.println("Still wait for response "+waittime);
		if(waittime > 0)
		    request_state = state.request.wait_for_response((int)waittime);
	    }
	    //System.out.println("waited for response");
	    break;
	case ClientRequest.STATE_COMPLETE:
	    // client side interceptors set reply, communication problem
	    // or syncState == 0
	    //System.out.println("STATE_COMPLETE");
	    break;
	default:
	    org.openorb.util.Trace.assertTrue(false, "Unexpected case");
	}

	switch(request_state) {
	case ClientRequest.STATE_UNMARSHAL:
	    //System.out.println("STATE_UNMARSHAL");
	    switch(state.request.reply_status()) {
	    case org.omg.PortableInterceptor.SUCCESSFUL.value:
		return state.request.receive_response();
	    case org.omg.PortableInterceptor.USER_EXCEPTION.value:
		throw new org.omg.CORBA.portable.ApplicationException(state.request.received_exception_id(),
								      state.request.receive_response());
	    default:
		org.openorb.util.Trace.assertTrue(false, "Unexpected case");
	    }
	    // break;
	case ClientRequest.STATE_COMPLETE:
	    //System.out.println("STATE_C");
	    switch(state.request.reply_status()) {
	    case org.omg.PortableInterceptor.SUCCESSFUL.value:
				// possible with a void response.
		//System.out.println("REC resp");
		return state.request.receive_response();
	    case org.omg.PortableInterceptor.USER_EXCEPTION.value:
				// possible with a bodyless exception.
		throw new org.omg.CORBA.portable.ApplicationException(state.request.received_exception_id(),
								      state.request.receive_response());
	    case org.omg.PortableInterceptor.SYSTEM_EXCEPTION.value:
		receive_sysexception(self, state, state.request.received_system_exception());
		throw new org.omg.CORBA.portable.RemarshalException();
	    case org.omg.PortableInterceptor.LOCATION_FORWARD.value:
		//_ior=state.request.forward_reference_ior();
		/*  org.omg.FT.Utils.print_IOR(_orb,_ior);*/
				local_ior=state.request.forward_reference_ior();
		local_v=org.omg.FT.Utils.extract_group_version(_orb,local_ior);
		if(local_v!=-1){
		    if(local_v!=org.omg.FT.Utils.extract_group_version(_orb,_ior)){
			_ior=state.request.forward_reference_ior();
			_bindings.clear();
			_bindingVersion=-1;
			receive_redirect(self, state, _ior, true/*false*/);
		    }
		    else
			failover_fatal(self,state,new org.omg.CORBA.COMM_FAILURE());
		}
		else
		    receive_redirect(self, state, state.request.forward_reference_ior(), false);
		throw new org.omg.CORBA.portable.RemarshalException();
	    case org.omg.PortableInterceptor.LOCATION_FORWARD_PERMANENT.value:
			//_ior=state.request.forward_reference_ior();
			/*	org.omg.FT.Utils.print_IOR(_orb,_ior);*/
				local_ior=state.request.forward_reference_ior();
		local_v=org.omg.FT.Utils.extract_group_version(_orb,local_ior);
		if(local_v!=-1){
		    if(local_v!=org.omg.FT.Utils.extract_group_version(_orb,_ior)){
			_ior=state.request.forward_reference_ior();
			_bindings.clear();
			_bindingVersion=-1;
			receive_redirect(self, state, _ior, true/*false*/);
		    }
		    else
			failover_fatal(self,state,new org.omg.CORBA.COMM_FAILURE());
		}
		else
		    receive_redirect(self, state, state.request.forward_reference_ior(), false);
		throw new org.omg.CORBA.portable.RemarshalException();
	    case org.omg.PortableInterceptor.TRANSPORT_RETRY.value:
		failover_transient(self, state, true);
		throw new org.omg.CORBA.portable.RemarshalException();
	    default:
		org.openorb.util.Trace.assertTrue(false, "Unexpected case");
	    }
	case ClientRequest.STATE_WAITING:
	    //System.out.println("STATE_W");
	    // timeout has expired (currently impossible)
	    org.omg.CORBA.TIMEOUT ex
		= new org.omg.CORBA.TIMEOUT(0, CompletionStatus.COMPLETED_MAYBE);
	    state.request.cancel(ex);
	    complete_invocation(state, false);
	    throw ex;
	default:
	    org.openorb.util.Trace.assertTrue(false, "Unexpected case");
	    return null;
	}
    }

    /**
   * This function is always called at the end of an invocation sequence
   * regardless of the outcome.
   *
   * @param self	the object to apply this operation
   * @param is	the input stream for marshalled data from remote object
   */
    public void releaseReply( org.omg.CORBA.Object self, org.omg.CORBA.portable.InputStream is )
    {
	RequestState state = locate_state();
	if(state != null && state.request != null) {
	    switch(state.request.state()) {
	    case ClientRequest.STATE_MARSHAL: {
		complete_invocation(state, false);
		org.omg.CORBA.MARSHAL ex = new org.omg.CORBA.MARSHAL("Exception thrown during marshal",
								     MinorCodes.MARSHAL_REQUEST_UNKNOWN, org.omg.CORBA.CompletionStatus.COMPLETED_NO);
		state.request.cancel(ex);
		// usualy caused by an exception during the request process, so don't
		// throw another.
	    } break;
	    case ClientRequest.STATE_UNMARSHAL: {
		complete_invocation(state, false);
		org.omg.CORBA.MARSHAL ex = new org.omg.CORBA.MARSHAL("Buffer Underread",
								     MinorCodes.MARSHAL_REPLY_UNKNOWN_OR_UNDERREAD, org.omg.CORBA.CompletionStatus.COMPLETED_YES);
		state.request.cancel(ex);
		throw ex;
	    }
	    case ClientRequest.STATE_WAITING: {
		// cancel a polling request. (not currently used)
		org.omg.CORBA.TIMEOUT ex
		    = new org.omg.CORBA.TIMEOUT(0, CompletionStatus.COMPLETED_MAYBE);
		state.request.cancel(ex);
		complete_invocation(state, false);
	    } break;
	    case ClientRequest.STATE_COMPLETE:
		// every state which reaches here will have a yes completion status,
		// only an exception during unmarshaling or thrown by interceptors
		// will have come through.
		complete_invocation(state, true);
		break;
	    }
	}
    }

    /**
   * This function is used for local invocation
   */
    public org.omg.CORBA.portable.ServantObject servant_preinvoke( org.omg.CORBA.Object self,
								   String operation,
								   Class expectedType )
    {
	RequestState state = begin_invocation(self, false);

	state.local_level++;

	boolean failed = true;
	try {
	    org.omg.CORBA.portable.ServantObject ret
		= state.binding.servant_preinvoke(operation, expectedType);
	    failed = false;
	    return ret;
	}
	catch(org.omg.CORBA.SystemException ex) {
	    receive_sysexception(self, state, ex);
	    failed = false;
	    return null;
	}
	catch(org.omg.PortableInterceptor.ForwardRequest ex) {
	    org.omg.IOP.IOR fwd = ((Delegate)((org.omg.CORBA.portable.ObjectImpl)ex.forward)._get_delegate()).ior();
	    receive_redirect(self, state, fwd, ex.permanent);
	    failed = false;
	    return null;
	}
	finally {
	    if(failed)
		complete_invocation(state, false);
	}
    }

    /**
   * This function is used after a local invocation
   */
    public void servant_postinvoke( org.omg.CORBA.Object self,
				    org.omg.CORBA.portable.ServantObject servant )
    {
	RequestState state = locate_state();

	try {
	    state.binding.servant_postinvoke(servant);
	}
	finally {
	    complete_invocation(state, true);
	}
    }

    /**
   * Return the policy associated to the policy type passed as parameter.
   * This will get
   *
   * @param policy_type	the policy type to search and to return
   * @return the policy associated to the policy type
   */
    public org.omg.CORBA.Policy get_policy( org.omg.CORBA.Object self, int policy_type )
    {
	org.omg.CORBA.Policy client_policy = client_policy(self, policy_type);

	int [] pt = new int[1];
	pt[0] = policy_type;
	org.omg.CORBA.PolicyListHolder policies = new org.omg.CORBA.PolicyListHolder();
	org.omg.CORBA.DomainManager[] managers = get_domain_managers_and_policies(self, pt, policies);

	org.omg.CORBA.Policy profile_policy = null;
	if(policies.value.length != 0)
	    profile_policy = policies.value[0];

	if(_orbReconciler == null) {
	    _orbReconciler = (org.openorb.policy.PolicyReconciler)((org.openorb.CORBA.ORB)_orb).getFeature("PolicyReconciler");
	    if(_orbReconciler == null)
		throw new org.omg.CORBA.INTERNAL();
	}

	return _orbReconciler.reconcile_policies(policy_type, client_policy, profile_policy, managers);
    }

    /**
   * Return the effective client side policy.
   *
   * @param policy_type	the policy type to search and to return
   * @return the policy associated to the policy type
   */
    public org.omg.CORBA.Policy get_client_policy( org.omg.CORBA.Object self, int policy_type ) {
	org.omg.CORBA.Policy policy = client_policy(self, policy_type);
	if(policy == null)
	    throw new org.omg.CORBA.INV_POLICY(0, CompletionStatus.COMPLETED_NO);
	return policy;
    }

    private org.omg.CORBA.Policy client_policy( org.omg.CORBA.Object self, int policy_type ) {
	RequestState state = locate_state();

	if(state != null && state.ignore_rebind_mode && policy_type == 23 /* org.omg.Messaging.REBIND_POLICY_TYPE.value */) {
	    // TODO: the RebindPolicy may be temporaraly overwritten with a TRANSPARENT
	    // value if in the context of an invocation which has state.ignore_rebind_mode
	    // as true.
	}

	int [] ptype = new int[1];
	ptype[0] = policy_type;
	org.omg.CORBA.Policy [] mpols;

	if(_policyOver != null &&
	   (mpols = _policyOver.get_policy_overrides(ptype)).length > 0)
	    return mpols[0];

	if(_policyCurr == null) {
	    try {
		_policyCurr = org.omg.CORBA.PolicyCurrentHelper.narrow(_orb.resolve_initial_references("PolicyCurrent"));
	    }
	    catch(org.omg.CORBA.ORBPackage.InvalidName ex) {
		throw new org.omg.CORBA.INTERNAL();
	    }
	}

	if((mpols = _policyCurr.get_policy_overrides(ptype)).length > 0)
	    return mpols[0];

	if(_orbPolicyManager == null) {
	    try {
		_orbPolicyManager = (org.openorb.policy.ORBPolicyManagerImpl)_orb.resolve_initial_references("ORBPolicyManager");
	    }
	    catch(org.omg.CORBA.ORBPackage.InvalidName ex) {
		throw new org.omg.CORBA.INTERNAL();
	    }
	}

	if((mpols = _orbPolicyManager.get_policy_overrides(ptype)).length > 0)
	    return mpols[0];

	return null;
    }

    /**
   * Return a copy of this object with the specified policies overriden.
	 * @throws NO_PERMISSION some of the overriden policies are invalid.
   */
    public org.omg.CORBA.Object set_policy_override( org.omg.CORBA.Object self,
						     org.omg.CORBA.Policy [] policies,
						     org.omg.CORBA.SetOverrideType set_add )
    {
	try {
	    return set_policy_overrides(self, policies, set_add);
	}
	catch(org.omg.CORBA.InvalidPolicies ex) {
	    throw new org.omg.CORBA.NO_PERMISSION(0, CompletionStatus.COMPLETED_NO);
	}
    }

    /**
   * Return a copy of this object with the specified policies overriden.
	 * @throws org.omg.CORBA.InvalidPolicies some of the overriden policies are invalid.
   */
    public org.omg.CORBA.Object set_policy_overrides( org.omg.CORBA.Object self,
						      org.omg.CORBA.Policy [] policies,
						      org.omg.CORBA.SetOverrideType set_add )
	throws org.omg.CORBA.InvalidPolicies
    {
	if(_policySetManager == null) {
	    _policySetManager = (org.openorb.policy.PolicySetManager)((org.openorb.CORBA.ORB)_orb).getFeature("PolicySetManager");
	    if(_policySetManager == null)
		throw new org.omg.CORBA.INTERNAL();
	}

	org.omg.CORBA.PolicyManagerOperations npols;
	if(policies.length == 0) {
	    // all old set
	    if(_policyOver == null || set_add == org.omg.CORBA.SetOverrideType.SET_OVERRIDE)
		npols = null;
	    else {
		npols = _policySetManager.create_policy_set(_policySetManager.CLIENT_POLICY_DOMAIN);
		npols.set_policy_overrides(_policyOver.get_policy_overrides(new int[0]),
					   org.omg.CORBA.SetOverrideType.SET_OVERRIDE);
	    }
	}
	else if(_policyOver == null || set_add == org.omg.CORBA.SetOverrideType.SET_OVERRIDE) {
	    // all new set
	    npols = _policySetManager.create_policy_set(_policySetManager.CLIENT_POLICY_DOMAIN);
	    npols.set_policy_overrides(policies, set_add);
	}
	else {
	    // union set.
	    npols = _policySetManager.create_policy_set(_policySetManager.CLIENT_POLICY_DOMAIN);
	    npols.set_policy_overrides(_policyOver.get_policy_overrides(new int[0]),
				       org.omg.CORBA.SetOverrideType.SET_OVERRIDE);
	    npols.set_policy_overrides(policies, set_add);
	}

	// we now have a property set to give to the new reference/delegate.


	// try to instantiate a new stub, of the same type as the old one
	org.omg.CORBA.portable.ObjectImpl nobj = null;
	if(!(nobj instanceof org.openorb.CORBA.ObjectStub)) {
	    try {
		nobj = (org.omg.CORBA.portable.ObjectImpl)self.getClass().newInstance();
	    }
	    catch(Exception ex) {
		org.openorb.util.Trace.assertTrue(false, "Unexpected case");
	    }
	}
	if(nobj == null)
	    nobj = new org.openorb.CORBA.ObjectStub();

	Delegate ndeleg = new Delegate(_orb, _effectiveIor);
	ndeleg._policyOver = npols;

	nobj._set_delegate(ndeleg);

	return nobj;
    }

    /**
	 * Returns the list of Policy overrides (of the specified policy types)
	 * set at the Object scope. If the specified sequence is empty, all
	 * Policy overrides at this scope will be returned. If none of the
	 * requested PolicyTypes are overridden at the Object scope,
	 * an empty sequence is returned.
	 */
    public org.omg.CORBA.Policy[] get_policy_overrides(int[] ts) {
	if(_policyOver == null)
	    return new org.omg.CORBA.Policy[0];
	return _policyOver.get_policy_overrides(ts);
    }

    /**
	 * Validate the connection to the client. This operation ignores the rebind
	 * policy.
	 */
    public boolean validate_connection( org.omg.CORBA.Object self,
					org.omg.CORBA.PolicyListHolder inconsistent_policies)
    {
	org.omg.CORBA.PolicyListHolder policies = new org.omg.CORBA.PolicyListHolder();

	// set the force rebind flag
	RequestState state = begin_invocation(self, true);

	try {
	    boolean ret = locate_and_get_policies(self, state, policies);
	    inconsistent_policies.value = new org.omg.CORBA.Policy[0];
	    return ret;
	}
	catch(org.omg.CORBA.INV_POLICY ex) {
	    // we have some invalid policies.
	    inconsistent_policies.value = new org.omg.CORBA.Policy[0];
	    // TODO: find which policies are inconsistent.
	    return false;
	}
    }

    /**
   * Return a reference to the ORB
   */
    public org.omg.CORBA.ORB orb( org.omg.CORBA.Object self )
    {
	return _orb;
    }

    /**
   * Pass to string this object
   */
    public String toString( org.omg.CORBA.Object self )
    {
	try {
	    Address [] addrs = getAddresses(self);

	    ArrayList iorList = new ArrayList();
	    for(int i = 0; i < addrs.length; ++i)
		if(addrs[i] != null) {
		    org.omg.GIOP.IORAddressingInfo info =
			addrs[i].getTargetAddress(org.omg.GIOP.ReferenceAddr.value).ior();
		    ArrayList [] profs = new ArrayList[info.ior.profiles.length];
		    iorList.add(profs);
		    ArrayList prof = profs[info.selected_profile_index] = new ArrayList();
		    prof.add(addrs[i]);
		    for(int j = i+1; j < addrs.length; ++j)
			if(addrs[j] != null) {
			    org.omg.GIOP.IORAddressingInfo info2 =
				addrs[j].getTargetAddress(org.omg.GIOP.ReferenceAddr.value).ior();
			    if(compareIORs(info.ior, info2.ior)) {
				prof = profs[info2.selected_profile_index];
				if(prof == null)
				    prof = profs[info2.selected_profile_index] = new ArrayList();
				prof.add(addrs[j]);
				addrs[j] = null;
			    }
			}
		}

	    ByteArrayOutputStream os = new ByteArrayOutputStream();
	    HexPrintStream hps = new HexPrintStream(os, HexPrintStream.FORMAT_MIXED);
	    PrintStream ps = new PrintStream(os);

	    ps.println("Binding information for object " + self.getClass() + ":" + Integer.toHexString(System.identityHashCode(self)));
	    String [] ids = ((org.omg.CORBA.portable.ObjectImpl)self)._ids();
	    if(ids.length != 0) {
		ps.println("Repository IDs from Object: ");
		for(int i = 0; i < ids.length; ++i)
		    ps.println("\t" + ids[i]);
	    }
	    if(iorList.size() > 1)
		ps.println("There have been " + (iorList.size()-1) + " redirections of this object");
	    for(int i = 0; i < iorList.size(); ++i) {
		ArrayList [] profs = (ArrayList[])iorList.get(i);
		int totProfs = 0;
		int profOne = -1;
		for(int j = profs.length-1; j >= 0; --j)
		    if(profs[j] != null) {
			totProfs++;
			profOne = j;
		    }
		if(totProfs == 0)
		    continue;
		org.omg.IOP.IOR ior = ((Address)profs[profOne].get(0)).getTargetAddress(org.omg.GIOP.ReferenceAddr.value).ior().ior;

		ps.print("IOR #" + i);
		if(compareIORs(ior, _ior))
		    ps.println(" (orginal IOR)");
		else if(compareIORs(ior, _effectiveIor))
		    ps.println(" (from latest permanent redirect)");
		else
		    ps.println(" (from redirect)");
		ps.println("Published RepoID: " + ior.type_id);
		ps.println("Bound profiles: " + totProfs);
		Address addr0 = null;
		for(int j = 0; j < profs.length; ++j)
		    if(profs[j] != null) {
			ps.print("Profile #" + j);
			addr0 = (Address)profs[j].get(0);
			switch(profs[j].size()) {
			case 0:
			    ps.println(" no endpoints found");
			    break;
			case 1:
			    ps.println(" endpoint: " + addr0.getEndpointDescription());
			    break;
			default:
			    ps.println(" endpoints:");
			    for(int k = 0; k < profs[j].size(); ++k)
				ps.println("\t" + ((Address)profs[j].get(k)).getEndpointDescription());
			}
			ps.print(addr0.getObjectKeyDescription());
			int profCpts = addr0.get_profile_components();
			if(profCpts > 0) {
			    ps.println("Components:");
			    for(int k = 0; k < profCpts; ++k) {
				java.lang.Object data = addr0.get_component_data(k);
				if(data == null) {
				    ps.println("Unknown Component. Tag: " + addr0.get_component(k).tag + " Data:");
				    ps.flush();
				    hps.write(addr0.get_component(k).component_data);
				    hps.flush();
				    ps.println();
				} else {
				    ps.println(data.toString());
				}
			    }
			}
		    }
		int profCpts = addr0.get_profile_components();
		int total = addr0.get_components().length;
		if(profCpts < total) {
		    ps.println("Components from Multi-Component Profile:");
		    for(int k = profCpts; k < total; ++k) {
			java.lang.Object data = addr0.get_component_data(k);
			if(data == null) {
			    ps.println("Unknown Component. Tag: " + addr0.get_component(k).tag + " Data:");
			    ps.flush();
			    hps.write(addr0.get_component(k).component_data);
			    hps.flush();
			    ps.println();
			} else {
			    ps.println(data.toString());
			}
		    }
		}
	    }

	    return os.toString();
	}
	catch(java.io.IOException ex) {
	    throw new RuntimeException("Unexpected Exception: " + ex);
	}
    }

    private boolean compareIORs(org.omg.IOP.IOR ior1, org.omg.IOP.IOR ior2) {
	if(ior1 == ior2)
	    return true;
	if(!ior1.type_id.equals(ior2.type_id) || ior1.profiles.length != ior2.profiles.length)
	    return false;
	for(int i = 0; i < ior1.profiles.length; ++i)
	    if(ior1.profiles[i].tag != ior2.profiles[i].tag ||
	       !Arrays.equals(ior1.profiles[i].profile_data, ior2.profiles[i].profile_data))
		return false;
	return true;
    }

    /**
   * Return codebase for stub class.
   */
    public String get_codebase(org.omg.CORBA.Object self) {
	return null;
    }

    // Public functions not in org.omg.CORBA.portable.Delegate

  /**
   * Get the ior of the object as it should be serialized / marshalled. This
	 * will change on a permanent redirection.
   */
    public org.omg.IOP.IOR ior() {
	return _effectiveIor;
    }

    /**
   * Get a snapshot list of the target addresses. This is subject to grow
   * when an object forward is received.
   */
    public Address [] getAddresses(org.omg.CORBA.Object self) {
	// locate any current state.
	java.lang.Object oldCtx = null;

	if( _piCurrent != null ) {
	    oldCtx = _piCurrent.get_invocation_ctx();
	    _piCurrent.set_invocation_ctx(null);
	}

	// begin a dummy invocation.
	RequestState state = begin_invocation(self, false);

	// ensure we have the entire binding list.
	if(state.bindings == null) {
	    state.binding = null;
	    refresh_binding_list(state, null);
	}
	// extract the addresses.
	Address [] addrs = new Address[state.bindings.length];
	for(int i = 0; i < addrs.length; ++i)
	    addrs[i] = state.bindings[i].getAddress();

	// throw away the dummy invocation.
	if( _piCurrent != null ) _piCurrent.set_invocation_ctx(oldCtx);

	return addrs;
    }

    // internal functions.


    private boolean locate_and_get_policies(org.omg.CORBA.Object self,
					    RequestState state,
					    org.omg.CORBA.PolicyListHolder policies)
    {
	int [] policy_types = new int[0];
		org.omg.IOP.IOR local_ior;
	int local_v;
	try {
	    while(true) {
		if(policies != null)
		    policies.value = state.binding.getAddress().get_target_policies(policy_types);
		if(state.binding.getObjectAdapter() == null) {
		    try {
			state.request = state.binding.create_locate_request(self);
		    }
		    catch(org.omg.CORBA.SystemException ex) {
			receive_sysexception(self, state, ex);
			continue;
		    }

		    // this will return null
		    state.request.begin_marshal();

		    int request_state = state.request.state();
		    if(request_state == ClientRequest.STATE_MARSHAL) {
			request_state = state.request.send_request();
			if(request_state == ClientRequest.STATE_WAITING) {
			    if(state.timeout < 0) {
				request_state = state.request.wait_for_response(0);
			    } else {
				long waittime = state.timeout - System.currentTimeMillis();
				if(waittime > 0)
				    request_state = state.request.wait_for_response((int)waittime);
			    }
			    if(request_state == ClientRequest.STATE_WAITING) {
				// timeout has expired (currently impossible)
				org.omg.CORBA.TIMEOUT ex
				    = new org.omg.CORBA.TIMEOUT(0, CompletionStatus.COMPLETED_MAYBE);
				state.request.cancel(ex);
				throw ex;
			    }
			}
		    }

		    org.openorb.util.Trace.assertTrue(request_state == ClientRequest.STATE_COMPLETE, "Unexpected case");

		    switch(state.request.reply_status()) {
		    case ClientRequest.OBJECT_HERE:
			complete_invocation(state, true);
			return false;
		    case ClientRequest.UNKNOWN_OBJECT:
			failover_permanent(self, state, new org.omg.CORBA.OBJECT_NOT_EXIST(0, CompletionStatus.COMPLETED_NO));
			break;
		    case org.omg.PortableInterceptor.SYSTEM_EXCEPTION.value:
			receive_sysexception(self, state, state.request.received_system_exception());
			break;
		    case org.omg.PortableInterceptor.LOCATION_FORWARD.value:
			//_ior=state.request.forward_reference_ior();
			/*	  org.omg.FT.Utils.print_IOR(_orb,_ior);*/
			local_ior=state.request.forward_reference_ior();
			local_v=org.omg.FT.Utils.extract_group_version(_orb,local_ior);
			if(local_v!=-1){
			    if(local_v!=org.omg.FT.Utils.extract_group_version(_orb,_ior)){
				_ior=state.request.forward_reference_ior();
				_bindings.clear();
				_bindingVersion=-1;
				receive_redirect(self, state, _ior, true/*false*/);
			    }
			    else
				failover_fatal(self,state,new org.omg.CORBA.COMM_FAILURE());
			}
			else
			    receive_redirect(self, state, state.request.forward_reference_ior(), false);

			break;
		    case org.omg.PortableInterceptor.LOCATION_FORWARD_PERMANENT.value:
			//_ior=state.request.forward_reference_ior();
			/*  org.omg.FT.Utils.print_IOR(_orb,_ior);*/
			local_ior=state.request.forward_reference_ior();
			local_v=org.omg.FT.Utils.extract_group_version(_orb,local_ior);
			if(local_v!=-1){
			    if(local_v!=org.omg.FT.Utils.extract_group_version(_orb,_ior)){
				_ior=state.request.forward_reference_ior();
				_bindings.clear();
				_bindingVersion=-1;
				receive_redirect(self, state, _ior, true/*false*/);
			    }
			    else
				failover_fatal(self,state,new org.omg.CORBA.COMM_FAILURE());
			}
			else
			    receive_redirect(self, state, state.request.forward_reference_ior(), false);
			break;
		    case org.omg.PortableInterceptor.TRANSPORT_RETRY.value:
			failover_transient(self, state, true);
			break;
		    default:
			org.openorb.util.Trace.assertTrue(false, "Unexpected case");
		    }
		}
		else
		    {
			try {
			    if(state.binding.locate()) {
				complete_invocation(state, true);
				return false;
			    } else
				failover_permanent(self, state, new org.omg.CORBA.OBJECT_NOT_EXIST(0, CompletionStatus.COMPLETED_NO));
			}
			catch(org.omg.PortableInterceptor.ForwardRequest ex) {
			    org.omg.IOP.IOR fwd = ((Delegate)((org.omg.CORBA.portable.ObjectImpl)ex.forward)._get_delegate()).ior();
			    receive_redirect(self, state, fwd, ex.permanent);
			}
		    }
	    }
	}
	catch(org.omg.CORBA.SystemException ex) {
	    if( (ex instanceof org.omg.CORBA.NO_RESPONSE) ||
		(ex instanceof org.omg.CORBA.COMM_FAILURE) ||
		(ex instanceof org.omg.CORBA.NO_IMPLEMENT) ||
		(ex instanceof org.omg.CORBA.OBJ_ADAPTER) ||
		(ex instanceof org.omg.CORBA.OBJECT_NOT_EXIST))
		return true;
	    throw ex;
	}
    }

    private static class RequestState {
	ClientBinding binding;
	ClientRequest request;
	boolean ignore_rebind_mode;

	int vers;
	int idx;
	ClientBinding [] bindings = null;
	boolean [] used = null;
	int remain;

	int ex_lvl = -1;
	org.omg.CORBA.SystemException ex = null;

	long timeout = -1;

	Delegate target;

	int local_level = 0;
	RequestState parentState = null;
    }

    private RequestState begin_invocation(org.omg.CORBA.Object obj,
					  boolean ignore_rebind_mode)
    {
	if(_piCurrent == null)
	    _piCurrent = ((org.openorb.CORBA.ORB)_orb).getPICurrent();

	RequestState state = (RequestState)_piCurrent.get_invocation_ctx();

	if(state == null) {
	    state = new RequestState();
	} else if(state.target == this) {
	    org.openorb.util.Trace.assertTrue(state.request == null, "Invalid state at begin_invocation");
	    if(ignore_rebind_mode)
		state.ignore_rebind_mode = true;
	    return state;
	} else if(state.request == null && state.local_level == 0) {
	    // is_local on one target, followed by invocation on another.
	    state = new RequestState();
	} else {
	    RequestState newState = new RequestState();
	    newState.parentState = state;
	    state = newState;
	}

	state.target = this;
	_piCurrent.set_invocation_ctx(state);

	ClientBinding [] blst;

	synchronized(_bindingSync) {
	    state.ignore_rebind_mode = ignore_rebind_mode;
	    state.vers = _bindingVersion;
	    state.idx = -1;

	    //System.out.println("Before test _bindingVersion is "+_bindingVersion);
	    if(_bindingVersion < 0) {
		// this is the first attempt to make a request
		// TODO: is this correct or do we force use of validate_connection
		// even for the first bind?
		state.ignore_rebind_mode = true;

		_clientManager = (ClientManager)((org.openorb.CORBA.ORB)_orb).getFeature("ClientCPCManager");
		//System.out.println("Before receive_redirect in Delegate.begin_invocation "+state.vers);
		//org.omg.FT.Utils.print_IOR(_orb,_ior);
		receive_redirect(obj, state, _ior, true);
		return state;
	    }

	    ClientChannel channel;

	    // first try the selected binding if there is one.
	    if(_selectedBinding != null) {
		state.binding = _selectedBinding;
		if(_selectedBinding.local_invoke() || _selectedBinding.getClientChannel().state() != ClientChannel.STATE_CLOSED)
		    return state;
		// TODO: if we have a NO_REBIND (but not NO_RECONNECT) policy and
		// the state of the channel is paused we want to at least attempt
		// to reopen the known good connection before rebinding.
		_selectedBinding = null;
	    }

	    blst = new ClientBinding[_bindings.size()];
	    blst = (ClientBinding[])_bindings.toArray(blst);
	}

	refresh_binding_list(state, blst);

	// try for the next available binding in the list
	find_next_binding(state);

	/* Modif. Diana June 11 2002*/

	//state.timeout=16;

	return state;
    }

    private RequestState locate_state() {
	if(_piCurrent == null)
	    _piCurrent = ((org.openorb.CORBA.ORB)_orb).getPICurrent();

	RequestState state = (RequestState)_piCurrent.get_invocation_ctx();

	while(state != null && state.target != this) {
	    if(state.request != null || state.local_level > 0)
		return null;
	    state = state.parentState;
	}

	return state;
    }

    /**
   * Update selected binding.
   */
    private void complete_invocation(RequestState state, boolean success) {
	if(success == (_selectedBinding == null))
	    synchronized(_bindingSync) {
		if(success) {
		    if(_selectedBinding == null)
			_selectedBinding = state.binding;
		} else {
		    if(_selectedBinding == state.binding)
			_selectedBinding = null;
		}
	    }

	if(--state.local_level <= 0)
	    _piCurrent.set_invocation_ctx(state.parentState);
    }

    private void find_next_binding(RequestState state) {
	// first look for locals or connected channels, then for
	// paused channels.
	try{
	    if(state==null) System.out.println("HMMMKM");
	    if(state.bindings==null) System.out.println("Wronggggg");
	    if(state.used==null) System.out.println("Wronger");
	    for(int j = 0, i = state.idx; j < state.used.length; ++j, i=(i+1)%state.used.length) {
		//System.out.println("state.used["+i+"]="+state.used[i]);
		if(!state.used[i] && (state.bindings[i].local_invoke() ||
				      state.bindings[i].getClientChannel().state() == ClientChannel.STATE_CONNECTED)) {
		    state.binding = state.bindings[i];
		    state.idx = i;
		    //System.out.println("State.idx when return 1 is "+state.idx+" "+state.binding);
		    return;
		}
	    }
	    for(int j = 0, i = state.idx; j < state.used.length; ++j, i=(i+1)%state.used.length) {
		//System.out.println("state.used["+i+"]="+state.used[i]);
		if(!state.used[i]) {
		    if(state.bindings[i].getClientChannel().state() != ClientChannel.STATE_CLOSED) {
		    state.binding = state.bindings[i];
		    state.idx = i;
		    //System.out.println("State.idx when return 2 is "+state.idx+" "+state.binding);
		    return;
		}
		    state.used[i] = true;
		    state.remain--;
		}
	}
	    org.openorb.util.Trace.assertTrue(state.remain == 0);
	    //System.out.println("Here state.remain should be 0 "+state.remain);
	    complete_invocation(state, false);
	}catch(java.lang.Exception e){System.out.println("In find_next_b exceptiuon");e.printStackTrace();}
	if(state.ex_lvl < 0){
	    //System.out.println("in find_next_binding throw COMM_FAILURE");
	    throw new org.omg.CORBA.COMM_FAILURE(0, CompletionStatus.COMPLETED_NO);
	}
	else{
	    //System.out.println("in find next bind throw "+state.ex);
	    throw state.ex;
	}
    }

    /**
   * Cleans up after recieving an exception. If a retry is indicated
   * this function returns normally, otherwise it throws the system exception.
   */
    private void receive_sysexception(org.omg.CORBA.Object self,
				      RequestState state,
				      org.omg.CORBA.SystemException ex) {

	if(ex instanceof org.omg.CORBA.TRANSACTION_ROLLEDBACK){
	    //System.out.println("It is rollback");
	    throw ex;
	}

	if(ex.completed == CompletionStatus.COMPLETED_NO) {
	    if(ex instanceof org.omg.CORBA.TRANSIENT) {
		failover_transient(self, state, false);
		return;
	    }
	    if((ex instanceof org.omg.CORBA.NO_RESPONSE) ||
	       (ex instanceof org.omg.CORBA.COMM_FAILURE) ||

	       // nonstandard failover conds.
	       (ex instanceof org.omg.CORBA.NO_IMPLEMENT) ||
	       (ex instanceof org.omg.CORBA.INV_POLICY) ||
	       (ex instanceof org.omg.CORBA.NO_PERMISSION)) {
		//state.bindings=null;
		//receive_redirect(self,state,_ior,true);
		failover_fatal(self, state, ex);
		return;
	    }
	    if((ex instanceof org.omg.CORBA.OBJ_ADAPTER) ||
	       (ex instanceof org.omg.CORBA.OBJECT_NOT_EXIST)) {
		failover_permanent(self, state, ex);
		return;
	    }
	}
	/* Added by Diana on Nov. 13 2001*/
	if(ex.completed == CompletionStatus.COMPLETED_MAYBE){
	    failover_transient(self,state,false);
	    return;
	}
	complete_invocation(state, ex.completed == CompletionStatus.COMPLETED_YES);
	throw ex;
    }

    /**
   * Request failed but a retry on the binding is indicated.
   */
    private void failover_transient(org.omg.CORBA.Object self, RequestState state,
				    boolean transport_retry)
    {
	// set the request to null
	state.request = null;

		// don't mark any binding as used.
	ClientBinding oldBinding = state.binding;
	if(state.bindings == null)
	    state.binding = null;

	refresh_binding_list(state, null);
	state.idx = -1;

    // TODO: if we have a NO_REBIND policy then set transport_retry to true

	if(transport_retry) {
	    // reuse the same connection.
	    state.binding = oldBinding;
	    for(int i = 0; i < state.bindings.length; ++i)
		if(oldBinding == state.bindings[i]) {
		    state.idx = i;
		    return;
		}
	    org.openorb.util.Trace.assertTrue(false);
	} else {
	    state.idx = (state.idx + 1) % state.bindings.length;
	    find_next_binding(state);
	}
    }

    private void failover_fatal(org.omg.CORBA.Object self, RequestState state,
				org.omg.CORBA.SystemException ex)
    {
	// set the request to null
	state.request = null;

	if(!state.ignore_rebind_mode) {
	    // TODO: check the rebind mode policy to ensure it is not
	    // equal to NO_REBIND or NO_RECONNECT.
	    if(false) {
		complete_invocation(state, false);
		throw new org.omg.CORBA.REBIND(0, CompletionStatus.COMPLETED_NO);
	    }
	}

	// mark selected binding as used.
	if(state.bindings != null) {
	    state.used[state.idx] = true;
	    state.remain--;
	}

	refresh_binding_list(state, null);

	if(state.ex_lvl <= 1) {
	    state.ex_lvl = 1;
	    state.ex = ex;
	}

	find_next_binding(state);
    }

    private void failover_permanent(org.omg.CORBA.Object self, RequestState state,
				    org.omg.CORBA.SystemException ex)
    {
	// set the request to null
	state.request = null;

	if(!state.ignore_rebind_mode) {
	    // TODO: check the rebind mode policy to ensure it is not
	    // equal to NO_REBIND or NO_RECONNECT.
	    if(false) {
		complete_invocation(state, false);
		throw new org.omg.CORBA.REBIND(0, CompletionStatus.COMPLETED_NO);
	    }
	}

	// lower the binding's priority to a last gasp level
	state.binding.setPriority((state.binding.getPriority() & 0xFFFF) | 0xFB0000);

	// mark as used.
	if(state.bindings != null) {
	    state.used[state.idx] = true;
	    state.remain--;
	}
	refresh_binding_list(state, null);

	if(state.ex_lvl <= 3) {
	    state.ex_lvl = 3;
	    state.ex = ex;
	}

	find_next_binding(state);
    }

    /**
   * Deals with a redirection.
   */
    private void receive_redirect(org.omg.CORBA.Object self, RequestState state,org.omg.IOP.IOR ior, boolean permanent)
    {
	// set the request to null
	state.request = null;
	/*	System.out.println("Pritn IOR in receive_Redirect");
	org.omg.FT.Utils.print_IOR(_orb,ior);
	*/
	if(!state.ignore_rebind_mode) {
	    // TODO: check the rebind mode policy to ensure it is not
	    // equal to NO_REBIND or NO_RECONNECT.
	    if(false) {
		complete_invocation(state, false);
		throw new org.omg.CORBA.REBIND(0, CompletionStatus.COMPLETED_NO);
	    }
	}

	// mark the binding which raised the redirect as used
	if(state.bindings != null) {
	    state.used[state.idx] = true;
	    state.remain--;
	}

	ClientBinding [] blst = _clientManager.create_bindings(self, ior);
	int vers = -1;
	synchronized(_bindingSync) {
	    if(permanent)
		_effectiveIor = ior;
	    // check to make sure the bindings aren't already present.
	    int tbind = blst.length;

	    for(int i = 0; i < blst.length; ++i)
		if(_bindings.contains(blst[i])) {
		    blst[i] = null;
		    tbind--;
		}

	    if(tbind > 0) {
		vers = ++_bindingVersion;
				// reshuffle the priorities.
		int pri;
		int npri = ((250-_bindingVersion)<<16);
		if(_bindingVersion < 250) {
		    for(int i = 0; i < blst.length; ++i) {
			if(blst[i]!=null){
			    pri = blst[i].getPriority();
			    if(pri >= 0) {
				pri = (pri & 0xFFFF) | npri;
				blst[i].setPriority(pri);
				_bindings.add(blst[i]);
			    }
			}
		    }
		}
		else {
		    for(Iterator itt = _bindings.iterator(); itt.hasNext();) {
			ClientBinding wk = (ClientBinding)itt.next();
			pri = wk.getPriority();
			if(pri >= 0) {
			    npri = (pri & 0xFF0000) + 0x10000;
			    pri = (pri & 0xFFFF) | ((npri > 0xFA0000)?0xFA0000:npri);
			    wk.setPriority(pri);
			}
		    }
		    for(int i = 0; i < blst.length; ++i)
			_bindings.add(blst[i]);
		}
		blst = (ClientBinding[])_bindings.toArray(blst);
	    }
	}

	if(vers >= 0) {
	    //  System.out.println("Refresh binding list");
	    refresh_binding_list(state, blst);
	    state.vers = vers;
	}
	find_next_binding(state);
	/*	System.out.println("Print effective IOR in receive_redirect");
		org.omg.FT.Utils.print_IOR(_orb,_effectiveIor);*/
    }


    /**
   * refreshes the binding list, adding the bindings in blst.
   */
    private void refresh_binding_list(RequestState state, ClientBinding [] blst) {
	if(blst == null)
	    synchronized(_bindingSync) {
		//System.out.println("Before test _bindingVersion is "+_bindingVersion);
		if(state.bindings != null && state.vers >= _bindingVersion)
		    return;
		blst = new ClientBinding[_bindings.size()];
		blst = (ClientBinding[])_bindings.toArray(blst);
		state.vers = _bindingVersion;
	    }

	boolean [] used = new boolean[blst.length];
	// not needed, default value is false
	// Arrays.fill(nused, false);

	state.remain = blst.length;

	// mark the previously used bindings as still used.
	if(state.bindings != null) {
	    for(int i = 0; i < state.bindings.length; ++i)
		if(state.used[i]){
		    //System.out.println("state.used("+i+")="+state.used[i]);
		    for(int j = 0; j < blst.length; ++j)
			if(blst[j].equals(state.bindings[i])) {
			    used[j] = true;
			    --state.remain;
			    break;
			}
		}
	} else if(state.binding != null) {
	    for(int j = 0; j < blst.length; ++j)
		if(blst[j].equals(state.binding)) {
		    used[j] = true;
		    --state.remain;
		    break;
		}
	}

	//	System.out.println("state.bindings will be blst");
	state.bindings = blst;
	state.used = used;

	if(state.remain > 1) {
	    // sort the states
	    ClientBinding [] blstsrc = (ClientBinding[])blst.clone();
	    boolean [] usedsrc = (boolean[])used.clone();
	    /* if(usedsrc==null) System.out.println("usedsrc is null");
	    else
	    System.out.println("usedsrc is not null");*/
	    mergeSort(blstsrc, usedsrc, blst, used, 0, blst.length);
	}

	state.idx = 0;
    }

    /**
   * Merge sort operation. Sorts used and binding list arrays together.
   */
    private static void mergeSort(ClientBinding [] blstsrc, boolean [] usedsrc,
				  ClientBinding [] blst, boolean [] used,
				  int low, int high) {
	int length = high - low;

	// Insertion sort on smallest arrays
	if (length < 7) {
	    ClientBinding t;
	    boolean ut;
	    for (int i=low; i<high; i++)
		for (int j=i; j>low && statecomp(blst[j-1], used[j-1], blst[j], used[j])>0; j--) {
		    t = blst[j];
		    ut = used[j];
		    blst[j] = blst[j-1];
		    used[j] = used[j-1];
		    blst[j-1] = t;
		    used[j-1] = ut;
		}
	    return;
	}

	// Recursively sort halves of dest into src
	int mid = (low + high)/2;
	mergeSort(blstsrc, usedsrc, blst, used, low, mid);
	mergeSort(blstsrc, usedsrc, blst, used, mid, high);

	// If list is already sorted, just copy from src to dest.  This is an
	// optimization that results in faster sorts for nearly ordered lists.
	if (statecomp(blstsrc[mid-1], usedsrc[mid-1], blstsrc[mid], usedsrc[mid]) <= 0) {
	    System.arraycopy(blstsrc, low, blst, low, length);
	    System.arraycopy(usedsrc, low, used, low, length);
	    return;
	}

	// Merge sorted halves (now in src) into dest
	for(int i = low, p = low, q = mid; i < high; i++) {
	    if (q>=high || p<mid && statecomp(blstsrc[p], usedsrc[p], blstsrc[q], usedsrc[q]) <= 0) {
		blst[i] = blstsrc[p];
		used[i] = usedsrc[p];
		p++;
	    } else {
		blst[i] = blstsrc[q];
		used[i] = usedsrc[q];
		q++;
	    }
	}
    }

    private static int statecomp(ClientBinding b1, boolean u1,
				 ClientBinding b2, boolean u2) {
	if(u1) {
	    if(!u2)
		return 1;
	    return 0;
	}
	if(u2)
	    return -1;

	return ClientBinding.priorityComp.compare(b1, b2);
    }
}
