/**
 * Redistribution and use of this software and associated
 * documentation ("Software"), with or without modification, are
 * permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright statements
 *    and notices.  Redistributions must also contain a copy of this
 *    document.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Intalio Inc.  For written permission, please
 *    contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Intalio Inc. Exolab is a registered trademark of
 *    Intalio Inc.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY INTALIO AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL INTALIO OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Copyright 2000 (C) Intalio Inc. All Rights Reserved.
 *
 * $Id: ORB.java,v 1.18 2001/07/31 13:45:33 wood Exp $
 *
 * Date         Author  Changes
 * 11/10/00	J.Daniel A small change in the connect operation to activate the object at the same time.
 */

/** Modified/extended by Diana Szentivanyi in June 2003
* contact diasz@ida.liu.se
*/

package org.openorb.CORBA;

import java.util.Map;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Iterator;

import org.openorb.util.RepoIDHelper;
import org.openorb.util.Trace;

import org.openorb.CORBA.MinorCodes;
import org.omg.CORBA.CompletionStatus;

import org.openorb.util.NamingUtils;

/**
 * This class provides all required operations and structures for the
 * org.openorb implementation.
 *
 * @author <a href="mailto:wood@intalio.com">Chris Wood &lt;wood@intalio.com&gt;</a>
 *         <a href="mailto:daniel@intalio.com">Jerome Daniel &lt;daniel@intalio.com&gt;</a>
 * @version $Revision: 1.18 $ $Date: 2001/07/31 13:45:33 $
 */
public class ORB extends org.openorb.CORBA.ORBSingleton
{
  // feature repository


    /* Modified by Diana on Oct. 12 2001*/

    private int retention_id=-1;

  /**
   * Synchronize for access to the below.
   */
  private java.lang.Object sync_state = new java.lang.Object();

  /**
   * This hash table contains all references to the OpenORB features
   */
  private Map _features = new HashMap();

  /**
   * Each ORB instance has its own initial references list added with
   * Portable Interceptors. Some other initial references are
   * available for all ORB instances.
   */
  private Map _initial_references = new HashMap();

  /**
   * The OpenORB loader is the entity which is responsible to load all
   * OpenORB parts. This entity also contains all OpenORB properties
   * specified from an XML profile, from the command line or from the
   * application.
   */
  private org.openorb.CORBA.kernel.ORBLoader _loader = null;

  /**
   * List of outstanding defered requests.
   */
  private LinkedList _deferredReq = new LinkedList();

  /**
   * This target is resolved very frequently for every operation,
   * so it is stored for quick access by the getPICurrent operation.
   */
  public org.openorb.PI.CurrentImpl _pi_current;

  /**
   * Orb instances must be created through ORB.init
   */
  public ORB() {
      setFeature("FTORB",new java.lang.Object());
  }

  // OpenORB Operations.

  /**
   * This operation is used to set an OpenORB feature reference.
   */
  public void setFeature( String feature, java.lang.Object reference )
  {
    synchronized(sync_state) {
      if(reference != null)
        _features.put( feature, reference );
      else
        _features.remove( feature );
    }
  }

  /**
   * This operation is used to get an OpenORB feature reference.
   */
  public java.lang.Object getFeature( String feature )
  {
    synchronized(sync_state) {
      return _features.get( feature );
    }
  }
    /* Modifications by Diana in Oct. 2001*/
    public synchronized int retention_id(){
	retention_id++;
	return retention_id;
    }

    public synchronized void retention_id(int value){
	retention_id=value;
    }

    public synchronized int current_retention_id(){
	return retention_id;
    }


    public void unblock_hanging(){
	org.openorb.PI.ServerManager _interceptorManager;
	_interceptorManager = (org.openorb.PI.ServerManager)getFeature("ServerInterceptorManager");
	if(_interceptorManager!=null)
	    _interceptorManager.send_exception(null,null);
    }

    /* Added by Diana on Jan. 25th 2002*/

    public void backup_set_to_primary(){
	org.openorb.PI.ServerManager _interceptorManager;
	_interceptorManager = (org.openorb.PI.ServerManager)getFeature("ServerInterceptorManager");
	if(_interceptorManager!=null)
	    _interceptorManager.send_other(null,null);
    }

    /****************************************************/
  /**
   * This operation returns the configurator used by this ORB instance
   */
  public org.openorb.CORBA.kernel.ORBLoader getLoader()
  {
    return _loader;
  }

  /**
   * Returns the PICurrent implementation. The PICurrent must be resolved
   * with every invocation both at the client and server end, so it's avalable
   * from this operation for optimization reasons.
   */
  public org.openorb.PI.CurrentImpl getPICurrent() {
    if(_pi_current == null)
      throw new org.omg.CORBA.BAD_INV_ORDER("Unable to make invocation, orb is not fully initialized",
        MinorCodes.BAD_INV_ORDER_ORB, CompletionStatus.COMPLETED_NO);
    return _pi_current;
  }

  /**
   * Register an initial reference.
   */
  public void addInitialReference( String name, org.omg.CORBA.Object target ) {
    if(name.equals("PICurrent"))
      _pi_current = (org.openorb.PI.CurrentImpl)target;

    synchronized(sync_state) {
      if(target != null)
        _initial_references.put(name, target);
      else
        _initial_references.remove(name);
    }
  }

  /**
   * Add service information
   */
  public void addServiceInformation( short service_type, org.omg.CORBA.ServiceInformation service_information ) {
    synchronized(sync_state) {
      Map service_info = (Map) getFeature("ServiceInfo");

      if(service_info == null) {
        if(service_information == null)
          return;

        service_info = new HashMap();
        setFeature("ServiceInfo", service_info );
      }

      if(service_information != null)
        service_info.put(new Short(service_type), service_information );
      else
        service_info.remove(new Short(service_type));
    }
  }

  // ORB Operations

  /**
   * Return an initial reference
   */
  public org.omg.CORBA.Object resolve_initial_references ( String object_name )
    throws org.omg.CORBA.ORBPackage.InvalidName
  {
    org.omg.CORBA.Object reference;
    synchronized(sync_state) {
      // look in initial referenced table
      reference = (org.omg.CORBA.Object)_initial_references.get(object_name);

      if(reference != null) {
        Trace.low("ORB::resolve_initial_references", object_name + " in reference table");
        return reference;
      }

      if(_pi_current == null) {
        Trace.low("ORB::resolve_initial_reference", "Unable during initialization to find the initial reference : " + object_name );
        throw new org.omg.CORBA.ORBPackage.InvalidName("Object not found : "+object_name);
      }
    }

    // try the InitRef
    String string_ref = _loader.getStringProperty("InitRef." + object_name , "");

    if(string_ref.length() > 0) {
      Trace.low("ORB::resolve_initial_references", "resolve with InitRef." + object_name);

      try {
        reference = string_to_object(string_ref);

        synchronized(sync_state) {
          _initial_references.put(object_name, reference);
        }

        Trace.low("ORB::resolve_initial_references", "resolve with InitRef." + object_name + " succeeded");
        return (org.omg.CORBA.Object)reference;
      }
      catch (org.omg.CORBA.SystemException ex) {}
    }

    String default_init_ref = _loader.getStringProperty("openorb.DefaultInitRef", "");
    if(default_init_ref.length() > 0) {
      try {
        // ensure we don't get a loop.
        //if(default_init_ref.substring(0, "corbaname:rir:"

        string_ref = default_init_ref + "/" + object_name;

        Trace.low("ORB::resolve_initial_references", "resolve with DefaultInitRef: " + string_ref);

        reference = string_to_object(string_ref);

        if(!reference._non_existent()) {
          Trace.low("ORB::resolve_initial_references", "resolve with DefaultInitRef succeeded");
          return reference;
        }
      }
      catch (org.omg.CORBA.SystemException ex) {}
    }

    // Is the reference available ?
    Trace.low("ORB::resolve_initial_reference", "Unable to find the initial reference : " + object_name );
    throw new org.omg.CORBA.ORBPackage.InvalidName("Object not found : "+object_name);
  }

  /**
   * Return as a string sequence all available initial services.
   */
  public String [] list_initial_services()
  {
    synchronized(sync_state) {
      return (String[])_initial_references.keySet().toArray(new String[_initial_references.size()]);
    }
  }

  /**
   * Run the server side.
   */
  public void run()
  {
    Trace.high("ORB", "running");

    org.openorb.net.ServerManager svrmgr = (org.openorb.net.ServerManager)getFeature("ServerCPCManager");
    if(svrmgr != null)
      svrmgr.startup(true, true);
  }

  /**
   * Stop the orb ( and the object adapter if required )
   */
  public void shutdown( boolean wait_for_completion )
  {
    // shutdown the server side first since it may throw an exception.
    org.openorb.net.ServerManager svrmgr = (org.openorb.net.ServerManager)getFeature("ServerCPCManager");
    if(svrmgr != null)
      svrmgr.shutdown(wait_for_completion);

    org.openorb.net.ClientManager cltmgr = (org.openorb.net.ClientManager)getFeature("ClientCPCManager");
    if(cltmgr != null)
      cltmgr.shutdown(wait_for_completion, false);

    Trace.high("ORB", "shutdown");
  }

  /**
   * A finalizer which will call shutdown.
   */
  protected void finalize() {
    shutdown(true);
  }

  /**
   * Check if some work are pending
   */
  public boolean work_pending()
  {
    org.openorb.net.ServerManager svrmgr = (org.openorb.net.ServerManager)getFeature("ServerCPCManager");
    if(svrmgr != null)
      return svrmgr.work_pending();
    return false;
  }

  /**
   * Perform all current work
   */
  public void perform_work()
  {
    org.openorb.net.ServerManager svrmgr = (org.openorb.net.ServerManager)getFeature("ServerCPCManager");
    if(svrmgr != null)
      svrmgr.serve_request(false);
  }

  /**
   * Create an Output stream
   */
  public org.omg.CORBA.portable.OutputStream create_output_stream()
  {
    return new org.openorb.io.ListOutputStream(this);
  }

  /**
   * Create a named value list
   */
  public org.omg.CORBA.NVList create_list( int count )
  {
    org.omg.CORBA.NVList nv = new org.openorb.CORBA.dii.NVList( this );

    for ( int i=0; i<count; i++ )
      nv.add(0);

    return nv;
  }

  /**
   * Create a named value
   */
  public org.omg.CORBA.NamedValue create_named_value ( String name,
                             org.omg.CORBA.Any value,
                             int flags )
  {
    return new org.openorb.CORBA.dii.NamedValue(name,value,flags);
  }

  /**
   * Create an exception list
   */
  public org.omg.CORBA.ExceptionList create_exception_list()
  {
    return new org.openorb.CORBA.dii.ExceptionList();
  }

  /**
   * Create a context list
   */
  public org.omg.CORBA.ContextList create_context_list()
  {
    return new org.openorb.CORBA.dii.ContextList();
  }

  /**
   * Return default context
   */
  public org.omg.CORBA.Context get_default_context()
  {
    return new org.openorb.CORBA.dii.Context("",null,this);
  }

  /**
   * Create an environment
   */
  public org.omg.CORBA.Environment create_environment()
  {
    return new org.openorb.CORBA.dii.Environment();
  }

  /**
   * Connect an object to the adapter
   */
  public void connect( org.omg.CORBA.Object obj )
  {
    if(obj == null)
      return;

    org.openorb.adapter.boa.BOA boa = (org.openorb.adapter.boa.BOA)getFeature("BOA");

    if(boa == null)
      throw new org.omg.CORBA.NO_IMPLEMENT();

    boa.connect((org.omg.CORBA.portable.ObjectImpl)obj, true);

    boa.obj_is_ready( (org.omg.CORBA.portable.ObjectImpl)obj );
  }

  /**
   * Disconnect an object from the adapter
   */
  public void disconnect( org.omg.CORBA.Object obj )
  {
    org.openorb.adapter.boa.BOA boa = (org.openorb.adapter.boa.BOA)getFeature("BOA");

    if(boa == null)
      throw new org.omg.CORBA.NO_IMPLEMENT();

    boa.disconnect((org.omg.CORBA.portable.ObjectImpl)obj);
  }

  /**
   * Convert an object reference to a string
   */
  public String object_to_string(org.omg.CORBA.Object obj)
  {

    byte [] encoded;

    try {
      org.omg.IOP.CodecFactory factory = (org.omg.IOP.CodecFactory)resolve_initial_references("CodecFactory");
      org.omg.IOP.Codec codec = factory.create_codec(new org.omg.IOP.Encoding(org.omg.IOP.ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));

      org.omg.CORBA.Any any = create_any();
      any.insert_Object(obj);
      encoded = codec.encode_value(any);

      java.io.ByteArrayOutputStream bos = new java.io.ByteArrayOutputStream(encoded.length*2);
      org.openorb.io.HexPrintStream hps = new org.openorb.io.HexPrintStream(bos);
      hps.write(encoded);

      return "IOR:" + bos.toString();
    }
    catch(org.omg.CORBA.ORBPackage.InvalidName ex) {}
    catch(org.omg.IOP.CodecFactoryPackage.UnknownEncoding ex) {}
    catch(org.omg.IOP.CodecPackage.InvalidTypeForEncoding ex) {}
    catch(java.io.IOException ex) {}
    // TODO: throw some kind of exception
    return null;
  }

  /**
   * Convert a string to an object reference
   */
  public org.omg.CORBA.Object string_to_object(String str)
  {
    int sch_idx = str.indexOf(':');
    if(sch_idx < 0)
      throw new org.omg.CORBA.BAD_PARAM( org.omg.CORBA.OMGVMCID.value | 7, org.omg.CORBA.CompletionStatus.COMPLETED_NO );

    String scheme = str.substring(0, sch_idx).toLowerCase();

    if ( scheme.equals("ior") )
    {
      int len = str.length();
      if((len % 2) != 0)
        throw new org.omg.CORBA.BAD_PARAM( org.omg.CORBA.OMGVMCID.value | 9, org.omg.CORBA.CompletionStatus.COMPLETED_NO );
      if(len < 28)
        throw new org.omg.CORBA.BAD_PARAM( org.omg.CORBA.OMGVMCID.value | 9, org.omg.CORBA.CompletionStatus.COMPLETED_NO );

      byte [] buf = new byte[(len-4)/2];
      int j = 0;
      char c;
      for(int i = 4; i < len; i+=2) {
        c = str.charAt(i);
        if(c >= '0' && c <= '9')
          buf[j] = (byte)((c-'0')<<4);
        else if(c >= 'a' && c <= 'f')
          buf[j] = (byte)((c-'a'+0xA)<<4);
        else if(c >= 'A' && c <= 'F')
          buf[j] = (byte)((c-'A'+0xA)<<4);
        else
          throw new org.omg.CORBA.BAD_PARAM( org.omg.CORBA.OMGVMCID.value | 9, org.omg.CORBA.CompletionStatus.COMPLETED_NO );
        c = str.charAt(i+1);
        if(c >= '0' && c <= '9')
          buf[j] += (byte)(c-'0');
        else if(c >= 'a' && c <= 'f')
          buf[j] += (byte)(c-'a'+0xA);
        else if(c >= 'A' && c <= 'F')
          buf[j] += (byte)(c-'A'+0xA);
        else
          throw new org.omg.CORBA.BAD_PARAM( org.omg.CORBA.OMGVMCID.value | 9, org.omg.CORBA.CompletionStatus.COMPLETED_NO );
        ++j;
      }

      try {
        org.omg.IOP.CodecFactory factory = (org.omg.IOP.CodecFactory)resolve_initial_references("CodecFactory");
        org.omg.IOP.Codec codec = factory.create_codec(new org.omg.IOP.Encoding(org.omg.IOP.ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));

        org.omg.CORBA.Any any = codec.decode_value(buf, get_primitive_tc(org.omg.CORBA.TCKind.tk_objref));
        return any.extract_Object();
      }
      catch(org.omg.CORBA.ORBPackage.InvalidName ex) {}
      catch(org.omg.IOP.CodecFactoryPackage.UnknownEncoding ex) {}
      catch(org.omg.IOP.CodecPackage.FormatMismatch ex) {}
      catch(org.omg.IOP.CodecPackage.TypeMismatch ex) {}
      throw new org.omg.CORBA.BAD_PARAM( org.omg.CORBA.OMGVMCID.value | 9, org.omg.CORBA.CompletionStatus.COMPLETED_NO );
    }
    else if ( scheme.equals("corbaloc") )
    {
      return scan_url_loc( str.substring(9) );
    }
    else if ( scheme.equals("corbaname") )
    {
      String loc = str.substring(10);

      return scan_url_name( loc );
    }
    else if ( scheme.equals("file") || scheme.equals("ftp") || scheme.equals("http") )
    {
      try
      {
        java.net.URL url = new java.net.URL( str );

        java.io.InputStream is = url.openStream();

        java.io.InputStreamReader rd = new java.io.InputStreamReader( is );

        java.io.BufferedReader inpt = new java.io.BufferedReader(rd);

        String string_ref = inpt.readLine();

        inpt.close();

        return string_to_object( string_ref );
      }
      catch ( java.net.MalformedURLException ex )
      {
        throw new org.omg.CORBA.BAD_PARAM( org.omg.CORBA.OMGVMCID.value | 9, org.omg.CORBA.CompletionStatus.COMPLETED_NO );
      }
      catch ( java.io.IOException ex )
      {
        throw new org.omg.CORBA.BAD_PARAM( org.omg.CORBA.OMGVMCID.value | 10, org.omg.CORBA.CompletionStatus.COMPLETED_NO );
      }
    }
    else
      throw new org.omg.CORBA.BAD_PARAM( org.omg.CORBA.OMGVMCID.value | 7, org.omg.CORBA.CompletionStatus.COMPLETED_NO );
  }

  /**
   * Send multiple oneway requests
   */
  public void send_multiple_requests_oneway ( org.omg.CORBA.Request [] req )
  {
    for ( int i=0; i<req.length; i++ )
      req[i].send_oneway();
  }

  /**
   * Send multiple deferred requests
   */
  public void send_multiple_requests_deferred ( org.omg.CORBA.Request [] req )
  {
    int count = 0;
    for ( int i=0; i<req.length; i++ )
      req[i].send_deferred();

    synchronized(_deferredReq) {
      for ( int i=0; i<req.length; i++ )
        _deferredReq.add(req[i]);
    }
  }

  /**
   * Poll next response
   */
  public boolean poll_next_response()
  {
    synchronized(_deferredReq) {
      if(_deferredReq.isEmpty())
        throw new org.omg.CORBA.BAD_INV_ORDER(org.omg.CORBA.OMGVMCID.value | 11, org.omg.CORBA.CompletionStatus.COMPLETED_NO);
      Iterator itt = _deferredReq.iterator();
      while(itt.hasNext()) {
        org.omg.CORBA.Request req = (org.omg.CORBA.Request)itt.next();
        if(req.poll_response())
          return true;
      }
      return false;
    }
  }

  /**
   * Get the next response. If poll_next_response would return true then
   * the first completed request will be returned, otherwise the oldest and
   * first request will be waited for and returned.
   */
  public org.omg.CORBA.Request get_next_response()
  {
    org.openorb.CORBA.dii.Request req = null;
    synchronized(_deferredReq) {
      if(_deferredReq.isEmpty())
        throw new org.omg.CORBA.BAD_INV_ORDER(org.omg.CORBA.OMGVMCID.value | 11, org.omg.CORBA.CompletionStatus.COMPLETED_NO);
      Iterator itt = _deferredReq.iterator();
      boolean found = false;
      while(itt.hasNext()) {
        req = (org.openorb.CORBA.dii.Request)itt.next();
        if(req.poll_response()) {
          itt.remove();
          found = true;
          break;
        }
      }
      if(!found)
        req = (org.openorb.CORBA.dii.Request)_deferredReq.removeFirst();
    }
    req.get_response();
    return req;
  }

  /**
   * Set the ORB parameters.
   */
  protected void set_parameters(String[] args, java.util.Properties properties )
  {
    String loader_name = scan_for_loader_name( args, properties );

    if ( loader_name == null )
      loader_name = "org.openorb.CORBA.kernel.OpenORBLoader";

    try
    {
      _loader = ( org.openorb.CORBA.kernel.ORBLoader ) java.lang.Class.forName( loader_name ).newInstance();
    }
    catch ( java.lang.Throwable ex )
    {
      throw new org.omg.CORBA.INITIALIZE("Unable to initialize orb loader");
    }
    /*    if(properties!=null){
	  String cl_id=properties.getProperty("client.id");
	  if(cl_id==null)

properties.put("client.id",this+"*"+Runtime.getRuntime().toString());
	  }
	  else{
	  properties=new java.util.Properties();
	  properties.put("client.id",this+"*"+Runtime.getRuntime().toString());
	  }*/
    _loader.init( args, properties, this );

    Trace.high("ORB", "created");
  }

  /**
   * This operation is used to get a loader name from the command line or properties arguments.
   */
  private String scan_for_loader_name( String [] args, java.util.Properties properties )
  {
    String name = null;

    if ( properties != null )
      name = ( String ) properties.get("ORBLoader");

    if(args != null)
      for ( int i=0; i<args.length; i++ )
        if ( args[i].startsWith("-ORBLoader") )
          name = args[i].substring( 11 );

    return name;
  }

  /**
   * Set the ORB parameters for applet.
   */
  protected void set_parameters(java.applet.Applet app, java.util.Properties properties )
  {
    String loader_name = scan_for_loader_name( app, properties );

    if ( loader_name == null )
      loader_name = "org.openorb.CORBA.kernel.OpenORBLoader";

    try
    {
      _loader = ( org.openorb.CORBA.kernel.ORBLoader ) java.lang.Class.forName( loader_name ).newInstance();
    }
    catch ( java.lang.Throwable ex )
    {
      throw new org.omg.CORBA.INITIALIZE( "Unable to initialize orb loader" );
    }

    _loader.init( null, properties, this );

    Trace.high("ORB", "created");
  }

  /**
   * This operation is used to get a loader name from the command line or properties arguments.
   */
  private String scan_for_loader_name( java.applet.Applet app, java.util.Properties properties )
  {
    String name = null;

    if ( properties != null )
      name = ( String ) properties.get("ORBLoader");

    if ( app != null && app.getParameter("ORBLoader") != null )
      name = app.getParameter("ORBLoader");

    return name;
  }

  /**
   * This function is used to register a value factory
   */
  public org.omg.CORBA.portable.ValueFactory register_value_factory ( String id,
                                    org.omg.CORBA.portable.ValueFactory factory )
  {
    synchronized(sync_state) {
      // Get the value factory list from the repository
      Map factories = ( Map ) getFeature("ValueFactory");

      if ( factories == null )
      {
        factories = new HashMap();

        setFeature("ValueFactory", factories);
      }

      factories.put( id, factory );
    }

    return factory;
  }

  /**
   * This function is used to unregister a value factory
   */
  public void unregister_value_factory( String id )
  {
    synchronized(sync_state) {
      Map factories = ( Map ) getFeature("ValueFactory");

      if ( factories != null )
        factories.remove( id );
    }
  }

  /**
   * This function is used to find a value factory
   */
  public org.omg.CORBA.portable.ValueFactory lookup_value_factory( String id )
  {
    org.omg.CORBA.portable.ValueFactory ret;
    synchronized(sync_state) {
      Map factories = ( Map ) getFeature("ValueFactory");

      if ( factories != null) {
        ret = (org.omg.CORBA.portable.ValueFactory)factories.get( id );
        if(ret != null)
          return ret;
      }
    }

    // load the default factory if it can be found
    String classname;
    try {
      classname = RepoIDHelper.idToClass(id, RepoIDHelper.TYPE_DEFAULT_FACTORY);
    }
    catch ( IllegalArgumentException ex ) {
      return null;
    }

    try {
      Class clz = Class.forName(classname);
      return (org.omg.CORBA.portable.ValueFactory)clz.newInstance();
    }
    catch(Exception ex) {}
    return null;
  }

  /**
   * This function is used to set a delegate
   */
  public void set_delegate( java.lang.Object wrapper )
  {
    if ( wrapper instanceof org.omg.PortableServer.Servant )
    {

      org.omg.PortableServer.portable.Delegate _delegate =
                                ( org.omg.PortableServer.portable.Delegate ) getFeature("POADelegate");

      if ( _delegate == null )
        throw new org.omg.CORBA.INITIALIZE( "Missing POA delegate" );

      ((org.omg.PortableServer.Servant)wrapper)._set_delegate( _delegate );
    }
    else
      throw new org.omg.CORBA.BAD_PARAM( 0, org.omg.CORBA.CompletionStatus.COMPLETED_NO );
  }

  /**
   * Return the service informaton
   */
  public boolean get_service_information( short service_type, org.omg.CORBA.ServiceInformationHolder service_information )
  {
    synchronized(sync_state) {
      Map service_info = (Map) getFeature("ServiceInfo");

      if ( service_info == null )
        return false;

      service_information.value = ( org.omg.CORBA.ServiceInformation ) service_info.get( new Short( service_type ) );

      return ( service_information.value != null );
    }
  }

  /**
   * This operations creates a CORBA policy.
   */
  public org.omg.CORBA.Policy create_policy( int policy_type, org.omg.CORBA.Any val )
    throws org.omg.CORBA.PolicyError
  {
    org.omg.PortableInterceptor.PolicyFactory policy_factory = ( org.omg.PortableInterceptor.PolicyFactory ) getFeature("PolicyFactory");
    if ( policy_factory != null )
      return policy_factory.create_policy( policy_type, val );
    throw new org.omg.CORBA.PolicyError( org.omg.CORBA.BAD_POLICY_VALUE.value );
  }

  // ----------------------------------------------------------------------------------------
  //
  // OpenORB implementation specific operations
  //
  // ----------------------------------------------------------------------------------------


  /**
   * This function scans a CORBA URL iioploc to extract reference information.
   */
  protected org.omg.CORBA.Object scan_url_loc( String loc )
  {
    if(loc.startsWith("rir:/"))
      try {
        return resolve_initial_references(loc.substring(5));
      }
      catch(org.omg.CORBA.ORBPackage.InvalidName ex) {
        throw new org.omg.CORBA.BAD_PARAM( org.omg.CORBA.OMGVMCID.value | 10, org.omg.CORBA.CompletionStatus.COMPLETED_NO );
      }

    // get key string
    int keyidx = loc.indexOf('/');
    if(keyidx < 0)
      throw new org.omg.CORBA.BAD_PARAM( org.omg.CORBA.OMGVMCID.value | 10, org.omg.CORBA.CompletionStatus.COMPLETED_NO );

    // get the addresses
    return scan_url_body(loc.substring(0, keyidx), getKeyFromString( loc.substring(keyidx+1) ));
  }

  protected org.omg.CORBA.Object scan_url_body(String addr, byte [] key) {
    java.util.Vector list = new java.util.Vector();
    int index;

    // Extract all addresses
    {
      int old = 0;
      String s = null;
      while ( true ) {
        index = addr.indexOf(",", old);

        if ( index < 0 ) {
          s = addr.substring( old );

          if ( s.startsWith("iiop:") )
            list.addElement( s.substring( 5 ) );
          else if( s.startsWith(":" ))
            list.addElement( s.substring( 1 ) );
          else
            throw new org.omg.CORBA.BAD_PARAM( org.omg.CORBA.OMGVMCID.value | 8, org.omg.CORBA.CompletionStatus.COMPLETED_NO );

          break;
        }
        else
        {
          s = addr.substring( old, index );

          if ( s.startsWith("iiop:") )
            list.addElement( s.substring( 5 ) );
          else if( s.startsWith(":" ))
            list.addElement( s.substring( 1 ) );
          else
            throw new org.omg.CORBA.BAD_PARAM( org.omg.CORBA.OMGVMCID.value | 8, org.omg.CORBA.CompletionStatus.COMPLETED_NO );

          old = index + 1;
        }
      }
    }

    if(list.isEmpty()) {
      // TODO: throw some kind of exception
      return null;
    }

    org.omg.IOP.Codec codec = null;
    try {
      org.omg.IOP.CodecFactory factory = (org.omg.IOP.CodecFactory)resolve_initial_references("CodecFactory");
      codec = factory.create_codec(new org.omg.IOP.Encoding(org.omg.IOP.ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));
    }
    catch(org.omg.CORBA.ORBPackage.InvalidName ex) {
    }
    catch(org.omg.IOP.CodecFactoryPackage.UnknownEncoding ex) {
    }

    if(codec == null) {
      // TODO: throw some kind of exception.
      return null;
    }

    org.omg.CORBA.Any any = create_any();

    String version;
    int port;
    org.omg.IIOP.Version iiop_version = new org.omg.IIOP.Version();

    org.omg.IIOP.ProfileBody_1_1 body_2 = null;
    byte [][] bodies = new byte[list.size()][];
    int lastbody = 0;
    int alternates = 0;

    for ( int i=0; i<list.size(); i++ )
    {
      addr = ( String ) list.elementAt(i);

      index = addr.indexOf("@");

      // Is there a version data ?
      if ( index != -1 )
      {
        version = addr.substring(0, index);
        addr = addr.substring( index + 1 );

        index = version.indexOf(".");

        if ( index == -1 )
          throw new org.omg.CORBA.BAD_PARAM( org.omg.CORBA.OMGVMCID.value | 9, org.omg.CORBA.CompletionStatus.COMPLETED_NO );

        try {
          iiop_version.major = Byte.parseByte( version.substring(0, index) );
          iiop_version.minor = Byte.parseByte( version.substring( index + 1 ) );
        }
        catch(NumberFormatException ex) {
          throw new org.omg.CORBA.BAD_PARAM( org.omg.CORBA.OMGVMCID.value | 9, org.omg.CORBA.CompletionStatus.COMPLETED_NO );
        }
      }
      else
      {
        iiop_version.major = ( byte ) 1;
        iiop_version.minor = ( byte ) 0;
      }

      index = addr.indexOf(":");

      // Is there a port number ?
      if ( index != -1 )
      {
        try {
          // ================ QUICK PATCH FOR MAC OS X : to improve ======================
          port = Integer.parseInt(addr.substring(index+1));
          if(port > 0xFFFF)
            throw new org.omg.CORBA.BAD_PARAM( org.omg.CORBA.OMGVMCID.value | 9, org.omg.CORBA.CompletionStatus.COMPLETED_MAYBE );
        }
        catch(NumberFormatException ex) {
          throw new org.omg.CORBA.BAD_PARAM( org.omg.CORBA.OMGVMCID.value | 9, org.omg.CORBA.CompletionStatus.COMPLETED_NO );
        }
        addr = addr.substring( 0, index );
      }
      else
        port = 683;

      org.openorb.util.Trace.low("ORB::scan_url_loc", "Search object => " + new String(key) + " / " + port + " / " + addr );

      switch(iiop_version.minor) {
      case 0:
        org.omg.IIOP.ProfileBody_1_0 body_0 = new org.omg.IIOP.ProfileBody_1_0();
        body_0.host = addr;
        body_0.port = (short)port;
        body_0.iiop_version = iiop_version;
        body_0.object_key = key;

        org.omg.IIOP.ProfileBody_1_0Helper.insert(any, body_0);
        try {
          bodies[lastbody++] = codec.encode_value(any);
        }
        catch(org.omg.IOP.CodecPackage.InvalidTypeForEncoding ex) {
          // TODO: throw some kind of exception.
        }
        break;
      case 1:
        org.omg.IIOP.ProfileBody_1_1 body_1 = new org.omg.IIOP.ProfileBody_1_1();
        body_1.host = addr;
        body_1.port = (short)port;
        body_1.iiop_version = iiop_version;
        body_1.object_key = key;
        body_1.components = new org.omg.IOP.TaggedComponent[0];

        org.omg.IIOP.ProfileBody_1_1Helper.insert(any, body_1);
        try {
          bodies[lastbody++] = codec.encode_value(any);
        }
        catch(org.omg.IOP.CodecPackage.InvalidTypeForEncoding ex) {
          // TODO: throw some kind of exception.
        }
        break;
      case 2:
        if(body_2 == null) {
          body_2 = new org.omg.IIOP.ProfileBody_1_1();
          body_2.host = addr;
          body_2.port = (short)port;
          body_2.iiop_version = iiop_version;
          body_2.object_key = key;
        } else {
          org.omg.IIOP.ListenPoint lp = new org.omg.IIOP.ListenPoint();
          lp.host = addr;
          lp.port = (short)port;

          org.omg.IIOP.ListenPointHelper.insert(any, lp);
          try {
            bodies[bodies.length-(alternates++)-1] = codec.encode_value(any);
          }
          catch(org.omg.IOP.CodecPackage.InvalidTypeForEncoding ex) {
            // TODO: throw some kind of exception.
          }
        }
        break;
      }
    }

    if(body_2 != null) {
      body_2.components = new org.omg.IOP.TaggedComponent[alternates];
      for(int i = 0; i < alternates; ++i)
        body_2.components[i] = new org.omg.IOP.TaggedComponent(org.omg.IOP.TAG_ALTERNATE_IIOP_ADDRESS.value, bodies[bodies.length-i]);
      org.omg.IIOP.ProfileBody_1_1Helper.insert(any, body_2);
      try {
        bodies[lastbody++] = codec.encode_value(any);
      }
      catch(org.omg.IOP.CodecPackage.InvalidTypeForEncoding ex) {
        // TODO: throw some kind of exception.
      }
    }

    org.omg.IOP.IOR ior = new org.omg.IOP.IOR("", new org.omg.IOP.TaggedProfile[lastbody]);
    for(int i = 0; i < lastbody; ++i)
      ior.profiles[i] = new org.omg.IOP.TaggedProfile(org.omg.IOP.TAG_INTERNET_IOP.value, bodies[i]);

    return new ObjectStub(this, ior);
  }

  /**
   * This function scans a CORBA URL iiopname to extract reference information.
   */
  protected org.omg.CORBA.Object scan_url_name( String loc )
  {
    // first resolve the name service.
    int hidx = loc.indexOf("#");
    if(hidx < 0)
      throw new org.omg.CORBA.BAD_PARAM( org.omg.CORBA.OMGVMCID.value | 9, org.omg.CORBA.CompletionStatus.COMPLETED_NO );
    int sidx = loc.indexOf("/");
    String url = loc.substring(0, hidx);
    if(sidx < 0 || sidx > hidx)
      url += "/NameService";

    String name = loc.substring(hidx+1);

    org.omg.CORBA.Object namingobj = scan_url_loc(url);

    // try namingExt first. We do this to attempt using the local
    // invocation path first...
    org.omg.CosNaming.NamingContextExt namingExt = null;
    try {
      namingExt = org.omg.CosNaming.NamingContextExtHelper.narrow(namingobj);
    }
    catch(org.omg.CORBA.BAD_PARAM ex) {
      // just use old naming context.
      org.omg.CosNaming.NamingContext naming;
      try {
        naming = org.omg.CosNaming.NamingContextHelper.narrow(namingobj);
      }
      catch(org.omg.CORBA.BAD_PARAM ex1) {
        throw new org.omg.CORBA.BAD_PARAM( org.omg.CORBA.OMGVMCID.value | 10, org.omg.CORBA.CompletionStatus.COMPLETED_NO );
      }

      if(name.length() == 0)
        return naming;

      try {
        return naming.resolve(NamingUtils.to_name(loc.substring(hidx+1)));
      }
      catch(org.omg.CORBA.UserException ex1) {
        throw new org.omg.CORBA.BAD_PARAM( org.omg.CORBA.OMGVMCID.value | 10, org.omg.CORBA.CompletionStatus.COMPLETED_NO );
      }
    }

    if(name.length() == 0)
      return namingExt;

    try {
      return namingExt.resolve_str(loc.substring(hidx+1));
    }
    catch(org.omg.CORBA.UserException ex1) {
      throw new org.omg.CORBA.BAD_PARAM( org.omg.CORBA.OMGVMCID.value | 10, org.omg.CORBA.CompletionStatus.COMPLETED_NO );
    }
  }

  /**
   * Get an object key from a stringified object key
   */
  protected byte [] getKeyFromString( String str )
  {
    String val = "";
    java.util.Vector list = new java.util.Vector();

    for ( int i=0; i<str.length(); i++ )
    {
      if ( str.charAt(i) == '%' )
      {
        // BUG Corrected by Martin DEVERA
        val =""+ str.charAt(i+1) + str.charAt(i+2);
        list.addElement( new Byte((byte)java.lang.Short.parseShort(val,16)) );

        i += 2;
      }
      else
        list.addElement( new java.lang.Byte( (byte)str.charAt(i) ) );
    }

    byte [] key = new byte[ list.size() ];

    for ( int i=0; i<key.length; i++ )
      key[i] = ( ( java.lang.Byte )list.elementAt(i) ).byteValue();

    return key;
  }
}
