/**
 * Redistribution and use of this software and associated
 * documentation ("Software"), with or without modification, are
 * permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright statements
 *    and notices.  Redistributions must also contain a copy of this
 *    document.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Intalio Inc.  For written permission, please
 *    contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Intalio Inc. Exolab is a registered trademark of
 *    Intalio Inc.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY INTALIO AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL INTALIO OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Copyright 2000 (C) Intalio Inc. All Rights Reserved.
 *
 * $Id: SystemExceptionHelper.java,v 1.2 2001/07/31 13:45:33 wood Exp $
 *
 * Date         Author  Changes
 */

/** Modified/extended by Diana Szentivanyi in June 2003
* contact diasz@ida.liu.se
*/

package org.openorb.CORBA;

import java.lang.reflect.Constructor;
import org.omg.CORBA.Any;
import org.omg.CORBA.TypeCode;
import org.omg.CORBA.SystemException;
import org.omg.CORBA.CompletionStatus;

import org.omg.CORBA.portable.InputStream;
import org.omg.CORBA.portable.OutputStream;

import java.util.Map;
import java.util.HashMap;
import org.openorb.CORBA.MinorCodes;
import org.omg.CORBA.CompletionStatus;

/**
 * This class should be used as the helper for all system exceptions.
 *
 * @author <a href="mailto:wood@intalio.com">Chris Wood &lt;wood@intalio.com&gt;</a>
 * @version $Revision: 1.2 $ $Date: 2001/07/31 13:45:33 $
 */
/*abstract*/ public class SystemExceptionHelper {
  final static private String [] sys_exceptions = {
      "UNKNOWN",
      "BAD_PARAM",
      "NO_MEMORY",
      "IMP_LIMIT",
      "COMM_FAILURE",
      "INV_OBJREF",
      "NO_PERMISSION",
      "INTERNAL",
      "MARSHAL",
      "INITIALIZE",
      "NO_IMPLEMENT",
      "BAD_TYPECODE",
      "BAD_OPERATION",
      "NO_RESOURCES",
      "NO_RESPONSE",
      "PERSIST_STORE",
      "BAD_INV_ORDER",
      "TRANSIENT",
      "FREE_MEM",
      "INV_IDENT",
      "INV_FLAG",
      "INTF_REPOS",
      "BAD_CONTEXT",
      "OBJ_ADAPTER",
      "DATA_CONVERSION",
      "OBJECT_NOT_EXIST",
      "TRANSACTION_REQUIRED",
      "TRANSACTION_ROLLEDBACK",
      "INVALID_TRANSACTION",
      "INV_POLICY",
      "CODESET_INCOMPATIBLE",
      "REBIND",
      "TIMEOUT",
      "TRANSACTION_UNAVAILABLE",
      "TRANSACTION_MODE",
      "BAD_QOS",
          /* Added by Diana on Oct. 27 2001*/
      "REQUEST_EXPIRED"
  };

  final private static Map finder = new HashMap(sys_exceptions.length*2);
  final private static MapEntry unknown;

  private static class MapEntry {
    MapEntry(String id, TypeCode type, Class clz, Class hclz, Constructor strCons) {
      this.id = id;
      this.type = type;
      this.clz = clz;
      this.hclz = hclz;
      this.strCons = strCons;
    }

    final String id;
    final TypeCode type;
    final Class clz;
    final Class hclz;
    final Constructor strCons;
  }

  static {
    try {
      unknown = new MapEntry(org.omg.CORBA.UNKNOWNHelper.id(), org.omg.CORBA.UNKNOWNHelper.type(),
                             org.omg.CORBA.UNKNOWN.class, org.omg.CORBA.UNKNOWNHelper.class,
                             org.omg.CORBA.UNKNOWN.class.getConstructor(new Class[] { String.class }));
      finder.put(unknown.id, unknown);
      finder.put(unknown.clz, unknown);
    }
    catch(NoSuchMethodException ex) {
      throw new InternalError(ex.toString());
    }

    MapEntry entry;
    for(int i = 1; i < sys_exceptions.length; ++i) {
      try {
        Class clz;
        Class hclz;
        try {
          clz = Class.forName("org.omg.CORBA."+sys_exceptions[i]);
          hclz = Class.forName("org.omg.CORBA."+sys_exceptions[i]+"Helper");
        }
        catch(ClassNotFoundException ex) {
          System.err.println("Warning: missing system exception class "+sys_exceptions[i]);
          continue;
        }

        String id = (String)hclz.getMethod("id", new Class[0]).invoke(null, new Object[0]);
        TypeCode type = (TypeCode)hclz.getMethod("type", new Class[0]).invoke(null, new Object[0]);
        Constructor strCons = clz.getConstructor(new Class[] { String.class });

        entry = new MapEntry(id, type, clz, hclz, strCons);
        finder.put(entry.id, entry);
        finder.put(entry.clz, entry);
      }
      catch(RuntimeException ex) {
        throw ex;
      }
      catch(Exception ex) {}
    }
  }

	/**
	 * Get a typecode from a repository ID.
	 * @param repo_id Repository ID of the system exception.
	 * @returns the required typecode, or typecode of UNKNOWN for an unknown
	 *          repository ID.
	 */
	public static TypeCode type(String repo_id)
  {
    MapEntry entry = (MapEntry)finder.get(repo_id);
    if(entry == null)
      entry = unknown;
    return entry.type;
  }

	/**
	 * Get a typecode from an exception.
	 * @param repo_id The exception to get the typecode of.
	 * @returns the required typecode, or typecode of UNKNOWN for an unknown
	 *          exception type.
	 */
  public static TypeCode type(SystemException ex)
  {
    MapEntry entry = (MapEntry)finder.get(ex.getClass());
    if(entry == null)
      entry = unknown;
    return entry.type;
  }

	/**
	 * Get the repository ID from an exception.
	 * @param ex the system exception.
	 * @returns the repository ID of the system exception, or UNKNOWN for an
	 *      unknown exception.
	 */
  public static String id(SystemException ex)
  {
    MapEntry entry = (MapEntry)finder.get(ex.getClass());
    if(entry == null)
      entry = unknown;
    return entry.id;
  }

	/**
	 * Insert a system exception into an any. If the exception type is unknown
	 * an unknown exception will be inserted into the any. UnknownExceptions can
	 * be successfully inserted/extracted from any types.
	 *
	 * @param any the any to get inserted into.
	 * @param ex the system exception to insert into the any.
	 */
  public static void insert(Any any, SystemException ex)
	{


	    MapEntry entry = (MapEntry)finder.get(ex.getClass());
	    if(entry == null) {
		if(ex instanceof org.omg.CORBA.portable.UnknownException &&
		   any instanceof org.openorb.CORBA.Any) {
		    ((org.openorb.CORBA.Any)any).unknown_ex = ex;
      }
      ex = new org.omg.CORBA.UNKNOWN(ex.minor, ex.completed);
      entry = unknown;
    }
    Class [] arg_clz = new Class[2];
    Object [] args = new Object[2];
    arg_clz[0] = Any.class;
    args[0] = any;
    arg_clz[1] = ex.getClass();
    args[1] = ex;
    try {
      entry.hclz.getMethod("insert", arg_clz).invoke(null, args);
      return;
    }
    catch(java.lang.reflect.InvocationTargetException iex) {
      Throwable tex = iex.getTargetException();
      if(tex instanceof java.lang.RuntimeException)
        throw (java.lang.RuntimeException)tex;
      if(tex instanceof java.lang.Error)
        throw (java.lang.Error)tex;
    }
    catch(RuntimeException rex) {
      throw rex;
    }
    catch(Exception rex) {}
	}

	/**
	 * Extract a system exception from an any. UnknownExceptions can
	 * be successfully inserted/extracted from any types.
	 *
	 * @param any the any to extract from.
	 * @throws org.omg.CORBA.BAD_OPERATION the any does not contain a system exception.
	 */
	public static SystemException extract(Any any)
  {
    MapEntry entry = null;
    try {
      entry = (MapEntry)finder.get(any.type().id());
    }
    catch(org.omg.CORBA.TypeCodePackage.BadKind ex) {}
    if(entry == null)
			throw new org.omg.CORBA.BAD_OPERATION("Wrong type for extraction",
				MinorCodes.BAD_OPERATION_ANY_TYPE, CompletionStatus.COMPLETED_MAYBE);

    Class [] arg_clz = new Class[1];
    Object [] args = new Object[1];
    arg_clz[0] = Any.class;
    args[0] = any;
    SystemException ret;
    try {
      ret = (SystemException)entry.hclz.getMethod("extract", arg_clz).invoke(null, args);

      if(ret instanceof org.omg.CORBA.UNKNOWN && any instanceof org.openorb.CORBA.Any) {
        SystemException tmp = new org.omg.CORBA.portable.UnknownException(((org.openorb.CORBA.Any)any).unknown_ex);
        tmp.minor = ret.minor;
        tmp.completed = ret.completed;
        ret = tmp;
      }

      return ret;
    }
    catch(java.lang.reflect.InvocationTargetException ex) {
      Throwable tex = ex.getTargetException();
      if(tex instanceof java.lang.RuntimeException)
        throw (java.lang.RuntimeException)tex;
      if(tex instanceof java.lang.Error)
        throw (java.lang.Error)tex;
    }
    catch(RuntimeException ex) {
      throw ex;
    }
    catch(Exception ex) {
    }
		throw new org.omg.CORBA.BAD_OPERATION("Wrong type for extraction",
			MinorCodes.BAD_OPERATION_ANY_TYPE, CompletionStatus.COMPLETED_MAYBE);
  }

	/**
	 * Read a system exception from an input stream.
	 */
  public static SystemException read(InputStream is)
	{
    return create(is.read_string(), null, is.read_ulong(), CompletionStatus.from_int(is.read_ulong()));
	}

	/**
	 * Read a system exception from an input stream.
   * @param reasonPrefix prefix to add to the reason string.
	 */
  public static SystemException read(String reasonPrefix, InputStream is)
	{
    return create(is.read_string(), reasonPrefix, is.read_ulong(), CompletionStatus.from_int(is.read_ulong()));
	}

	/**
	 * Create a system exception from it's repository ID, minor value and
	 * completion status.
	 */
  public static SystemException create(String repo_id, int minor, CompletionStatus completed)
  {
    return create(repo_id, null, minor, completed);
  }

	/**
	 * Create a system exception from it's repository ID, minor value and
	 * completion status.
   *
   * @param reasonPrefix prefix to add to the reason string.
	 */
  public static SystemException create(String repo_id, String reason, int minor, CompletionStatus completed)
  {
    MapEntry entry = (MapEntry)finder.get(repo_id);
    if(entry == null)
      entry = unknown;

    if(reason == null)
      reason = getReasonString(entry, minor);
    else
      reason = reason + ": " + getReasonString(entry, minor);

    try {
      SystemException ret = (SystemException)entry.strCons.newInstance(new Object[] {reason});
      ret.minor = minor;
      ret.completed = completed;
      return ret;
    }
    catch(RuntimeException ex) {
      throw ex;
    }
    catch(Throwable ex) {
      return null;
    }
  }

	/**
	 * Write a system excpetion to an output stream.
	 */
	public static void write(OutputStream out, SystemException val)
  {
		out.write_string(id(val));
		out.write_ulong(val.minor);
		out.write_ulong(val.completed.value());
  }

  private static String getReasonString(MapEntry entry, int minor) {
    switch(minor & 0xFFFFF000) {
    case org.omg.CORBA.OMGVMCID.value:
      return "OMG Standard Exception #" + (minor &0xFFF);
    case org.openorb.policy.OPENORB_VPVID.value:
      return "OpenORB exception #" + (minor &0xFFF);
    case 0:
      return "Unregistered vendor exception #" + (minor &0xFFF);
    default:
      return "Unknown vendor exception #" + (minor &0xFFF);
    }
  }
}
