/**
 * Redistribution and use of this software and associated
 * documentation ("Software"), with or without modification, are
 * permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright statements
 *    and notices.  Redistributions must also contain a copy of this
 *    document.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Intalio Inc.  For written permission, please
 *    contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Intalio Inc. Exolab is a registered trademark of
 *    Intalio Inc.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY INTALIO AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL INTALIO OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Copyright 2000 (C) Intalio Inc. All Rights Reserved.
 *
 * $Id: Request.java,v 1.4 2001/04/03 05:36:23 wood Exp $
 *
 * Date         Author  Changes
 */

/** Modified/extended by Diana Szentivanyi in June 2003
* contact diasz@ida.liu.se
*/

package org.openorb.CORBA.dii ;

import org.omg.GIOP.* ;
import org.omg.CORBA.portable.* ;
import org.openorb.CORBA.MinorCodes;
import org.omg.CORBA.CompletionStatus;

/**
 * This class provides a way to use dynamic request.
 *
 * @author <a href="mailto:jdaniel@intalio.com">Jerome Daniel &lt;daniel@intalio.com&gt;</a>
 * @version $Revision: 1.4 $ $Date: 2001/04/03 05:36:23 $
 */
public class Request extends org.omg.CORBA.Request
{
	// ----------------------------------------------------------------------------------------
	//
	// Internal states
	//
	// ----------------------------------------------------------------------------------------

	/**
	 * Reference to the target object reference
	 */
	private org.omg.CORBA.Object _target ;

    //private boolean _multiple_target=false;

  /**
   * Reference to the target delegate
   */
  private org.openorb.CORBA.Delegate _delegate;

	/**
	 * Operation to invoke
	 */
	private String _operation ;

	/**
	 * Operation arguments
	 */
	private org.omg.CORBA.NVList _arguments ;

	/**
	 * Result
	 */
	private org.omg.CORBA.NamedValue _result ;

	/**
	 * Environment variable for exception
	 */
	private org.omg.CORBA.Environment _env ;

	/**
	 * An exception list to describe what kind of user exception could be raised
	 */
	private org.omg.CORBA.ExceptionList _exceptions ;

	/**
	 * Context list
	 */
	private org.omg.CORBA.ContextList _contexts ;

	/**
	 * Context object
	 */
	private org.omg.CORBA.Context _context;

	/**
	 * Reference to the ORB
	 */
	private org.openorb.CORBA.ORB _orb;

	/**
	 * True if send_deferred was used.
	 */
	private boolean _deferred = false;

	/**
	 * True if response was available during send_deferred
	 */
    private boolean _response = false;


    /* Fields added by Diana*/
    private int _retention_id=-1;

    private String _client_id="";

    private long _expiration_time=0;

    private org.omg.IOP.ServiceContext _one_service_context=null;
  /**
   * Becomes true after first poll_response or get_response
   */
  private boolean _firstPoll = false;

	// ----------------------------------------------------------------------------------------
	//
	// Constructors
	//
	// ----------------------------------------------------------------------------------------

	/**
	 * Constructor
	 */
	public Request(	org.omg.CORBA.Object target, String operation, org.omg.CORBA.ORB orb )
	{
		_orb = ( org.openorb.CORBA.ORB ) orb;

		_target = target;

    _delegate = (org.openorb.CORBA.Delegate)((org.omg.CORBA.portable.ObjectImpl)target)._get_delegate();

		_operation = operation;

		_arguments = _orb.create_list(0);
		_env = new org.openorb.CORBA.dii.Environment();
		_exceptions = new org.openorb.CORBA.dii.ExceptionList();
		_contexts = new org.openorb.CORBA.dii.ContextList();
		_context = null;

		//	_multiple_target=false;

		org.omg.CORBA.Any any = _orb.create_any();
		any.type ( _orb.get_primitive_tc( org.omg.CORBA.TCKind.tk_void ) );
		_result = new org.openorb.CORBA.dii.NamedValue("",any,0);
	}

    /* Added by Diana on feb. 4 2002*/
	public Request(	org.omg.CORBA.Object target, org.omg.IOP.ServiceContext one_sc,String operation, org.omg.CORBA.ORB orb )
	{
		_orb = ( org.openorb.CORBA.ORB ) orb;

		_target = target;

		_delegate = (org.openorb.CORBA.Delegate)((org.omg.CORBA.portable.ObjectImpl)target)._get_delegate();

		_operation = operation;

_one_service_context=one_sc;
		_arguments = _orb.create_list(0);
		_env = new org.openorb.CORBA.dii.Environment();
		_exceptions = new org.openorb.CORBA.dii.ExceptionList();
		_contexts = new org.openorb.CORBA.dii.ContextList();
		_context = null;

		//	_multiple_target=false;

		org.omg.CORBA.Any any = _orb.create_any();
		any.type ( _orb.get_primitive_tc( org.omg.CORBA.TCKind.tk_void ) );
		_result = new org.openorb.CORBA.dii.NamedValue("",any,0);
	}

	/**
	 * Constructor modified by Diana in Nov. 2001*/

	public Request(	org.omg.CORBA.Object target,
			int retention_id,
			String client_id,
			long expiration_time,
			String operation,
			org.omg.CORBA.NVList arguments,
			org.omg.CORBA.NamedValue result,
			org.omg.CORBA.Environment env,
			org.omg.CORBA.ExceptionList exceptions,
			org.omg.CORBA.ContextList contexts,
			org.omg.CORBA.ORB orb )
    {
		_orb = ( org.openorb.CORBA.ORB ) orb;

		_target = target;

		_delegate = (org.openorb.CORBA.Delegate)((org.omg.CORBA.portable.ObjectImpl)target)._get_delegate();

		_operation = operation;
		_arguments = _orb.create_list(0);
		_env = env;
		_exceptions = exceptions;
		_contexts = contexts;
		_result = result;
		_context = null;
		_retention_id=retention_id;
		_client_id=client_id;
		_expiration_time=expiration_time;
		for(int i=0;i<arguments.count();i++)
		{
			try
			{
				String s = arguments.item(i).name() ;
				org.omg.CORBA.Any a = arguments.item(i).value() ;
				int f = arguments.item(i).flags() ;

				_arguments.add_value(s,a,f) ;
			}
			catch(org.omg.CORBA.Bounds ex)
			{ }
		}
	}


       public Request(org.omg.CORBA.Object target,
		   int retention_id,
		   String client_id,
		   long expiration_time,
		   org.omg.IOP.ServiceContext one_sc,
		   String operation,
		   org.omg.CORBA.NVList arguments,
		   org.omg.CORBA.NamedValue result,
		   org.omg.CORBA.Environment env,
		   org.omg.CORBA.ExceptionList exceptions,
		   org.omg.CORBA.ContextList contexts,
		   org.omg.CORBA.ORB orb )
    {
		_orb = ( org.openorb.CORBA.ORB ) orb;

		_target = target;

		//	_multiple_target=multiple_target;

		_delegate = (org.openorb.CORBA.Delegate)((org.omg.CORBA.portable.ObjectImpl)target)._get_delegate();

		_operation = operation;
		_arguments = _orb.create_list(0);
		_env = env;
		_exceptions = exceptions;
		_contexts = contexts;
		_result = result;
		_context = null;
		_retention_id=retention_id;
		_client_id=client_id;
		_one_service_context=one_sc;
		_expiration_time=expiration_time;
		for(int i=0;i<arguments.count();i++)
		{
			try
			{
				String s = arguments.item(i).name() ;
				org.omg.CORBA.Any a = arguments.item(i).value() ;
				int f = arguments.item(i).flags() ;

				_arguments.add_value(s,a,f) ;
			}
			catch(org.omg.CORBA.Bounds ex)
			{ }
		}
	}

    	public Request(	org.omg.CORBA.Object target,
			String operation,
			org.omg.CORBA.NVList arguments,
			org.omg.CORBA.NamedValue result,
			org.omg.CORBA.Environment env,
			org.omg.CORBA.ExceptionList exceptions,
			org.omg.CORBA.ContextList contexts,
			org.omg.CORBA.ORB orb )
    {
		_orb = ( org.openorb.CORBA.ORB ) orb;

		_target = target;

		//	_multiple_target=false;

    _delegate = (org.openorb.CORBA.Delegate)((org.omg.CORBA.portable.ObjectImpl)target)._get_delegate();

    _operation = operation;
		_arguments = _orb.create_list(0);
		_env = env;
		_exceptions = exceptions;
		_contexts = contexts;
		_result = result;
		_context = null;
		_retention_id=-1;
		_client_id="";
		_expiration_time=0;
		for(int i=0;i<arguments.count();i++)
		{
			try
			{
				String s = arguments.item(i).name() ;
				org.omg.CORBA.Any a = arguments.item(i).value() ;
				int f = arguments.item(i).flags() ;

				_arguments.add_value(s,a,f) ;
			}
			catch(org.omg.CORBA.Bounds ex)
			{ }
		}
	}

     	public Request(	org.omg.CORBA.Object target,
			String operation,
			org.omg.IOP.ServiceContext one_service_ctx,
			org.omg.CORBA.NVList arguments,
			org.omg.CORBA.NamedValue result,
			org.omg.CORBA.Environment env,
			org.omg.CORBA.ExceptionList exceptions,
			org.omg.CORBA.ContextList contexts,
			org.omg.CORBA.ORB orb )
    {
		_orb = ( org.openorb.CORBA.ORB ) orb;

		_target = target;

		//	_multiple_target=false;

		_delegate = (org.openorb.CORBA.Delegate)((org.omg.CORBA.portable.ObjectImpl)target)._get_delegate();

		_operation = operation;
		_one_service_context=one_service_ctx;
		_arguments = _orb.create_list(0);
		_env = env;
		_exceptions = exceptions;
		_contexts = contexts;
		_result = result;
		_context = null;
		_retention_id=-1;
		_client_id="";
		_expiration_time=0;
		for(int i=0;i<arguments.count();i++)
		{
			try
			{
				String s = arguments.item(i).name() ;
				org.omg.CORBA.Any a = arguments.item(i).value() ;
				int f = arguments.item(i).flags() ;

				_arguments.add_value(s,a,f) ;
			}
			catch(org.omg.CORBA.Bounds ex)
			{ }
		}
	}

	// ----------------------------------------------------------------------------------------
	//
	// Request interface operations
	//
	// ----------------------------------------------------------------------------------------

	/**
	 * Return the target object reference
	 */
	public org.omg.CORBA.Object target()
	{
		return _target ;
	}

	/**
	 * Return the operation name
	 */
	public String operation()
	{
		return _operation ;
	}

	/**
	 * Return the operation arguments
	 */
	public org.omg.CORBA.NVList arguments()
	{
		return _arguments ;
	}

	/**
	 * Return the result value
	 */
	public org.omg.CORBA.NamedValue result()
	{
    if(!_response)
      throw new org.omg.CORBA.BAD_INV_ORDER("Operation not completed",
				org.omg.CORBA.OMGVMCID.value | 5, CompletionStatus.COMPLETED_MAYBE);
		return _result ;
	}

	/**
	 * Return the environment value
	 */
	public org.omg.CORBA.Environment env()
	{
		return _env ;
	}

	/**
	 * Return the exception list
	 */
	public org.omg.CORBA.ExceptionList exceptions()
	{
		return _exceptions;
	}

	/**
	 * Return the context list
	 */
	public org.omg.CORBA.ContextList contexts()
	{
		return _contexts ;
	}

	/**
	 * Return the context object
	 */
	public org.omg.CORBA.Context ctx()
	{
		return _context ;
	}

	/**
	 * Set the context object
	 */
	public void ctx( org.omg.CORBA.Context c )
	{
    if(_output != null)
      throw new org.omg.CORBA.BAD_INV_ORDER("Operation already invoked",
				org.omg.CORBA.OMGVMCID.value | 5, CompletionStatus.COMPLETED_MAYBE);
		_context = c ;
	}

	/**
	 * Add an IN argument
	 */
	public org.omg.CORBA.Any add_in_arg()
	{
    if(_output != null)
      throw new org.omg.CORBA.BAD_INV_ORDER("Operation already invoked",
				org.omg.CORBA.OMGVMCID.value | 5, CompletionStatus.COMPLETED_MAYBE);
		return ( _arguments.add(org.omg.CORBA.ARG_IN.value) ).value();
	}

	/**
	 * Add an IN argument
	 */
	public org.omg.CORBA.Any add_named_in_arg(String name)
	{
    if(_output != null)
      throw new org.omg.CORBA.BAD_INV_ORDER("Operation already invoked",
				org.omg.CORBA.OMGVMCID.value | 5, CompletionStatus.COMPLETED_MAYBE);
		return ( _arguments.add_item(name,org.omg.CORBA.ARG_IN.value) ).value();
	}

	/**
	 * Add an INOUT argument
	 */
	public org.omg.CORBA.Any add_inout_arg()
	{
    if(_output != null)
      throw new org.omg.CORBA.BAD_INV_ORDER("Operation already invoked",
				org.omg.CORBA.OMGVMCID.value | 5, CompletionStatus.COMPLETED_MAYBE);
		return ( _arguments.add(org.omg.CORBA.ARG_INOUT.value) ).value();
	}

	/**
	 * Add an INOUT argument
	 */
	public org.omg.CORBA.Any add_named_inout_arg(String name)
	{
    if(_output != null)
      throw new org.omg.CORBA.BAD_INV_ORDER("Operation already invoked",
				org.omg.CORBA.OMGVMCID.value | 5, CompletionStatus.COMPLETED_MAYBE);
		return ( _arguments.add_item(name,org.omg.CORBA.ARG_INOUT.value)).value() ;
	}

	/**
	 * Add an OUT argument
	 */
	public org.omg.CORBA.Any add_out_arg()
	{
    if(_output != null)
      throw new org.omg.CORBA.BAD_INV_ORDER("Operation already invoked",
				org.omg.CORBA.OMGVMCID.value | 5, CompletionStatus.COMPLETED_MAYBE);
		return ( _arguments.add(org.omg.CORBA.ARG_OUT.value) ).value();
	}

	/**
	 * Add an OUT argument
	 */
	public org.omg.CORBA.Any add_named_out_arg(String name)
	{
    if(_output != null)
      throw new org.omg.CORBA.BAD_INV_ORDER("Operation already invoked",
				org.omg.CORBA.OMGVMCID.value | 5, CompletionStatus.COMPLETED_MAYBE);
		return ( _arguments.add_item(name,org.omg.CORBA.ARG_OUT.value)).value() ;
	}

	/**
	 * Set the return type
	 */
	public void set_return_type( org.omg.CORBA.TypeCode tc )
	{
    if(_output != null)
      throw new org.omg.CORBA.BAD_INV_ORDER("Operation already invoked",
				org.omg.CORBA.OMGVMCID.value | 5, CompletionStatus.COMPLETED_MAYBE);
		_result.value().type(tc) ;
	}

	/**
	 * Return the value
	 */
	public org.omg.CORBA.Any return_value()
	{
		return _result.value() ;
	}

  private void marshal(org.omg.CORBA.portable.OutputStream argument_stream) {
		for ( int i=0; i<_arguments.count(); i++ )
		{
			try
			{
				if ( _arguments.item(i).flags() != org.omg.CORBA.ARG_OUT.value )
					_arguments.item(i).value().write_value( argument_stream ) ;
			}
			catch(org.omg.CORBA.Bounds e)
			{
			}
		}

		java.util.Vector allCtx = new java.util.Vector();

		if ( _context != null )
  		for ( int i=0; i<_contexts.count(); i++ ) {
        try {
          org.omg.CORBA.NVList list = _context.get_values("",0,_contexts.item(i));
          for ( int j=0; j<list.count(); j++ ) {
            allCtx.addElement( list.item(j).name() );

            allCtx.addElement ( list.item(j).value().extract_string() );
          }
        }
        catch ( org.omg.CORBA.Bounds ex )
        { }
      }

		if ( allCtx.size() != 0 )
		{
			argument_stream.write_ulong(allCtx.size());
			for ( int j=0; j<allCtx.size(); j++ )
				argument_stream.write_string( (String)allCtx.elementAt(j) );
		}
		else
		if ( _context != null )
		   argument_stream.write_ulong(0);
  }

  private void unmarshal(org.omg.CORBA.portable.InputStream response_stream) {
    _result.value().read_value(response_stream, _result.value().type());

		for ( int i=0; i<_arguments.count(); i++ ) {
			try
			{
        org.omg.CORBA.NamedValue nv = _arguments.item(i);
				if ( nv.flags() != org.omg.CORBA.ARG_IN.value )
					nv.value().read_value( response_stream, nv.value().type() ) ;
			}
			catch(org.omg.CORBA.Bounds e)
			{
			}
		}
  }

  private void unmarshalException(String ex_id, org.omg.CORBA.portable.InputStream response_stream) {
    for (int i = 0; i < _exceptions.count(); ++i) {
      try {
        if(_exceptions.item(i).id().equals(ex_id)) {
          org.omg.CORBA.UnknownUserException ex = new org.omg.CORBA.UnknownUserException(_orb.create_any());
          ex.except.read_value(response_stream, _exceptions.item(i));
          env().exception(ex);
          return;
        }
      }
      catch(org.omg.CORBA.TypeCodePackage.BadKind ex) {}
      catch(org.omg.CORBA.Bounds ex) {}
    }
    env().exception(new org.omg.CORBA.UNKNOWN("Unexcepected User Exception: "+ ex_id,
			org.omg.CORBA.OMGVMCID.value | 1, CompletionStatus.COMPLETED_YES));
  }

	/**
	 * Invoke an operation
	 */
	public void invoke() {
    invoke(true);
  }

	/**
	 * Send a oneway request
	 */
	public void send_oneway() {
    invoke(false);
  }

  private org.omg.CORBA.portable.InputStream _input;
  private org.omg.CORBA.portable.OutputStream _output;

  private void invoke(boolean response_expected)
	{
    if(_response)
      throw new org.omg.CORBA.BAD_INV_ORDER("Response already recieved.",
				org.omg.CORBA.OMGVMCID.value | 5, CompletionStatus.COMPLETED_MAYBE);
    if(_output != null)
      throw new org.omg.CORBA.BAD_INV_ORDER("Multiple request sending",
				org.omg.CORBA.OMGVMCID.value | 10, CompletionStatus.COMPLETED_MAYBE);

		_response = true;

    try {
      while( true )
      {
        try
        {
	    if(_one_service_context!=null)
		_output = _delegate.request(_target,_retention_id, _client_id, _expiration_time,_one_service_context, _operation, response_expected);
	    else{
		//System.out.println("Before _output=request "+_operation);
		_output = _delegate.request(_target,_retention_id, _client_id, _expiration_time, _operation, response_expected);
		//System.out.println("after _output=request "+_operation);
	    }
          marshal(_output);
	  //System.out.println("after marshal _output "+_operation);
          _input = _delegate.invoke(_target,_output);
	  //System.out.println("After invoke "+_operation);
          if(response_expected)
            unmarshal(_input);
          return;
        }
        catch( org.omg.CORBA.portable.RemarshalException _exception )
        {
          continue;
        }
        catch( org.omg.CORBA.portable.ApplicationException _exception )
        {
          unmarshalException(_exception.getId(), _exception.getInputStream());
          return;
        }
        finally
        {
          _delegate.releaseReply(_target, _input);
        }
      }
    }
    catch(org.omg.CORBA.SystemException ex) {
      env().exception(ex);
			throw ex;
    }
	}

	/**
	 * Send a deferred request
	 */
	public void send_deferred()
	{
    if(_response)
      throw new org.omg.CORBA.BAD_INV_ORDER("Response already recieved.",
				org.omg.CORBA.OMGVMCID.value | 5, CompletionStatus.COMPLETED_MAYBE);
    if(_output != null)
      throw new org.omg.CORBA.BAD_INV_ORDER("Multiple request sending",
				org.omg.CORBA.OMGVMCID.value | 10, CompletionStatus.COMPLETED_MAYBE);
		_deferred = true;

    try {
      _input = null;
      _output = _delegate.request(_target, _operation, true);
      marshal(_output);
      _delegate.invoke_deferred(_target, _output);
    }
    catch(org.omg.CORBA.SystemException ex) {
      env().exception(ex);
			throw ex;
    }
  }

	/**
	 * Return TRUE if a response is available
	 */
	public boolean poll_response()
	{
    if(_output == null)
      throw new org.omg.CORBA.BAD_INV_ORDER("Operation not sent",
				org.omg.CORBA.OMGVMCID.value | 11, CompletionStatus.COMPLETED_MAYBE);
    if(!_deferred)
      throw new org.omg.CORBA.BAD_INV_ORDER("Operation not sent deferred",
				org.omg.CORBA.OMGVMCID.value | 13, CompletionStatus.COMPLETED_MAYBE);
    if(_response)
      throw new org.omg.CORBA.BAD_INV_ORDER("Response already recieved.",
				org.omg.CORBA.OMGVMCID.value | 12, CompletionStatus.COMPLETED_MAYBE);

    return _delegate.poll_response(_target, _output);
  }

	/**
	 * Get response
	 */
	public void get_response()
	{
    if(_output == null)
      throw new org.omg.CORBA.BAD_INV_ORDER("Operation not sent",
				org.omg.CORBA.OMGVMCID.value | 11, CompletionStatus.COMPLETED_MAYBE);
    if(!_deferred)
      throw new org.omg.CORBA.BAD_INV_ORDER("Operation not sent deferred",
				org.omg.CORBA.OMGVMCID.value | 13, CompletionStatus.COMPLETED_MAYBE);
    if(_response)
      throw new org.omg.CORBA.BAD_INV_ORDER("Response already recieved.",
				org.omg.CORBA.OMGVMCID.value | 12, CompletionStatus.COMPLETED_MAYBE);

		_response = true;

    try {
      try
      {
        _input = _delegate.invoke(_target, _output);
        unmarshal(_input);
				return;
      }
      catch( org.omg.CORBA.portable.RemarshalException _exception )
      {
				_output = null;
				_response = false;
      }
      catch( org.omg.CORBA.portable.ApplicationException _exception )
      {
        unmarshalException(_exception.getId(), _exception.getInputStream());
				return;
      }
      finally
      {
				_delegate.releaseReply(_target, _input);
      }
    }
    catch(org.omg.CORBA.SystemException ex) {
      env().exception(ex);
			throw ex;
    }

		invoke(true);
	}

	public org.omg.CORBA.Object sendp() {
		throw new org.omg.CORBA.NO_IMPLEMENT();
	}

	public void prepare(org.omg.CORBA.Object p) {
		throw new org.omg.CORBA.NO_IMPLEMENT();
	}

	void sendc(org.omg.CORBA.Object handler) {
		throw new org.omg.CORBA.NO_IMPLEMENT();
	}
}
