/**
 * Redistribution and use of this software and associated
 * documentation ("Software"), with or without modification, are
 * permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright statements
 *    and notices.  Redistributions must also contain a copy of this
 *    document.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Intalio Inc.  For written permission, please
 *    contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Intalio Inc. Exolab is a registered trademark of
 *    Intalio Inc.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY INTALIO AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL INTALIO OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Copyright 2000 (C) Intalio Inc. All Rights Reserved.
 *
 * $Id: IdlCompiler.java,v 1.10 2001/07/31 22:56:16 daniel Exp $
 *
 * Date         Author  Changes
 * 11/2         J.Daniel    Bug #222 fixed ( cross #define #ifndef bug )
 * 06/3         O. Modica   Translated the remaining French comments
 */

/** Modified/extended by Diana Szentivanyi in June 2003
* contact diasz@ida.liu.se
*/

package org.openorb.compiler;

import org.openorb.compiler.object.*;
import org.openorb.compiler.parser.*;
import org.openorb.compiler.generator.*;
import java.util.*;
import java.net.URL;
import java.net.MalformedURLException;

import org.openorb.ReleaseInfo;

/**
 * This class is the IDL compiler implementation.
 *
 * @author <a href="mailto:jdaniel@intalio.com">Jerome Daniel &lt;daniel@intalio.com&gt;</a>
 * @version $Revision: 1.10 $ $Date: 2001/07/31 22:56:16 $
 */
public class IdlCompiler
{
	/**
	 * Include directories
	 */
	public static Vector includeList  = new java.util.Vector();

	/**
	 * Maps #defined macros to their expansions
	 */
	public static Hashtable definedMacros = new Hashtable();

	/**
	 * Pseudo Import list
	 */
	public static Vector importLink = new java.util.Vector();

	/**
	 * Native descriptions
	 */
	public static Vector nativeDefinition = new java.util.Vector();

	/**
	 * IDL file name to compile
	 */
	public static String [] idl_file_name;

	/**
	 * IDL files list
	 */
	public static java.util.Vector idl_file_name_list = new java.util.Vector();

	/**
	 * Indicates if the stub must be generated
	 */
	public static boolean map_stub = true;

	/**
	 * Indicates using local stubs
	 */
	public static boolean local_stub = true;

	/**
	 * Indicates if the skeleton must be generated
	 */
	public static boolean map_skeleton = true;

	/**
	 * Indicates if the included elements must be mapped
	 */
	public static boolean map_all = false;

	/**
	 * Indicates if TIE approach must be generated
	 */
	public static boolean map_tie = true;

	/**
	 * Indicates if the user code must be generated
	 */
	public static boolean map_user = false;

	/**
	 * Indicates if the compilation uses PIDL
	 */
	public static boolean pidl = false;

	/**
	 * Indicates if the POA adapter is used
	 */
	public static boolean map_poa = true;

	/**
	 * Indicates if verbose mode is activated
	 */
	public static boolean verbose = false;

	/**
	 * Package name for the generated code
	 */
	public static String packageName = null;

	/**
	 * Indicates if a package must be used
	 */
	public static boolean use_package = false;

	/**
	 * Use portable helper. This is only needed if the class must be compiled
   * without openorb.
   */
	public static boolean portableHelper = false;

	/**
	 * Indicates the output directory
	 */
	public static String outdir = null;

	/**
	 * Indicates if the the stub and skeleton uses DII and DSI
	 */
	public static boolean dynamic = false;

	/**
	 * Indicates if the prefix is used to name the packages
	 *
	 */
	public static boolean usePrefix = true;

	/**
	 * Active silent mode
	 */
	public static boolean silentMode = false;

	/**
	 * Display bad flag
	 */
	public static boolean displayBadFlag = true;

        /**
         * Reverse the prefix
         */
        public static boolean reversePrefix = true;

	/**
	 * Display a help message
	 */
	public static void display_help()
	{
		System.out.println("");
		System.out.println("");
		System.out.println("---------------------------------------------------------------");
		System.out.println("OpenORB                                                        ");
		System.out.println("(c) 2000-2001, Exolab.org ( www.openorb.org )                  ");
		System.out.println("Java IDL Compiler                                              ");
		System.out.println("---------------------------------------------------------------");
		System.out.println("");
		System.out.println("");
		System.out.println("Options");
		System.out.println("\t-release");
		System.out.println("\t\tShow version number");
		System.out.println("\t-d directory name");
		System.out.println("\t\tProvide a way to specify the ouput dir. This option");
		System.out.println("\t\twill not use the 'generated' directory.");
    System.out.println("\t\tFor example :");
    System.out.println("\t\t	idl2java demo.idl -d /home/me/");
		System.out.println("\t-package package_name");
		System.out.println("\t\tGenerate files in package_name");
		System.out.println("\t\tExample:");
		System.out.println("\t\t	idl2java demo.idl -package exemple");
		System.out.println("\t-I");
		System.out.println("\t\tAllow specification of include directory");
		System.out.println("\t\tExample:");
		System.out.println("\t\t  idl2java demo.idl -I /home/me/idl -I ../autre");
		System.out.println("\t-D");
		System.out.println("\t\tDefine a symbol. It is equivalent to #define");
		System.out.println("\t-nostub");
		System.out.println("\t\tDon't generate stub.");
		System.out.println("\t-nolocalstub");
		System.out.println("\t\tGenerate stubs without local invocation path.");
		System.out.println("\t-noskeleton");
		System.out.println("\t\tDon't generate skeleton.");
		System.out.println("\t-notie");
		System.out.println("\t\tDon't generate TIE classes for delegation mode.");
		System.out.println("\t-boa");
		System.out.println("\t\tGenerate skeleton for BOA");
		System.out.println("\t-dynamic");
		System.out.println("\t\tGenerate stub with DII and skeleton with DSI");
		System.out.println("\t\t( portable way before CORBA 2.3 )");
		System.out.println("\t-portablehelper");
		System.out.println("\t\tGenerate portable helper classes. Nonportable helpers must be");
		System.out.println("\t\tcompiled using OpenORB, but the generated class files will work");
    System.out.println("\t\tanywhere. Portable helpers can be compiled anywhere and will");
		System.out.println("\t\twork identically on all orbs with Any.extract_Streamable.");
		System.out.println("\t-all");
		System.out.println("\t\tGenerate mapping for included files.");
		System.out.println("\t-noprefix");
		System.out.println("\t\tDon't use prefixes as packages names.");
    System.out.println("\t-noreverseprefix");
    System.out.println("\t\tThe prefixes are used as package name but they are not reversed");
		System.out.println("\t-native native name native mapping ");
		System.out.println("\t\tDefine native type mapping.");
		System.out.println("\t\tFor example : ");
		System.out.println("\t\t	idl2java demo.idl -native cookie java.lang.Object");
		System.out.println("\t\t	this command implies the mapping of");
		System.out.println("\t\t	cookie in java.lang.Object.");
		System.out.println("");
		System.out.println("");
		System.exit(1);
	}

	/**
	 * Scan command line arguments
	 */
	public static void scan_arguments( String[] args )
	{
		for ( int i=0; i<args.length; i++ )
		{
		    if(args[i].equals("-application")) continue;
		    if ( args[i].charAt(0) != '-' )
			idl_file_name_list.addElement( args[i] );
		    else
			if ( args[i].equals("-release") )
			    {
				System.out.println(ReleaseInfo.RELEASE);
				System.exit(1);
			    }
			else if ( args[i].equals("-h") || args[i].equals("-help") )
			    display_help();
			else if ( args[i].equals("-silence") || args[i].equals("-quiet") )
			    silentMode = true;
			else if ( args[i].equals("-verbose") )
			    verbose = true;
			else if ( args[i].equals("-noprefix") )
			    usePrefix = false;
			else if ( args[i].equals("-noreverseprefix") )
			    reversePrefix = false;
			else if ( args[i].equals("-nostub") )
			    map_stub = false;
			else if ( args[i].equals("-nolocalstub") )
			    local_stub = false;
			else if ( args[i].equals("-noskeleton") )
			    map_skeleton = false;
			else if ( args[i].equals("-notie") )
			    map_tie = false;
			else if ( args[i].equals("-portablehelper") )
			    portableHelper = true;
			else if ( args[i].equals("-boa") )
			    map_poa = false;
			else if ( args[i].equals("-dynamic") )
			    dynamic = true;
			else if ( args[i].equals("-all") )
			    map_all = true;
			else if ( args[i].equals("-pidl") )
			    {
				pidl = true;
				map_stub = false;
				map_skeleton = false;
				map_tie = false;
			    }
			else if ( args[i].equals("-package") )
			    {
				if ( i+1 == args.length  )
				    {
					System.out.println("Argument expected after '-package'");
					System.exit(2);
				    }
				packageName = args[++i];
				use_package = false;
			    }
			else if ( args[i].equals("-importlink") )
			    {
				if ( i+1 == args.length  )
				    {
					System.out.println("Argument expected after '-importlink'");
					System.exit(2);
				    }
				importLink.addElement( args[++i] );
			    }
			else if ( args[i].equals("-d") )
			    {
				if ( i+1 == args.length  )
				    {
					System.out.println("Argument expected after '-d'");
					System.exit(2);
				    }
				outdir = args[++i];
				if ( packageName == null )
				    {
					packageName = "";
					use_package = false;
				    }
			    }
			else if ( args[i].equals("-I") )
			    {
				if ( ++i == args.length  )
				    {
					System.out.println("Argument expected after '-I'");
					System.exit(2);
				    }
				URL url = null;
				try {
				    url = new URL(args[i]);
				}
				catch(MalformedURLException ex) {
				    try {
					url = new java.io.File(args[i]).toURL();
				    }
				    catch(MalformedURLException ex1) {}
				}
				if(url != null)
				    includeList.addElement( url );
			    }
			else if ( args[i].startsWith("-I") )
			    {
				String path = args[i].substring( 2 );

				URL url = null;
				try {
				    url = new URL(path);
				}
				catch(MalformedURLException ex) {
				    try {
					url = new java.io.File(path).toURL();
				    }
				    catch(MalformedURLException ex1) {}
				}
				if(url != null)
				    includeList.addElement( url );
			    }
			else if ( args[i].startsWith("-D") )
			    {
				try
				    {
					int idx = args[i].indexOf('=');
					if(idx < 0)
					    definedMacros.put( args[i].substring(2,args[i].length()) , "");
					else
					    definedMacros.put( args[i].substring(2,idx) , args[i].substring(idx+1));
				    }
				catch ( StringIndexOutOfBoundsException ex )
				    { }
			    }
			else if ( args[i].equals("-native") )
			    {
				if ( i+2 == args.length  )
				    {
					System.out.println("Argument expected after '-native'");
					System.exit(2);
				    }
				nativeDefinition.addElement( args[++i] +":"+ args[++i] );
			    }
			else if ( displayBadFlag )
			    {
				System.out.println("Bad parameter : " + args[i]);
				System.out.println("Please, use no flag to display all compiler option flags");
				System.exit(2);
			    }
		}
		if ( outdir == null || outdir.equals("") )
		    {
			outdir = new String("generated");
			if ( packageName == null )
			    {
				packageName = "";
				use_package = false;
			    }
		    }
		idl_file_name = new String[ idl_file_name_list.size() ];
		for ( int i=0; i<idl_file_name_list.size(); i++ )
		    idl_file_name[i] = ( String ) idl_file_name_list.elementAt(i);
	}

    /**
     * Get argument from configuration file
     */
    public static void configFile()
    {
	Configurator config = null;

	// -- try to load the configurator --

	try
	    {
		java.lang.Object obj = Class.forName("org.openorb.compiler.orb.DefaultConfigurator").newInstance();

		config = ( org.openorb.compiler.Configurator ) obj;
	    }
	catch ( java.lang.Exception ex )
	    {
		return;
	    }

	// -- Invoke the configurator --

	config.updateInfo( includeList, importLink );
    }

    /**
     * Prepare the compilation process
     */
    public static void prepare()
    {
	// define some standard macros
	definedMacros.put("__IDL_TO_JAVA__", "");
	definedMacros.put("__OPENORB__", ReleaseInfo.RELEASE);
	definedMacros.put("__OPENORB_MAJOR__", ""+ReleaseInfo.OPENORB_MAJOR);
	definedMacros.put("__OPENORB_MINOR__", ""+ReleaseInfo.OPENORB_MINOR);

	definedMacros.put("__CORBA_IDL__", ReleaseInfo.CORBA_VERSION);
	definedMacros.put("__CORBA_MAJOR__", ""+ReleaseInfo.CORBA_MAJOR);
	definedMacros.put("__CORBA_MINOR__", ""+ReleaseInfo.CORBA_MINOR);
    }

    /**
     * This operation is used to compile an IDL file
     */
    /* Modified by Diana on Aug. 31 2001 - added the boolean param.*/
    public static void compile_file( String file_name, IdlParser Parser,boolean application)
	throws org.openorb.compiler.parser.CompilationException
    {
	IdlObject CompilationGraph = Parser.compile_idl( file_name );

	if ( Parser.totalError != 0 )
	    {
		System.out.println("there are errors...");
		System.out.println("compilation process stopped !");
		System.exit(2);
	    }

	// -- Start to generate Java code --

	if ( verbose )
	    System.out.println("Now translate to Java..." );



	IdlToJava toJava = new IdlToJava(application);

	toJava.translateData(CompilationGraph,packageName);

	if ( map_stub == true )
	    toJava.translateStub(CompilationGraph,packageName);

	if ( map_skeleton == true )
	    toJava.translateSkeleton(CompilationGraph,packageName);

	if ( map_tie == true )
	    toJava.translateTIE(CompilationGraph,packageName);
    }

    public static IdlParser createIDLParser( String [] args )
    {
	prepare();

	scan_arguments(args);

	configFile();

	return new IdlParser();
    }

    // ===========================
    // Class entry point
    // ===========================
    /**
     * The IDL compiler starts here
     *
     * @param	args  command line arguments list
     */
    public static void main( String[] args )
    {
	SymboleDef s;
	boolean application=false;
	int k;

	if ( args.length == 0 )
	    display_help();


	IdlParser Parser = createIDLParser( args );

	if(idl_file_name.length == 0)
	    display_help();

	// -- Welcome message --
	if ( !silentMode )	 {
	    System.out.println("OpenORB IDL Compiler / (c) 2000-2001 Exolab.org");			 }

	// -- Starts the compilation --

	try
	    {
		java.util.Hashtable definedCopy = (java.util.Hashtable)definedMacros.clone();

		for ( int i=0; i<idl_file_name.length; i++ )
		    {
			definedMacros = definedCopy;

			System.out.println("compile : " + idl_file_name[i] );

			/* Modified by Diana on Aug.31 2001 to permit generation (or not) of
			   stub files with retention_id and client_id variables
			   - for infrastructure class stubs, there will not be such variables;
			   - for application class stubs the option will be used*/

			for(k=0;k<args.length;k++){
			    if(args[k].equals("-application")){
				application=true;
			    }
			}

			compile_file(idl_file_name[i], Parser,application);
		    }
		System.exit(0);
	    }
	catch ( org.openorb.compiler.parser.CompilationException ex )
	    {
		System.out.println("there are errors...");
		System.out.println("compilation process stopped !");
		System.exit(2);
	    }
    }
}
