/**
 * Redistribution and use of this software and associated
 * documentation ("Software"), with or without modification, are
 * permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright statements
 *    and notices.  Redistributions must also contain a copy of this
 *    document.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Intalio Inc.  For written permission, please
 *    contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Intalio Inc. Exolab is a registered trademark of
 *    Intalio Inc.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY INTALIO AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL INTALIO OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Copyright 2000 (C) Intalio Inc. All Rights Reserved.
 *
 * $Id: ClientChannel.java,v 1.2 2001/02/27 00:41:59 wood Exp $
 *
 * Date         Author  Changes
 */

/** Modified/extended by Diana Szentivanyi in June 2003
* contact diasz@ida.liu.se
*/
package org.openorb.net;

/**
 * Client channel interface. A client channel creates and sends requests
 * and recieves replies.
 *
 * @author <a href="mailto:wood@intalio.com">Chris Wood &lt;wood@intalio.com&gt;</a>
 * @version $Revision: 1.2 $ $Date: 2001/02/27 00:41:59 $
 */
public interface ClientChannel extends Channel {

  /**
   * An orb reference.
   */
  public org.omg.CORBA.ORB orb();

  /**
   * Connection state. Note that state values are 0x1000000 apart,
   * Allowing this value to be ORed with the binding priority on the
   * client side.
   */
	public int state();

	/**
	 * Connection open. In this state requests can be created without
   * checking the rebind policy.
	 */
	public final static int STATE_CONNECTED = 0x11000000;

	/**
	 * Channel paused. The channel has been shut down temporaraly to conserve
   * resources or has recently been created. Creating new requests with
   * with rebind policy set to NO_RECONNECT will result in a REBIND exception.
   * Channels are created in this state.
	 */
	public final static int STATE_PAUSED = 0x12000000;

	/**
	 * Channel closed. This state is typicaly reached if transmission dificulties
   * are experienced. Attempting to create requests on closed channels will
   * cause a system exception. The channel can be reopened only using the open
   * operation.
	 */
	public final static int STATE_CLOSED = 0x13000000;

	// state management

	/**
	 * Pause the channel. If no requests are currently active enter the
   * paused state.
	 */
	public void pause();

	/**
	 * Change to the CLOSED state and reject new requests by throwing a system
   * exception with status COMPLETED_NO. If kill_requests is true, call cancel
   * on any active ClientRequests, otherwise wait until active requests
   * complete before closing the commmunication channel.
	 */
	public void close(boolean kill_requests, org.omg.CORBA.SystemException ex);

	// request creation.

	/**
   * Create a request. If this is the first request on this channel
   * then client_connect will be called on all ChannelInterceptor
   * before returning the request. This may throw a system exception if the
   * channel cannot establish a connection for some reason, for example
   * INV_POLICY if client side policies prevent a successfull invocation,
   * COMM_FAILURE if a communication problem occours, or REBIND if channel
   * is temporaraly closed and a NO_REBIND policy is in effect.
   *
   * @param target The target of the request.
   * @param address The target address. If the target has been redirected
   *            this may not correspond to the target's ior.
   */
	public ClientRequest create_request(org.omg.CORBA.Object target,Address address,Object transportAssociation,String operation,boolean response_expected) throws RebindChannelException;

	/**
   * Create a locate request. This may throw a system exception if the
   * channel cannot establish a connection for some reason, for example
   * INV_POLICY if client side policies prevent a successfull invocation,
   * COMM_FAILURE if a communication problem occours, or REBIND if channel
   * is temporaraly closed and a NO_REBIND policy is in effect.
   *
   * @param target The target of the request.
   * @param address The target address. If the target has been redirected
   *            this may not correspond to the target's ior.
   */
	public ClientRequest create_locate_request(org.omg.CORBA.Object target,Address address,Object transportAssociation)throws RebindChannelException;




	public ClientRequest create_request(org.omg.CORBA.Object target,int retention_id,String client_id,Address address,Object transportAssociation,String operation,boolean response_expected) throws RebindChannelException;
    /* Modif. by Diana on Oct. 28 2001*/
    	public ClientRequest create_request(org.omg.CORBA.Object target,int retention_id,String client_id,long request_start_time,Address address,Object transportAssociation,String operation,boolean response_expected) throws RebindChannelException;
	public ClientRequest create_request(org.omg.CORBA.Object target,int retention_id,String client_id,org.omg.Dynamic.Parameter[] args,Address address,Object transportAssociation,String operation,boolean response_expected) throws RebindChannelException;

    /* new version by Diana on Oct 10 2001*/
    public ClientRequest create_request(org.omg.CORBA.Object target,int retention_id,String client_id,org.omg.Dynamic.Parameter[] args,org.omg.CORBA.TypeCode[] exceptions,org.omg.CORBA.Any result,Address address,Object transportAssociation,String operation,boolean response_expected) throws RebindChannelException;

    /* new version by Diana on Oct 28 2001*/

    public ClientRequest create_request(org.omg.CORBA.Object target,int retention_id,String client_id,long request_start_time,org.omg.Dynamic.Parameter[] args,org.omg.CORBA.TypeCode[] exceptions,org.omg.CORBA.Any result,Address address,Object transportAssociation,String operation,boolean response_expected) throws RebindChannelException;

    /* Added by diana on 2 feb. 2002*/
   public ClientRequest create_request(org.omg.CORBA.Object target,int retention_id,String client_id,org.omg.IOP.ServiceContext one_sc,Address address, Object tpAssc,String operation,boolean response_expected) throws RebindChannelException;

    public ClientRequest create_request(org.omg.CORBA.Object target,int retention_id,String client_id,long req_st_t,org.omg.IOP.ServiceContext one_sc,Address address, Object tpAssc,String operation,boolean response_expected) throws RebindChannelException;


}
