/**
 * Redistribution and use of this software and associated
 * documentation ("Software"), with or without modification, are
 * permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright statements
 *    and notices.  Redistributions must also contain a copy of this
 *    document.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Intalio Inc.  For written permission, please
 *    contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Intalio Inc. Exolab is a registered trademark of
 *    Intalio Inc.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY INTALIO AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL INTALIO OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Copyright 2000 (C) Intalio Inc. All Rights Reserved.
 *
 * $Id: ClientManagerImpl.java,v 1.3 2001/08/08 13:37:26 wood Exp $
 *
 * Date         Author  Changes
 */

/** Modified/extended by Diana Szentivanyi in June 2003
* contact diasz@ida.liu.se
*/
package org.openorb.net;

import java.util.Map;
import java.util.HashMap;

import java.util.ArrayList;
import java.util.Arrays;

import java.util.Iterator;

import org.openorb.policy.*;
import org.openorb.CORBA.MinorCodes;
import org.omg.CORBA.CompletionStatus;

/**
 * @author <a href="mailto:wood@intalio.com">Chris Wood &lt;wood@intalio.com&gt;</a>
 * @version $Revision: 1.3 $ $Date: 2001/08/08 13:37:26 $
 */
public class ClientManagerImpl implements ClientManager {

    public ClientManagerImpl(org.omg.CORBA.ORB orb) {
	this(orb, null);
    }

    public ClientManagerImpl(org.omg.CORBA.ORB orb,ServerManager serverManager) {
	_orb = orb;
	_serverManager = serverManager;
	_rootGroup = Thread.currentThread().getThreadGroup();

	_pause = ((org.openorb.CORBA.ORB)orb).getLoader().getIntProperty("openorb.client.reapPauseDelay", DEFAULT_PAUSE_TIME);
	_useStaticThreadGroup = ((org.openorb.CORBA.ORB)orb).getLoader().getBooleanProperty("openorb.useStaticThreadGroup", false);
    }

    private org.omg.CORBA.ORB _orb;

    private Map _protocols = new HashMap();
    private Map _channels = new HashMap();

    private ServerManager _serverManager;

    private Object _syncState = new Object();

    private ThreadGroup _rootGroup;
    private boolean _useStaticThreadGroup = false;
    private static ThreadGroup _staticWorkThreads;
    private ThreadGroup _workThreads;
    private Thread _channelReaper;

    private boolean _shutdown = false;

    private int _pause;
    // _channels unused two minuites are paused
    private static final int DEFAULT_PAUSE_TIME = 2*60*1000;

    /**
     * Return an orb reference.
     */
    public org.omg.CORBA.ORB orb() {
	return _orb;
    }

    /**
     * Get a reference to the server manager. This is used to find the adapter
     * for local requests.
     */
    public ServerManager getServerManager() {
	return _serverManager;
    }

    /**
     * Create bindings for a given IOR.
     */
    public ClientBinding [] create_bindings(org.omg.CORBA.Object obj, org.omg.IOP.IOR ior) {

	int pref_profile_nr=0, profiles = ior.profiles.length,version=-1;

	/* modifications by Diana on August 23 2001*/

	if(profiles>1){
	    /* It might be a group reference ior*/
	    //System.out.println("Nr. of profiles is bigger than 1");
	    if(org.omg.FT.Utils.is_group_ref(_orb,ior)){
		//System.out.println("ior is a group reference");
		//org.omg.FT.Utils.print_IOR(_orb,ior);
		pref_profile_nr=org.omg.FT.Utils.primary_prof_ind(_orb,ior);
		version=org.omg.FT.Utils.extract_group_version(_orb,ior);
		if(pref_profile_nr!=-1){
		    System.out.println("Primary profile is "+pref_profile_nr+" and version is "+version);
		    org.omg.FT.Utils.print_profile(_orb,ior.profiles[pref_profile_nr]);
		}
	    }
	}

	if(pref_profile_nr==-1) {
	    pref_profile_nr=0;
	}

	org.omg.GIOP.IORAddressingInfo info = new org.omg.GIOP.IORAddressingInfo(pref_profile_nr, ior);
	// lookup the profile priority policy
	ProfilePriorityPolicy policy = null;
	try {
	    org.openorb.CORBA.Delegate delegate = (org.openorb.CORBA.Delegate)((org.omg.CORBA.portable.ObjectImpl)obj)._get_delegate();
	    policy = ProfilePriorityPolicyHelper.narrow(delegate.get_client_policy(obj, PROFILE_PRIORITY_POLICY_ID.value));
	}
	catch(org.omg.CORBA.INV_POLICY ex) {}
	ArrayList allbindings = new ArrayList(8);
	for(int i = 0; i < profiles; ++i) {
	    int profile_tag = ior.profiles[i].tag;
	    byte priority = (policy==null)?8:policy.find_priority(profile_tag);
	    ClientProtocol protocol;
	    // lookup the protocol
	    if(priority != DISABLE_PROFILE.value && (protocol = (ClientProtocol)_protocols.get(new Integer(profile_tag))) != null) {
		//System.out.println("I am here "+i);
		info.selected_profile_index = i;
				// create the addresses
		Address [] addresses = protocol.createAddresses(info);
				// create the bindings
		ClientBinding binding;

		for(int j = 0; j < addresses.length; ++j) {

		    if(version!=-1)
			binding = protocol.createBinding(addresses[j],version);
		    else
			binding = protocol.createBinding(addresses[j]);
		    binding.setPriority((binding.getPriority() & ~ClientBinding.MASK_PROFILE_PRIORITY)
					| (((int)priority)<<12 & ClientBinding.MASK_PROFILE_PRIORITY));
		    allbindings.add(binding);
		}
	    }
	}
	if(allbindings.isEmpty())
	    throw new org.omg.CORBA.INV_OBJREF();
	ClientBinding [] ret = new ClientBinding[allbindings.size()];
	ret = (ClientBinding [])allbindings.toArray(ret);
	//System.out.println("Before return from create_bindings");
	return ret;
    }
    /**
     * Register a client protocol.
     */
    public void register_protocol(int profile_tag,
				  ClientProtocol protocol)
    {
	_protocols.put(new Integer(profile_tag), protocol);
    }

    /**
     * Register a channel to join the work queue. This is called when the channel
     * enters the open state.
     */
    public boolean register_channel(ClientChannel channel) {
	synchronized(_syncState) {
	    if(_shutdown)
		return false;

	    synchronized(_channels) {
		if(_channels.containsKey(channel))
		    return true;

		if(_workThreads == null || _workThreads.isDestroyed()) {
		    if(_useStaticThreadGroup) {
			if(_staticWorkThreads == null || _staticWorkThreads.isDestroyed())
			    _staticWorkThreads = new ThreadGroup(_rootGroup, "Client IO");
			_workThreads = _staticWorkThreads;
		    } else
			_workThreads = new ThreadGroup(_rootGroup, "Client IO");
		}

		if(_channelReaper == null && _pause > 0) {
		    // create the channel reaper.
		    _channelReaper = new Thread(_workThreads, new Runnable() {
			public void run() { channel_reaper(); }
		    }, "Channel Reaper");
		    _channelReaper.setDaemon(true);
		    _channelReaper.start();
		}

		Thread thread = new Thread(_workThreads, new ChannelRecvRunnable(channel), "Recieve Worker for " + channel.toString());
		thread.setDaemon(true);
		thread.start();

		_channels.put(channel, thread);

		return true;
	    }
	}
    }

    /**
     * Called when a channel enters the closed or paused state. Returns once
     * all channel threads have completed their work cycles.
     */
    public void unregister_channel(ClientChannel channel) {
	Thread thread;
	synchronized(_channels) {
	    thread = (Thread) _channels.remove(channel);
	}

	if(thread != null) {
	    Thread curr = Thread.currentThread();
	    thread.interrupt();
	    while(thread != curr && thread.isAlive()) {
		try {
		    thread.join();
		}
		catch(InterruptedException ex) {}
	    }
	}
    }

    private class ChannelRecvRunnable implements Runnable {
	public ChannelRecvRunnable(ClientChannel chan) {
	    _chan = chan;
	}
	ClientChannel _chan;
	public void run() {
	    // this returns when the thread is interrupted or the channel is closed.
	    _chan.run_recv();
	}
    }

    public void shutdown(boolean wait_for_complete, boolean kill_requests) {
	boolean shut = false;
	Thread chanReap = null;
	synchronized(_syncState) {
	    if(!_shutdown) {
		_shutdown = true;
		shut = true;
		if(_channelReaper != null) {
		    _channelReaper.interrupt();
		    chanReap = _channelReaper;
		}
	    }
	}
	if(chanReap != null) {
	    try {
		chanReap.join();
	    }
	    catch(InterruptedException ex) {}
	}

	if(shut) {
	    ClientChannel [] cchans;
	    synchronized(_channels) {
		cchans = new ClientChannel[_channels.size()];
		_channels.keySet().toArray(cchans);
	    }

	    for(int i = 0; i < cchans.length; ++i)
		cchans[i].close(kill_requests, new org.omg.CORBA.BAD_INV_ORDER(org.omg.CORBA.OMGVMCID.value | 4, org.omg.CORBA.CompletionStatus.COMPLETED_MAYBE));
	}

	if(!_useStaticThreadGroup) {
	    if(_workThreads != null) {
		_workThreads.setDaemon(true);
		try {
		    _workThreads.destroy();
		}
		catch(IllegalThreadStateException ex) {}
	    }
	}
    }

    private void channel_reaper() {
	int lasttime = RequestIDAllocator.get_request_id();
	//System.out.println("lasttime = "+lasttime);
	ClientChannel [] chans = null;
	while(!_shutdown) {
	    //System.out.println("Go to sleep for "+_pause);
	    try {
		Thread.sleep(_pause);
	    }
	    catch(InterruptedException ex) {
		break;
	    }

	    //System.out.println("Wake up from sleep");
	    synchronized(_channels) {
		if(_channels.isEmpty()) {
		    //      System.out.println("_cannels.isEmpty()");
		    _channelReaper = null;
		    break;
		}
		//System.out.println("_channles is not empty");
		if(chans == null || chans.length < _channels.size())
		    chans = new ClientChannel[_channels.size()];

		chans = (ClientChannel[])_channels.keySet().toArray(chans);
	    }

	    int age;
	    //System.out.println("chans.length = "+chans.length);
	    for(int i = 0; i < chans.length && chans[i] != null; ++i) {
		//  System.out.println("i = "+i);
		age=chans[i].channel_age();
		//System.out.println("Channel age is "+age+" and lasttime = "+lasttime);
		if(chans[i].state() == ClientChannel.STATE_CONNECTED && age < lasttime){
		    //  System.out.println("PAusing channel");
		    chans[i].pause();
		}
		chans[i] = null;
	    }

	    lasttime = RequestIDAllocator.get_request_id();
	    //System.out.println("Now lasttime is "+lasttime);
	}
    }
}
