/**
 * Redistribution and use of this software and associated
 * documentation ("Software"), with or without modification, are
 * permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright statements
 *    and notices.  Redistributions must also contain a copy of this
 *    document.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Intalio Inc.  For written permission, please
 *    contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Intalio Inc. Exolab is a registered trademark of
 *    Intalio Inc.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY INTALIO AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL INTALIO OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Copyright 2000 (C) Intalio Inc. All Rights Reserved.
 *
 * $Id: IdlToJava.java,v 1.20 2001/08/02 17:55:01 daniel Exp $
 *
 * Date         Author  Changes
 * 12/14        J.Daniel    Changed the helper generator to fully qualify the narrow method name
 *                          in the extract operation.
 */

 /** Modified/extended by Diana Szentivanyi in June 2003
 * contact diasz@ida.liu.se
*/

package org.openorb.compiler.generator;

import org.openorb.compiler.object.*;
import org.openorb.compiler.parser.*;

/**
 * This class generates all mapping for IDL descriptions.
 *
 * @author <a href="mailto:jdaniel@intalio.com">Jerome Daniel &lt;daniel@intalio.com&gt;</a>
 * @version $Revision: 1.20 $ $Date: 2001/08/02 17:55:01 $
 */
public class IdlToJava
{
	/**
	 * File path separator
	 */
	private String sep;

	/**
	 * Current package
	 */
	public String current_pkg = adaptToDot( org.openorb.compiler.IdlCompiler.packageName );

	/**
	 * Reference to the compilation graph
	 */
	public IdlObject _root = null;

	/**
	 * Reference to the initial directory
	 */
	private java.io.File initial = null;
     /* Field introduced by Diana on Aug. 31 2001*/
    private boolean application=false;


	/**
	 * Default constructor
	 */
	public IdlToJava(boolean application)
	{
		sep = System.getProperty("file.separator");
		this.application=application; /* Modified by Diana on aug. 31 2001*/
	}

	/**
	 * Retur true if a definition exists for a native type
	 *
	 * @param	obj	native object
	 * @return	true if a definition exists
	 */
	public boolean isNativeDefinition( IdlObject obj )
	{
		for ( int i=0; i<org.openorb.compiler.IdlCompiler.nativeDefinition.size(); i++ )
		{
			String s = ( String )org.openorb.compiler.IdlCompiler.nativeDefinition.elementAt(i);

			int index = s.indexOf(':');

			String word = s.substring(0,index);

			if ( obj.name().equals(word) )
				return true;
		}

		return false;
	}

	/**
	 * Print the type translated corresponding to a natibe type definition
	 *
	 * @param	obj		native object
	 * @param	output	file where definition is added
	 * @return	true if a definition exists
	 */
	public void printNativeDefinition( IdlObject obj, java.io.PrintWriter output )
	{
		for ( int i=0; i<org.openorb.compiler.IdlCompiler.nativeDefinition.size(); i++ )
		{
			String s = ( String )org.openorb.compiler.IdlCompiler.nativeDefinition.elementAt(i);

			int index = s.indexOf(':');

			String word = s.substring(0,index);

			if ( obj.name().equals(word) )
			{
				index = s.lastIndexOf(':');

				word = s.substring(index+1,s.length());

				output.print(word);
			}
		}
	}

	/**
	 * Allows to get an access on write to a target file
	 *
	 * @param	writeInto	Target file descriptor
	 * @return	write access
	 */
	public java.io.PrintWriter fileAccess( java.io.File writeInto )
	{
		// Deprecated
		// java.io.PrintStream printout = null;
		java.io.PrintWriter printout = null;

		try
		{
			java.io.FileOutputStream output = new java.io.FileOutputStream(writeInto);
			java.io.DataOutputStream dataout = new java.io.DataOutputStream(output);
			printout = new java.io.PrintWriter(dataout, true);
		}
		catch ( java.io.IOException e )
		{
			e.printStackTrace();
		}

		return printout;
	}

	/**
	 * Change a prefix order : omg.org -> org.omg
	 *
	 * @param	prefix		the prefix to inverse
	 * @return	the inversed prefix
	 */
	public String inversedPrefix ( String prefix )
	{
		int index = 0;
		int previous_index = 0;
		java.util.Vector seq = new java.util.Vector();
		String inversed = new String("");

		try
		{
			while ( index != -1 )
			{
				index = prefix.indexOf('.',previous_index);
				if ( index != -1 )
				{
					seq.addElement( new String(prefix.substring(previous_index,index)) );
					previous_index = index + 1;
				}
			}
		}
		catch ( StringIndexOutOfBoundsException ex )
		{ }

		seq.addElement( new String(prefix.substring(previous_index,prefix.length())) );

		for ( int i=seq.size()-1; i>=0; i-- )
		{
			if ( !inversed.equals("") )
				inversed = inversed + ".";

			inversed = inversed + (String)seq.elementAt(i);
		}

		return inversed;
	}

	/**
	 * Creates a new Java file
	 *
	 * @param	writeInto	the directory where the file must be created
	 * @param	name		file name ( without .java extension )
	 * @return	write access
	 */
	public java.io.PrintWriter newFile( java.io.File writeInto, String name )
	{
		String path;

		if ( !writeInto.exists() )
			writeInto.mkdirs();

		path = new String(writeInto.getPath() +  sep + name + ".java");

		java.io.File file = new java.io.File(path);

		return fileAccess(file);
	}

	/**
	 * Creates a directory
	 *
	 * @param	name	the directory name to create
	 * @param	writeInto	the directory in which the search has to appear
	 */
	public java.io.File createDirectory( String name, java.io.File writeInto )
	{
		String path;
		String fname;
		boolean init = false;

		char [] tab = new char [ name.length() + 20 ];

		int j = 0;
		for ( int i=0; i<name.length(); i++ )
		{
			if ( name.charAt(i) == '.' )
			{
				tab[j++] = sep.charAt(0);
				init = true;
			}
			else
				tab[j++] = name.charAt(i);
		}

		fname = new String(tab,0,j);

		if ( writeInto != null )
		{
			path = new String(writeInto.getPath() +  sep + fname);
		}
		else
			path = fname;

		java.io.File file = new java.io.File(path);

		if ( file.exists() == false )
			file.mkdirs();

		if ( init == true )
			org.openorb.compiler.IdlCompiler.packageName = fname;

		return file;
	}

	/**
	 * Creates the directories corresponding to a CORBA ID prefix
	 *
	 * @param	prefix		the prefix
	 * @param	writeInto	the directory in which the search has to appear
	 */
	public java.io.File createPrefixDirectories( String prefix, java.io.File writeInto )
	{
		String path;

		char [] tab = new char [ prefix.length() + 20 ];

		String name = null;
		if ( org.openorb.compiler.IdlCompiler.reversePrefix )
			name = new String( inversedPrefix(prefix) );
		else
			name = prefix;

		int j = 0;
		for ( int i=0; i<name.length(); i++ )
		{
			if ( name.charAt(i) == '.' )
				tab[j++] = sep.charAt(0);
			else
				tab[j++] = name.charAt(i);
		}

		name = new String(tab,0,j);

		if ( writeInto != null )
		{
			path = new String(writeInto.getPath() +  sep + name);
		}
		else
			path = name;

		java.io.File file = new java.io.File(path);

		if ( file.exists() == false )
			file.mkdirs();

		return file;
	}

	/**
	 * Get a write access in a directory
	 *
	 * @param	name the directory name
	 * @param	writeInto	the directory in which the search has to appear
	 */
	public java.io.File getDirectory( String name, java.io.File writeInto )
	{
		String path;

		if ( writeInto != null )
		{
			path = new String(writeInto.getPath() +  sep + name);
		}
		else
			path = name;

		java.io.File file = new java.io.File(path);

		return file;
	}

	/**
	 * Get a writing access to a directory corresponding to a CORBA ID prefix
	 *
	 * @param	prefix	a object prefix
	 * @param	writeInto	the directory in which the search has to appear
	 */
	public java.io.File getPrefixDirectories( String prefix, java.io.File writeInto )
	{
		String path;

		char [] tab = new char [ prefix.length() + 20 ];

		String name = null;
		if ( org.openorb.compiler.IdlCompiler.reversePrefix )
			name = new String( inversedPrefix(prefix) );
		else
			name = prefix;

		int j = 0;
		for ( int i=0; i<name.length(); i++ )
		{
			if ( name.charAt(i) == '.' )
				tab[j++] = sep.charAt(0);
			else
				tab[j++] = name.charAt(i);
		}

		name = new String(tab,0,j);

		if ( writeInto != null )
		{
			path = new String(writeInto.getPath() +  sep + name);
		}
		else
			path = name;

		java.io.File file = new java.io.File(path);

		return file;
	}

	/**
	 * Construct a package name
	 *
	 * @param	name	name of the superior level
	 */
	public void addToPkg( IdlObject obj, String name )
	{
		if ( org.openorb.compiler.IdlCompiler.use_package == false )
		{
			if ( !current_pkg.equals("generated") )
			{
				if ( !current_pkg.equals("") )
					current_pkg = current_pkg + "." + name;
				else
					current_pkg = name;
			}
			else
				current_pkg = name;
		}
		else
		{
			if ( !current_pkg.equals("") )
				current_pkg = current_pkg + "." + name;
			else
				current_pkg = name;
		}
	}

	/**
	 * Add a descriptive header in a Java file
	 *
	 * @param	output		the target file
	 * @param	obj	object which descriptive header has to be added
	 */
	public void cartouche ( java.io.PrintWriter output, IdlObject obj )
	{
		if ( current_pkg != null )
		{
			if ( current_pkg.equals("generated") )
			{
				if ( org.openorb.compiler.IdlCompiler.use_package == true )
				{
					output.println("package "+current_pkg+";");
					output.println("");
				}
			}
			else
			{
				if ( !current_pkg.equals("") )
				{
					output.println("package "+current_pkg+";");
					output.println("");
				}
			}
		}

		switch ( obj.kind() )
		{
		case IdlType.e_const :
			if ( obj.hasComment() )
				javadoc(output,obj);
			else
			{
				output.println("//");
				output.println("// Constant definition : " +obj.name());
				output.println("//");
				output.println("// @author OpenORB Compiler");
				output.println("//");
			}
			break;
		case IdlType.e_enum :
			if ( obj.hasComment() )
				javadoc(output,obj);
			else
			{
				output.println("//");
				output.println("// Enum definition : " +obj.name());
				output.println("//");
				output.println("// @author OpenORB Compiler");
				output.println("//");
			}
			break;
		case IdlType.e_struct :
			if ( obj.hasComment() )
				javadoc(output,obj);
			else
			{
				output.println("//");
				output.println("// Struct definition : " +obj.name());
				output.println("//");
				output.println("// @author OpenORB Compiler");
				output.println("//");
			}
			break;
		case IdlType.e_union :
			if ( obj.hasComment() )
				javadoc(output,obj);
			else
			{
				output.println("//");
				output.println("// Union definition : " +obj.name());
				output.println("//");
				output.println("// @author OpenORB Compiler");
				output.println("//");
			}
			break;
		case IdlType.e_exception :
			if ( obj.hasComment() )
				javadoc(output,obj);
			else
			{
				output.println("//");
				output.println("// Exception definition : " +obj.name());
				output.println("//");
				output.println("// @author OpenORB Compiler");
				output.println("//");
			}
			break;
		case IdlType.e_interface :
			if ( obj.hasComment() )
				javadoc(output,obj);
			else
			{
				output.println("//");
				output.println("// Interface definition : " +obj.name());
				output.println("//");
				output.println("// @author OpenORB Compiler");
				output.println("//");
			}
			break;
		case IdlType.e_value_box :
			if ( obj.hasComment() )
				javadoc(output,obj);
			else
			{
				output.println("//");
				output.println("// Value box definition : " +obj.name());
				output.println("//");
				output.println("// @author OpenORB Compiler");
				output.println("//");
			}
			break;
		case IdlType.e_value :
			if ( obj.hasComment() )
				javadoc(output,obj);
			else
			{
				output.println("//");
				output.println("// Value Type definition : " +obj.name());
				output.println("//");
				output.println("// @author OpenORB Compiler");
				output.println("//");
			}
			break;
		case IdlType.e_factory :
			if ( obj.hasComment() )
				javadoc(output,obj);
			else
			{
				output.println("//");
				output.println("// Factory definition : " +obj.name());
				output.println("//");
				output.println("// @author OpenORB Compiler");
				output.println("//");
			}
			break;
		}

	}

	/**
	 * Translate a JavaDoc comments section
	 */
	public void translate_comment_section( java.io.PrintWriter output, String description, IdlObject obj )
	{
		int i = 0;

		while ( i < description.length() )
		{
			if ( description.charAt(i) == '\n' )
			{
				if ( i != description.length()-1 )
				{
					output.println("");
					if ( ( obj.kind() == IdlType.e_attribute ) || ( obj.kind() == IdlType.e_operation ) || ( obj.kind() == IdlType.e_state_member ) )
						output.print("\t");
					output.print(" * ");
				}
				else
				{
					output.println("");
					return;
				}
			}
			else
				output.print( description.charAt(i) );

			i++;
		}
	}

	/**
	 * Add a JavaDoc comment
	 *
	 * @param	output		the target file
	 * @param	obj			the object the header has to be added
	 */
	public void javadoc ( java.io.PrintWriter output, IdlObject obj )
	{
		IdlComment comment = obj.getComment();
		String description = null;

		if ( comment != null )
		{
			description = comment.get_description();

			if ( ( obj.kind() == IdlType.e_attribute ) || ( obj.kind() == IdlType.e_operation ) || ( obj.kind() == IdlType.e_state_member ) )
				output.print("\t");

			output.println("/**");

			if ( ( obj.kind() == IdlType.e_attribute ) || ( obj.kind() == IdlType.e_operation ) || ( obj.kind() == IdlType.e_state_member ) )
				output.print("\t");

			output.print(" * ");

			translate_comment_section( output, description, obj );

			IdlCommentSection [] sections = comment.get_sections();

			for ( int i=0; i<sections.length; i++ )
			{
				switch ( sections[i].kind().value() )
				{
				case IdlCommentField._author_field :
					if ( ( obj.kind() == IdlType.e_attribute ) || ( obj.kind() == IdlType.e_operation ) )
						output.print("\t");
					output.print(" * @author ");
					break;
				case IdlCommentField._deprecated_field :
					if ( ( obj.kind() == IdlType.e_attribute ) || ( obj.kind() == IdlType.e_operation ) )
						output.print("\t");
					output.print(" * @deprecated ");
					break;
				case IdlCommentField._exception_field :
					if ( ( obj.kind() == IdlType.e_attribute ) || ( obj.kind() == IdlType.e_operation ) )
						output.print("\t");
					output.print(" * @exception ");
					break;
				case IdlCommentField._return_field :
					if ( ( obj.kind() == IdlType.e_attribute ) || ( obj.kind() == IdlType.e_operation ) )
						output.print("\t");
					output.print(" * @return ");
					break;
				case IdlCommentField._param_field :
					if ( ( obj.kind() == IdlType.e_attribute ) || ( obj.kind() == IdlType.e_operation ) )
						output.print("\t");
					output.print(" * @param ");
					break;
				case IdlCommentField._see_field :
					if ( ( obj.kind() == IdlType.e_attribute ) || ( obj.kind() == IdlType.e_operation ) )
						output.print("\t");
					output.print(" * @see ");
					break;
				case IdlCommentField._version_field :
					if ( ( obj.kind() == IdlType.e_attribute ) || ( obj.kind() == IdlType.e_operation ) )
						output.print("\t");
					output.print(" * @version ");
					break;
				case IdlCommentField._unknown_field :
					if ( ( obj.kind() == IdlType.e_attribute ) || ( obj.kind() == IdlType.e_operation ) )
						output.print("\t");
					output.print(" * @" + sections[i].get_title() + " ");
					break;
				}

				description = sections[i].get_description();
				translate_comment_section( output, description, obj);
			}

			if ( ( obj.kind() == IdlType.e_attribute ) || ( obj.kind() == IdlType.e_operation ) || ( obj.kind() == IdlType.e_state_member )  )
				output.print("\t");
			output.println(" */");
		}
	}

	/**
	 * Returns the final type of a data type
	 *
	 * @param	obj	the object
	 * @return	the final type
	 */
	public int final_kind( IdlObject obj )
	{
		switch ( obj.kind() )
		{
		case IdlType.e_ident :
			return final_kind(((IdlIdent)obj).internalObject());
		case IdlType.e_typedef :
		case IdlType.e_union_member :
		case IdlType.e_struct_member :
		case IdlType.e_param :
			return final_kind(obj.current());
		default :
			return obj.kind();
		}
	}

	/**
	 * Returns the final definition of a data type
	 *
	 * @param	obj	the object
	 * @return	the final definition
	 */
	public IdlObject final_type( IdlObject obj )
	{
		switch ( obj.kind() )
		{
		case IdlType.e_ident :
			return final_type(((IdlIdent)obj).internalObject());
		case IdlType.e_typedef :
		case IdlType.e_union_member :
		case IdlType.e_struct_member :
		case IdlType.e_param :
			return final_type(obj.current());
		default :
			return obj;
		}
	}

	/**
	 * This method replaces in a path the file separator by '.'
	 */
	private String adaptToDot( String path )
	{
		char [] tmp = new char[ path.length() ];

		for ( int i=0; i<path.length(); i++ )
		{
			if ( ( path.charAt(i) == '/' ) || ( path.charAt(i) == '\\' ) )
				tmp[i] = '.';
			else
				tmp[i] = path.charAt(i);
		}

		return new String(tmp);
	}

	/**
	 * Returns the complete name of a CORBA object
	 *
	 * @param	obj	the object the name has to be retrieved
	 * @return	the complete name
	 */
	public String fullname ( IdlObject obj )
	{
		java.util.Vector v = new java.util.Vector();
		IdlObject obj2 = obj;
		String name = new String("");
		String s;
		boolean first = false;

		while ( obj2 != null )
		{
			if ( first )
			{
				if ( ( obj2.kind() == IdlType.e_interface ) ||
						 ( obj2.kind() == IdlType.e_value ) ||
						 ( obj2.kind() == IdlType.e_struct ) ||
						 ( obj2.kind() == IdlType.e_union ) ||
						 ( obj2.kind() == IdlType.e_exception ) )
					v.addElement( (obj2.name()+"Package") );
				else
					v.addElement(obj2.name());
			}
			else
				if ( obj2.kind() != IdlType.e_union_member )
					v.addElement(obj2.name());

			if ( obj2.upper() != null )
				if ( obj2.upper().kind() == IdlType.e_root )
					break;
			obj2 = obj2.upper();
			first = true;
		}


		if ( org.openorb.compiler.IdlCompiler.packageName != null )
		{
			if ( !obj.included() )
			{
				if ( !org.openorb.compiler.IdlCompiler.packageName.equals("") )
				{
					if ( !((org.openorb.compiler.IdlCompiler.packageName.equals("generated")) && ( org.openorb.compiler.IdlCompiler.use_package == false )) )
						name = adaptToDot(org.openorb.compiler.IdlCompiler.packageName);
				}
			}
		}

		if ( org.openorb.compiler.IdlCompiler.usePrefix )
			if ( obj.getPrefix() != null )
			{
				if ( !name.equals("") )
					name = name + ".";
				if ( org.openorb.compiler.IdlCompiler.reversePrefix )
					name = name  + inversedPrefix(obj.getPrefix());
				else
					name = name  + obj.getPrefix();
			}

		for ( int i=v.size()-1; i>=0; i-- )
		{
			s = (String)v.elementAt(i);

			if ( s != null )
			{
				if ( !name.equals("")  )
					name = name + ".";

				name = name +  s;
			}
		}
		return name;
	}

	/**
	 * Delete the package of an id
	 */
	public String removePackageName( String expr )
	{
		int last = 0;
		int index = 0;
		String tmp = "";

		while ( true )
		{
			index = expr.indexOf("Package",last);
			if ( index == -1 )
			{
				tmp = tmp + expr.substring(last, expr.length());
				break;
			}
			else
			{
				tmp = tmp + expr.substring(last, index);
				last = index + 7;
			}
		}

		return tmp;
	}

	/**
	 * Check if the the id is an enum member
	 */
	public boolean isEnumCase( String expr )
	{
		boolean isEnum = false;

		if ( expr.indexOf("@") != -1 )
			isEnum = true;

		return isEnum;
	}

	/**
	 * Returns true if the element passed as argument is in the same
	 * scope as the second argument.
	 */
	public boolean isSameScope( String ident, IdlObject obj )
	{
		IdlObject obj2 = obj.upper().returnVisibleObject( ident, false );

		if ( obj2 == null )
			return false;

		if ( obj2.upper().equals( obj.upper() ) )
			return true;

		return false;
	}

	/**
	 * Translate a Scope IDL Ident::Ident into Scoped Java Ident.Ident.value
	 *
	 * @param	expr	the IDL expression
	 * @return	the equivalent Java expression
	 */
	public String IdlScopeToJavaScope( String expr, boolean complete, boolean fixed, IdlObject obj )
	{
		java.util.Vector s = new java.util.Vector();
		String mot = new String();
		int deb;
		boolean hexaValue;
		int last = 0;
		boolean isEnum = false;

		for ( int i=0; i<expr.length(); i++ )
		{
			if ( ( Character.isDigit( expr.charAt(i) ) ) ||
					 ( expr.charAt(i) == '-' ) )
			{
				if ( expr.charAt(i) == '-' )
				{
					s.addElement( new Character('-') );

					if ( expr.charAt(i++) == ' ' )
						continue;
				}

				hexaValue = false;
				while ( ( i != expr.length() ) && ( expr.charAt(i) != ' ' ) )
				{

					if ( expr.charAt(i) == 'x' )
						hexaValue = true;

					s.addElement( new Character( expr.charAt(i) ) );
					i++;
				}

				if ( !fixed )
					s.addElement( new Character('l') );
			}
			else
				if ( expr.charAt(i) == '\"' )
				{
					i++;
					s.addElement( new Character('\"') );
					boolean prev = false;
					boolean stop = false;
					while ( ( i != expr.length() ) && ( stop == false ) )
					{
						if ( ( expr.charAt(i) == '\"' ) && ( prev == false ) )
							stop = true;
						else
						{
							prev = false;
							if ( expr.charAt(i) == '\\' )
								prev = true;
							s.addElement( new Character( expr.charAt(i) ) );
							i++;
						}
					}
					s.addElement( new Character('\"') );
				}
				else
					if ( expr.charAt(i) == '\'' )
					{
						i++;
						s.addElement( new Character('\'') );
						boolean prev = false;
						boolean stop = false;
						while ( ( i != expr.length() ) && ( stop == false ) )
						{
							if ( ( expr.charAt(i) == '\'') && ( prev == false ) )
								stop = true;
							else
							{
								prev = false;
								if ( expr.charAt(i) == '\\' )
									prev = true;
								s.addElement( new Character( expr.charAt(i) ) );
								i++;
							}
						}
						s.addElement( new Character('\'') );
					}
					else
						if ( Character.isLetter(expr.charAt(i)) )
						{
							mot = "";
							deb = i;
							while ( ( i != expr.length() ) && ( expr.charAt(i) != ' ' ) )
							{
								if ( expr.charAt(i) == ':' )
								{
									i++;
									if ( i < expr.length() )
										if ( expr.charAt(i) == ':' )
										{
											s.addElement( new Character('.') );
											last = s.size();
											i++;
										}
										else
											s.addElement( new Character(':') );
								}

								if ( expr.charAt(i) == '.' )
									last = s.size() + 1;

								s.addElement( new Character( expr.charAt(i) ) );
								mot = mot + expr.charAt(i);
								i++;
							}
							if ( !( ( mot.equals("true") ) || ( mot.equals("false") ) ) )
							{
								isEnum = isEnumCase( expr );
								boolean isInSameScope = isSameScope( mot, obj );
								if ( isEnum )
									s.removeElementAt( s.size()-1 );
								if ( ( isEnum == false ) && ( isInSameScope == false ) )
								{
									if ( !mot.endsWith(".value") )
									{
										s.addElement( new Character('.') );
										s.addElement( new Character('v') );
										s.addElement( new Character('a') );
										s.addElement( new Character('l') );
										s.addElement( new Character('u') );
										s.addElement( new Character('e') );
									}
								}
								else
									if ( ( complete == true ) && ( isInSameScope == false ) )
									{
										s.insertElementAt( new Character('_'), last );
									}

							}
						}
						else
							if ( expr.charAt(i) != ' ' )
								s.addElement ( new Character(expr.charAt(i)) );
		}

		String newExpr = new String();

		for ( int i=0; i<s.size(); i++ )
			newExpr = newExpr + ((Character)s.elementAt(i)).charValue();

		return newExpr;
	}

	/**
	 * Allow to test if the character is an hexa number
	 *
	 * @param	c	the character to test
	 * @return	true if the character is an hexa number
	 */
	public boolean isHexaCar( char c )
	{
		switch ( c )
		{
		case '0' :
		case '1' :
		case '2' :
		case '3' :
		case '4' :
		case '5' :
		case '6' :
		case '7' :
		case '8' :
		case 'a' :
		case 'b' :
		case 'c' :
		case 'd' :
		case 'e' :
		case 'f' :
			return true;
		}

		return false;
	}

	/**
	 * Change the IDL escape characters into CORBA escape characters
	 *
	 * @param	expr	the IDL expression
	 * @return	the equivalent Java expression
	 */
	public String IdlEspaceCharToJavaEscapeChar( String expr )
	{
		java.util.Vector s = new java.util.Vector();

		for ( int i=0; i<expr.length(); i++ )
		{

			if ( expr.charAt(i) == '\\' )
			{
				s.addElement ( new Character('\\') );
				i++;

				switch ( expr.charAt(i) )
				{
				case '\\':
					break;
				case 'a' :
					s.addElement( new Character('0') );
					s.addElement( new Character('0') );
					s.addElement( new Character('7') );
					break;
				case 'v' :
					s.addElement( new Character('0') );
					s.addElement( new Character('1') );
					s.addElement( new Character('3') );
					break;
				case 'x' :
					i++;
					while ( isHexaCar( expr.charAt(i) ) )
					{
						i++;
					}
					s.addElement( new Character('3') );
					s.addElement( new Character('7') );
					s.addElement( new Character('7') );
				default :
					s.addElement( new Character(expr.charAt(i)) );
				}
			}
			else
				s.addElement ( new Character(expr.charAt(i)) );
		}

		String newExpr = new String();

		for ( int i=0; i<s.size(); i++ )
			newExpr = newExpr + ((Character)s.elementAt(i)).charValue();

		return newExpr;
	}

	/**
	 * Translate an IDL expression into a Java expression
	 *
	 * @param	expr	the IDL expression
	 * @return	the equivalent Java expression
	 */
	public String translate_to_java_expression( String expr, boolean fixed, IdlObject obj )
	{

		String newExpr = IdlScopeToJavaScope( expr, true, fixed, obj );

		newExpr = IdlEspaceCharToJavaEscapeChar( newExpr );

		if ( fixed )
			newExpr = "new java.math.BigDecimal( \"" + newExpr + "\" )";

		return newExpr;
	}

	/**
	 * Translate an IDL expression into an union expression
	 *
	 * @param	expr	the IDL expression
	 * @return	the equivalent Java expression
	 */
	public String translate_to_union_case_expression( IdlUnionMember disc, String expr )
	{
		String header = "";
		String newExpr = IdlScopeToJavaScope( expr, false, false, disc );
		newExpr = IdlEspaceCharToJavaEscapeChar( newExpr );

		disc.reset();
		switch ( final_type( disc.current() ).kind() )
		{
		case IdlType.e_simple :
			IdlSimple simple = ( IdlSimple )final_type( disc.current() );
			switch ( simple.internal() )
			{
			case Token.t_short :
			case Token.t_ushort :
				header = "short)";
				break;
			case Token.t_long :
			case Token.t_ulong :
				header = "int)";
				break;
			case Token.t_longlong :
			case Token.t_ulonglong :
				header = "long)";
				break;
			case Token.t_char :
			case Token.t_wchar :
				header = "char)";
				break;
			case Token.t_boolean :
				header =  "boolean)";
				break;
			case Token.t_octet :
				header = "byte)";
				break;
			}
			break;
		default :
			break;
		}

		if ( header.equals("") )
			return newExpr;
		else
			return "("+header + "(" + newExpr + ")";
	}

	/**
	 * Translate a data type
	 *
	 * @param	obj	the object to translate
	 * @param	output	the write access
	 */
	public void translate_type( IdlObject obj, java.io.PrintWriter output,boolean is_for_real )
	{
		IdlSimple simple = null;
		switch ( obj.kind() )
		{
		case IdlType.e_simple :
			simple = ( IdlSimple )obj;
			switch ( simple.internal() )
			{
			case Token.t_void :
				output.print("void");
				break;
			case Token.t_float :
				output.print("float");
				break;
			case Token.t_double :
				output.print("double");
				break;
			case Token.t_short :
			case Token.t_ushort :
				output.print("short");
				break;
			case Token.t_long :
			case Token.t_ulong :
				output.print("int");
				break;
			case Token.t_longlong :
			case Token.t_ulonglong :
				output.print("long");
				break;
			case Token.t_char :
			case Token.t_wchar :
				output.print("char");
				break;
			case Token.t_boolean :
				output.print("boolean");
				break;
			case Token.t_octet :
				output.print("byte");
				break;
			case Token.t_any :
				output.print("org.omg.CORBA.Any");
				break;
			case Token.t_typecode :
				output.print("org.omg.CORBA.TypeCode");
				break;
			case Token.t_object :
				output.print("org.omg.CORBA.Object");
				break;
			case Token.t_ValueBase :
				output.print("java.io.Serializable");
				break;
			}
			break;
		case IdlType.e_fixed :
			output.print("java.math.BigDecimal");
			break;
		case IdlType.e_string :
		case IdlType.e_wstring :
			output.print("String");
			break;
		case IdlType.e_struct :
		case IdlType.e_union :
		case IdlType.e_enum :
		case IdlType.e_interface :
		case IdlType.e_forward_interface :
		case IdlType.e_exception :
		case IdlType.e_value :
		case IdlType.e_forward_value :
			output.print(fullname(obj));
			break;
		case IdlType.e_native :
			if ( isNativeDefinition(obj) )
				printNativeDefinition(obj,output);
			else
				output.print(fullname(obj));
			break;
		case IdlType.e_typedef :
			obj.reset();
			translate_type(obj.current(),output,true);
			break;
		case IdlType.e_sequence :
		case IdlType.e_array :
		    if(is_for_real){

			translate_type(obj.current(),output,true);
			output.print("[]");
		    }
		    else{
			output.print(fullname(obj));
		    }
			break;
		case IdlType.e_ident :
		    //System.out.println("E ident first");
		    if(!is_for_real)
		    	output.print(fullname(((IdlIdent)obj).internalObject()));
		    else
			translate_type(((IdlIdent)obj).internalObject(),output,true);
		    break;
		case IdlType.e_value_box :
			if ( ((IdlValueBox)obj).simple() )
				output.print(fullname(obj));
			else
			{
				obj.reset();
				translate_type(obj.current(),output,true);
			}
			break;
		}
	}

	/**
	 * Translate a parameter
	 *
	 * @param	obj		param object to translate
	 * @param	output	write access
	 * @param	attr	parameter attribute
	 */
	public void translate_parameter( IdlObject obj, java.io.PrintWriter output, int attr )
	{
		IdlSimple simple = null;

		switch ( obj.kind() )
		{
		case IdlType.e_simple :
			simple = ( IdlSimple )obj;
			switch ( simple.internal() )
			{
			case Token.t_float :
				if ( attr == 0 )
					output.print("float");
				else
					output.print("org.omg.CORBA.FloatHolder");
				break;
			case Token.t_double :
				if ( attr == 0 )
					output.print("double");
				else
					output.print("org.omg.CORBA.DoubleHolder");
				break;
			case Token.t_short :
			case Token.t_ushort :
				if ( attr == 0 )
					output.print("short");
				else
					output.print("org.omg.CORBA.ShortHolder");
				break;
			case Token.t_long :
			case Token.t_ulong :
				if ( attr == 0 )
					output.print("int");
				else
					output.print("org.omg.CORBA.IntHolder");
				break;
			case Token.t_longlong :
			case Token.t_ulonglong :
				if ( attr == 0 )
					output.print("long");
				else
					output.print("org.omg.CORBA.LongHolder");
				break;
			case Token.t_char :
			case Token.t_wchar :
				if ( attr == 0 )
					output.print("char");
				else
					output.print("org.omg.CORBA.CharHolder");
				break;
			case Token.t_boolean :
				if ( attr == 0 )
					output.print("boolean");
				else
					output.print("org.omg.CORBA.BooleanHolder");
				break;
			case Token.t_octet :
				if ( attr == 0 )
					output.print("byte");
				else
					output.print("org.omg.CORBA.ByteHolder");
				break;
			case Token.t_any :
				if ( attr == 0 )
					output.print("org.omg.CORBA.Any");
				else
					output.print("org.omg.CORBA.AnyHolder");
				break;
			case Token.t_typecode :
				if ( attr == 0 )
					output.print("org.omg.CORBA.TypeCode");
				else
					output.print("org.omg.CORBA.TypeCodeHolder");
				break;
			case Token.t_object :
				if ( attr == 0 )
					output.print("org.omg.CORBA.Object");
				else
					output.print("org.omg.CORBA.ObjectHolder");
				break;
			case Token.t_ValueBase :
				if ( attr == 0 )
					output.print("java.io.Serializable");
				else
					output.print("org.omg.CORBA.ValueBaseHolder");
				break;
			}
			break;
		case IdlType.e_fixed :
			if ( attr == 0 )
				output.print("java.math.BigDecimal");
			else
				output.print("org.omg.CORBA.FixedHolder");
			break;
		case IdlType.e_string :
		case IdlType.e_wstring :
			if ( attr == 0 )
				output.print("String");
			else
				output.print("org.omg.CORBA.StringHolder");
			break;
		case IdlType.e_value_box :
			if ( attr == 0 )
			{
				if ( ((IdlValueBox)obj).simple() )
					output.print(fullname(obj));
				else
				{
					obj.reset();
					translate_parameter( obj.current() , output, attr );
				}
			}
			else
				output.print(fullname(obj)+"Holder");
			break;
		case IdlType.e_struct :
		case IdlType.e_union :
		case IdlType.e_enum :
		case IdlType.e_interface :
		case IdlType.e_forward_interface :
		case IdlType.e_value :
		case IdlType.e_forward_value :
			if ( attr == 0 )
				output.print(fullname(obj));
			else
				output.print(fullname(obj)+"Holder");
			break;
		case IdlType.e_native :
			if ( attr != 0 )
			{
                                /*
																	if ( isNativeDefinition(obj) )
																	{
																	printNativeDefinition(obj,output);
																	output.print("Holder");
																	}
																	else*/
				output.print(fullname(obj)+"Holder");
			}
			else
			{
				if ( isNativeDefinition(obj) )
					printNativeDefinition(obj,output);
				else
					output.print(fullname(obj));
			}
			break;
		case IdlType.e_typedef :
			obj.reset();
			if ( attr != 0 )
			{
                                /*
																	if ( ( final_type( obj ).kind() == IdlType.e_simple ) ||
																	( final_type( obj ).kind() == IdlType.e_string ) ||
																	( final_type( obj ).kind() == IdlType.e_wstring ) )
																	translate_parameter( obj.current(), output, attr );
																	else
																	output.print(fullname(obj)+"Holder");*/
				if ( ( final_type( obj ).kind() == IdlType.e_sequence ) ||
						 ( final_type( obj ).kind() == IdlType.e_array ) )
					output.print(fullname(obj)+"Holder");
				else
					translate_parameter( obj.current(), output, attr );
			}
			else
				translate_parameter(obj.current(),output,attr);
			break;
		case IdlType.e_sequence :
		case IdlType.e_array :
			if ( attr == 0 )
			{
				translate_parameter(obj.current(),output,attr);
				output.print("[]");
			}
			else
			{
				output.print(fullname(obj.upper())+"Holder");
			}
			break;
		case IdlType.e_ident :
			translate_parameter( ((IdlIdent)obj).internalObject(), output, attr);
			break;
		}
	}

	/**
	 * Find include depth for an object
	 *
	 * @param	obj object to test
	 * @return	include depth
	 */
	public int recursion( IdlObject obj )
	{
		String name = final_type(obj.current()).name();
		int i = 0;

		if ( final_type(obj.current()) instanceof IdlSimple )
			return 0;

		obj = obj.upper();
		while ( obj != null )
		{

			if ( final_type(obj).name() != null )
				if ( final_type(obj).name().equals( name ) )
					return i;

			i++;
			obj = obj.upper();
			if( obj == null )
				return i;

			if ( ( obj.kind() == IdlType.e_interface ) ||
					 ( obj.kind() == IdlType.e_module ) ||
					 ( obj.kind() == IdlType.e_root ) )
				break;
		}

		return 0;

	}

	/**
	 * Translate a TypeCode
	 *
	 * @param	obj	object to translate
	 * @param	output	write access
	 */
	public void translate_typecode ( IdlObject obj, java.io.PrintWriter output )
	{
		IdlSimple simple = null;

		switch ( obj.kind() )
		{
		case IdlType.e_simple :
			simple = ( IdlSimple )obj;
			output.print("orb.get_primitive_tc(org.omg.CORBA.TCKind.");
			switch ( simple.internal() )
			{
			case Token.t_float :
				output.print("tk_float)");
				break;
			case Token.t_double :
				output.print("tk_double)");
				break;
			case Token.t_short :
				output.print("tk_short)");
				break;
			case Token.t_ushort :
				output.print("tk_ushort)");
				break;
			case Token.t_long :
				output.print("tk_long)");
				break;
			case Token.t_ulong :
				output.print("tk_ulong)");
				break;
			case Token.t_longlong :
				output.print("tk_longlong)");
				break;
			case Token.t_ulonglong :
				output.print("tk_ulonglong)");
				break;
			case Token.t_char :
				output.print("tk_char)");
				break;
			case Token.t_wchar :
				output.print("tk_wchar)");
				break;
			case Token.t_boolean :
				output.print("tk_boolean)");
				break;
			case Token.t_octet :
				output.print("tk_octet)");
				break;
			case Token.t_any :
				output.print("tk_any)");
				break;
			case Token.t_typecode :
				output.print("tk_TypeCode)");
				break;
			case Token.t_object :
				output.print("tk_objref)");
				break;
			case Token.t_ValueBase :
				output.print("tk_value)");
				break;
			}
			break;
		case IdlType.e_fixed :
			output.print("orb.create_fixed_tc((short)"+((IdlFixed)obj).digits()+",(short)"+((IdlFixed)obj).scale()+")");
			break;
		case IdlType.e_string :
			if ( ((IdlString)obj).maxSize() == 0 )
				output.print("orb.get_primitive_tc( org.omg.CORBA.TCKind.tk_string )");
			else
				output.print("orb.create_string_tc("+((IdlString)obj).maxSize()+")");
			break;
		case IdlType.e_wstring :
			if ( ((IdlWString)obj).maxSize() == 0 )
				output.print("orb.get_primitive_tc( org.omg.CORBA.TCKind.tk_wstring )");
			else
				output.print("orb.create_wstring_tc("+((IdlWString)obj).maxSize()+")");
			break;
		case IdlType.e_struct :
		case IdlType.e_union :
		case IdlType.e_enum :
		case IdlType.e_interface :
		case IdlType.e_native :
		case IdlType.e_forward_value :
		case IdlType.e_forward_interface :
		case IdlType.e_value_box :
		case IdlType.e_value :
			output.print(fullname(obj));
			output.print("Helper.type()");
			break;
		case IdlType.e_typedef :
			output.print(fullname(obj));
			output.print("Helper.type()");
			break;
		case IdlType.e_sequence :
			/*	int rec = recursion(obj);
					if ( rec != 0 )
					{
					output.print("orb.create_recursive_sequence_tc(");
					output.print(((IdlSequence)obj).getSize()+","+rec);
					output.print(")");
					}
					else
					{*/
			output.print("orb.create_sequence_tc(");
			output.print(((IdlSequence)obj).getSize()+",");
			obj.reset();
			translate_typecode(obj.current(),output);
			output.print(")");
			//}
			break;
		case IdlType.e_array :
			output.print("orb.create_array_tc(");
			output.print(((IdlArray)obj).getDimension()+",");
			obj.reset();
			translate_typecode(obj.current(),output);
			output.print(")");
			break;
		case IdlType.e_ident :
			translate_typecode(((IdlIdent)obj).internalObject(),output);
			break;
		default :
			System.out.println("Unexpected... : " + obj.kind() );
			break;
		}
	}

	/**
	 * Translate an any insert
	 *
	 * @param	obj	object to insert
	 * @param	output	write access
	 * @param	aname	any type name
	 * @param	tname	type name to insert
	 */
	public void translate_any_insert ( IdlObject obj, java.io.PrintWriter output, String aname, String tname )
	{
		IdlSimple simple = null;

		switch ( obj.kind() )
		{
		case IdlType.e_simple :
			simple = ( IdlSimple )obj;
			if ( simple.internal() == Token.t_ValueBase )
			{
				output.print("org.omg.CORBA.portable.ValueBaseHelper.insert("+aname+",(org.omg.CORBA.portable.ValueBase)"+tname+")");
				return;
			}
			output.print(aname+".insert_");
			switch ( simple.internal() )
			{
			case Token.t_float :
				output.print("float((float)"+tname+")");
				break;
			case Token.t_double :
				output.print("double((double)"+tname+")");
				break;
			case Token.t_short :
				output.print("short((short)"+tname+")");
				break;
			case Token.t_ushort :
				output.print("ushort((short)"+tname+")");
				break;
			case Token.t_long :
				output.print("long((int)"+tname+")");
				break;
			case Token.t_ulong :
				output.print("ulong((int)"+tname+")");
				break;
			case Token.t_longlong :
				output.print("longlong((long)"+tname+")");
				break;
			case Token.t_ulonglong :
				output.print("ulonglong((long)"+tname+")");
				break;
			case Token.t_char :
				output.print("char("+tname+")");
				break;
			case Token.t_wchar :
				output.print("wchar("+tname+")");
				break;
			case Token.t_boolean :
				output.print("boolean("+tname+")");
				break;
			case Token.t_octet :
				output.print("octet((byte)"+tname+")");
				break;
			case Token.t_any :
				output.print("any("+tname+")");
				break;
			case Token.t_typecode :
				output.print("TypeCode("+tname+")");
				break;
			case Token.t_object :
				output.print("Object("+tname+")");
				break;
			}
			break;
		case IdlType.e_fixed :
			output.print(aname+".insert_fixed("+tname+")");
			break;
		case IdlType.e_string :
			output.print(aname+".insert_string("+tname+")");
			break;
		case IdlType.e_wstring :
			output.print(aname+".insert_wstring("+tname+")");
			break;
		case IdlType.e_native :
		case IdlType.e_interface :
		case IdlType.e_forward_interface :
		case IdlType.e_struct :
		case IdlType.e_union :
		case IdlType.e_enum :
		case IdlType.e_value_box :
		case IdlType.e_value :
			output.print(fullname(obj));
			output.print("Helper.insert("+aname+","+tname+")");
			break;
		case IdlType.e_typedef :
			output.print(fullname(obj));
			output.print("Helper.insert("+aname+","+tname+")");
			break;
		case IdlType.e_sequence :
		case IdlType.e_array :
			break;
		case IdlType.e_ident :
			translate_any_insert(((IdlIdent)obj).internalObject(),output,aname,tname);
			break;
		}
	}

	/**
	 * This function looks for a truncatable interface in value type
	 * inheritance
	 */
	public boolean isTruncatable( IdlObject obj )
	{
		IdlValue value = (IdlValue) obj;

		java.util.Vector inheritance = value.getInheritanceList();

		for ( int i=0; i<inheritance.size(); i++ )
		{
			if ( ((IdlValueInheritance)(inheritance.elementAt(i))).truncatable_member() )
				return true;
		}

		return false;
	}

	/**
	 * This function prints the concrete base value type TypeCode
	 */
	public void printConcreteTypeCode( IdlObject obj, java.io.PrintWriter output )
	{
		IdlValue value = (IdlValue) obj;

		IdlValue [] inheritance = value.getInheritance();

		for ( int i=0; i<inheritance.length; i++ )
		{
			if ( inheritance[i].abstract_value() == false )
			{
				translate_typecode(inheritance[i],output);
				return;
			}
		}

		output.print("orb.get_primitive_tc( org.omg.CORBA.TCKind.tk_null )");
	}

	/**
	 * Construct a new TypeCode
	 *
	 * @param	obj	the typecode
	 * @param	output write access
	 */
	public void translate_new_typecode ( IdlObject obj, java.io.PrintWriter output )
	{
		int i;
		int idx;

		switch ( obj.kind() )
		{
		case IdlType.e_enum :
			output.println("\t\t\torg.omg.CORBA.ORB orb = org.omg.CORBA.ORB.init();");
			output.println("\t\t\tString []_members = new String["+obj.length()+"];");

			obj.reset();
			i = 0;
			while ( obj.end() != true )
			{
				output.println("\t\t\t_members["+i+"] = \"" + obj.current().name() + "\";");
				i++;
				obj.next();
			}

			output.println("\t\t\t_tc = orb.create_enum_tc(id(),\""+obj.name()+"\",_members);");
			break;
		case IdlType.e_struct :
			output.println("\t\t\t\torg.omg.CORBA.ORB orb = org.omg.CORBA.ORB.init();");
			output.println("\t\t\t\torg.omg.CORBA.StructMember []_members = new org.omg.CORBA.StructMember["+obj.length()+"];");
			output.println("");

			obj.reset();
			i = 0;
			while ( obj.end() != true )
			{
				output.println("\t\t\t\t_members["+i+"] = new org.omg.CORBA.StructMember();");

				output.println("\t\t\t\t_members["+i+"].name = \"" +obj.current().name()+"\";");

				output.print("\t\t\t\t_members["+i+"].type = ");

				obj.current().reset();
				translate_typecode(obj.current().current(),output);

				output.println(";");

				obj.next();
				i++;
			}

			output.println("\t\t\t\t_tc = orb.create_struct_tc(id(),\""+obj.name()+"\",_members);");
			break;
		case IdlType.e_union :
			output.println("\t\t\t\torg.omg.CORBA.ORB orb = org.omg.CORBA.ORB.init();");
			output.println("\t\t\t\torg.omg.CORBA.UnionMember []_members = new org.omg.CORBA.UnionMember["+(obj.length()-1)+"];");
			output.println("\t\t\t\torg.omg.CORBA.Any any;");
			output.println("");

			idx = ((IdlUnion)obj).index();

			obj.reset();
			IdlUnionMember disc = ( IdlUnionMember ) obj.current();
			disc.reset();
			obj.next();
			i = 0;
			while ( obj.end() != true )
			{
				output.println("\t\t\t\tany = orb.create_any();");
				if ( i != idx )
				{
					output.print("\t\t\t\t");
					translate_any_insert(disc.current(),output,"any",translate_to_union_case_expression(disc,((IdlUnionMember)obj.current()).getExpression()));
					output.println(";");
				}
				else
					output.println("\t\t\t\tany.insert_octet((byte)0);");

				output.println("\t\t\t\t_members["+i+"] = new org.omg.CORBA.UnionMember();");
				output.println("\t\t\t\t_members["+i+"].name = \"" +obj.current().name()+"\";");
				output.println("\t\t\t\t_members["+i+"].label = any;");
				output.print("\t\t\t\t_members["+i+"].type = ");

				obj.current().reset();
				translate_typecode(obj.current().current(),output);

				output.println(";");

				obj.next();
				i++;
			}

			obj.reset();
			output.print("\t\t\t\t_tc = orb.create_union_tc(id(),\""+obj.name()+"\",");
			translate_typecode(obj.current().current(),output);
			output.println(",_members);");
			break;
		case IdlType.e_typedef :
			output.println("\t\t\torg.omg.CORBA.ORB orb = org.omg.CORBA.ORB.init();");
			output.print("\t\t\t_tc = orb.create_alias_tc(id(),\""+obj.name()+"\",");

			obj.reset();
			translate_typecode(obj.current(),output);

			output.println(");");
			break;
		case IdlType.e_exception :
			output.println("\t\t\t\torg.omg.CORBA.ORB orb = org.omg.CORBA.ORB.init();");
			output.println("\t\t\t\torg.omg.CORBA.StructMember []_members = new org.omg.CORBA.StructMember["+obj.length()+"];");
			output.println("");

			obj.reset();
			i = 0;
			while ( obj.end() != true )
			{
				output.println("\t\t\t\t_members["+i+"] = new org.omg.CORBA.StructMember();");

				output.println("\t\t\t\t_members["+i+"].name = \"" +obj.current().name()+"\";");

				output.print("\t\t\t\t_members["+i+"].type = ");

				obj.current().reset();
				translate_typecode(obj.current().current(),output);

				output.println(";");

				obj.next();
				i++;
			}

			output.println("\t\t\t\t_tc = orb.create_exception_tc(id(),\""+obj.name()+"\",_members);");
			break;
		case IdlType.e_interface :
		case IdlType.e_forward_interface :
			output.println("\t\t\torg.omg.CORBA.ORB orb = org.omg.CORBA.ORB.init();");
			if ( ((IdlInterface)obj).abstract_interface() )
				output.println("\t\t\t_tc = orb.create_abstract_interface_tc(id(),\""+obj.name()+"\");");
			else
				output.println("\t\t\t_tc = orb.create_interface_tc(id(),\""+obj.name()+"\");");
			break;
		case IdlType.e_native :
			output.println("\t\t\torg.omg.CORBA.ORB orb = org.omg.CORBA.ORB.init();");
			output.println("\t\t\t_tc = orb.create_native_tc(id(),\""+obj.name()+"\");");
			break;
		case IdlType.e_value_box :
			output.println("\t\t\torg.omg.CORBA.ORB orb = org.omg.CORBA.ORB.init();");
			output.print("\t\t\t_tc = orb.create_value_box_tc(id(),\""+obj.name()+"\",");
			obj.reset();
			translate_typecode(obj.current(),output);
			output.println(");");
			break;
		case IdlType.e_value :
			output.println("\t\t\t\torg.omg.CORBA.ORB orb = org.omg.CORBA.ORB.init();");

			int nbvalue = 0;

			obj.reset();
			while ( obj.end() != true )
			{
				if ( obj.current().kind() == IdlType.e_state_member )
					nbvalue++;

				obj.next();
			}

			output.println("\t\t\t\torg.omg.CORBA.ValueMember []_members = new org.omg.CORBA.ValueMember["+nbvalue+"];");
			output.println("");

			obj.reset();
			i = 0;
			while ( obj.end() != true )
			{
				if ( obj.current().kind() == IdlType.e_state_member )
				{
					output.println("\t\t\t\t_members["+i+"] = new org.omg.CORBA.ValueMember();");

					output.println("\t\t\t\t_members["+i+"].name = \"" +obj.current().name()+"\";");

					output.print("\t\t\t\t_members["+i+"].type = ");

					obj.current().reset();
					translate_typecode(obj.current().current(),output);

					output.println(";");

					output.print("\t\t\t\t_members["+i+"].access = ");

					if ( ((IdlStateMember)(obj.current())).public_member() )
						output.println("org.omg.CORBA.PUBLIC_MEMBER.value;");
					else
						output.println("org.omg.CORBA.PRIVATE_MEMBER.value;");
					i++;
				}

				obj.next();
			}

			output.println("");
			output.print("\t\t\t\torg.omg.CORBA.TypeCode _concrete_tc = ");

			printConcreteTypeCode(obj,output);

			output.println(";");
			output.println("");

			output.print("\t\t\t\t_tc = orb.create_value_tc(id(),\""+obj.name()+"\",");

			if ( ((IdlValue)obj).abstract_value() )
				output.print("org.omg.CORBA.VM_ABSTRACT.value");
			else
				if ( ((IdlValue)obj).custom_value() )
					output.print("org.omg.CORBA.VM_CUSTOM.value");
				else
					if ( isTruncatable(obj) )
						output.print("org.omg.CORBA.VM_TRUNCATABLE.value");
					else
						output.print("org.omg.CORBA.VM_NONE.value");

			output.println(",_concrete_tc,_members);");
			break;
		}
	}

	/**
	 * Return simple array name or null else
	 */
	public String get_array_name( IdlObject obj )
	{
		switch ( final_kind(obj) )
		{
		case IdlType.e_simple :
			IdlSimple simple = ( IdlSimple )final_type(obj);
			switch ( simple.internal() )
			{
			case Token.t_float :
				return "float";
			case Token.t_double :
				return "double";
			case Token.t_short :
				return "short";
			case Token.t_ushort :
				return "ushort";
			case Token.t_long :
				return "long";
			case Token.t_ulong :
				return "ulong";
			case Token.t_longlong :
				return "longlong";
			case Token.t_ulonglong :
				return "ulonglong";
			case Token.t_char :
				return "char";
			case Token.t_wchar :
				return "wchar";
			case Token.t_boolean :
				return "boolean";
			case Token.t_octet :
				return "octet";
			default :
				return null;
			}
		default :
			return null;
		}
	}

	/**
	 * Encode a member data type
	 *
	 * @param	obj member to encode
	 * @param	output	write access
	 * @param	outname	outputstream name
	 * @param	tname	data type name
	 * @param	space	indent space
	 */
	public void translate_marshalling_member ( IdlObject obj, java.io.PrintWriter output, String outname, String tname, String space )
	{
		IdlSimple simple = null;
		String array_name = null;
		int val;

		switch ( obj.kind() )
		{
		case IdlType.e_simple :
			simple = ( IdlSimple )obj;
			if ( simple.internal() == Token.t_ValueBase )
			{
				output.println(space+"((org.omg.CORBA_2_3.portable.OutputStream)"+outname+").write_value( "+tname+" );");
				return;
			}
			output.print(space+outname+".write");
			switch ( simple.internal() )
			{
			case Token.t_float :
				output.println("_float("+tname+");");
				break;
			case Token.t_double :
				output.println("_double("+tname+");");
				break;
			case Token.t_short :
				output.println("_short("+tname+");");
				break;
			case Token.t_ushort :
				output.println("_ushort("+tname+");");
				break;
			case Token.t_long :
				output.println("_long("+tname+");");
				break;
			case Token.t_ulong :
				output.println("_ulong("+tname+");");
				break;
			case Token.t_longlong :
				output.println("_longlong("+tname+");");
				break;
			case Token.t_ulonglong :
				output.println("_ulonglong("+tname+");");
				break;
			case Token.t_char :
				output.println("_char("+tname+");");
				break;
			case Token.t_wchar :
				output.println("_wchar("+tname+");");
				break;
			case Token.t_boolean :
				output.println("_boolean("+tname+");");
				break;
			case Token.t_octet :
				output.println("_octet("+tname+");");
				break;
			case Token.t_any :
				output.println("_any("+tname+");");
				break;
			case Token.t_typecode :
				output.println("_TypeCode("+tname+");");
				break;
			case Token.t_object :
				output.println("_Object("+tname+");");
				break;
			}
			break;
		case IdlType.e_fixed :
			output.println(space+outname+".write_fixed("+tname+");");
			break;
		case IdlType.e_string :
			output.println(space+outname+".write_string("+tname+");");
			break;
		case IdlType.e_wstring :
			output.println(space+outname+".write_wstring("+tname+");");
			break;
		case IdlType.e_native :
		case IdlType.e_interface :
		case IdlType.e_forward_interface :
		case IdlType.e_struct :
		case IdlType.e_union :
		case IdlType.e_enum :
		case IdlType.e_value_box :
		case IdlType.e_value :
		case IdlType.e_forward_value :
			output.print(space+fullname(obj));
			output.println("Helper.write("+outname+","+tname+");");
			break;
		case IdlType.e_typedef :
			output.print(space+fullname(obj));
			output.println("Helper.write("+outname+","+tname+");");
			break;
		case IdlType.e_array :
			array_name = get_array_name( obj.current() );
			if ( array_name == null )
			{
				val = space.length()-1;
				output.println(space+"if ( "+tname+".length != "+ ((IdlArray)obj).getDimension() + " ) ");
				output.println(space+"   throw new org.omg.CORBA.MARSHAL();");
				output.println(space+"for ( int i"+ val +"=0; i"+ val +"<"+tname+".length; i"+val+"++)");
				output.println(space+"{");
				translate_marshalling_member(obj.current(),output,outname,tname+"[i"+val+"]",space+"\t");
				output.println("");
				output.println(space+"}");
			}
			else
			{
				output.println(space+outname+".write_"+ array_name + "_array("+tname+", 0," + tname + ".length);");
			}
			break;
		case IdlType.e_sequence :
			array_name = get_array_name( obj.current() );
			if ( array_name == null )
			{
				val = space.length()-1;
				output.println(space+outname+".write_ulong("+tname+".length);");
				output.println(space+"for ( int i"+ val +"=0; i"+ val +"<"+tname+".length; i"+val+"++)");
				output.println(space+"{");
				translate_marshalling_member(obj.current(),output,outname,tname+"[i"+val+"]",space+"\t");
				output.println("");
				output.println(space+"}");
			}
			else
			{
				output.println(space+outname+".write_ulong("+tname+".length);");
				output.println(space+outname+".write_"+ array_name + "_array("+tname+", 0," + tname + ".length);");
			}
			break;
		case IdlType.e_ident :
			translate_marshalling_member(((IdlIdent)obj).internalObject(),output,outname,tname,space);
			break;
		}
	}

	/**
	 * Encode a member data type
	 *
	 * @param	obj	the member to encode
	 * @param	output	write access
	 * @param	inname	inputstream name
	 * @param	tname	data type name
	 * @param	space	indent space
	 */
	public void translate_unmarshalling_member ( IdlObject obj, java.io.PrintWriter output, String inname, String tname, String space )
	{
		IdlSimple simple = null;
		String array_name = null;
		IdlObject o = null;
		int val;
		int next;

		switch ( obj.kind() )
		{
		case IdlType.e_simple :
			simple = ( IdlSimple )obj;
			if ( simple.internal() == Token.t_ValueBase )
			{
				output.println(space+tname+" = ((org.omg.CORBA_2_3.portable.InputStream)"+inname+").read_value();");
				return;
			}
			output.print(space+tname+" = "+inname+".read");
			switch ( simple.internal() )
			{
			case Token.t_float :
				output.println("_float();");
				break;
			case Token.t_double :
				output.println("_double();");
				break;
			case Token.t_short :
				output.println("_short();");
				break;
			case Token.t_ushort :
				output.println("_ushort();");
				break;
			case Token.t_long :
				output.println("_long();");
				break;
			case Token.t_ulong :
				output.println("_ulong();");
				break;
			case Token.t_longlong :
				output.println("_longlong();");
				break;
			case Token.t_ulonglong :
				output.println("_ulonglong();");
				break;
			case Token.t_char :
				output.println("_char();");
				break;
			case Token.t_wchar :
				output.println("_wchar();");
				break;
			case Token.t_boolean :
				output.println("_boolean();");
				break;
			case Token.t_octet :
				output.println("_octet();");
				break;
			case Token.t_any :
				output.println("_any();");
				break;
			case Token.t_typecode :
				output.println("_TypeCode();");
				break;
			case Token.t_object :
				output.println("_Object();");
				break;
			case Token.t_ValueBase :
				output.println("_value();");
				break;
			}
			break;
		case IdlType.e_fixed :
			output.println(space+tname+" ="+inname+".read_fixed();");
			break;
		case IdlType.e_string :
			output.println(space+tname+" = "+inname+".read_string();");
			break;
		case IdlType.e_wstring :
			output.println(space+tname+" = "+inname+".read_wstring();");
			break;
		case IdlType.e_struct :
		case IdlType.e_union :
		case IdlType.e_enum :
		case IdlType.e_native :
		case IdlType.e_interface :
		case IdlType.e_forward_interface :
		case IdlType.e_value_box :
		case IdlType.e_value :
		case IdlType.e_forward_value :
			output.print(space+tname+" = "+fullname(obj));
			output.println("Helper.read("+inname+");");
			break;
		case IdlType.e_typedef :
			output.print(space+tname+" = "+fullname(obj));
			output.println("Helper.read("+inname+");");
			break;
		case IdlType.e_array :
			val = space.length()-1;
			output.println(space+"{");
			output.println(space+"int size"+val+" = "+((IdlArray)obj).getDimension()+";");
			output.print(space+tname+" = new ");

			obj.reset();
			o = final_type( obj.current() );
			next = 0;
			while ( ( o.kind() == IdlType.e_array ) ||
							( o.kind() == IdlType.e_sequence ) )
			{
				o.reset();
				next++;
				o = final_type(o.current());
			}
			translate_type(o,output,true);

			output.print("[size"+val+"]");

			for ( int i=0; i<next; i++ )
				output.print("[]");

			output.println(";");

			array_name = get_array_name( obj.current() );
			if ( array_name == null )
			{
				output.println(space+"for ( int i"+ val +"=0; i"+ val +"<"+tname+".length; i"+val+"++)");
				output.println(space+" {");
				translate_unmarshalling_member(obj.current(),output,inname,tname+"[i"+val+"]",space+"\t");
				output.println("");
				output.println(space+" }");
			}
			else
			{
				output.println(space+inname+".read_" + array_name + "_array(" + tname + ", 0, " + tname + ".length);");
			}
			output.println(space+"}");
			break;
		case IdlType.e_sequence :
			val = space.length()-1;
			output.println(space+"{");
			output.println(space+"int size"+val+" = "+inname+".read_ulong();");
			output.print(space+tname+" = new ");

			obj.reset();
			o = final_type( obj.current() );
			next = 0;
			while ( ( o.kind() == IdlType.e_array ) ||
							( o.kind() == IdlType.e_sequence ) )
			{
				o.reset();
				next++;
				o = final_type(o.current());
			}
			translate_type(o,output,true);

			output.print("[size"+val+"]");

			for ( int i=0; i<next; i++ )
				output.print("[]");

			output.println(";");

			array_name = get_array_name( obj.current() );
			if ( array_name == null )
			{
				output.println(space+"for ( int i"+ val +"=0; i"+ val +"<"+tname+".length; i"+val+"++)");
				output.println(space+" {");
				translate_unmarshalling_member(obj.current(),output,inname,tname+"[i"+val+"]",space+"\t");
				output.println("");
				output.println(space+" }");
			}
			else
			{
				output.println(space+inname+".read_" + array_name + "_array(" + tname + ", 0, " + tname + ".length);");
			}
			output.println(space+"}");
			break;
		case IdlType.e_ident :
			translate_unmarshalling_member(((IdlIdent)obj).internalObject(),output,inname,tname,space);
			break;
		}
	}

	/**
	 * Encode a data type
	 *
	 * @param	obj the member to encode
	 * @param	output	write access
	 * @param	outname	outputstream	name
	 * @param   tname data type name
	 */
	public void translate_marshalling_data ( IdlObject obj, java.io.PrintWriter output, String outname, String tname )
	{
		IdlSimple simple = null;

		switch ( obj.kind() )
		{
		case IdlType.e_simple :
			simple = ( IdlSimple )obj;
			if ( simple.internal() == Token.t_ValueBase )
			{
				output.println("((org.omg.CORBA_2_3.portable.OutputStream)"+outname+").write_value( " + tname +" );");
				return;
			}
			output.print(outname+".write");
			switch ( simple.internal() )
			{
			case Token.t_float :
				output.print("_float");
				break;
			case Token.t_double :
				output.print("_double");
				break;
			case Token.t_short :
				output.print("_short");
				break;
			case Token.t_ushort :
				output.print("_ushort");
				break;
			case Token.t_long :
				output.print("_long");
				break;
			case Token.t_ulong :
				output.print("_ulong");
				break;
			case Token.t_longlong :
				output.print("_longlong");
				break;
			case Token.t_ulonglong :
				output.print("_ulonglong");
				break;
			case Token.t_char :
				output.print("_char");
				break;
			case Token.t_wchar :
				output.print("_wchar");
				break;
			case Token.t_boolean :
				output.print("_boolean");
				break;
			case Token.t_octet :
				output.print("_octet");
				break;
			case Token.t_any :
				output.print("_any");
				break;
			case Token.t_typecode :
				output.print("_TypeCode");
				break;
			case Token.t_object :
				output.print("_Object");
				break;
			case Token.t_ValueBase :
				output.print("_value");
				break;
			}
			output.println("("+tname+");");
			break;
		case IdlType.e_fixed :
			output.println(outname+".write_fixed("+tname+");");
			break;
		case IdlType.e_string :
			output.println(outname+".write_string("+tname+");");
			break;
		case IdlType.e_wstring :
			output.println(outname+".write_wstring("+tname+");");
			break;
		case IdlType.e_struct :
		case IdlType.e_union :
		case IdlType.e_enum :
		case IdlType.e_interface :
		case IdlType.e_native :
		case IdlType.e_forward_interface :
		case IdlType.e_value_box :
		case IdlType.e_value :
		case IdlType.e_forward_value :
			output.print(fullname(obj));
			output.println("Helper.write("+outname+","+tname+");");
			break;
		case IdlType.e_typedef :
			output.print(fullname(obj));
			output.println("Helper.write("+outname+","+tname+");");
			break;
		case IdlType.e_array :
		case IdlType.e_sequence :
			break;
		case IdlType.e_ident :
			translate_marshalling_data(((IdlIdent)obj).internalObject(),output,outname,tname);
			break;
		}
	}

	/**
	 * Decode a data type
	 *
	 * @param	obj	the member to decode
	 * @param	output	write access
	 * @param	inname	inputstream	name
	 */
	public void translate_unmarshalling_data ( IdlObject obj, java.io.PrintWriter output, String inname )
	{
		IdlSimple simple = null;

		switch ( obj.kind() )
		{
		case IdlType.e_simple :
			simple = ( IdlSimple )obj;
			if ( simple.internal() == Token.t_ValueBase )
			{
				output.println("((org.omg.CORBA_2_3.portable.InputStream)"+inname+").read_value();");
				return;
			}
			output.print(inname+".read");
			switch ( simple.internal() )
			{
			case Token.t_float :
				output.println("_float();");
				break;
			case Token.t_double :
				output.println("_double();");
				break;
			case Token.t_short :
				output.println("_short();");
				break;
			case Token.t_ushort :
				output.println("_ushort();");
				break;
			case Token.t_long :
				output.println("_long();");
				break;
			case Token.t_ulong :
				output.println("_ulong();");
				break;
			case Token.t_longlong :
				output.println("_longlong();");
				break;
			case Token.t_ulonglong :
				output.println("_ulonglong();");
				break;
			case Token.t_char :
				output.println("_char();");
				break;
			case Token.t_wchar :
				output.println("_wchar();");
				break;
			case Token.t_boolean :
				output.println("_boolean();");
				break;
			case Token.t_octet :
				output.println("_octet();");
				break;
			case Token.t_any :
				output.println("_any();");
				break;
			case Token.t_typecode :
				output.println("_TypeCode();");
				break;
			case Token.t_object :
				output.println("_Object();");
				break;
			}
			break;
		case IdlType.e_fixed :
			output.println(inname+".read_fixed();");
			break;
		case IdlType.e_string :
			output.println(inname+".read_string();");
			break;
		case IdlType.e_wstring :
			output.println(inname+".read_wstring();");
			break;
		case IdlType.e_struct :
		case IdlType.e_union :
		case IdlType.e_enum :
		case IdlType.e_interface :
		case IdlType.e_native :
		case IdlType.e_forward_interface :
		case IdlType.e_value_box :
		case IdlType.e_value :
		case IdlType.e_forward_value :
			output.print(fullname(obj));
			//System.out.println("it is struct, union, enum,interf,native, fw_int,value_b,value,fw_v "+fullname(obj)+" "+obj.kind());
			output.println("Helper.read("+inname+");");
			break;
		case IdlType.e_typedef :
			output.print(fullname(obj));
			//System.out.println("it is typedef "+fullname(obj)+" "+obj.kind());
			output.println("Helper.read("+inname+");");
			break;
		case IdlType.e_array :
		case IdlType.e_sequence :
			/*
				int val = space.length()-1;
				output.println(space+"int size"+val+" = "+inname+".read_ulong();");
				output.print(" = new ");

				obj.reset();
				translate_type(obj.current(),output);

				output.println("[size"+val+"];");

				output.println(space+"for ( int i"+ val +"=0; i"+ val +"<"+tname+".length; i"+val+"++)");
				output.println(space+"{");
				translate_unmarshalling_member(obj.current(),output,inname,tname+"[i"+val+"]",space+"\t");
				output.println("");
				output.println(space+"}");
			*/
			break;
		case IdlType.e_ident :
			translate_unmarshalling_data(((IdlIdent)obj).internalObject(),output,inname);
			break;
		}
	}

	/**
	 * Encode a data type
	 *
	 * @param	obj	the data type to encode
	 * @param	output	write access
	 * @param	outname	outputstream name
	 * @param	tname	data type name
	 */
	public void translate_marshalling ( IdlObject obj, java.io.PrintWriter output, String outname, String tname )
	{
		int i;
		int idx;

		switch ( obj.kind() )
		{
		case IdlType.e_simple :
			if ( ((IdlSimple)obj).internal() == Token.t_typecode )
				output.println("\t\t"+outname+".write_TypeCode(" + tname + ");");
			break;
		case IdlType.e_enum :
			output.println("\t\t"+outname+".write_ulong("+tname+".value());");
			break;
		case IdlType.e_struct :
			obj.reset();
			while ( obj.end() != true )
			{
				obj.current().reset();
				translate_marshalling_member(obj.current().current(),output,outname,tname+"."+obj.current().name(),"\t\t");
				obj.next();
			}
			break;
		case IdlType.e_union :
			idx = ((IdlUnion)obj).index();
			obj.reset();

			obj.current().reset();
			boolean enumeration = false;
			if ( final_kind( obj.current().current() ) == IdlType.e_enum )
				enumeration = true;

			IdlObject d = obj.current().current();

			translate_marshalling_member(obj.current().current(),output,outname,tname+"._"+obj.current().name(),"\t\t");

			obj.next();

			String discrim = null;
			if ( ((IdlUnionMember)obj.current()).getExpression().equals("true ") ||
					 ((IdlUnionMember)obj.current()).getExpression().equals("false ") )
			{
				discrim = tname + ".toInt()";
			}
			else
			{
				if ( enumeration )
					discrim = tname + ".__d.value()";
				else
					discrim = tname + ".__d";
			}

			i=0;
			while ( obj.end() != true )
			{
				if ( i != idx )
				{
					output.print("\t\tif ( " + discrim + " ==");
					if ( ((IdlUnionMember)obj.current()).getExpression().equals("true ") )
						output.println(" 1 )");
					else
						if ( ((IdlUnionMember)obj.current()).getExpression().equals("false ") )
							output.println(" 0 )");
						else
						{
							if ( !enumeration )
							{
								output.print("(");
								translate_type( d, output,true );
								output.print(")");
							}
							output.println(" "+translate_to_java_expression(((IdlUnionMember)obj.current()).getExpression(),false, ((IdlUnionMember)obj.current()))+" )");
						}

					output.println("\t\t{");
					if ( ((IdlUnionMember)obj.current()).isAsNext() == false )
					{
						obj.current().reset();
						translate_marshalling_member(obj.current().current(),output,outname,tname+"._"+obj.current().name(),"\t\t\t");
					}
					else
					{
						IdlObject next = getAsNext( obj );

						next.reset();
						translate_marshalling_member(next.current(),output,outname,tname+"._"+obj.current().name(),"\t\t\t");
					}
					output.println("\t\t}");
				}
				obj.next();

				if ( ( obj.end() != true ) && ( (i+1) != idx ) )
					output.println("\t\telse");

				i++;
			}

			if ( idx != -1 )
			{
				i=0;
				obj.reset();
				obj.next();
				while ( obj.end() != true )
				{
					if ( i == idx )
					{
						if ( obj.length() != 2 )
							output.println("\t\telse");
						output.println("\t\t{");
						translate_marshalling_member(obj.current().current(),output,outname,tname+"._"+obj.current().name(),"\t\t\t");
						output.println("\t\t}");
					}

					obj.next();
					i++;
				}
			}

			break;
		case IdlType.e_typedef :
			obj.reset();
			switch ( obj.current().kind() )
			{
			case IdlType.e_string :
			case IdlType.e_wstring :
			case IdlType.e_simple :
			case IdlType.e_sequence :
			case IdlType.e_array :
			case IdlType.e_fixed :
				translate_marshalling_member(obj.current(),output,outname,tname,"\t\t");
				break;
			default :
				translate_marshalling(obj.current(),output,outname,tname);
			}
			break;
		case IdlType.e_ident :
			translate_marshalling(((IdlIdent)obj).internalObject(),output,outname,tname);
			break;
		case IdlType.e_exception :
			output.println("\t\t"+outname+".write_string( id() );");
			obj.reset();
			while ( obj.end() != true )
			{
				obj.current().reset();
				translate_marshalling_member(obj.current().current(),output,outname,tname+"."+obj.current().name(),"\t\t");
				obj.next();
			}
			break;
		case IdlType.e_native :
			output.println("\t\tthrow new org.omg.CORBA.MARSHAL();");
			break;
		case IdlType.e_interface :
		case IdlType.e_forward_interface :
			if ( ((IdlInterface)obj).local_interface() )
			{
				output.println("\t\tthrow new org.omg.CORBA.MARSHAL();");
				break;
			}
			if ( ((IdlInterface)obj).abstract_interface() )
				output.println("\t\t((org.omg.CORBA_2_3.portable.OutputStream)"+outname+").write_abstract_interface("+tname+");");
			else
				output.println("\t\t"+outname+".write_Object((org.omg.CORBA.portable.ObjectImpl)"+tname+");");
			break;
		case IdlType.e_value_box :
			obj.reset();
			if ( ((IdlValueBox)obj).simple() )
			{
				output.println("\t\t"+obj.name()+" _box = (" + obj.name() + ")"+tname+";");
				translate_marshalling_member(obj.current(),output,outname,"_box.value","\t\t");
			}
			else
			{
				output.print("\t\t");
				translate_type(obj.current(),output,true);
				output.print(" _box = (");
				translate_type(obj.current(),output,true);
				output.println(")"+tname+";");
				translate_marshalling_member(obj.current(),output,outname,"_box","\t\t");
			}
			break;
		case IdlType.e_value :
			if ( ((IdlValue)obj).abstract_value() )
				output.println("\t\t((org.omg.CORBA_2_3.portable.OutputStream)"+ outname + ").write_value("+tname+", _id );");
			else
				output.println("\t\t((org.omg.CORBA_2_3.portable.OutputStream)"+ outname + ").write_value("+tname+", _id );");
			break;
		}
	}


    public void translate_insert_any(IdlObject obj,java.io.PrintWriter output,String anyname,String tname){
	IdlSimple simple = null;

	switch ( obj.kind() )
	    {
	    case IdlType.e_simple :
		simple = ( IdlSimple )obj;
		if ( simple.internal() == Token.t_ValueBase )
		    {
			output.println(anyname+".insert_value( " + tname +" );");
			return;
			}
			output.print(anyname+".insert");
			switch ( simple.internal() )
			{
			case Token.t_float :
				output.print("_float");
				if(tname==null)
				    tname="new java.lang.Float()";
				break;
			case Token.t_double :
				output.print("_double");
					if(tname==null)
				    tname="new java.lang.Double()";
				break;
			case Token.t_short :
				output.print("_short");
					if(tname==null)
				    tname="new java.lang.Short()";
				break;
			case Token.t_ushort :
				output.print("_ushort");
				if(tname==null)
				    tname="new java.lang.Short()";
				break;
			case Token.t_long :
				output.print("_long");
				if(tname==null)
				    tname="new java.lang.Integer()";
				break;
			case Token.t_ulong :
				output.print("_ulong");
				if(tname==null)
				    tname="new java.lang.Integer()";
				break;
			case Token.t_longlong :
				output.print("_longlong");
				if(tname==null)
				    tname="new java.lang.Long()";
				break;
			case Token.t_ulonglong :
			    output.print("_ulonglong");
			    if(tname==null)
				tname="new java.lang.Long()";
			    break;
			case Token.t_char :
			    output.print("_char");
			    if(tname==null)
				tname="new java.lang.Character()";
			    break;
			case Token.t_wchar :
			    output.print("_wchar");
			    if(tname==null)
				tname="new java.lang.Character()";
			    break;
			case Token.t_boolean :
			    output.print("_boolean");
			    if(tname==null)
				tname="new java.lang.Boolean()";
			    break;
			case Token.t_octet :
			    output.print("_octet");
			    if(tname==null)
				tname="new java.lang.Byte()";
			    break;
			case Token.t_any :
			    output.print("_any");
			    if(tname==null)
				tname="_orb().create_any()";
			    break;
			case Token.t_typecode :
				output.print("_TypeCode");
				/*	if(tname==null)
					tname="new java.lang.Float()";*/
				break;
			case Token.t_object :
				output.print("_Object");
				if(tname==null)
				    tname="new org.omg.CORBA.portable.ObjectImpl()";
				break;
			case Token.t_ValueBase :
				output.print("_value");
				break;
			}
			output.println("("+tname+");");
			break;
		case IdlType.e_fixed :
		    //output.println(anyname+".insert_fixed("+tname+");");
			break;
		case IdlType.e_string :
		    if(tname==null)
			tname="new java.lang.String()";
		    output.println(anyname+".insert_string("+tname+");");
		    break;
		case IdlType.e_wstring :
		    if(tname==null)
			tname="new java.lang.String()";
		    output.println(anyname+".insert_wstring("+tname+");");
		    break;
		case IdlType.e_struct :
		case IdlType.e_union :
		case IdlType.e_enum :
		case IdlType.e_interface :
		case IdlType.e_native :
		case IdlType.e_forward_interface :
		case IdlType.e_value_box :
		case IdlType.e_value :
		case IdlType.e_forward_value :
			output.print(fullname(obj));
			if(tname==null)
			    tname="new "+fullname(obj)+"POA()";
			output.println("Helper.insert("+anyname+","+tname+");");
			break;
		case IdlType.e_typedef :
			output.print(fullname(obj));
			if(tname==null)
			    tname="new "+fullname(obj)+"POA()";
			output.println("Helper.insert("+anyname+","+tname+");");
			break;
		case IdlType.e_array :
		case IdlType.e_sequence :
			break;
		case IdlType.e_ident :
			translate_insert_any(((IdlIdent)obj).internalObject(),output,anyname,tname);
			break;
	}
	}
	/**
	 * Return the typed member for an union member
	 */
	private IdlObject getAsNext( IdlObject obj )
	{
		int p = obj.pos();

		while ( obj.end() != true )
		{
			IdlUnionMember member = ( IdlUnionMember ) obj.current();

			if ( member.isAsNext() == false )
			{
				obj.pos( p );
				return member;
			}

			obj.next();
		}

		obj.pos( p );
		return null;
	}

	/**
	 * Decode a data type
	 *
	 * @param	obj	the data type to decode
	 * @param	output write access
	 * @param	inname	inputstream name
	 */
	public void translate_unmarshalling ( IdlObject obj, java.io.PrintWriter output, String inname )
	{
		int i;
		int idx;

		switch ( obj.kind() )
		{
		case IdlType.e_simple :
			if ( ((IdlSimple)obj).internal() == Token.t_typecode )
				output.println("\t\treturn "+inname+".read_TypeCode();");
			break;
		case IdlType.e_enum :
			output.println("\t\treturn "+obj.name()+".from_int("+inname+".read_ulong());");
			break;
		case IdlType.e_struct :
			obj.reset();
			output.println("\t\t"+fullname(obj)+" new_one = new "+fullname(obj)+"();");
			output.println("");
			while ( obj.end() != true )
			{
				obj.current().reset();
				translate_unmarshalling_member(obj.current().current(),output,inname,"new_one."+obj.current().name(),"\t\t");
				obj.next();
			}
			output.println("");
			output.println("\t\treturn new_one;");
			break;
		case IdlType.e_union :
			idx = ((IdlUnion)obj).index();
			obj.reset();
			output.println("\t\t"+fullname(obj)+" new_one = new "+fullname(obj)+"();");
			output.println("");

			obj.current().reset();
			boolean enumeration = false;
			if ( final_kind( obj.current().current() ) == IdlType.e_enum )
				enumeration = true;

			IdlObject d =  obj.current().current();
			translate_unmarshalling_member(obj.current().current(),output,inname,"new_one._"+obj.current().name(),"\t\t");
			obj.next();

			String discrim = null;
			if ( ((IdlUnionMember)obj.current()).getExpression().equals("true ") ||
					 ((IdlUnionMember)obj.current()).getExpression().equals("false ") )
			{
				discrim = "new_one.toInt()";
			}
			else
			{
				if ( enumeration )
					discrim = "new_one.__d.value()";
				else
					discrim = "new_one.__d";
			}

			i=0;
			while ( obj.end() != true )
			{
				if ( i != idx )
				{
					output.print("\t\tif ( " + discrim + " == " );

					if ( ((IdlUnionMember)obj.current()).getExpression().equals("true ") )
						output.println("1 )");
					else
						if ( ((IdlUnionMember)obj.current()).getExpression().equals("false ") )
							output.println("0 )");
						else
						{
							if ( !enumeration )
							{
								output.print("(");
								translate_type( d, output,true );
								output.print(")");
							}
							output.println( translate_to_java_expression(((IdlUnionMember)obj.current()).getExpression(),false,((IdlUnionMember)obj.current()))+" )");
						}

					output.println("\t\t{");
					if ( ((IdlUnionMember)obj.current()).isAsNext() == false )
					{
						obj.current().reset();
						translate_unmarshalling_member(obj.current().current(),output,inname,"new_one._"+obj.current().name(),"\t\t\t");
					}
					else
					{
						IdlObject next = getAsNext( obj );

						next.reset();
						translate_unmarshalling_member(next.current(),output,inname,"new_one._"+obj.current().name(),"\t\t\t");
					}
					output.println("\t\t}");

				}
				obj.next();

				if ( ( obj.end() != true ) && ( (i+1) != idx ) )
					output.println("\t\telse");

				i++;
			}

			i=0;
			obj.reset();
			obj.next();
			while ( obj.end() != true )
			{
				if ( i == idx )
				{
					if ( obj.length() != 2 )
						output.println("\t\telse" );

					output.println("\t\t{");
					obj.current().reset();
					translate_unmarshalling_member(obj.current().current(),output,inname,"new_one._"+obj.current().name(),"\t\t\t");
					output.println("\t\t}");

				}
				obj.next();

				i++;
			}

			output.println("");
			output.println("\t\treturn new_one;");
			break;
		case IdlType.e_typedef :
			obj.reset();
			switch ( obj.current().kind() )
			{
			case IdlType.e_string :
			case IdlType.e_wstring :
			case IdlType.e_simple :
			case IdlType.e_sequence :
			case IdlType.e_fixed :
			case IdlType.e_array :
				output.print("\t\t");
				translate_type(obj.current(),output,true);
				output.println(" new_one;");
				translate_unmarshalling_member(obj.current(),output,inname,"new_one","\t\t");
				output.println("");
				output.println("\t\treturn new_one;");
				break;
			default :
				translate_unmarshalling(obj.current(),output,inname);
			}
			break;
		case IdlType.e_ident :
			translate_unmarshalling(((IdlIdent)obj).internalObject(),output,inname);
			break;
		case IdlType.e_exception :
			obj.reset();
			output.println("\t\t"+fullname(obj)+" new_one = new "+fullname(obj)+"();");
			output.println("");
			output.println("\t\tif ( !"+inname+".read_string().equals( id() ) )");
			output.println("\t\t	throw new org.omg.CORBA.MARSHAL();");
			while ( obj.end() != true )
			{
				obj.current().reset();
				translate_unmarshalling_member(obj.current().current(),output,inname,"new_one."+obj.current().name(),"\t\t");
				obj.next();
			}
			output.println("");
			output.println("\t\treturn new_one;");
			break;
		case IdlType.e_native :
			output.println("\t\tthrow new org.omg.CORBA.MARSHAL();");
			break;
		case IdlType.e_interface :
		case IdlType.e_forward_interface :
			if ( ((IdlInterface)obj).local_interface() )
			{
				output.println("\t\tthrow new org.omg.CORBA.MARSHAL();");
				break;
			}
			if ( ((IdlInterface)obj).abstract_interface() )
			{
				output.print("\t\tObject new_one = ((org.omg.CORBA_2_3.portable.InputStream)"+inname+").read_abstract_interface( ");

				String stubname = fullname(obj);

				if ( stubname.lastIndexOf(".") != -1 )
					stubname = stubname.substring(0, stubname.lastIndexOf(".") + 1);
				else stubname = "";

				stubname = stubname + "_" + obj.name() + "Stub";

				output.println( stubname + ".class );");

				output.println("\t\treturn (" + fullname(obj) + ") new_one;");
			}
			else
			{
				String stubname = fullname(obj);

				if ( stubname.lastIndexOf(".") != -1 )
					stubname = stubname.substring(0, stubname.lastIndexOf(".") + 1);
				else stubname = "";

				stubname = stubname + "_" + obj.name() + "Stub";

				output.println("\t\treturn( " + fullname(obj) + " )" + inname + ".read_Object( " + stubname + ".class );");
			}
			break;
		case IdlType.e_value :
			if ( ((IdlValue)obj).abstract_value() )
				output.println("\t\treturn ( "+ obj.name() +" ) ((org.omg.CORBA_2_3.portable.InputStream)istream).read_value( _id );");
			else
				output.println("\t\treturn ( "+ obj.name() +" ) ((org.omg.CORBA_2_3.portable.InputStream)istream).read_value( _id );");
			break;
		case IdlType.e_value_box :
			obj.reset();
			if ( ((IdlValueBox)obj).simple() )
			{
				if ( is_boolean( obj.current() ) )
					output.println("\t\t" + fullname(obj) + " _box = new " + fullname(obj) + "(false);");
				else
					output.println("\t\t" + fullname(obj) + " _box = new " + fullname(obj) + "(0);");

				translate_unmarshalling_member(obj.current(),output,inname,"_box.value","\t\t");
			}
			else
			{
				output.print("\t\t");
				translate_type(obj.current(),output,true);
				output.println(" _box = null;");

				translate_unmarshalling_member(obj.current(),output,inname,"_box","\t\t");
			}
			output.println("\t\treturn _box;");
			break;
		}
	}

	private boolean is_boolean( IdlObject obj )
	{
		switch ( final_kind( obj ) )
		{
		case IdlType.e_simple :
			if ( ((IdlSimple)obj).internal() == Token.t_boolean )
				return true;
		default :
			return false;
		}
	}

	/**
	 * Translate a constant
	 *
	 * @param	obj	the constant to translate
	 * @param	writeInto	the directory where the constant must be defined
	 */
	public void translate_constant( IdlObject obj, java.io.File writeInto, java.io.PrintWriter print )
	{
		IdlConst const_obj = ( IdlConst )obj;

		java.io.PrintWriter output = null;

		if ( ( const_obj.into( IdlType.e_interface ) == true ) || ( const_obj.into( IdlType.e_value ) ) )
		{

			print.println("\t//");
			print.println("\t// Constant value");
			print.println("\t//");
			print.print("\tpublic static final ");

			boolean fixed = false;
			if ( final_kind( obj.current() ) == IdlType.e_fixed )
				fixed = true;

			translate_type(obj.current(),print,true);

			print.print(" "+ obj.name() +" = ");

			if ( !fixed )
			{
				print.print("( ");
				obj.reset();
				translate_type(obj.current(),print,true);
				print.print(" ) ");
			}

			print.println("( "+translate_to_java_expression(const_obj.expression(),fixed,obj)+" );");
			print.println("");

		}
		else
		{
			output = newFile(writeInto,obj.name());

			cartouche(output,obj);

			output.println("public interface "+obj.name());
			output.println("{");
			output.println("\t//");
			output.println("\t// Constant value");
			output.println("\t//");
			output.print("\tpublic static final ");

			translate_type(obj.current(),output,true);

			boolean fixed = false;
			if ( final_kind( obj.current() ) == IdlType.e_fixed )
				fixed = true;

			output.print(" value = ");

			if ( !fixed )
			{
				output.print("( ");
				obj.reset();
				translate_type(obj.current(),output,true);
				output.print(" ) ");
			}

			output.println("( "+translate_to_java_expression(const_obj.expression(),fixed,const_obj)+" );");

			output.println("}");

			output.close();
		}
	}

	/**
	 * This function is used to test if an interface has some inherits from bases abstracts
	 * interfaces
	 */
	public boolean isAbstractBaseInterface( IdlObject obj )
	{
		IdlInterface itf = ( IdlInterface ) obj;

		java.util.Vector base = itf.getInheritance();

		for ( int i=0; i<base.size(); i++ )
			if ( ((IdlInterface)(base.elementAt(i))).abstract_interface() )
				return true;

		for ( int i=0; i<base.size(); i++ )
			if ( isAbstractBaseInterface( (IdlInterface)(base.elementAt(i)) ) )
				return true;

		return false;
	}

	/**
	 * Add a Helper for a data type
	 *
	 * @param	obj	the object to translate
	 * @param	writeInto	the directory where the object must be defined
	 */
	public void write_helper( IdlObject obj, java.io.File writeInto )
	{
		java.io.PrintWriter output = newFile(writeInto,obj.name()+"Helper");
		boolean real_corba_object = false;
		boolean abstract_object = false;

		if ( current_pkg != null )
		{
			if ( current_pkg.equals("generated") )
			{
				if ( org.openorb.compiler.IdlCompiler.use_package == true )
				{
					output.println("package "+current_pkg+";");
					output.println("");
				}
			}
			else
				if ( !current_pkg.equals("") )
				{
					output.println("package "+current_pkg+";");
					output.println("");
				}
		}

		output.println("//");
		output.println("// Helper class for : " +obj.name());
		output.println("//");
		output.println("// @author OpenORB Compiler");
		output.println("//");

		// Define the Helper class
		if ( obj.kind() == IdlType.e_value_box )
			output.println("public class "+obj.name()+"Helper implements org.omg.CORBA.portable.BoxedValueHelper");
		else
			output.println("public class "+obj.name()+"Helper");
		output.println("{");

		switch ( final_kind(obj) ) {
		case IdlType.e_struct :
		case IdlType.e_union :
		case IdlType.e_exception :
		case IdlType.e_sequence :
		case IdlType.e_array :
		case IdlType.e_fixed :
			// Test for presence of OpenORB any class.
			if(!org.openorb.compiler.IdlCompiler.portableHelper) {
				output.println("\tprivate static final boolean HAS_OPENORB;");
				output.println("\tstatic {");
				output.println("\t\tboolean hasOpenORB = false;");
				output.println("\t\ttry {");
				output.println("\t\t\tClass.forName(\"org.openorb.CORBA.Any\");");
				output.println("\t\t\thasOpenORB = true;");
				output.println("\t\t}");
				output.println("\t\tcatch(ClassNotFoundException ex) {");
				output.println("\t\t}");
				output.println("\t\tHAS_OPENORB = hasOpenORB;");
				output.println("\t}");
			} else {
				output.println("\tprivate static Object [] _extractMethods;");
				output.println("\tstatic {");
				output.println("\t\ttry {");
				output.println("\t\t\tClass clz = Class.forName(\"org.openorb.CORBA.Any\");");
				output.println("\t\t\tjava.lang.reflect.Method meth = clz.getMethod(\"extract_Streamable\", null);");
				output.println("\t\t\t_extractMethods = new Object[] { clz, meth };");
				output.println("\t\t}");
				output.println("\t\tcatch(Exception ex) {");
				output.println("\t\t}");
				output.println("\t\tif(_extractMethods == null)");
				output.println("\t\t\t_extractMethods = new Object[0];");
				output.println("\t}");
				output.println();
				output.println("\tprivate static java.lang.reflect.Method getExtract(Class clz) {");
				output.println("\t\tint len = _extractMethods.length;");
				output.println("\t\tfor(int i = 0; i < len; i += 2)");
				output.println("\t\t\tif(clz.equals(_extractMethods[i]))");
				output.println("\t\t\t\treturn (java.lang.reflect.Method)_extractMethods[i+1];");
				output.println("\t\t// unknown class, look for method.");
				output.println("\t\tsynchronized(org.omg.CORBA.Any.class) {");
				output.println("\t\t\tfor(int i = len; i < _extractMethods.length; i += 2)");
				output.println("\t\t\t\tif(clz.equals(_extractMethods[i]))");
				output.println("\t\t\t\t\treturn (java.lang.reflect.Method)_extractMethods[i+1];");
				output.println("\t\t\tObject [] tmp = new Object[_extractMethods.length+2];");
				output.println("\t\t\tSystem.arraycopy(_extractMethods, 0, tmp, 0, _extractMethods.length);");
				output.println("\t\t\ttmp[_extractMethods.length] = clz;");
				output.println("\t\t\ttry {");
				output.println("\t\t\t\ttmp[_extractMethods.length+1] = clz.getMethod(\"extract_Streamable\", null);");
				output.println("\t\t\t}");
				output.println("\t\t\tcatch(Exception ex) {}");
				output.println("\t\t\t_extractMethods = tmp;");
				output.println("\t\t\treturn (java.lang.reflect.Method)_extractMethods[_extractMethods.length-1];");
				output.println("\t\t}");
				output.println("\t}");
			}
		}

		// The method  insert
		output.println("\t//");
		output.println("\t// Insert "+obj.name()+" into an any");
		output.println("\t// @param	a an any");
		output.println("\t// @param	t "+obj.name()+" value");
		output.println("\t//");
		output.print("\tpublic static void insert( org.omg.CORBA.Any a, ");

		translate_type(obj,output,true);

		output.println(" t )");
		output.println("\t{");
		switch ( final_kind(obj) ) {
		case IdlType.e_interface :
			if ( ((IdlInterface)final_type(obj)).abstract_interface() ) {
				output.println("\t\tif(t instanceof org.omg.CORBA.Object)");
				output.println("\t\t\ta.insert_Object( (org.omg.CORBA.Object)t , type() );");
				output.println("\t\telse if(t instanceof java.io.Serializable)");
				output.println("\t\t\ta.insert_Value( (java.io.Serializable)t, type() );");
				output.println("\t\telse");
				output.println("\t\t\tthrow new org.omg.CORBA.BAD_PARAM();");
			} else
				output.println("\t\ta.insert_Object( t , type() );");
			break;
		case IdlType.e_value:
		case IdlType.e_value_box:
			output.println("\t\ta.insert_Value( t, type() );");
			break;
		case IdlType.e_struct :
		case IdlType.e_union :
		case IdlType.e_exception :
			output.print("\t\ta.insert_Streamable( new ");
			translate_type(obj, output,true);
			output.println("Holder( t ));");
			break;
		case IdlType.e_sequence :
		case IdlType.e_array :
		case IdlType.e_fixed :
			output.println("\t\ta.insert_Streamable( new " + fullname(obj) + "Holder( t ));");
			break;
		default:
			output.println("\t\ta.type( type() );");
			output.println("\t\twrite(a.create_output_stream(),t);");
		}
		output.println("\t}");
		output.println("");

		// The method  extract
		output.println("\t//");
		output.println("\t// Extract "+obj.name()+" from an any");
		output.println("\t// @param	a an any");
		output.println("\t// @return the extracted "+obj.name()+" value");
		output.println("\t//");
		output.print("\tpublic static ");

		translate_type(obj,output,true);

		output.println(" extract( org.omg.CORBA.Any a )");
		output.println("\t{");
		output.println("\t\tif ( !a.type().equal( type() ) )");
		output.println("\t\t\tthrow new org.omg.CORBA.MARSHAL();");
		switch( final_kind(obj) ) {
		case IdlType.e_interface :
			output.println("\t\ttry {");
			output.print("\t\t\treturn ");
			translate_type(obj,output,true);
			output.println("Helper.narrow( a.extract_Object() );");
			output.println("\t\t}");
			output.println("\t\tcatch ( org.omg.CORBA.BAD_PARAM ex ) {");
			output.println("\t\t\tthrow new org.omg.CORBA.MARSHAL();");
			output.println("\t\t}");
			if ( !((IdlInterface)final_type(obj)).abstract_interface() )
				break;
			output.println("\t\tcatch ( org.omg.CORBA.BAD_OPERATION ex ) {}");
			// fallthrough
		case IdlType.e_value:
		case IdlType.e_value_box:
			output.println("\t\ttry {");
			output.print("\t\t\treturn (");
			translate_type(obj, output,true);
			output.println(")a.extract_Value();");
			output.println("\t\t}");
			output.println("\t\tcatch( ClassCastException ex ) {");
			output.println("\t\t\tthrow new org.omg.CORBA.MARSHAL();");
			output.println("\t\t}");
			break;
		case IdlType.e_struct :
		case IdlType.e_union :
		case IdlType.e_exception :
			if(!org.openorb.compiler.IdlCompiler.portableHelper) {
				output.println("\t\tif(HAS_OPENORB && a instanceof org.openorb.CORBA.Any) {");
				output.println("\t\t\t// streamable extraction. The jdk stubs incorrectly define the Any stub");
				output.println("\t\t\torg.openorb.CORBA.Any any = (org.openorb.CORBA.Any)a;");
				output.println("\t\t\ttry {");
				output.println("\t\t\t\torg.omg.CORBA.portable.Streamable s = any.extract_Streamable();");
			} else {
				output.println("\t\t\t// streamable extraction. The jdk stubs incorrectly define the Any stub");
				output.println("\t\tjava.lang.reflect.Method meth = getExtract(a.getClass());");
				output.println("\t\tif(meth != null) {");
				output.println("\t\t\ttry {");
				output.println("\t\t\t\torg.omg.CORBA.portable.Streamable s ");
				output.println("\t\t\t\t\t= (org.omg.CORBA.portable.Streamable)meth.invoke(a, null);");
			}
			output.print("\t\t\t\tif(s instanceof ");
			translate_type(obj, output,true);
			output.println("Holder)");
			output.print("\t\t\t\t\treturn ((");
			translate_type(obj, output,true);
			output.println("Holder)s).value;");
			output.println("\t\t\t}");
			if(!org.openorb.compiler.IdlCompiler.portableHelper)
				output.println("\t\t\tcatch ( org.omg.CORBA.BAD_INV_ORDER ex ) {}");
			else {
				output.println("\t\t\tcatch ( IllegalAccessException ex ) {");
				output.println("\t\t\t\tthrow new org.omg.CORBA.INTERNAL(ex.toString());");
				output.println("\t\t\t}");
				output.println("\t\t\tcatch ( IllegalArgumentException ex ) {");
				output.println("\t\t\t\tthrow new org.omg.CORBA.INTERNAL(ex.toString());");
				output.println("\t\t\t}");
				output.println("\t\t\tcatch ( java.lang.reflect.InvocationTargetException ex ) {");
				output.println("\t\t\t\tThrowable rex = ex.getTargetException();");
				output.println("\t\t\t\tif(rex instanceof org.omg.CORBA.BAD_INV_ORDER)");
				output.println("\t\t\t\t\t; // do nothing");
				output.println("\t\t\t\telse if(rex instanceof Error)");
				output.println("\t\t\t\t\tthrow (Error)rex;");
				output.println("\t\t\t\telse if(rex instanceof RuntimeException)");
				output.println("\t\t\t\t\tthrow (RuntimeException)rex;");
				output.println("\t\t\t\tthrow new org.omg.CORBA.INTERNAL(rex.toString());");
				output.println("\t\t\t}");
			}
			output.print("\t\t\t");
			translate_type(obj, output,true);
			output.print("Holder h = new ");
			translate_type(obj, output,true);
			output.println("Holder(read(a.create_input_stream()));");
			output.println("\t\t\ta.insert_Streamable(h);");
			output.println("\t\t\treturn h.value;");
			output.println("\t\t}");
			output.println("\t\treturn read(a.create_input_stream());");
			break;
		case IdlType.e_sequence :
		case IdlType.e_array :
		case IdlType.e_fixed :
			if(!org.openorb.compiler.IdlCompiler.portableHelper) {
				output.println("\t\tif(HAS_OPENORB && a instanceof org.openorb.CORBA.Any) {");
				output.println("\t\t\t// streamable extraction. The jdk stubs incorrectly define the Any stub");
				output.println("\t\t\torg.openorb.CORBA.Any any = (org.openorb.CORBA.Any)a;");
				output.println("\t\t\ttry {");
				output.println("\t\t\t\torg.omg.CORBA.portable.Streamable s = any.extract_Streamable();");
			} else {
				output.println("\t\t\t// streamable extraction. The jdk stubs incorrectly define the Any stub");
				output.println("\t\tjava.lang.reflect.Method meth = getExtract(a.getClass());");
				output.println("\t\tif(meth != null) {");
				output.println("\t\t\ttry {");
				output.println("\t\t\t\torg.omg.CORBA.portable.Streamable s ");
				output.println("\t\t\t\t\t= (org.omg.CORBA.portable.Streamable)meth.invoke(a, null);");
			}
			output.println("\t\t\t\tif(s instanceof " + fullname(obj) + "Holder)");
			output.println("\t\t\t\t\treturn ((" + fullname(obj) + "Holder)s).value;");
			output.println("\t\t\t}");
			if(!org.openorb.compiler.IdlCompiler.portableHelper)
				output.println("\t\t\tcatch ( org.omg.CORBA.BAD_INV_ORDER ex ) {}");
			else {
				output.println("\t\t\tcatch ( IllegalAccessException ex ) {");
				output.println("\t\t\t\tthrow new org.omg.CORBA.INTERNAL(ex.toString());");
				output.println("\t\t\t}");
				output.println("\t\t\tcatch ( IllegalArgumentException ex ) {");
				output.println("\t\t\t\tthrow new org.omg.CORBA.INTERNAL(ex.toString());");
				output.println("\t\t\t}");
				output.println("\t\t\tcatch ( java.lang.reflect.InvocationTargetException ex ) {");
				output.println("\t\t\t\tThrowable rex = ex.getTargetException();");
				output.println("\t\t\t\tif(rex instanceof org.omg.CORBA.BAD_INV_ORDER)");
				output.println("\t\t\t\t\t; // do nothing");
				output.println("\t\t\t\telse if(rex instanceof Error)");
				output.println("\t\t\t\t\tthrow (Error)rex;");
				output.println("\t\t\t\telse if(rex instanceof RuntimeException)");
				output.println("\t\t\t\t\tthrow (RuntimeException)rex;");
				output.println("\t\t\t\tthrow new org.omg.CORBA.INTERNAL(rex.toString());");
				output.println("\t\t\t}");
			}
			output.println("\t\t\t" + fullname(obj) + "Holder h = new " + fullname(obj) + "Holder(read(a.create_input_stream()));");
			output.println("\t\t\ta.insert_Streamable(h);");
			output.println("\t\t\treturn h.value;");
			output.println("\t\t}");
			output.println("\t\treturn read(a.create_input_stream());");
			break;
		default:
			output.println("\t\treturn read(a.create_input_stream());");
		}
		output.println("\t}");
		output.println("");

		// The method static _tc
		output.println("\t//");
		output.println("\t// Internal TypeCode value");
		output.println("\t//");
		output.println("\tprivate static org.omg.CORBA.TypeCode _tc = null;");
		switch( final_kind(obj) ) {
		case IdlType.e_value:
		case IdlType.e_struct :
		case IdlType.e_union :
		case IdlType.e_exception :
			output.println("\tprivate static boolean _working = false;");
		}
		output.println("");

		// The method type
		output.println("\t//");
		output.println("\t// Return the "+obj.name()+" TypeCode");
		output.println("\t// @return a TypeCode");
		output.println("\t//");
		output.println("\tpublic static org.omg.CORBA.TypeCode type()");
		output.println("\t{");
		output.println("\t\tif ( _tc == null ) {");
		switch( final_kind(obj) ) {
		case IdlType.e_value:
		case IdlType.e_struct :
		case IdlType.e_union :
		case IdlType.e_exception :
			output.println("\t\t\tsynchronized(org.omg.CORBA.TypeCode.class) {");
			output.println("\t\t\t\tif ( _tc != null )");
			output.println("\t\t\t\t\treturn _tc;");
			output.println("\t\t\t\tif ( _working )");
			output.println("\t\t\t\t\treturn org.omg.CORBA.ORB.init().create_recursive_tc( id() );");
			output.println("\t\t\t\t_working = true;");
		}

		translate_new_typecode(obj,output);

		switch( final_kind(obj) ) {
		case IdlType.e_value:
		case IdlType.e_struct :
		case IdlType.e_union :
		case IdlType.e_exception :
			output.println("\t\t\t\t_working = false;");
			output.println("\t\t\t}");
		}

		output.println("\t\t}");
		output.println("\t\treturn _tc;");
		output.println("\t}");
		output.println("");

		// The method id
		output.println("\t//");
		output.println("\t// Return the "+obj.name()+" IDL ID");
		output.println("\t// @return an ID");
		output.println("\t//");
		output.println("\tpublic static String id()");
		output.println("\t{");
		output.println("\t\treturn _id;");
		output.println("\t}");
		output.println("");
		output.println("\tprivate final static String _id = \""+obj.getId()+"\";");
		output.println("");

		// The method read
		output.println("\t//");
		output.println("\t// Read "+obj.name()+" from a marshalled stream");
		output.println("\t// @param	istream the input stream");
		output.println("\t// @return the readed "+obj.name()+" value");
		output.println("\t//");
		output.print("\tpublic static ");

		translate_type(obj,output,true);

		output.println(" read( org.omg.CORBA.portable.InputStream istream )");
		output.println("\t{");

		if ( final_kind(obj) == IdlType.e_value_box )
		{
			output.print("\t\treturn ( ");

			translate_type(final_type(obj),output,true);

			output.println(" ) ((org.omg.CORBA_2_3.portable.InputStream)istream).read_value( new " + fullname(final_type(obj)) + "Helper() );");
		}
		else
			if ( final_kind( obj ) == IdlType.e_forward_interface )
			{
				if ( ( ((IdlInterface)final_type(obj)).getInterface().local_interface() ) || ( org.openorb.compiler.IdlCompiler.pidl ) )
					output.println("\t\tthrow new org.omg.CORBA.MARSHAL();");
				else
					translate_unmarshalling(obj,output,"istream");
			}
			else
				if ( final_kind(obj) == IdlType.e_interface )
				{
					if ( ( ((IdlInterface)final_type(obj)).local_interface() ) || ( org.openorb.compiler.IdlCompiler.pidl ) )
						output.println("\t\tthrow new org.omg.CORBA.MARSHAL();");
					else
						translate_unmarshalling(obj,output,"istream");
				}
				else
					if ( final_kind( obj ) == IdlType.e_fixed )
					{
						output.println("\t\tjava.math.BigDecimal _f = istream.read_fixed();");
						if ( (( IdlFixed ) final_type(obj)).scale() != 0 )
							output.println("\t\t_f.movePointLeft(" + (( IdlFixed ) final_type(obj)).scale() +  ");");
						output.println("\t\treturn _f;");
					}
					else
						if ( ( obj.kind() == IdlType.e_typedef ) &&
								 ( final_kind( obj ) != IdlType.e_sequence ) &&
								 ( final_kind( obj ) != IdlType.e_array ) &&
								 ( final_kind( obj ) != IdlType.e_string ) &&
								 ( final_kind( obj ) != IdlType.e_wstring ) &&
								 ( final_kind( obj ) != IdlType.e_simple ) )
						{
							output.print("\t\treturn ");

							translate_type( final_type(obj), output,true );

							output.println("Helper.read( istream );");
						}
						else
							translate_unmarshalling(obj,output,"istream");

		output.println("\t}");
		output.println("");

		// La fonction write
		output.println("\t//");
		output.println("\t// Write "+obj.name()+" into a marshalled stream");
		output.println("\t// @param	ostream the output stream");
		output.println("\t// @param	value "+obj.name()+" value");
		output.println("\t//");
		output.print("\tpublic static void write( org.omg.CORBA.portable.OutputStream ostream, ");

		translate_type(obj,output,true);

		output.println(" value )");
		output.println("\t{");

		if ( final_kind(obj) == IdlType.e_value_box )
		{
			output.println("\t\t((org.omg.CORBA_2_3.portable.OutputStream)ostream).write_value( value, new " + fullname(final_type(obj)) + "Helper() );");
		}
		else
			if ( final_kind(obj) == IdlType.e_interface )
			{
				if ( ( ((IdlInterface)final_type(obj)).local_interface() ) || ( org.openorb.compiler.IdlCompiler.pidl ) )
					output.println("\t\tthrow new org.omg.CORBA.MARSHAL();");
				else
					translate_marshalling(obj,output,"ostream","value");
			}
			else
				if ( final_kind( obj ) == IdlType.e_fixed )
				{
					if ( (( IdlFixed ) final_type(obj)).scale() != 0 )
					{
						output.println("\t\tif ( value.scale() != " + (( IdlFixed ) final_type(obj)).scale() +  " )");
						output.println("\t\t\tthrow new org.omg.CORBA.DATA_CONVERSION();");
					}

					output.println("\t\tostream.write_fixed( value );");
				}
				else
					if ( ( obj.kind() == IdlType.e_typedef ) &&
							 ( final_kind( obj ) != IdlType.e_sequence ) &&
							 ( final_kind( obj ) != IdlType.e_array ) &&
							 ( final_kind( obj ) != IdlType.e_string ) &&
							 ( final_kind( obj ) != IdlType.e_wstring ) &&
							 ( final_kind( obj ) != IdlType.e_simple ) )
					{
						output.print("\t\t");

						translate_type( final_type(obj), output,true );

						output.println("Helper.write( ostream, value );");
					}
					else
						translate_marshalling(obj,output,"ostream","value");

		output.println("\t}");
		output.println("");

		// The narrow function
		if ( obj.kind() == IdlType.e_interface )
		{

			if ( ((IdlInterface)obj).abstract_interface() )
				abstract_object = true;
			else
				real_corba_object = true;

			if ( isAbstractBaseInterface( obj ) )
				abstract_object = true;

			if ( abstract_object )
			{
				output.println("\t//");
				output.println("\t// Narrow CORBA::Object to "+obj.name());
				output.println("\t// @param	obj the abstract Object");
				output.println("\t// @return "+obj.name()+" Object");
				output.println("\t//");
				output.println("\tpublic static "+obj.name()+" narrow( Object obj)");
				output.println("\t{");
				output.println("\t\tif ( obj == null )");
				output.println("\t\t\treturn null;");
				output.println("\t\tif ( obj instanceof "+obj.name()+" )");
				output.println("\t\t\treturn ( "+obj.name()+")obj;");
				if ( !org.openorb.compiler.IdlCompiler.pidl && !((IdlInterface)obj).local_interface())
				{
					output.println();
					output.println("\t\tif ( obj instanceof org.omg.CORBA.portable.ObjectImpl ) {");
					output.println("\t\t\torg.omg.CORBA.portable.ObjectImpl objimpl = (org.omg.CORBA.portable.ObjectImpl)obj;");
					output.println("\t\t\tif ( objimpl._is_a( id() ) ) {");
					output.println("\t\t\t\t_"+obj.name()+"Stub stub = new _"+obj.name()+"Stub();");
					output.println("\t\t\t\tstub._set_delegate( objimpl._get_delegate() );");
					output.println("\t\t\t\treturn stub;");
					output.println("\t\t\t}");
					output.println("\t\t}");
					output.println();
				}
				output.println("");
				output.println("\t\tthrow new org.omg.CORBA.BAD_PARAM();");
				output.println("\t}");
				output.println("");

				// Unchecked narrow
				output.println("\t//");
				output.println("\t// Unchecked Narrow CORBA::Object to "+obj.name());
				output.println("\t// @param	obj the abstract Object");
				output.println("\t// @return "+obj.name()+" Object");
				output.println("\t//");
				output.println("\tpublic static "+obj.name()+" unchecked_narrow( Object obj )");
				output.println("\t{");
				output.println("\t\tif ( obj == null )");
				output.println("\t\t\treturn null;");
				output.println("\t\tif ( obj instanceof "+obj.name()+" )");
				output.println("\t\t\treturn ( "+obj.name()+")obj;");
				if ( !org.openorb.compiler.IdlCompiler.pidl && !((IdlInterface)obj).local_interface())
				{
					output.println();
					output.println("\t\tif ( obj instanceof org.omg.CORBA.portable.ObjectImpl ) {");
					output.println("\t\t\torg.omg.CORBA.portable.ObjectImpl objimpl = (org.omg.CORBA.portable.ObjectImpl)obj;");
					output.println("\t\t\t_"+obj.name()+"Stub stub = new _"+obj.name()+"Stub();");
					output.println("\t\t\tstub._set_delegate( objimpl._get_delegate() );");
					output.println("\t\t\treturn stub;");
					output.println("\t\t}");
					output.println();
				}
				output.println("");
				output.println("\t\tthrow new org.omg.CORBA.BAD_PARAM();");
				output.println("\t}");
				output.println("");
			}

			if ( real_corba_object )
			{
				output.println("\t//");
				output.println("\t// Narrow CORBA::Object to "+obj.name());
				output.println("\t// @param	obj the CORBA Object");
				output.println("\t// @return "+obj.name()+" Object");
				output.println("\t//");
				if ( !org.openorb.compiler.IdlCompiler.pidl )
				    output.println("\tpublic static "+obj.name()+" narrow( org.omg.CORBA.Object obj )");
				else
				    output.println("\tpublic static "+obj.name()+" narrow( Object obj )");
				output.println("\t{");
				output.println("\t\tif ( obj == null )");
				output.println("\t\t\treturn null;");
				output.println("\t\tif ( obj instanceof "+obj.name()+" )");
				output.println("\t\t\treturn ( "+obj.name()+")obj;");
				output.println("");
				if ( !org.openorb.compiler.IdlCompiler.pidl && !((IdlInterface)obj).local_interface())
				    {
					output.println("\t\tif ( obj._is_a( id() ) )");
					output.println("\t\t{");
					output.println("\t\t\t_"+obj.name()+"Stub stub = new _"+obj.name()+"Stub();");
					output.println("\t\t\tstub._set_delegate( ((org.omg.CORBA.portable.ObjectImpl)obj)._get_delegate() );");
					output.println("\t\t\treturn stub;");
					output.println("\t\t}");
					output.println("");
				    }
				output.println("\t\tthrow new org.omg.CORBA.BAD_PARAM();");
				output.println("\t}");
				output.println("");
				if(application ){

				}
				// Unchecked narrow
				output.println("\t//");
				output.println("\t// Unchecked Narrow CORBA::Object to "+obj.name());
				output.println("\t// @param	obj the CORBA Object");
				output.println("\t// @return "+obj.name()+" Object");
				output.println("\t//");
				if ( org.openorb.compiler.IdlCompiler.pidl == false )
				    output.println("\tpublic static "+obj.name()+" unchecked_narrow( org.omg.CORBA.Object obj )");
				else
				    output.println("\tpublic static "+obj.name()+" unchecked_narrow( Object obj )");
				output.println("\t{");
				output.println("\t\tif ( obj == null )");
				output.println("\t\t\treturn null;");
				output.println("\t\tif ( obj instanceof "+obj.name()+" )");
				output.println("\t\t\treturn ( "+obj.name()+")obj;");
				output.println("");
				if ( !org.openorb.compiler.IdlCompiler.pidl && !((IdlInterface)obj).local_interface())
				    {
					output.println("\t\t_"+obj.name()+"Stub stub = new _"+obj.name()+"Stub();");
					output.println("\t\tstub._set_delegate( ((org.omg.CORBA.portable.ObjectImpl)obj)._get_delegate() );");
					output.println("\t\treturn stub;");
					output.println();
				} else {
					output.println("\t\tthrow new org.omg.CORBA.BAD_PARAM();");
				}
				output.println("\t}");
				output.println("");
			}
		}

		// Special for value type
		if ( obj.kind() == IdlType.e_value )
		{
			if ( ((IdlValue)obj).abstract_value() == false )
			{
				obj.reset();
				while ( obj.end() != true )
				{
					if ( obj.current().kind() == IdlType.e_factory )
					{
						output.println("\t//");
						output.println("\t// Create a value type ( by its factory way )");
						output.println("\t//");
						output.print("\tpublic static " + obj.name() + " " +obj.name()+"( ");

						output.print("org.omg.CORBA.ORB orb");

						obj.current().reset();
						while ( obj.current().end() != true )
						{
							output.print(", ");

							IdlFactoryMember member = ( IdlFactoryMember ) obj.current().current();

							member.reset();
							translate_type(member.current(), output,true);
							output.print(" "+member.name());

							obj.current().next();
						}

						output.println(" )");
						output.println("\t{");
						output.println("\t\torg.omg.CORBA.portable.ValueFactory _factory = ((org.omg.CORBA_2_3.ORB)orb).lookup_value_factory(id());");
						output.println("\t\tif ( _factory == null )");
						output.println("\t\t\tthrow new org.omg.CORBA.BAD_INV_ORDER();");
						output.print("\t\treturn ((" + fullname(obj) +"ValueFactory)(_factory))." + obj.current().name() + "(");

						obj.current().reset();
						while ( obj.current().end() != true )
						{
							IdlFactoryMember member = ( IdlFactoryMember ) obj.current().current();

							member.reset();
							output.print(" "+member.name());

							obj.current().next();

							if ( obj.current().end() != true )
								output.print(", ");
						}

						output.println(" );");
						output.println("\t}");
						output.println("");
					}

					obj.next();
				}
			}
		}

		// Special for value box
		if ( obj.kind() == IdlType.e_value_box )
		{
			output.println("\t//");
			output.println("\t// Read a value from an input stream");
			output.println("\t//");
			output.println("\tpublic java.io.Serializable read_value( org.omg.CORBA.portable.InputStream is )");
			output.println("\t{");

			translate_unmarshalling(obj,output,"is");

			output.println("\t}");
			output.println("");

			output.println("\t//");
			output.println("\t// Write a value into an output stream");
			output.println("\t//");
			output.println("\tpublic void write_value( org.omg.CORBA.portable.OutputStream os, java.io.Serializable value )");
			output.println("\t{");

			translate_marshalling(obj,output,"os","value");

			output.println("\t}");
			output.println("");

			output.println("\t//");
			output.println("\t// Return the value id");
			output.println("\t//");
			output.println("\tpublic String get_id()");
			output.println("\t{");
			output.println("\t\treturn id();");
			output.println("\t}");
			output.println("");
		}

		/*	if(application){
		    describe_write_to_file(obj,output);
		    describe_read_from_file(obj,output);
		    }*/

		output.println("}");
		output.close();
	}

	/**
	 * Add an holder for a data type
	 *
	 * @param	obj	the object to translate
	 * @param	writeInto	the directory where the object must be defined
	 */
	public void write_holder( IdlObject obj, java.io.File writeInto )
	{
		java.io.PrintWriter output = newFile(writeInto,obj.name()+"Holder");

		if ( current_pkg != null )
		{
			if ( current_pkg.equals("generated") )
			{
				if ( org.openorb.compiler.IdlCompiler.use_package == true )
				{
					output.println("package "+current_pkg+";");
					output.println("");
				}
			}
			else
				if ( !current_pkg.equals("") )
				{
					output.println("package "+current_pkg+";");
					output.println("");
				}
		}

		output.println("//");
		output.println("// Holder class for : " +obj.name());
		output.println("//");
		output.println("// @author OpenORB Compiler");
		output.println("//");

		output.println("final public class "+obj.name()+"Holder");
		output.println("\t\timplements org.omg.CORBA.portable.Streamable");
		output.println("{");

		// The internal value
		output.println("\t//");
		output.println("\t// Internal "+obj.name()+" value");
		output.println("\t//");

		output.print("\tpublic ");

		if ( obj.kind() == IdlType.e_value_box )
		{
			if ( ((IdlValueBox)obj).simple() )
			{
				obj.reset();
				translate_type(obj.current(),output,true);
			}
			else
				translate_type(obj,output,true);
		}
		else
			translate_type(obj,output,true);

		output.println(" value;");

		output.println("");

		// Default constructor
		output.println("\t//");
		output.println("\t// Default constructor");
		output.println("\t//");
		output.println("\tpublic "+obj.name()+"Holder()");
		output.println("\t{ }");
		output.println("");

		// Constructor with init
		output.println("\t//");
		output.println("\t// Constructor with value initialisation");
		output.println("\t// @param	initial	the initial value");
		output.println("\t//");

		output.print("\tpublic "+obj.name()+"Holder( ");

		translate_type(obj,output,true);

		output.println(" initial )");
		output.println("\t{");
		if ( obj.kind() == IdlType.e_value_box )
		{
			if ( ((IdlValueBox)obj).simple() )
				output.println("\t\tvalue = initial.value;");
			else
				output.println("\t\tvalue = initial;");
		}
		else
			output.println("\t\tvalue = initial;");

		output.println("\t}");
		output.println("");

		// The method _read
		output.println("\t//");
		output.println("\t// Read "+obj.name()+" from a marshalled stream");
		output.println("\t// @param	istream the input stream");
		output.println("\t//");
		output.println("\tpublic void _read( org.omg.CORBA.portable.InputStream istream )");
		output.println("\t{");
		if ( obj.kind() == IdlType.e_value_box )
		{
			if ( ((IdlValueBox)obj).simple() )
				output.println("\t\tvalue = ("+obj.name()+"Helper.read(istream)).value;");
			else
				output.println("\t\tvalue = "+obj.name()+"Helper.read(istream);");
		}
		else
			if ( obj.kind() == IdlType.e_interface )
			{
				if ( ((IdlInterface)obj).local_interface() )
					output.println("\t\tthrow new org.omg.CORBA.NO_IMPLEMENT();");
				else
					output.println("\t\tvalue = "+obj.name()+"Helper.read(istream);");
			}
			else
				output.println("\t\tvalue = "+obj.name()+"Helper.read(istream);");
		output.println("\t}");
		output.println("");

		// The method _write
		output.println("\t//");
		output.println("\t// Write "+obj.name()+" into a marshalled stream");
		output.println("\t// @param	ostream the output stream");
		output.println("\t//");
		output.println("\tpublic void _write( org.omg.CORBA.portable.OutputStream ostream )");
		output.println("\t{");
		if ( obj.kind() == IdlType.e_value_box )
		{
			if ( ((IdlValueBox)obj).simple() )
				output.println("\t\t"+obj.name()+"Helper.write(ostream, new " + obj.name() + "(value));");
			else
				output.println("\t\t"+obj.name()+"Helper.write(ostream,value);");
		}
		else
			if ( obj.kind() == IdlType.e_interface )
			{
				if ( ((IdlInterface)obj).local_interface() )
					output.println("\t\tthrow new org.omg.CORBA.NO_IMPLEMENT();");
				else
					output.println("\t\t"+obj.name()+"Helper.write(ostream,value);");
			}
			else
				output.println("\t\t"+obj.name()+"Helper.write(ostream,value);");
		output.println("\t}");
		output.println("");

		// The method _type
		output.println("\t//");
		output.println("\t// Return the "+obj.name()+" TypeCode");
		output.println("\t// @return a TypeCode");
		output.println("\t//");
		output.println("\tpublic org.omg.CORBA.TypeCode _type()");
		output.println("\t{");
		if ( obj.kind() == IdlType.e_interface )
		{
			if ( ((IdlInterface)obj).local_interface() )
				output.println("\t\tthrow new org.omg.CORBA.NO_IMPLEMENT();");
			else
				output.println("\t\treturn "+obj.name()+"Helper.type();");
		}
		else
			output.println("\t\treturn "+obj.name()+"Helper.type();");
		output.println("\t}");
		output.println("");

		output.println("}");
		output.close();
	}

	/**
	 * Translate an enumeration
	 *
	 * @param	obj	the enum to be translated
	 * @param	writeInto	the directory where the enum must be defined
	 */
	public void translate_enum( IdlObject obj, java.io.File writeInto )
	{
		// Deprecated
		// java.io.PrintStream output = ...
		java.io.PrintWriter output = newFile(writeInto,obj.name());
		IdlEnumMember member_obj;

		cartouche(output,obj);


		// Define the class
		output.println("public final class "+obj.name()+" implements org.omg.CORBA.portable.IDLEntity");
		output.println("{");

        // The internal member
		output.println("\t//");
		output.println("\t// Internal member value ");
		output.println("\t//");
		output.println("\tprivate int _"+ obj.name() +"_value;");
		output.println("");

		// Value of each member
		obj.reset();
		while ( obj.end() != true )
		{
			member_obj = (IdlEnumMember)obj.current();

			output.println("\t//");
			output.println("\t// Enum member " + member_obj.name() + " value ");
			output.println("\t//");
			output.print("\tpublic static final int _"+ member_obj.name());
			output.println(" = "+member_obj.getValue()+";");
			output.println("");

			output.println("\t//");
			output.println("\t// Enum member " + member_obj.name());
			output.println("\t//");
			output.print("\tpublic static final "+ obj.name()+" "+member_obj.name());
			output.println(" = new "+obj.name()+"(_"+member_obj.name()+");");
			output.println("");

			obj.next();
		}

		// the method value
		output.println("\t//");
		output.println("\t// Return the internal member value");
		output.println("\t// @return	the member value");
		output.println("\t//");
		output.println("\tpublic int value()");
		output.println("\t{");
		output.println("\t\treturn _"+obj.name()+"_value;");
		output.println("\t}");
		output.println("");

		// the method from_int
		output.println("\t//");
		output.println("\t// Return a enum member from its value");
		output.println("\t// @param		an enum value");
		output.println("\t// @return	an enum member");
		output.println("\t//");
		output.println("\tpublic static "+obj.name()+" from_int( int value )");
		output.println("\t{");
		output.println("\t\tswitch ( value )");
		output.println("\t\t{");

		obj.reset();
		while ( obj.end() != true )
		{
			member_obj = (IdlEnumMember)obj.current();

			output.println("\t\tcase "+member_obj.getValue()+" :");
			output.println("\t\t\treturn "+member_obj.name()+";");

			obj.next();
		}

		output.println("\t\t}");
		output.println("\t\tthrow new org.omg.CORBA.BAD_OPERATION();");
		output.println("\t}");
		output.println("");

		// The constructor
		output.println("\t//");
		output.println("\t// Private constructor");
		output.println("\t// @param		the enum value for this new member");
		output.println("\t//");
		output.println("\tprivate "+obj.name()+"( int value )");
		output.println("\t{");
		output.println("\t\t_"+obj.name()+"_value = value;");
		output.println("\t}");
		output.println("");

		output.println("}");

		output.close();

		write_helper(obj,writeInto);
		write_holder(obj,writeInto);
	}

	/**
	 * Translate the structure
	 *
	 * @param	obj	le module to be translated
	 * @param	writeInto	the directory where the structure must be defined
	 */
	public void translate_struct( IdlObject obj, java.io.File writeInto )
	{
		// Deprecated
		// java.io.PrintStream output = ...
		java.io.PrintWriter output = newFile(writeInto,obj.name());
		java.io.File sub = writeInto;
		String old_pkg = current_pkg;

		IdlStructMember member_obj;

		cartouche(output,obj);

		if ( isEmpty( obj ) == false )
		{
			sub = createDirectory( obj.name() + "Package" , writeInto );
		}

		addToPkg(obj,obj.name()+"Package");

        // Define the sub-types
		obj.reset();
		while ( obj.end() != true )
		{
			obj.current().reset();
			switch ( obj.current().current().kind() )
			{
			case IdlType.e_union :
				translate_union(obj.current().current(),sub);
				break;
			case IdlType.e_struct :
				translate_struct(obj.current().current(),sub);
				break;
			case IdlType.e_enum :
				translate_enum(obj.current().current(),sub);
				break;
			}
			obj.next();
		}

		current_pkg = old_pkg;

        // Define the class
		output.println("public final class "+obj.name()+" implements org.omg.CORBA.portable.IDLEntity");
		output.println("{");

        // Declare each member of the structure
		obj.reset();
		while ( obj.end() != true )
		{
			member_obj = (IdlStructMember)obj.current();

			output.println("\t//");
			output.println("\t// Struct member " + member_obj.name());
			output.println("\t//");
			output.print("\tpublic ");

			member_obj.reset();
			translate_type(member_obj.current(),output,true);

			output.println(" "+member_obj.name()+";");
			output.println("");

			obj.next();
		}

		// Default constructor
		output.println("\t//");
		output.println("\t// Default constructor");
		output.println("\t//");
		output.println("\tpublic "+obj.name()+"()");
		output.println("\t{ }");
		output.println("");


		output.println("\t//");
		output.println("\t// Constructor with fields initialization");

		obj.reset();
		while ( obj.end() != true )
		{
			member_obj = (IdlStructMember)obj.current();
			output.println("\t// @param	"+member_obj.name()+"	"+ member_obj.name()+" struct member");
			obj.next();
		}

		output.println("\t//");
		output.print("\tpublic "+obj.name()+"( ");

		obj.reset();
		while ( obj.end() != true )
		{
			member_obj = (IdlStructMember)obj.current();

			translate_type(member_obj.current(),output,true);
			output.print(" "+member_obj.name());
			obj.next();

			if ( obj.end() != true )
				output.print(", ");
		}

		output.println(" )");
		output.println("\t{");

		obj.reset();
		while ( obj.end() != true )
		{
			member_obj = (IdlStructMember)obj.current();

			output.println("\t\tthis."+member_obj.name()+" = "+member_obj.name()+";");

			obj.next();
		}

		output.println("\t}");
		output.println("");

		output.println("}");

		output.close();

		write_helper(obj,writeInto);
		write_holder(obj,writeInto);
	}

	/**
	 * Search the default value for an union
	 *
	 * @param	obj	the union
	 * @return	default value
	 */
	public String find_default_value( IdlObject obj )
	{
		IdlUnionMember disc;
		IdlUnionMember member_obj;
		IdlObject en;
		int idx = ((IdlUnion)obj).index();
		int i;
		int l;
		int p = obj.pos();
		String s;
		String ts;

		obj.reset();
		disc = (IdlUnionMember)obj.current();

		disc.reset();

		switch ( final_kind(disc.current()) )
		{
		case IdlType.e_enum :
			en =  final_type(disc.current());
			obj.next();
			l = 0;
			i = 0;
			en.reset();
			s = fullname(en.current());
			ts = s + "@ ";
			while ( obj.end() != true )
			{
				if ( i != idx )
				{
					member_obj = (IdlUnionMember)obj.current();
					if ( ts.equals( member_obj.getExpression() )  )
					{
						l++;
						i = -1;
						en.next();
						s = fullname(en.current());
						ts = s + "@ ";
						obj.reset();
					}
				}
				i++;
				obj.next();
			}

			obj.pos(p);
			return s;
		default :
			l = 0;
			i = 0;
			obj.next();
			while ( obj.end() != true )
			{
				if ( i != idx )
				{
					member_obj = (IdlUnionMember)obj.current();
					s = member_obj.getExpression();
					if ( s.charAt(0) == '\'' )
						s = "" + Character.digit( s.charAt( 1 ), 10 );
					if ( l == (new Integer(s)).intValue() )
					{
						l++;
						i = -1;
						obj.reset();
					}
				}
				i++;
				obj.next();
			}

			if ( final_kind(disc.current()) == IdlType.e_simple )
			{
			}
			break;
		}

		obj.pos(p);
		return ""+l;
	}

	/**
	 * Translate an union
	 *
	 * @param	obj	the union to translate
	 * @param	writeInto	the directory where the union must be defined
	 */
	public void translate_union( IdlObject obj, java.io.File writeInto )
	{
		// Deprecated
		// java.io.PrintStream output = ...
		java.io.PrintWriter output = newFile(writeInto,obj.name());
		java.io.File sub = writeInto;
		String old_pkg = current_pkg;

		IdlUnionMember member_obj;
		IdlObject disc;
		int i;
		int idx = ((IdlUnion)obj).index();
		boolean doMap = false;

		cartouche(output,obj);

		if ( isEmpty( obj ) == false )
			sub = createDirectory(obj.name()+"Package", writeInto);

		addToPkg(obj,obj.name()+"Package");

		// Define the sub-types
		obj.reset();
		while ( obj.end() != true )
		{
			obj.current().reset();
			switch ( obj.current().current().kind() )
			{
			case IdlType.e_union :
				translate_union(obj.current().current(),sub);
				break;
			case IdlType.e_struct :
				translate_struct(obj.current().current(),sub);
				break;
			case IdlType.e_enum :
				translate_enum(obj.current().current(),sub);
				break;
			}
			obj.next();
		}

		current_pkg = old_pkg;

        // Define the class
		output.println("public final class "+obj.name()+" implements org.omg.CORBA.portable.IDLEntity");
		output.println("{");

        // Declare union members
		obj.reset();
		while ( obj.end() != true )
		{
			member_obj = (IdlUnionMember)obj.current();

			if ( member_obj.isAsNext() == false )
			{
				output.println("\t//");
				output.println("\t// Union member " + member_obj.name());
				output.println("\t//");
				output.print("\tprotected ");

				member_obj.reset();
				translate_type(member_obj.current(),output,true);

				output.println(" _"+member_obj.name()+";");
				output.println("");

			}
			obj.next();
		}

		// Default constructor
		output.println("\t//");
		output.println("\t// Default constructor");
		output.println("\t//");
		output.println("\tpublic "+obj.name()+"()");
		output.println("\t{");
		if ( idx != -1 )
		{
			output.println("\t\t__d = "+find_default_value(obj)+";");
		}
		output.println("\t}");
		output.println("");

		obj.reset();

		output.println("\t//");
		output.println("\t// Get discriminator value");
		output.println("\t//");
		output.print("\tpublic ");

		member_obj = (IdlUnionMember)obj.current();

		IdlObject discri = member_obj;
		member_obj.reset();
		disc =  member_obj.current();
		translate_type(member_obj.current(),output,true);

		output.println(" discriminator()");
		output.println("\t{");
		output.println("\t\treturn __d;");
		output.println("\t}");
		output.println("");


		obj.next();
		i = 0;
		while ( obj.end() != true )
		{
			member_obj = (IdlUnionMember)obj.current();

			if ( member_obj.isAsNext() == false )
			{
				doMap = false;
				output.println("\t//");
				output.println("\t// Set " + member_obj.name()+" value");
				output.println("\t//");
				output.print("\tpublic void "+member_obj.name()+"( ");

				member_obj.reset();
				translate_type(member_obj.current(),output,true);

				output.println(" value )");
				output.println("\t{");
				if ( i != idx )
					output.println("\t\t__d = "+ translate_to_union_case_expression((IdlUnionMember)discri, member_obj.getExpression())+";");
				else
					output.println("\t\t__d = "+find_default_value(obj) +";");

				output.println("\t\t_"+member_obj.name()+" = value;");
				output.println("\t}");
				output.println("");

				if ( i == idx )
				{
					output.println("\t//");
					output.println("\t// Set " + member_obj.name()+" value");
					output.println("\t//");
					output.print("\tpublic void "+member_obj.name()+"( ");

					translate_type(disc,output,true);

					output.print(" dvalue, ");

					member_obj.reset();
					translate_type(member_obj.current(),output,true);

					output.println(" value )");
					output.println("\t{");
					output.println("\t\t__d = dvalue;");

					output.println("\t\t_"+member_obj.name()+" = value;");
					output.println("\t}");
					output.println("");

				}

				output.println("\t//");
				output.println("\t// Get " + member_obj.name()+" value");
				output.println("\t//");
				output.print("\tpublic ");

				member_obj.reset();
				translate_type(member_obj.current(),output,true);

				output.println(" "+member_obj.name()+"()");
				output.println("\t{");
				output.println("\t\treturn _"+member_obj.name()+";");
				output.println("\t}");
				output.println("");
			}
			else
			{
				if ( doMap == false )
				{
					output.println("\t//");
					output.println("\t// Set " + member_obj.name()+" value");
					output.println("\t//");
					output.print("\tpublic void "+member_obj.name()+"( ");

					translate_type(disc,output,true);

					output.print(" dvalue, ");

					member_obj.reset();
					translate_type(member_obj.current(),output,true);

					output.println(" value )");
					output.println("\t{");
					output.println("\t\t__d = dvalue;");

					output.println("\t\t_"+member_obj.name()+" = value;");
					output.println("\t}");
					output.println("");

					doMap = true;
				}

			}

			obj.next();

			i++;
		}

		if ( idx == -1 )
		{
			output.println("\t//");
			output.println("\t// default access");
			output.println("\t//");
			output.println("\tpublic void __default()");
			output.println("\t{");
			output.println("\t}");
		}

		obj.reset();
		obj.next();

		if ( ((IdlUnionMember)obj.current()).getExpression().equals("true ") ||
				 ((IdlUnionMember)obj.current()).getExpression().equals("false ") )
		{
			output.println("\t//");
			output.println("\t// Return an int value for discriminator");
			output.println("\t//");
			output.println("\tpublic int toInt()");
			output.println("\t{");
			output.println("\tif ( __d == true )");
			output.println("\t	return 1;");
			output.println("\treturn 0;");
			output.println("\t}");
		}

		output.println("}");

		output.close();

		write_helper(obj,writeInto);
		write_holder(obj,writeInto);
	}

	/**
	 * Translate a typedef
	 *
	 * @param	obj	the typedef to translate
	 * @param	writeInto	the directory where the typedef must be defined
	 */
	public void translate_typedef( IdlObject obj, java.io.File writeInto )
	{
		obj.reset();
		switch ( obj.current().kind() )
		{
		case IdlType.e_string :
		case IdlType.e_wstring :
		case IdlType.e_simple :
			write_helper(obj,writeInto);
			break;
		case IdlType.e_union :
		case IdlType.e_struct :
		case IdlType.e_enum :
			write_helper(obj,writeInto);
			break;
		case IdlType.e_fixed :
		case IdlType.e_sequence :
		case IdlType.e_array :
			write_helper(obj,writeInto);
			write_holder(obj,writeInto);
			break;
		case IdlType.e_ident :
			write_helper(obj,writeInto);
			if ( ( final_type( obj.current() ).kind() == IdlType.e_sequence ) ||
					 ( final_type( obj.current() ).kind() == IdlType.e_array ) )
				write_holder(obj, writeInto);
			break;
		case IdlType.e_typedef :
			write_helper(obj,writeInto);
			break;
		}
	}

	/**
	 * Translate an exception
	 *
	 * @param	obj	exception to translate
	 * @param	writeInto	the directory where the exception must be defined
	 */
	public void translate_exception( IdlObject obj, java.io.File writeInto )
	{
		// Deprecated
		// java.io.PrintStream output = ...
		java.io.PrintWriter output = newFile(writeInto,obj.name());
		java.io.File sub = writeInto;
		String old_pkg = current_pkg;

		IdlStructMember member_obj;

		cartouche(output,obj);

		if ( isEmpty( obj ) == false )
			sub = createDirectory( obj.name() + "Package" , writeInto );

		addToPkg(obj,obj.name()+"Package");

		// Define sub-types
		obj.reset();
		while ( obj.end() != true )
		{
			obj.current().reset();
			switch ( obj.current().current().kind() )
			{
			case IdlType.e_union :
				translate_union(obj.current().current(),sub);
				break;
			case IdlType.e_struct :
				translate_struct(obj.current().current(),sub);
				break;
			case IdlType.e_enum :
				translate_enum(obj.current().current(),sub);
				break;
			}
			obj.next();
		}

		current_pkg = old_pkg;

        // Define the class
		output.println("public final class "+obj.name()+" extends org.omg.CORBA.UserException");
		output.println("{");

        // Declare exception members
		obj.reset();
		while ( obj.end() != true )
		{
			member_obj = (IdlStructMember)obj.current();

			output.println("\t//");
			output.println("\t// Exception member " + member_obj.name());
			output.println("\t//");
			output.print("\tpublic ");

			member_obj.reset();
			translate_type(member_obj.current(),output,true);

			output.println(" "+member_obj.name()+";");
			output.println("");

			obj.next();
		}

		// Default constructor
		output.println("\t//");
		output.println("\t// Default constructor");
		output.println("\t//");
		output.println("\tpublic "+obj.name()+"()");
		output.println("\t{");
		output.println("\t\tsuper( "+obj.name()+"Helper.id() );");
		output.println("\t}");
		output.println("");

		if ( obj.length() != 0 )
		{
			output.println("\t//");
			output.println("\t// Constructor with fields initialization");

			obj.reset();
			while ( obj.end() != true )
			{
				member_obj = (IdlStructMember)obj.current();
				output.println("\t// @param	"+member_obj.name()+"	"+ member_obj.name()+" exception member");
				obj.next();
			}

			output.println("\t//");
			output.print("\tpublic "+obj.name()+"( ");

			obj.reset();
			while ( obj.end() != true )
			{
				member_obj = (IdlStructMember)obj.current();

				translate_type(member_obj.current(),output,true);
				output.print(" "+member_obj.name());
				obj.next();

				if ( obj.end() != true )
					output.print(", ");
			}

			output.println(" )");
			output.println("\t{");
			output.println("\t\tsuper( "+obj.name()+"Helper.id() );");

			obj.reset();
			while ( obj.end() != true )
			{
				member_obj = (IdlStructMember)obj.current();

				output.println("\t\tthis."+member_obj.name()+" = "+member_obj.name()+";");

				obj.next();
			}

			output.println("\t}");
			output.println("");
		}

		// Le constructeur le plus complet avec initialisation de chaque membre
		output.println("\t//");
		output.println("\t// Full constructor with fields initialization");

		obj.reset();
		while ( obj.end() != true )
		{
			member_obj = (IdlStructMember)obj.current();
			output.println("\t// @param	"+member_obj.name()+"	"+ member_obj.name()+" exception member");
			obj.next();
		}

		output.println("\t//");
		output.print("\tpublic "+obj.name()+"( String orb_reason");

		obj.reset();
		while ( obj.end() != true )
		{
			output.print(", ");

			member_obj = (IdlStructMember)obj.current();

			translate_type(member_obj.current(),output,true);
			output.print(" "+member_obj.name());
			obj.next();
		}

		output.println(" )");
		output.println("\t{");

		obj.reset();
		output.println("\t\tsuper( "+obj.name()+"Helper.id() +\" \" +  orb_reason );");
		while ( obj.end() != true )
		{
			member_obj = (IdlStructMember)obj.current();

			output.println("\t\tthis."+member_obj.name()+" = "+member_obj.name()+";");

			obj.next();
		}

		output.println("\t}");
		output.println("");

		output.println("}");

		output.close();

		write_helper(obj,writeInto);
		write_holder(obj,writeInto);
	}

	/**
	 * Translate an attribute
	 *
	 * @param	obj	attribute to translate
	 * @param	output	write access
	 */
	public void translate_attribute ( IdlObject obj, java.io.PrintWriter output )
	{
		if ( obj.hasComment() )
			javadoc(output,obj);
		else
		{

			output.println("\t//");
			output.println("\t// Read accessor for "+obj.name()+" attribute");
			output.println("\t// @return	the attribute value");
			output.println("\t//");
		}

		output.print("\tpublic ");

		if ( obj.upper().kind() == IdlType.e_value )
			output.print("abstract ");

		obj.reset();
		translate_type(obj.current(),output,true);

		output.println(" "+obj.name()+"();");
		output.println("");

		if ( ((IdlAttribute)obj).readOnly() == false )
		{
			if ( obj.hasComment() )
				javadoc(output,obj);
			else
			{
				output.println("\t//");
				output.println("\t// Write accessor for "+obj.name()+" attribute");
				output.println("\t// @param	value	the attribute value");
				output.println("\t//");
			}

			output.print("\tpublic ");

			if ( obj.upper().kind() == IdlType.e_value )
				output.print("abstract ");

			output.print("void "+obj.name()+"( ");

			translate_type(obj.current(),output,true);

			output.println(" value );");
			output.println("");
		}
	}

	/**
	 * Translate an attribute for the user code
	 *
	 * @param	obj	attribute to translate
	 * @param	output	write access
	 */
	public void translate_user_attribute ( IdlObject obj, java.io.PrintWriter output )
	{
		output.println("\t//");
		output.println("\t// "+obj.name()+" read attribute");
		output.println("\t//");
		output.print("\tpublic ");

		obj.reset();
		translate_type(obj.current(),output,true);

		output.println(" "+obj.name()+"()");
		output.println("\t{");
		output.println("\t\t//TODO: put your code here");
		output.println("\t}");
		output.println("");

		if ( ((IdlAttribute)obj).readOnly() == false )
		{
			output.println("\t//");
			output.println("\t// "+obj.name()+" write attribute");
			output.println("\t//");
			output.print("\tpublic void "+obj.name()+"( ");

			translate_type(obj.current(),output,true);

			output.println(" value )");
			output.println("\t{");
			output.println("\t\t//TODO: put your code here");
			output.println("\t}");
			output.println("");
		}
	}

	/**
	 * Translate an attribute for a TIE approach
	 *
	 * @param	obj	attribute to translate
	 * @param	output	write access
	 */
	public void translate_attribute_tie ( IdlObject obj, java.io.PrintWriter output )
	{
		output.println("\t//");
		output.println("\t// Read accessor for "+obj.name()+" attribute");
		output.println("\t//");
		output.print("\tpublic ");

		obj.reset();
		translate_type(obj.current(),output,true);

		output.println(" "+obj.name()+"()");
		output.println("\t{");
		output.println("\t\treturn _tie."+obj.name()+"();");
		output.println("\t}");
		output.println("");

		if ( ((IdlAttribute)obj).readOnly() == false )
		{
			output.println("\t//");
			output.println("\t// Write accessor for "+obj.name()+" attribute");
			output.println("\t//");
			output.print("\tpublic void "+obj.name()+"( ");

			translate_type(obj.current(),output,true);

			output.println(" value )");
			output.println("\t{");
			output.println("\t\t_tie."+obj.name()+"(value);");
			output.println("\t}");
			output.println("");
		}
	}

	/**
	 * Translate an attribute
	 *
	 * @param	obj	attribute to translate
	 * @param	output	write access
	 */
	public void translate_attribute_stub ( IdlObject obj, java.io.PrintWriter output )
	{
		output.println("\t//");
		output.println("\t// Read accessor for "+obj.name()+" attribute");
		output.println("\t// @return	the attribute value");
		output.println("\t//");
		output.print("\tpublic ");

		obj.reset();
		translate_type(obj.current(),output,true);

		output.println(" "+obj.name()+"()");
		output.println("\t{");
		output.println("\t\torg.omg.CORBA.Request _arg_request = _request(\"_get_" +obj.name() +"\");");
		output.println("\t\t\torg.omg.CORBA.ORB orb = org.omg.CORBA.ORB.init();");
		output.println("");
		output.print("\t\t_arg_request.set_return_type(");

		obj.reset();
		translate_typecode(obj.current(),output);

		output.println(");");
		output.println("");

		output.println("\t\t_arg_request.invoke();");
		output.println("");

		output.println("\t\tException _except = _arg_request.env().exception();");
		output.println("\t\tif ( _except != null )");
		output.println("\t\t	throw (org.omg.CORBA.SystemException)_except;");
		output.println("");

		output.println("\t\torg.omg.CORBA.Any _arg_result = _arg_request.return_value();");
		output.print("\t\treturn ");

		translate_unmarshalling_data(obj.current(),output,"_arg_result.create_input_stream()");

		output.println("\t}");
		output.println("");

		if ( ((IdlAttribute)obj).readOnly() == false )
		{
			output.println("\t//");
			output.println("\t// Write accessor for "+obj.name()+" attribute");
			output.println("\t// @param	value	the attribute value");
			output.println("\t//");
			output.print("\tpublic void "+obj.name()+"( ");

			translate_type(obj.current(),output,true);

			output.println(" value )");
			output.println("\t{");
			output.println("\t\torg.omg.CORBA.Request _arg_request = _request(\"_set_" +obj.name() +"\");");
			output.println("\t\t\torg.omg.CORBA.ORB orb = org.omg.CORBA.ORB.init();");
			output.println("");
			output.println("\t\torg.omg.CORBA.Any _arg = _arg_request.add_in_arg();");

			output.print("\t\t");
			translate_any_insert(obj.current(),output,"_arg","value");
			output.println(";");

			output.println("\t\t_arg_request.invoke();");
			output.println("");

			output.println("\t\tException _except = _arg_request.env().exception();");
			output.println("\t\tif ( _except != null )");
			output.println("\t\t	throw (org.omg.CORBA.SystemException)_except;");
			output.println("");

			output.println("\t}");
			output.println("");
		}
	}

	/**
	 * Translate an attribute with Stream mode
	 *
	 * @param	obj	attribute to translate
	 * @param	output	write access
	 */
	public void translate_attribute_stub_stream ( IdlObject obj, java.io.PrintWriter output )
	{
	       IdlRaises r;
    IdlContext c;
    boolean someParams = false;
    boolean noReturn = false;
    int p,i,k;
		output.println("\t//");
		output.println("\t// Read accessor for "+obj.name()+" attribute");
		output.println("\t// @return	the attribute value");
		output.println("\t//");
		output.print("\tpublic ");
		obj.reset();
		translate_type(obj.current(),output,true);
		output.println(" "+obj.name()+"()");
		output.println("\t{");
		if(application){
		    	output.println("\n\tint local_retention_id;");
	output.println("\tif(((org.openorb.CORBA.ORB)_orb()).getFeature(\"FTORB\")!=null)");
	output.println("\t\t\tlocal_retention_id=((org.openorb.CORBA.ORB)_orb()).retention_id();\n");
	output.println("\telse");
	output.println("\t\t\tlocal_retention_id=-1;");
output.println("\tlong request_start_time=System.currentTimeMillis();");
	output.println("\n\tif(client_id==null)");
	output.println("\t\tclient_id();");

		}
		output.println("\t\twhile( true )");
		output.println("\t\t{");

		if ( org.openorb.compiler.IdlCompiler.local_stub )
		{
			output.println("\t\t\tif (!this._is_local() )");
			output.println("\t\t\t{");
		}

		output.println("\t\t\t\torg.omg.CORBA.portable.InputStream _input = null;");
		if(application){
		    	obj.reset();
			i=0;
			k=nr_of_args(obj);
			output.println("\t\t\t\t\torg.omg.Dynamic.Parameter[] arguments=new org.omg.Dynamic.Parameter["+k+"];");
			output.println("\t\t\t\t\torg.omg.CORBA.Any any;");
			output.println("\t\t\t\torg.omg.CORBA.Any reply=_orb().create_any();");
			org.omg.CORBA.StringHolder result_s=new org.omg.CORBA.StringHolder("");
			translate_ret_out_typecode(obj.current(),result_s);
			String reply_t_code=result_s.value;
			output.println("\t\t\t\treply.insert_TypeCode("+reply_t_code+");");
			if(k>0)
			    obj.next();
			if ( obj.end() != true )
			    {
				if ( obj.current().kind() == IdlType.e_param )
				    while ( obj.end() != true )
					{
					    output.println("\t\t\t\t\tany=_orb().create_any();");
					    obj.current().reset();
					    switch ( ((IdlParam)obj.current()).param_attr() )
						{
						case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_IN :
						    output.print("\t\t\t\t\t");
						    obj.current().reset();
						    translate_insert_any(obj.current().current(),output,"any",obj.current().name());
						    break;
						case org.openorb.compiler.idl.reflect.idlParameter.PARAM_OUT:
						    output.print("\t\t\t\t\t");
						    obj.current().reset();
						    result_s=new org.omg.CORBA.StringHolder("");
						    translate_ret_out_typecode(obj.current().current(),result_s);
						    reply_t_code=result_s.value;
						    output.println("any.insert_TypeCode("+reply_t_code+");");
						    //translate_marshalling_argument(obj.current().current(),output,"any",null);
						    break;
						case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_INOUT :
						    output.print("\t\t\t\t\t");
						    obj.current().reset();
						    translate_insert_any(obj.current().current(),output,"any",obj.current().name()+".value");
						    break;
						}
					    output.print("\t\t\t\t\targuments["+i+"]=new org.omg.Dynamic.Parameter(any,org.omg.CORBA.ParameterMode.");
					    switch(((IdlParam)obj.current()).param_attr()){
					    case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_IN:output.println("PARAM_IN);");break;
					    case org.openorb.compiler.idl.reflect.idlParameter.PARAM_OUT:output.println("PARAM_OUT);");break;
					    case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_INOUT:output.println("PARAM_INOUT);");break;
					    }
					    obj.next();
					    i++;
					    if ( obj.end() != true )
						{
						    if ( obj.current().kind() != IdlType.e_param )
							{

							    if(obj.current().kind()==IdlType.e_raises )
								{
								    r = (IdlRaises)obj.current();
								    r.reset();
								    output.println("\t\t\t\t\torg.omg.CORBA.TypeCode exceptions[]=new org.omg.CORBA.TypeCode["+r.length()+"];");
								    i=0;
								    while ( r.end() != true )
									{
									    output.println("exceptions["+i+"]="+fullname(r.current())+"Helper.type();");
									    i++;
									    r.next();
									}
								}
							    else
								output.println("\t\t\t\t\torg.omg.CORBA.TypeCode exceptions[]=new org.omg.CORBA.TypeCode[0];");
							    break;
							}
						}
					    else
						output.println("\t\t\t\t\torg.omg.CORBA.TypeCode exceptions[]=new org.omg.CORBA.TypeCode[0];");
					}
				else
				    if(k<1){
					obj.next();
					if(!obj.end()&&obj.current().kind()==IdlType.e_raises )
					    {
						r = (IdlRaises)obj.current();
						r.reset();
						output.println("\t\t\t\t\torg.omg.CORBA.TypeCode exceptions[]=new org.omg.CORBA.TypeCode["+r.length()+"];");
						i=0;
						while ( r.end() != true )
						    {
							output.println("exceptions["+i+"]="+fullname(r.current())+"Helper.type();");
							i++;
							r.next();
						    }
					    }
					else
					    output.println("\t\t\t\t\torg.omg.CORBA.TypeCode exceptions[]=new org.omg.CORBA.TypeCode[0];");
				    }
	    }
		}
		output.println("\t\t\t\ttry");
		output.println("\t\t\t\t{");
		if(application)
		    output.println("\t\t\t\t\torg.omg.CORBA.portable.OutputStream _output = this._request(local_retention_id,client_id,request_start_time,arguments,exceptions,reply,\"_get_" +obj.name() +"\",true);");
		else
		    output.println("\t\t\t\t\torg.omg.CORBA.portable.OutputStream _output = this._request(\"_get_" +obj.name() +"\",true);");

		output.println("\t\t\t\t\t_input = this._invoke(_output);");

		output.print("\t\t\t\t\treturn ");
	obj.reset();
		translate_unmarshalling_data(obj.current(),output,"_input");

		output.println("\t\t\t\t}");
		output.println("\t\t\t\tcatch( org.omg.CORBA.portable.RemarshalException _exception )");
		output.println("\t\t\t\t{");
		output.println("\t\t\t\t\tcontinue;");
		output.println("\t\t\t\t}");
		output.println("\t\t\t\tcatch( org.omg.CORBA.portable.ApplicationException _exception )");
		output.println("\t\t\t\t{");
		output.println("\t\t\t\t\tString _exception_id = _exception.getId();");
		output.println("\t\t\t\t\tthrow new org.omg.CORBA.UNKNOWN(\"Unexcepected User Exception: \"+ _exception_id);");
		output.println("\t\t\t\t}");
		output.println("\t\t\t\tfinally");
		output.println("\t\t\t\t{");
		output.println("\t\t\t\t\tthis._releaseReply(_input);");
		output.println("\t\t\t\t}");

		boolean isAbstract = obj.upper().kind() == IdlType.e_interface &&
                    ((IdlInterface)(obj.upper())).abstract_interface();

    if ( org.openorb.compiler.IdlCompiler.local_stub )
		{
			output.println("\t\t\t}");
      output.println("\t\t\telse");
      output.println("\t\t\t{");
      output.println("\t\t\t\torg.omg.CORBA.portable.ServantObject _so = _servant_preinvoke(\"_get_" +obj.name() +"\",_opsClass);");
      output.println("\t\t\t\tif ( _so == null )");
      output.println("\t\t\t\t   continue;");
      if(isAbstract)
        output.println("\t\t\t\t" + fullname(obj.upper()) + " _self = ( " + fullname(obj.upper()) + " ) _so.servant;");
      else
        output.println("\t\t\t\t" + fullname(obj.upper()) + "Operations _self = ( " + fullname(obj.upper()) + "Operations ) _so.servant;");
      output.println("\t\t\t\ttry");
      output.println("\t\t\t\t{");
      output.println("\t\t\t\t\treturn _self." + obj.name() + "();");
      output.println("\t\t\t\t}");
      output.println("\t\t\t\tfinally");
      output.println("\t\t\t\t{");
      output.println("\t\t\t\t\t_servant_postinvoke(_so);");
      output.println("\t\t\t\t}");
      output.println("\t\t\t}");
		}

		output.println("\t\t}");
		output.println("\t}");
		output.println("");

		if ( ((IdlAttribute)obj).readOnly() == false )
		{
			output.println("\t//");
			output.println("\t// Write accessor for "+obj.name()+" attribute");
			output.println("\t// @param	value	the attribute value");
			output.println("\t//");
			output.print("\tpublic void "+obj.name()+"( ");

			translate_type(obj.current(),output,true);

			output.println(" value )");
			output.println("\t{");
				if(application){
		    	output.println("\n\tint local_retention_id;");
	output.println("\tif(((org.openorb.CORBA.ORB)_orb()).getFeature(\"FTORB\")!=null)");
	output.println("\t\t\tlocal_retention_id=((org.openorb.CORBA.ORB)_orb()).retention_id();\n");
	output.println("\telse");
	output.println("\t\t\tlocal_retention_id=-1;");
output.println("\tlong request_start_time=System.currentTimeMillis();");
	output.println("\n\tif(client_id==null)");
	output.println("\t\tclient_id();");

		}

			output.println("\t\twhile( true )");
			output.println("\t\t{");
			if ( org.openorb.compiler.IdlCompiler.local_stub )
			{
				output.println("\t\t\tif (!this._is_local() )");
				output.println("\t\t\t{");
			}
			output.println("\t\t\t\t\torg.omg.CORBA.portable.InputStream _input = null;");
					if(application){
		    	obj.reset();
			i=0;
			k=nr_of_args(obj);
			output.println("\t\t\t\t\torg.omg.Dynamic.Parameter[] arguments=new org.omg.Dynamic.Parameter["+k+"];");
			output.println("\t\t\t\t\torg.omg.CORBA.Any any;");
			output.println("\t\t\t\torg.omg.CORBA.Any reply=_orb().create_any();");
			org.omg.CORBA.StringHolder result_s=new org.omg.CORBA.StringHolder("");
			translate_ret_out_typecode(obj.current(),result_s);
			String reply_t_code=result_s.value;

			output.println("\t\t\t\treply.insert_TypeCode("+reply_t_code+");");
			if(k>0)
			    obj.next();
			if ( obj.end() != true )
			    {
				if ( obj.current().kind() == IdlType.e_param )
				    while ( obj.end() != true )
					{
					    output.println("\t\t\t\t\tany=_orb().create_any();");
					    obj.current().reset();
					    switch ( ((IdlParam)obj.current()).param_attr() )
						{
						case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_IN :
						    output.print("\t\t\t\t\t");
						    obj.current().reset();
						    translate_insert_any(obj.current().current(),output,"any",obj.current().name());
						    break;
						case org.openorb.compiler.idl.reflect.idlParameter.PARAM_OUT:
						    output.print("\t\t\t\t\t");
						    obj.current().reset();
						    result_s=new org.omg.CORBA.StringHolder("");
						    translate_ret_out_typecode(obj.current().current(),result_s);
						    reply_t_code=result_s.value;
						    output.println("any.insert_TypeCode("+reply_t_code+");");
						    //translate_marshalling_argument(obj.current().current(),output,"any",null);
						    break;
						case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_INOUT :
						    output.print("\t\t\t\t\t");
						    obj.current().reset();
						    translate_insert_any(obj.current().current(),output,"any",obj.current().name()+".value");
						    break;
						}
					    output.print("\t\t\t\t\targuments["+i+"]=new org.omg.Dynamic.Parameter(any,org.omg.CORBA.ParameterMode.");
					    switch(((IdlParam)obj.current()).param_attr()){
					    case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_IN:output.println("PARAM_IN);");break;
					    case org.openorb.compiler.idl.reflect.idlParameter.PARAM_OUT:output.println("PARAM_OUT);");break;
					    case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_INOUT:output.println("PARAM_INOUT);");break;
					    }
					    obj.next();
					    i++;
					    if ( obj.end() != true )
						{
						    if ( obj.current().kind() != IdlType.e_param )
							{

							    if(obj.current().kind()==IdlType.e_raises )
								{
								    r = (IdlRaises)obj.current();
								    r.reset();
								    output.println("\t\t\t\t\torg.omg.CORBA.TypeCode exceptions[]=new org.omg.CORBA.TypeCode["+r.length()+"];");
								    i=0;
								    while ( r.end() != true )
									{
									    output.println("exceptions["+i+"]="+fullname(r.current())+"Helper.type();");
									    i++;
									    r.next();
									}
								}
							    else
								output.println("\t\t\t\t\torg.omg.CORBA.TypeCode exceptions[]=new org.omg.CORBA.TypeCode[0];");
							    break;
							}
						}
					    else
						output.println("\t\t\t\t\torg.omg.CORBA.TypeCode exceptions[]=new org.omg.CORBA.TypeCode[0];");
					}
				else
				    if(k<1){
					obj.next();
					if(!obj.end()&&obj.current().kind()==IdlType.e_raises )
					    {
						r = (IdlRaises)obj.current();
						r.reset();
						output.println("\t\t\t\t\torg.omg.CORBA.TypeCode exceptions[]=new org.omg.CORBA.TypeCode["+r.length()+"];");
						i=0;
						while ( r.end() != true )
						    {
							output.println("exceptions["+i+"]="+fullname(r.current())+"Helper.type();");
							i++;
							r.next();
						    }
					    }
					else
					    output.println("\t\t\t\t\torg.omg.CORBA.TypeCode exceptions[]=new org.omg.CORBA.TypeCode[0];");
				    }
	    }
		}
			output.println("\t\t\t\ttry");
			output.println("\t\t\t\t{");
			if(application)
			    output.println("\t\t\t\t\torg.omg.CORBA.portable.OutputStream _output = this._request(local_retention_id,client_id,request_start_time,arguments,exceptions,reply,\"_set_" +obj.name() +"\",true);");
			else
			    output.println("\t\t\t\t\torg.omg.CORBA.portable.OutputStream _output = this._request(\"_set_" +obj.name() +"\",true);");
			output.print("\t\t\t\t\t");
			obj.reset();
			translate_marshalling_data(obj.current(),output,"_output","value");

			output.println("\t\t\t\t\t_input = this._invoke(_output);");

			output.println("\t\t\t\t\treturn;");

			output.println("\t\t\t\t}");
			output.println("\t\t\t\tcatch( org.omg.CORBA.portable.RemarshalException _exception )");
			output.println("\t\t\t\t{");
			output.println("\t\t\t\t\tcontinue;");
			output.println("\t\t\t\t}");
			output.println("\t\t\t\tcatch( org.omg.CORBA.portable.ApplicationException _exception )");
			output.println("\t\t\t\t{");
			output.println("\t\t\t\t\tString _exception_id = _exception.getId();");
			output.println("\t\t\t\t\tthrow new org.omg.CORBA.UNKNOWN(\"Unexcepected User Exception: \"+ _exception_id);");
			output.println("\t\t\t\t}");
			output.println("\t\t\t\tfinally");
			output.println("\t\t\t\t{");
			output.println("\t\t\t\t\tthis._releaseReply(_input);");
			output.println("\t\t\t\t}");

			if ( org.openorb.compiler.IdlCompiler.local_stub )
			{
				output.println("\t\t\t}");

        output.println("\t\t\telse");
        output.println("\t\t\t{");
        output.println("\t\t\t\torg.omg.CORBA.portable.ServantObject _so = _servant_preinvoke(\"_set_" +obj.name() +"\",_opsClass);");
        output.println("\t\t\t\tif ( _so == null )");
        output.println("\t\t\t\t   continue;");
        if(isAbstract)
          output.println("\t\t\t\t" + fullname(obj.upper()) + "Operations _self = ( " + fullname(obj.upper()) + "Operations ) _so.servant;");
        else
          output.println("\t\t\t\t" + fullname(obj.upper()) + " _self = ( " + fullname(obj.upper()) + " ) _so.servant;");
        output.println("\t\t\t\ttry");
        output.println("\t\t\t\t{");
        output.println("\t\t\t\t\t_self." + obj.name() + "(value);");
        output.println("\t\t\t\t\treturn;");
        output.println("\t\t\t\t}");
        output.println("\t\t\t\tfinally");
        output.println("\t\t\t\t{");
        output.println("\t\t\t\t\t_servant_postinvoke(_so);");
        output.println("\t\t\t\t}");
        output.println("\t\t\t}");
			}

			output.println("\t\t}");
			output.println("\t}");
			output.println("");
		}
	}

	/**
	 * Translate a read attribute for a skeleton
	 *
	 * @param	obj	attribute to translate
	 * @param	output	write access
	 */
	public void translate_read_attribute_skel ( IdlObject obj, java.io.PrintWriter output )
	{
		obj.reset();

		output.print("\t\t\t");

		translate_type(obj.current(),output,true);

		output.println(" arg = "+obj.name()+"();");

		output.println("");
		output.println("\t\t\torg.omg.CORBA.NVList argList = orb.create_list(0);");
		output.println("\t\t\torg.omg.CORBA.Any result = orb.create_any();");
		output.println("\t\t\trequest.arguments(argList);");
		output.print("\t\t\t");

		translate_any_insert(obj.current(),output,"result","arg");
		output.println(";");
		output.println("\t\t\trequest.set_result( result );");
	}

	/**
	 * Translate a read attribute for a Stream mode skeleton
	 *
	 * @param	obj	attribute to translate
	 * @param	output	write access
	 */
	public void translate_read_attribute_skel_stream ( IdlObject obj, java.io.PrintWriter output )
	{
		obj.reset();

		output.print("\t\t\t");

		translate_type(obj.current(),output,true);

		output.println(" arg = "+obj.name()+"();");

		output.println("\t\t\t_output = handler.createReply();");

		output.print("\t\t\t");
		translate_marshalling_data(obj.current(),output,"_output","arg");

		output.println("\t\t\treturn _output;");
	}

	/**
	 * Translate a write attribute for a skeleton
	 *
	 * @param	obj	attribute to translate
	 * @param	output	write access
	 */
	public void translate_write_attribute_skel ( IdlObject obj, java.io.PrintWriter output )
	{
		obj.reset();

		output.println("\t\t\torg.omg.CORBA.NVList argList = orb.create_list(0);");
		output.println("\t\t\torg.omg.CORBA.Any arg = orb.create_any();");

		output.print("\t\t\targ.type( ");

		translate_typecode(obj.current(),output);

		output.println(" );");

		output.println("\t\t\targList.add_value(\"\", arg, org.omg.CORBA.ARG_IN.value);");
		output.println("\t\t\trequest.arguments(argList);");
		output.println("");

		output.print("\t\t\t");
		translate_type(obj.current(),output,true);

		output.print(" result = ");

		translate_unmarshalling_data(obj.current(),output,"arg.create_input_stream()");

		output.println("");
		output.println("\t\t\t"+obj.name()+"( result );");
	}

	/**
	 * Translate a write attribute for a Stream mode skeleton
	 *
	 * @param	obj	attribute to translate
	 * @param	output	write access
	 */
	public void translate_write_attribute_skel_stream ( IdlObject obj, java.io.PrintWriter output )
	{
		obj.reset();

		output.print("\t\t\t");
		translate_type(obj.current(),output,true);

		output.print(" result = ");

		translate_unmarshalling_data(obj.current(),output,"_is");
		output.println("");

		output.println("\t\t\t"+obj.name()+"( result );");

		output.println("\t\t\t_output = handler.createReply();");
		output.println("\t\t\treturn _output;");
	}

	/**
	 * Return the context associated with an operation
	 *
	 * @param	obj	the operation
	 * @return associated context object
	 */
	public IdlContext getContext( IdlObject obj )
	{
		int p = obj.pos();
		IdlObject find = null;

		obj.reset();
		while( obj.end() != true )
		{
			if ( obj.current().kind() == IdlType.e_context )
			{
				find = obj.current();
				break;
			}

			obj.next();
		}

		obj.pos(p);

		return ( IdlContext) find;
	}

	/**
	 * Translate an operation
	 *
	 * @param	obj	operation to translate
	 * @param	output	write access
	 */
	public void translate_operation( IdlObject obj, java.io.PrintWriter output )
	{
		IdlRaises r;
		IdlContext c;
		boolean someParams = false;

		if ( obj.hasComment() )
			javadoc(output,obj);
		else
		{
			output.println("\t//");
			output.println("\t// Operation "+obj.name());
			output.println("\t//");
		}
		output.print("\tpublic ");

		if ( obj.upper().kind() == IdlType.e_value )
			output.print("abstract ");

		obj.reset();
		translate_type(obj.current(),output,true);

		output.print(" "+obj.name()+"(");

		obj.next();
		if ( obj.end() != true )
		{
			if ( obj.current().kind() == IdlType.e_param )
			{
				someParams = true;
				while ( obj.end() != true )
				{

					obj.current().reset();
					translate_parameter(obj.current().current(),output,((IdlParam)obj.current()).param_attr());

					output.print(" "+obj.current().name());

					obj.next();

					if ( obj.end() != true )
					{
						if ( obj.current().kind() == IdlType.e_param )
							output.print(", ");
						else
							break;
					}
				}
			}
		}

		c = getContext(obj);
		if ( c != null )
		{
			if ( someParams == true )
				output.print(", ");
			output.print("org.omg.CORBA.Context ctx");
		}

		output.print(")");

		if ( obj.end() != true )
			if ( obj.current().kind() == IdlType.e_raises )
			{
				output.println("");
				output.print("\t\tthrows ");
				r = (IdlRaises)obj.current();

				r.reset();

				while ( r.end() != true )
				{

					output.print(fullname(r.current()));

					r.next();

					if ( r.end() != true )

						output.print(", ");

				}
			}

		output.println(";");
		output.println("");
	}

	/**
	 * Translate an operation for the user code
	 *
	 * @param	obj	operation to translate
	 * @param	output	write access
	 */
	public void translate_user_operation( IdlObject obj, java.io.PrintWriter output )
	{
		IdlRaises r;
		IdlContext c;
		boolean someParams = false;

		output.println("\t//");
		output.println("\t// Operation "+obj.name());
		output.println("\t//");
		output.print("\tpublic ");

		obj.reset();
		translate_type(obj.current(),output,true);

		output.print(" "+obj.name()+"(");

		obj.next();
		if ( obj.end() != true )
		{
			if ( obj.current().kind() == IdlType.e_param )
			{
				someParams = true;
				while ( obj.end() != true )
				{

					obj.current().reset();
					translate_parameter(obj.current().current(),output,((IdlParam)obj.current()).param_attr());

					output.print(" "+obj.current().name());

					obj.next();

					if ( obj.end() != true )
					{
						if ( obj.current().kind() == IdlType.e_param )
							output.print(", ");
						else
							break;
					}
				}
			}
		}

		c = getContext(obj);
		if ( c != null )
		{
			if ( someParams == true )
				output.print(", ");
			output.print("org.omg.CORBA.Context ctx");
		}

		output.print(")");

		if ( obj.end() != true )
			if ( obj.current().kind() == IdlType.e_raises )
			{
				output.println("");
				output.print("\t\tthrows ");
				r = (IdlRaises)obj.current();

				r.reset();

				while ( r.end() != true )
				{

					output.print(fullname(r.current()));

					r.next();

					if ( r.end() != true )

						output.print(", ");

				}
			}

		output.println("");
		output.println("\t{");
		output.println("\t\t//TODO: put your code here");
		output.println("\t}");
		output.println("");
	}

	/**
	 * Translate an operation for the TIE approach
	 *
	 * @param	obj	operation to translate
	 * @param	output	write access
	 */
	public void translate_operation_tie( IdlObject obj, java.io.PrintWriter output )
	{
		IdlRaises r;
		IdlContext c;
		boolean someParams = false;

		output.println("\t//");
		output.println("\t// Operation "+obj.name());
		output.println("\t//");
		output.print("\tpublic ");

		obj.reset();
		translate_type(obj.current(),output,true);

		output.print(" "+obj.name()+"(");

		obj.next();
		if ( obj.end() != true )
		{
			if ( obj.current().kind() == IdlType.e_param )
			{
				someParams = true;
				while ( obj.end() != true )
				{

					obj.current().reset();
					translate_parameter(obj.current().current(),output,((IdlParam)obj.current()).param_attr());

					output.print(" "+obj.current().name());

					obj.next();

					if ( obj.end() != true )
					{
						if ( obj.current().kind() == IdlType.e_param )
							output.print(", ");
						else
							break;
					}
				}
			}
		}

		c = getContext(obj);
		if ( c != null )
		{
			if ( someParams == true )
				output.print(", ");
			output.print("org.omg.CORBA.Context ctx");
		}

		output.print(")");

		if ( obj.end() != true )
			if ( obj.current().kind() == IdlType.e_raises )
			{
				output.println("");
				output.print("\t\tthrows ");
				r = (IdlRaises)obj.current();

				r.reset();

				while ( r.end() != true )
				{

					output.print(fullname(r.current()));

					r.next();

					if ( r.end() != true )

						output.print(", ");

				}
			}

		output.println("");
		output.println("\t{");

		obj.reset();

		output.print("\t\t");

		if ( obj.current().kind() == IdlType.e_simple )
		{
			if ( ((IdlSimple)obj.current()).internal() != Token.t_void )
				output.print("return ");
		}
		else
			output.print("return ");

		obj.next();
		output.print("_tie."+obj.name()+"(");

		if ( obj.end() != true )
		{
			if ( obj.current().kind() == IdlType.e_param )
				while ( obj.end() != true )
				{
					output.print(" "+obj.current().name());

					obj.next();

					if ( obj.end() != true )
					{
						if ( obj.current().kind() == IdlType.e_param )
							output.print(", ");
						else
							break;
					}
				}
		}

		c = getContext(obj);
		if ( c != null )
		{
			if ( someParams == true )
				output.print(", ");
			output.print("ctx");
		}

		output.println(");");

		output.println("\t}");
		output.println("");
	}

	/**
	 * Translate an operation for a skeleton
	 *
	 * @param	obj	operation to translate
	 * @param	output	write access
	 */
	public void translate_operation_skel( IdlObject obj, java.io.PrintWriter output )
	{
		IdlRaises r;
		IdlContext c;
		int i=0;
		boolean raises = false;
		boolean someParams = false;

		output.println("\t\t\torg.omg.CORBA.NVList argList = orb.create_list(0);");

		obj.reset();
		obj.next();
		if ( obj.end() != true )
		{
			if ( obj.current().kind() == IdlType.e_param )
				while ( obj.end() != true )
				{
					output.println("\t\t\torg.omg.CORBA.Any arg"+i+" = orb.create_any();");
					obj.current().reset();
					switch ( ((IdlParam)obj.current()).param_attr() )
					{
					case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_IN:
						output.print("\t\t\targ"+i+".type( ");
						translate_typecode(obj.current().current(),output);
						output.println(" );");
						output.println("\t\t\targList.add_value(\"\",arg"+i+",org.omg.CORBA.ARG_IN.value);");
						break;
					case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_OUT :
						output.println("\t\t\targList.add_value(\"\",arg"+i+",org.omg.CORBA.ARG_OUT.value);");
						break;
					case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_INOUT :
						output.print("\t\t\targ"+i+".type( ");
						translate_typecode(obj.current().current(),output);
						output.println(" );");
						output.println("\t\t\targList.add_value(\"\",arg"+i+",org.omg.CORBA.ARG_INOUT.value);");
						break;

					}
					i++;

					obj.next();

					if ( obj.end() != true )
					{
						if ( obj.current().kind() != IdlType.e_param )
							break;
					}
				}
		}

		i = 0;
		output.println("");
		output.println("\t\t\trequest.arguments(argList);");
		output.println("");

		c = getContext(obj);
		if ( c != null )
		{
			output.println("\t\t\torg.omg.CORBA.Context arg_ctx = request.ctx();");
			output.println("");
		}

		obj.reset();
		obj.next();
		if ( obj.end() != true )
		{
			if ( obj.current().kind() == IdlType.e_param )
				while ( obj.end() != true )
				{
					obj.current().reset();
					switch ( ((IdlParam)obj.current()).param_attr() )
					{
					case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_IN :
						output.print("\t\t\t");
						translate_type(obj.current().current(),output,true);
						output.print(" arg"+i+"_in = ");
						translate_unmarshalling_data(obj.current().current(),output,"arg"+i+".create_input_stream()");
						break;
					case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_OUT :
						output.print("\t\t\t");
						translate_parameter(obj.current().current(),output,((IdlParam)obj.current()).param_attr());
						output.print(" arg"+i+"_out = new ");
						translate_parameter(obj.current().current(),output,((IdlParam)obj.current()).param_attr());
						output.println("();");
						break;
					case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_INOUT :
						output.print("\t\t\t");
						translate_parameter(obj.current().current(),output,((IdlParam)obj.current()).param_attr());
						output.print(" arg"+i+"_inout = new ");
						translate_parameter(obj.current().current(),output,((IdlParam)obj.current()).param_attr());
						output.println("();");
						output.print("\t\t\targ"+i+"_inout.value = ");
						translate_unmarshalling_data(obj.current().current(),output,"arg"+i+".create_input_stream()");
						break;
					}
					i++;

					obj.next();

					if ( obj.end() != true )
					{
						if ( obj.current().kind() != IdlType.e_param )
							break;
					}
				}
		}

		i = 0;
		output.println("");

		if ( obj.end() != true )
			if ( obj.current().kind() == IdlType.e_raises )
			{
				output.println("\t\t\ttry");
				output.println("\t\t\t{");
				raises = true;
			}

		obj.reset();

		if ( obj.current().kind() == IdlType.e_simple )
		{
			if ( ((IdlSimple)obj.current()).internal() != Token.t_void )
			{
				if ( raises == true )
					output.print("\t");
				output.print("\t\t\t");
				translate_type(obj.current(),output,true);
				output.print(" _arg_result = ");
			}
			else
			{
				if ( raises == true )
					output.print("\t");
				output.print("\t\t\t");
			}
		}
		else
		{
			if ( raises == true )
				output.print("\t");
			output.print("\t\t\t");
			translate_type(obj.current(),output,true);
			output.print(" _arg_result = ");
		}



		output.print(obj.name()+"(");
		obj.next();
		if ( obj.end() != true )
		{
			if ( obj.current().kind() == IdlType.e_param )
			{
				someParams = true;
				while ( obj.end() != true )
				{
					obj.current().reset();
					switch ( ((IdlParam)obj.current()).param_attr() )
					{
					case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_IN :
						output.print("arg"+i+"_in");
						break;
					case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_OUT :
						output.print("arg"+i+"_out");
						break;
					case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_INOUT :
						output.print("arg"+i+"_inout");
						break;
					}
					i++;

					obj.next();

					if ( obj.end() != true )
					{
						if ( obj.current().kind() != IdlType.e_param )
							break;
						else
							output.print(", ");
					}
				}
			}
		}
		i = 0;

		if ( c != null )
		{
			if ( someParams == true )
				output.print(", ");
			output.print("arg_ctx");
		}

		output.println(");");
		output.println("");

		obj.reset();

		if ( obj.current().kind() == IdlType.e_simple )
		{
			if ( ((IdlSimple)obj.current()).internal() != Token.t_void )
			{
				if ( raises == true )
					output.print("\t");
				output.println("\t\t\torg.omg.CORBA.Any any_result = orb.create_any();");
				if ( raises == true )
					output.print("\t");
				output.print("\t\t\t");
				translate_any_insert(obj.current(),output,"any_result","_arg_result");
				output.println(";");
				if ( raises == true )
					output.print("\t");
				output.println("\t\t\trequest.set_result( any_result );");
			}
		}
		else
		{
			if ( raises == true )
				output.print("\t");
			output.println("\t\t\torg.omg.CORBA.Any any_result = orb.create_any();");
			if ( raises == true )
				output.print("\t");
			output.print("\t\t\t");
			translate_any_insert(obj.current(),output,"any_result","_arg_result");
			output.println(";");
			if ( raises == true )
				output.print("\t");
			output.println("\t\t\trequest.set_result( any_result );");
		}

		output.println("");

		obj.next();
		if ( obj.end() != true )
		{
			if ( obj.current().kind() == IdlType.e_param )
				while ( obj.end() != true )
				{
					obj.current().reset();
					switch ( ((IdlParam)obj.current()).param_attr() )
					{
					case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_IN :
						break;
					case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_OUT :
						if( raises == true )
							output.print("\t");
						output.print("\t\t\t");
						translate_any_insert(obj.current().current(),output,"arg"+i,"arg"+i+"_out.value");
						output.println(";");
						break;
					case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_INOUT :
						if( raises == true )
							output.print("\t");
						output.print("\t\t\t");
						translate_any_insert(obj.current().current(),output,"arg"+i,"arg"+i+"_inout.value");
						output.println(";");
						break;
					}
					i++;

					obj.next();

					if ( obj.end() != true )
					{
						if ( obj.current().kind() != IdlType.e_param )
							break;
					}
				}
		}

		if ( obj.end() != true )
			if ( obj.current().kind() == IdlType.e_raises )
			{

				r = (IdlRaises)obj.current();

				r.reset();

				output.println("\t\t\t}");
				while ( r.end() != true )
				{

					output.println("\t\t\tcatch ( " + fullname(r.current()) + " ex )");
					output.println("\t\t\t{");
					output.println("\t\t\t\torg.omg.CORBA.Any any_ex = orb.create_any();");
					output.println("\t\t\t\t"+fullname(r.current())+"Helper.insert(any_ex,ex);");
					output.println("\t\t\t\trequest.set_exception(any_ex);");
					output.println("\t\t\t}");

					r.next();

				}

			}
	}

	/**
	 * Translate an operation for a skeleton with Stream mode
	 *
	 * @param	obj	operation to translate
	 * @param	output	write access
	 */


public void translate_operation_skel_stream( IdlObject obj, java.io.PrintWriter output )
{
    IdlRaises r;
    IdlContext c;
    int i=0,j,nr_of_params,nr_of_params_in,nr_of_params_out_inout;
    boolean raises = false;
    boolean someParams = false;


    // Extract the parameters
       nr_of_params=0;
       nr_of_params_in=0;
    obj.reset();
    obj.next();
    if ( obj.end() != true )
	{

	    if ( obj.current().kind() == IdlType.e_param )
		while ( obj.end() != true )
		    {
			nr_of_params++;
			obj.current().reset();
			switch ( ((IdlParam)obj.current()).param_attr() )
			    {
			    case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_IN :
				nr_of_params_in++;
				output.print("\t\t\t");
				translate_type(obj.current().current(),output,true);
				output.print(" arg"+i+"_in = ");
				translate_unmarshalling_data(obj.current().current(),output,"_is");
				break;
			    case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_OUT :
				output.print("\t\t\t");
				translate_parameter(obj.current().current(),output,((IdlParam)obj.current()).param_attr());
				output.print(" arg"+i+"_out = new ");
				translate_parameter(obj.current().current(),output,((IdlParam)obj.current()).param_attr());
				output.println("();");
				break;
			    case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_INOUT :
				output.print("\t\t\t");
				translate_parameter(obj.current().current(),output,((IdlParam)obj.current()).param_attr());
				output.print(" arg"+i+"_inout = new ");
				translate_parameter(obj.current().current(),output,((IdlParam)obj.current()).param_attr());
				output.println("();");
				output.print("\t\t\targ"+i+"_inout.value = ");
				translate_unmarshalling_data(obj.current().current(),output,"_is");
				break;
			    }

			i++;

			obj.next();

			if ( obj.end() != true )
			    {
				if ( obj.current().kind() != IdlType.e_param )
				    break;
			    }
		    }
	}

    i = 0;
    output.println("");

    c = getContext(obj);
    if ( c != null )
	{
	    output.println("\t\t\torg.omg.CORBA.Context arg_ctx = _is.read_Context();");
	    output.println("");
	}

    // Check if the exceptions can be managed
    if ( obj.end() != true )
	if ( obj.current().kind() == IdlType.e_raises )
	    {
		output.println("\t\t\ttry");
		output.println("\t\t\t{");
		raises = true;
	    }

    obj.reset();

    // Do the invoke
    if ( obj.current().kind() == IdlType.e_simple )
	{
	    if ( ((IdlSimple)obj.current()).internal() != Token.t_void )
		{
		    if ( raises == true )
			output.print("\t");
		    output.print("\t\t\t");
		    translate_type(obj.current(),output,true);
		    output.print(" _arg_result = ");
		}
	    else
		{
		    if ( raises == true )
			output.print("\t");
		    output.print("\t\t\t");
		}
	}
    else
	{
	    if ( raises == true )
		output.print("\t");
	    output.print("\t\t\t");
	    translate_type(obj.current(),output,true);
	    output.print(" _arg_result = ");
	}
    output.print(obj.name()+"(");

    obj.next();
    if ( obj.end() != true )
	{
	    if ( obj.current().kind() == IdlType.e_param )
		{
		    someParams = true;
		    while ( obj.end() != true )
			{
			    obj.current().reset();
			    switch ( ((IdlParam)obj.current()).param_attr() )
				{
				case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_IN :
				    output.print("arg"+i+"_in");
				    break;
				case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_OUT :
				    output.print("arg"+i+"_out");
				    break;
				case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_INOUT :
				    output.print("arg"+i+"_inout");
				    break;
				}
			    i++;

			    obj.next();

			    if ( obj.end() != true )
				{
				    if ( obj.current().kind() != IdlType.e_param )
					break;
				    else
					output.print(", ");
				}
			}
		}
	}


    if ( c != null )
	{
	    if ( someParams == true )
		output.print(", ");
	    output.print("arg_ctx");
	}

    output.println(");");
    output.println("");

    if(application){
	obj.reset();
	if ( raises == true )
			    output.print("\t");
	output.print("\t\t\t");
	output.println("org.omg.CORBA.Any any=_orb().create_any();");
	if ( obj.current().kind() == IdlType.e_simple )
	    if ( ((IdlSimple)obj.current()).internal() != Token.t_void )
		translate_insert_any(obj.current(),output,"any","_arg_result");
		else
			output.println("\t\t\t\tany.insert_TypeCode(_orb().get_primitive_tc(org.omg.CORBA.TCKind.tk_void));");
	else
		translate_insert_any(obj.current(),output,"any","_arg_result");
	output.println("\t\t\t\t((org.openorb.net.ServerRequest)handler).result(any);");

	nr_of_params_out_inout=nr_of_params-nr_of_params_in;

	output.println("org.omg.CORBA.Any[] result_args=new org.omg.CORBA.Any["+nr_of_params_out_inout+"];");
	if(nr_of_params_out_inout>0){
	    j=0;
	    i=0;
	    obj.next();
	    if ( obj.end() != true )
		{
		    if ( obj.current().kind() == IdlType.e_param )
			{
			    while ( obj.end() != true )
				{
				    obj.current().reset();
				    switch ( ((IdlParam)obj.current()).param_attr() )
					{
					case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_IN :
					    break;
					case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_OUT :

					    output.println("result_args["+j+"]=_orb().create_any();");
					    translate_insert_any(obj.current().current(),output,"result_args["+j+"]","arg"+i+"_out.value");
					    j++;
					    break;
					case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_INOUT :
					    output.println("result_args["+j+"]=_orb().create_any();");
					    translate_insert_any(obj.current().current(),output,"result_args["+j+"]","arg"+i+"_inout.value");
					    j++;
					    break;
					}
				    i++;

				    obj.next();

				    if ( obj.end() != true )
					{
					    if ( obj.current().kind() != IdlType.e_param )
						break;
					}
				}
			}
		}
	}
	output.println("\t\t\t\t((org.openorb.net.ServerRequest)handler).result_arguments(result_args);");
    }/*end if(application)*/

    if( raises == true )
	output.print("\t");
    output.println("\t\t\t_output = handler.createReply();");

    // Encode return parameters
    obj.reset();
    i = 0;
    if ( obj.current().kind() == IdlType.e_simple )
	{
	    if ( ((IdlSimple)obj.current()).internal() != Token.t_void )
		{
		    if( raises == true )
			output.print("\t");
		    output.print("\t\t\t");
		    translate_marshalling_data(obj.current(), output, "_output", "_arg_result");
		}
	}
    else
	{
	    if ( raises == true )
		output.print("\t");
	    output.print("\t\t\t");
	    translate_marshalling_data(obj.current(), output, "_output", "_arg_result");
	}

    output.println("");

    obj.next();
    if ( obj.end() != true )
	{
	    if ( obj.current().kind() == IdlType.e_param )
		while ( obj.end() != true )
		    {
			obj.current().reset();
			switch ( ((IdlParam)obj.current()).param_attr() )
			    {
			    case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_IN :
				break;
			    case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_OUT :
				if( raises == true )
				    output.print("\t");
				output.print("\t\t\t");
				translate_marshalling_data(obj.current().current(),output,"_output","arg"+i+"_out.value");
				break;
			    case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_INOUT :
				if( raises == true )
				    output.print("\t");
				output.print("\t\t\t");
				translate_marshalling_data(obj.current().current(),output,"_output","arg"+i+"_inout.value");
				break;
			    }
			i++;

			obj.next();

			if ( obj.end() != true )
			    {
				if ( obj.current().kind() != IdlType.e_param )
				    break;
			    }
		    }
	}

    // Catch th exceptions
    if ( obj.end() != true )
	if ( obj.current().kind() == IdlType.e_raises )
	    {

		r = (IdlRaises)obj.current();

		r.reset();

		output.println("\t\t\t}");
		while ( r.end() != true )
		    {

			output.println("\t\t\tcatch ( " + fullname(r.current()) + " _exception )");
			output.println("\t\t\t{");
			output.println("\t\t\t\t_output = handler.createExceptionReply();");
			output.println("\t\t\t\t"+fullname(r.current())+"Helper.write(_output,_exception);");
			output.println("\t\t\t}");

			r.next();
		    }
	    }

    // Retourne le output
    output.println("\t\t\treturn _output;");
}

	/**
	 * Translate an operation
	 *
	 * @param	obj	operation to translate
	 * @param	output	write access
	 */
	public void translate_operation_stub( IdlObject obj, java.io.PrintWriter output )
	{
		IdlRaises r;
		IdlContext c;
		boolean someParams = false;
		int i = 0;

		output.println("\t//");
		output.println("\t// Operation "+obj.name());
		output.println("\t//");
		output.print("\tpublic ");

		obj.reset();
		translate_type(obj.current(),output,true);

		output.print(" "+obj.name()+"(");

		obj.next();
		if ( obj.end() != true )
		{
			if ( obj.current().kind() == IdlType.e_param )
			{
				someParams = true;
				while ( obj.end() != true )
				{

					obj.current().reset();
					translate_parameter(obj.current().current(),output,((IdlParam)obj.current()).param_attr());

					output.print(" "+obj.current().name());

					obj.next();

					if ( obj.end() != true )
					{
						if ( obj.current().kind() == IdlType.e_param )
							output.print(", ");
						else
							break;
					}
				}
			}
		}

		c = getContext(obj);
		if ( c != null )
		{
			if ( someParams == true )
				output.print(", ");
			output.print("org.omg.CORBA.Context ctx");
		}

		output.print(")");

		if ( obj.end() != true )
			if ( obj.current().kind() == IdlType.e_raises )
			{
				output.println("");
				output.print("\t\tthrows ");
				r = (IdlRaises)obj.current();

				r.reset();

				while ( r.end() != true )
				{

					output.print(fullname(r.current()));

					r.next();

					if ( r.end() != true )

						output.print(", ");

				}
			}

		output.println("");
		output.println("\t{");
		output.println("\t\torg.omg.CORBA.Request request = _request(\"" + initialName(obj.name()) +"\");");
		output.println("\t\t\torg.omg.CORBA.ORB orb = org.omg.CORBA.ORB.init();");
		output.println("");

		if ( c != null )
		{
			java.util.Vector ctx = c.getValues();

			for ( int j=0; j<ctx.size(); j++ )
			{
				output.print("\t\trequest.contexts().add(\"");
				output.print((String)ctx.elementAt(j));
				output.println("\");");
			}

			output.println("\t\trequest.ctx(ctx);");
			output.println("");
		}


		obj.reset();

		if ( obj.current().kind() == IdlType.e_simple )
		{
			if ( ((IdlSimple)obj.current()).internal() != Token.t_void )
			{
				output.print("\t\trequest.set_return_type(");
				translate_typecode(obj.current(),output);
				output.println(");");
				output.println("");
			}
		}
		else
		{
			output.print("\t\trequest.set_return_type(");
			translate_typecode(obj.current(),output);
			output.println(");");
			output.println("");
		}

		obj.next();

		if ( obj.end() != true )
		{
			if ( obj.current().kind() == IdlType.e_param )
				while ( obj.end() != true )
				{

					obj.current().reset();
					switch ( ((IdlParam)obj.current()).param_attr() )
					{
					case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_IN :
						output.println("\t\torg.omg.CORBA.Any arg"+i+" = request.add_in_arg();");
						output.print("\t\t");
						obj.current().reset();
						translate_any_insert(obj.current().current(),output,"arg"+i,obj.current().name());
						output.println(";");
						break;
					case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_OUT :
						output.println("\t\torg.omg.CORBA.Any arg"+i+" = request.add_out_arg();");
						output.print("\t\targ"+i+".type(");
						translate_typecode(obj.current().current(),output);
						output.println(");");
						break;
					case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_INOUT :
						output.println("\t\torg.omg.CORBA.Any arg"+i+" = request.add_inout_arg();");
						output.print("\t\t");
						obj.current().reset();
						translate_any_insert(obj.current().current(),output,"arg"+i,obj.current().name()+".value");
						output.println(";");
						break;
					}
					i++;

					obj.next();

					if ( obj.end() != true )
					{
						if ( obj.current().kind() != IdlType.e_param )
						{
							break;
						}
					}
				}
		}

		if ( obj.end() != true )
			if ( obj.current().kind() == IdlType.e_raises )
			{

				output.println("");

				r = (IdlRaises)obj.current();

				r.reset();

				while ( r.end() != true )
				{

					output.print("\t\trequest.exceptions().add( ");
					output.println(fullname(r.current())+"Helper.type() );");

					r.next();
				}

			}

		output.println("");
		if ( ((IdlOp)obj).oneway() == true )
			output.println("\t\trequest.send_oneway();");
		else
			output.println("\t\trequest.invoke();");
		output.println("");

		if ( obj.end() != true )
			if ( obj.current().kind() == IdlType.e_raises )
			{
				output.println("\t\tException except = request.env().exception();");
				output.println("\t\tif ( except != null )");
				output.println("\t\t{");
				output.println("\t\t	org.omg.CORBA.UnknownUserException unk_except;");
				output.println("\t\t	try");
				output.println("\t\t	{");
				output.println("\t\t		unk_except = (org.omg.CORBA.UnknownUserException)except;");
				output.println("\t\t	}");
				output.println("\t\t	catch ( ClassCastException ex )");
				output.println("\t\t	{");
				output.println("\t\t		throw (org.omg.CORBA.SystemException)except;");
				output.println("\t\t	}");
				output.println("");

				r = (IdlRaises)obj.current();

				r.reset();

				while ( r.end() != true )
				{

					output.println("\t\t	try");
					output.println("\t\t	{");
					output.println("\t\t		throw "+fullname(r.current())+"Helper.extract(unk_except.except);");
					output.println("\t\t	}");
					output.println("\t\t	catch ( org.omg.CORBA.MARSHAL ex )");
					output.println("\t\t	{ }");
					output.println("");

					r.next();
				}

				output.println("\t\tthrow new org.omg.CORBA.UNKNOWN();");
				output.println("\t\t}");
				output.println("");
			}


		obj.reset();
		obj.next();

		i = 0;
		if ( obj.end() != true )
		{
			if ( obj.current().kind() == IdlType.e_param )
				while ( obj.end() != true )
				{

					obj.current().reset();
					switch ( ((IdlParam)obj.current()).param_attr() )
					{
					case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_IN :

						break;
					case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_OUT :
						output.print("\t\t"+obj.current().name()+".value = ");
						obj.current().reset();
						translate_unmarshalling_data(obj.current().current(),output,"arg"+i+".create_input_stream()");
						break;
					case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_INOUT :
						output.print("\t\t"+obj.current().name()+".value = ");
						obj.current().reset();
						translate_unmarshalling_data(obj.current().current(),output,"arg"+i+".create_input_stream()");
						break;
					}
					i++;

					obj.next();

					if ( obj.end() != true )
					{
						if ( obj.current().kind() != IdlType.e_param )
							break;
					}
				}
		}
		output.println("");
		obj.reset();
		if ( obj.current().kind() == IdlType.e_simple )
		{
			if ( ((IdlSimple)obj.current()).internal() != Token.t_void )
			{
				output.println("\t\torg.omg.CORBA.Any _arg_result = request.return_value();");
				output.print("\t\treturn ");

				translate_unmarshalling_data(obj.current(),output,"_arg_result.create_input_stream()");
			}
		}
		else
		{
			output.println("\t\torg.omg.CORBA.Any _arg_result = request.return_value();");
			output.print("\t\treturn ");

			translate_unmarshalling_data(obj.current(),output,"_arg_result.create_input_stream()");
		}

		output.println("\t}");
		output.println("");
	}
    /*Written by Diana on Oct. 9 2001*/
    private int nr_of_args(IdlObject obj){
	int i=0;
	obj.reset();
	/* skip return value*/
	obj.next();
	while(!obj.end()&&obj.current().kind()==IdlType.e_param){
	    i++;
	    obj.next();
	}
	obj.reset();
	return i;
    }


    private String get_TCKind(int internal){
	switch(internal){
	case Token.t_boolean:return "org.omg.CORBA.TCKind.tk_boolean";
	case Token.t_char:return "org.omg.CORBA.TCKind.tk_char";
	case Token.t_double:return "org.omg.CORBA.TCKind.tk_double";
	case Token.t_float:return "org.omg.CORBA.TCKind.tk_float";
	case Token.t_long:return "org.omg.CORBA.TCKind.tk_long";
	case Token.t_longlong:return "org.omg.CORBA.TCKind.tk_longlong";
	case Token.t_void:return "org.omg.CORBA.TCKind.tk_void";
	case Token.t_short:return "org.omg.CORBA.TCKind.tk_short";
	case Token.t_ulonglong:return "org.omg.CORBA.TCKind.tk_ulonglong";
	case Token.t_ulong:return "org.omg.CORBA.TCKind.tk_ulong";
	case Token.t_wchar:return "org.omg.CORBA.TCKind.tk_wchar";
	case Token.t_ushort:return "org.omg.CORBA.TCKind.tk_ushort";
	case Token.t_typecode: return "org.omg.CORBA.TCKind.tk_TypeCode";
	case Token.t_object: return "org.omg.CORBA.TCKind.tk_objref";
	}
	return null;
    }

    private void  translate_ret_out_typecode(IdlObject obj,org.omg.CORBA.StringHolder toReturn){
	switch ( obj.kind() )
	    {
	    case IdlType.e_simple :
		toReturn.value+="_orb().get_primitive_tc("+get_TCKind(((IdlSimple)obj).internal())+")";break;
	    case IdlType.e_string :
	    case IdlType.e_wstring :
		if ( ((IdlString)obj).maxSize() == 0 )
		    toReturn.value+="_orb().get_primitive_tc(org.omg.CORBA.TCKind.tk_string )";
		else
		    toReturn.value+="_orb().create_string_tc("+((IdlString)obj).maxSize()+")";
		break;
	    case IdlType.e_struct :
	    case IdlType.e_union :
	    case IdlType.e_enum :
	    case IdlType.e_interface :
	    case IdlType.e_forward_interface :
	    case IdlType.e_exception :
	    case IdlType.e_value :
	    case IdlType.e_forward_value :
	    case IdlType.e_native :
	    case IdlType.e_typedef :
	    case IdlType.e_value_box :
		toReturn.value+=fullname(obj)+"Helper.type()";
		break;
		/*		obj.reset();
				translate_ret_out_typecode(obj.current(),toReturn);
				break; */
	    case IdlType.e_sequence :
	    case IdlType.e_array :
		/*int pos=obj.pos();
		  obj.next();
		  toReturn.value+=rec_fullname(obj.current())+"Helper.type()";
		  obj.pos(pos);*/
		break;
	    case IdlType.e_ident :
		translate_ret_out_typecode(((IdlIdent)obj).internalObject(),toReturn);
		break;
		/*
		  if ( ((IdlValueBox)obj).simple() )
		  toReturn.value+=fullname(obj)+"Helper.type()";
		  else
		  {
		  obj.reset();
		  translate_ret_out_typecode(obj.current(),toReturn);
		  }
		  break;*/
	    }
    }

private String rec_fullname(IdlObject obj){
    if(obj.kind()!=IdlType.e_ident)
	return fullname(obj);
    else
	return rec_fullname(((IdlIdent)obj).internalObject());
}

    /**
	 * Translate an operation in Stream mode
	 *
	 * @param	obj	operation to translate
	 * @param	output	write access
	 */

public void translate_operation_stub_stream( IdlObject obj, java.io.PrintWriter output )
{
    IdlRaises r;
    IdlContext c;
    boolean someParams = false;
    boolean noReturn = false;
    int p,i,k;

    output.println("\t//");
    output.println("\t// Operation "+obj.name());
    output.println("\t//");
    output.print("\tpublic ");

    obj.reset();
    translate_type(obj.current(),output,true);

    output.print(" "+obj.name()+"(");

    obj.next();
    if ( obj.end() != true )
	{
	    if ( obj.current().kind() == IdlType.e_param )
		{
		    someParams = true;
		    while ( obj.end() != true )
			{

			    obj.current().reset();
			    translate_parameter(obj.current().current(),output,((IdlParam)obj.current()).param_attr());

			    output.print(" "+obj.current().name());

			    obj.next();

			    if ( obj.end() != true )
				{
				    if ( obj.current().kind() == IdlType.e_param )
					output.print(", ");
				    else
					break;
				}
			}
		}
	}

    c = getContext(obj);
    if ( c != null )
	{
	    if ( someParams == true )
		output.print(", ");
	    output.print("org.omg.CORBA.Context ctx");
	}

    output.print(")");

    if ( obj.end() != true )
	if ( obj.current().kind() == IdlType.e_raises )
	    {
		output.println("");
		output.print("\t\tthrows ");
		r = (IdlRaises)obj.current();

		r.reset();

		while ( r.end() != true )
		    {

			output.print(fullname(r.current()));

			r.next();

			if ( r.end() != true )

			    output.print(", ");

		    }
	    }

    output.println("");
    output.println("\t{");
    if(application){
	output.println("\n\tint local_retention_id;");
	output.println("\tif(((org.openorb.CORBA.ORB)_orb()).getFeature(\"FTORB\")!=null)");
	output.println("\t\t\tlocal_retention_id=((org.openorb.CORBA.ORB)_orb()).retention_id();\n");
	output.println("\telse");
	output.println("\t\t\tlocal_retention_id=-1;");
output.println("\tlong request_start_time=System.currentTimeMillis();");
	output.println("\n\tif(client_id==null)");
	output.println("\t\tclient_id();");
    }
    output.println("\t\twhile( true )");
    output.println("\t\t{");

    if ( org.openorb.compiler.IdlCompiler.local_stub )
	{
	    output.println("\t\t\tif (!this._is_local() )");
	    output.println("\t\t\t{");
	}

    if ( c != null )
	{
	    output.println("\t\t\t\torg.omg.CORBA.ContextList ctxList = org.omg.CORBA.ORB.init().create_context_list();");
	}

    output.println("\t\t\t\torg.omg.CORBA.portable.InputStream _input = null;");
    output.println("\t\t\t\ttry");
    output.println("\t\t\t\t{");
    if(application){
	obj.reset();
	i=0;
	k=nr_of_args(obj);
	output.println("\t\t\t\t\torg.omg.Dynamic.Parameter[] arguments=new org.omg.Dynamic.Parameter["+k+"];");
	output.println("\t\t\t\t\torg.omg.CORBA.Any any;");
	output.println("\t\t\t\torg.omg.CORBA.Any reply=_orb().create_any();");
	org.omg.CORBA.StringHolder result_s=new org.omg.CORBA.StringHolder("");
	translate_ret_out_typecode(obj.current(),result_s);
	String reply_t_code=result_s.value;
	output.println("\t\t\t\treply.insert_TypeCode("+reply_t_code+");");
	if(k>0)
	    obj.next();
	if ( obj.end() != true )
	    {
		if ( obj.current().kind() == IdlType.e_param )
		    while ( obj.end() != true )
			{
			    output.println("\t\t\t\t\tany=_orb().create_any();");
			    obj.current().reset();
			    switch ( ((IdlParam)obj.current()).param_attr() )
				{
				case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_IN :
				    output.print("\t\t\t\t\t");
				    obj.current().reset();
				    translate_insert_any(obj.current().current(),output,"any",obj.current().name());
				    break;
				case org.openorb.compiler.idl.reflect.idlParameter.PARAM_OUT:
				    output.print("\t\t\t\t\t");
				    obj.current().reset();
				    result_s=new org.omg.CORBA.StringHolder("");
				    translate_ret_out_typecode(obj.current().current(),result_s);
				    reply_t_code=result_s.value;
				    output.println("any.insert_TypeCode("+reply_t_code+");");
				    //translate_marshalling_argument(obj.current().current(),output,"any",null);
				    break;
				case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_INOUT :
				    output.print("\t\t\t\t\t");
				    obj.current().reset();
				    translate_insert_any(obj.current().current(),output,"any",obj.current().name()+".value");
				    break;
				}
			    output.print("\t\t\t\t\targuments["+i+"]=new org.omg.Dynamic.Parameter(any,org.omg.CORBA.ParameterMode.");
			    switch(((IdlParam)obj.current()).param_attr()){
			    case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_IN:output.println("PARAM_IN);");break;
			    case org.openorb.compiler.idl.reflect.idlParameter.PARAM_OUT:output.println("PARAM_OUT);");break;
			    case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_INOUT:output.println("PARAM_INOUT);");break;
			    }
			    obj.next();
			    i++;
			    if ( obj.end() != true )
				{
				    if ( obj.current().kind() != IdlType.e_param )
					{

					    if(obj.current().kind()==IdlType.e_raises )
						{
						    r = (IdlRaises)obj.current();
						    r.reset();
						    output.println("\t\t\t\t\torg.omg.CORBA.TypeCode exceptions[]=new org.omg.CORBA.TypeCode["+r.length()+"];");
						    i=0;
						    while ( r.end() != true )
							{
							    output.println("exceptions["+i+"]="+fullname(r.current())+"Helper.type();");
							    i++;
							    r.next();
							}
						}
					    else
						output.println("\t\t\t\t\torg.omg.CORBA.TypeCode exceptions[]=new org.omg.CORBA.TypeCode[0];");
					    break;
					}
				}
			    else
				output.println("\t\t\t\t\torg.omg.CORBA.TypeCode exceptions[]=new org.omg.CORBA.TypeCode[0];");
			}
		else
		    if(k<1){
			obj.next();
			if(!obj.end()&&obj.current().kind()==IdlType.e_raises )
			    {
				r = (IdlRaises)obj.current();
				r.reset();
				output.println("\t\t\t\t\torg.omg.CORBA.TypeCode exceptions[]=new org.omg.CORBA.TypeCode["+r.length()+"];");
				i=0;
				while ( r.end() != true )
				    {
					output.println("exceptions["+i+"]="+fullname(r.current())+"Helper.type();");
					i++;
					r.next();
				    }
			    }
			else
			    output.println("\t\t\t\t\torg.omg.CORBA.TypeCode exceptions[]=new org.omg.CORBA.TypeCode[0];");
		    }
	    }
    }
    if ( ((IdlOp)obj).oneway() == true )
	if(application)
	    output.println("\t\t\t\t\torg.omg.CORBA.portable.OutputStream _output = this._request(local_retention_id,client_id,request_start_time,arguments,exceptions,reply,\"" + initialName(obj.name()) +"\",false);");
	else
	    output.println("\t\t\t\t\torg.omg.CORBA.portable.OutputStream _output = this._request(\"" + initialName(obj.name()) +"\",false);");
    else
	if(application)
	    output.println("\t\t\t\t\torg.omg.CORBA.portable.OutputStream _output = this._request(local_retention_id,client_id,request_start_time,arguments,exceptions,reply,\"" + initialName(obj.name()) +"\",true);");
	else
	    output.println("\t\t\t\t\torg.omg.CORBA.portable.OutputStream _output = this._request(\"" + initialName(obj.name()) +"\",true);");

    // Encode the parameters

    obj.reset();
    obj.next();
    if ( obj.end() != true )
	{
	    if ( obj.current().kind() == IdlType.e_param )
		while ( obj.end() != true )
		    {
			obj.current().reset();
			switch ( ((IdlParam)obj.current()).param_attr() )
			    {
			    case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_IN :
				output.print("\t\t\t\t\t");
				obj.current().reset();
				translate_marshalling_data(obj.current().current(),output,"_output",obj.current().name());
				break;
			    case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_INOUT :
				output.print("\t\t\t\t\t");
				obj.current().reset();
				translate_marshalling_data(obj.current().current(),output,"_output",obj.current().name()+".value");
				break;
			    }

			obj.next();

			if ( obj.end() != true )
			    {
				if ( obj.current().kind() != IdlType.e_param )
				    {
					break;
				    }
			    }
		    }
	}

    // Encode the contexts
    if ( c != null )
	{
	    java.util.Vector ctx = c.getValues();

	    for ( int j=0; j<ctx.size(); j++ )
		{
		    output.println("\t\t\t\t\tctxList.add(\"" + (String)ctx.elementAt(j) + "\");");
		}

	    output.println("\t\t\t\t\t_output.write_Context(ctx,ctxList);");

	    output.println("");
	}

    // Do the invoke

    output.println("\t\t\t\t\t_input = this._invoke(_output);");

    // Decode the params

    obj.reset();

    if ( obj.current().kind() == IdlType.e_simple )
	{
	    if ( ((IdlSimple)obj.current()).internal() != Token.t_void )
		{
		    output.print("\t\t\t\t\t");
		    translate_type(obj.current(),output,true);
		    output.print(" _arg_ret = ");
		    translate_unmarshalling_data(obj.current(),output,"_input");
		}
	}
    else
	{
	    output.print("\t\t\t\t\t");
	    translate_type(obj.current(),output,true);
	    output.print(" _arg_ret = ");
	    translate_unmarshalling_data(obj.current(),output,"_input");
	}

    obj.next();
    if ( obj.end() != true )
	{
	    if ( obj.current().kind() == IdlType.e_param )
		while ( obj.end() != true )
		    {

			obj.current().reset();
			switch ( ((IdlParam)obj.current()).param_attr() )
			    {
			    case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_OUT :
			    case  org.openorb.compiler.idl.reflect.idlParameter.PARAM_INOUT :
				output.print("\t\t\t\t\t"+obj.current().name()+".value = ");
				obj.current().reset();
				translate_unmarshalling_data(obj.current().current(),output,"_input");
				break;
			    }

			obj.next();

			if ( obj.end() != true )
			    {
				if ( obj.current().kind() != IdlType.e_param )
				    break;
			    }
		    }
	}

    p = obj.pos();

    // Return the result
    obj.reset();
    if ( obj.current().kind() == IdlType.e_simple )
	{
	    if ( ((IdlSimple)obj.current()).internal() != Token.t_void )
		{
		    output.println("\t\t\t\t\treturn _arg_ret;");
		}
	    else
		{
		    output.println("\t\t\t\t\treturn;");
		    noReturn = true;
		}
	}
    else
	{
	    output.println("\t\t\t\t\treturn _arg_ret;");
	}

    output.println("\t\t\t\t}");
    output.println("\t\t\t\tcatch( org.omg.CORBA.portable.RemarshalException _exception )");
    output.println("\t\t\t\t{");
    output.println("\t\t\t\t\tcontinue;");
    output.println("\t\t\t\t}");

    // Get the user exceptions
    output.println("\t\t\t\tcatch( org.omg.CORBA.portable.ApplicationException _exception )");
    output.println("\t\t\t\t{");
    output.println("\t\t\t\t\tString _exception_id = _exception.getId();");

    obj.pos(p);

    if ( obj.end() != true )
	if ( obj.current().kind() == IdlType.e_raises )
	    {
		r = (IdlRaises)obj.current();

		r.reset();

		while ( r.end() != true )
		    {

			output.println("\t\t\t\t\tif ( _exception_id.equals("+ fullname(r.current()) + "Helper.id()) )");
			output.println("\t\t\t\t\t{");
			output.println("\t\t\t\t\t\tthrow "+fullname(r.current())+"Helper.read(_exception.getInputStream());");
			output.println("\t\t\t\t\t}");
			output.println("");

			r.next();
		    }

	    }

    output.println("\t\t\t\t\tthrow new org.omg.CORBA.UNKNOWN(\"Unexcepected User Exception: \"+ _exception_id);");
    output.println("\t\t\t\t}");

    output.println("\t\t\t\tfinally");
    output.println("\t\t\t\t{");
    output.println("\t\t\t\t\tthis._releaseReply(_input);");
    output.println("\t\t\t\t}");

    boolean isAbstract = obj.upper().kind() == IdlType.e_interface &&
	((IdlInterface)(obj.upper())).abstract_interface();

    if ( org.openorb.compiler.IdlCompiler.local_stub )
	{
	    output.println("\t\t\t}");
	    output.println("\t\t\telse");
	    output.println("\t\t\t{");
	    output.println("\t\t\t\torg.omg.CORBA.portable.ServantObject _so = _servant_preinvoke(\"" +obj.name() +"\",_opsClass);");
	    output.println("\t\t\t\tif ( _so == null )");
	    output.println("\t\t\t\t   continue;");
	    if(isAbstract)
		output.println("\t\t\t\t" + fullname(obj.upper()) + " _self = ( " + fullname(obj.upper()) + " ) _so.servant;");
	    else
		output.println("\t\t\t\t" + fullname(obj.upper()) + "Operations _self = ( " + fullname(obj.upper()) + "Operations ) _so.servant;");
	    output.println("\t\t\t\ttry");
	    output.println("\t\t\t\t{");

	    if ( noReturn )
		output.print("\t\t\t\t\t_self." + obj.name()+"(");
	    else
		output.print("\t\t\t\t\treturn _self." + obj.name()+"(");

	    obj.reset();
	    obj.next();
	    if ( obj.end() != true )
		{
		    if ( obj.current().kind() == IdlType.e_param )
			{
			    someParams = true;
			    while ( obj.end() != true )
				{

				    obj.current().reset();
				    output.print(" "+obj.current().name());

				    obj.next();

				    if ( obj.end() != true )
					{
					    if ( obj.current().kind() == IdlType.e_param )
						output.print(", ");
					    else
						break;
					}
				}
			}
		}

	    c = getContext(obj);
	    if ( c != null )
		{
		    if ( someParams )
			output.print(", ");

		    output.print("ctx");
		}

	    output.println(");");

	    if ( noReturn )
		output.println("\t\t\t\t\treturn;");

	    output.println("\t\t\t\t}");
	    output.println("\t\t\t\tfinally");
	    output.println("\t\t\t\t{");
	    output.println("\t\t\t\t\t_servant_postinvoke(_so);");
	    output.println("\t\t\t\t}");
	    output.println("\t\t\t}");
	}

    output.println("\t\t}");

    output.println("\t}");
    output.println("");
}
	/**
	 * Translate the operations declaration of an interface
	 *
	 * @param	obj		interface from which operations must be translated
	 * @param	writeInto	the directory where the interface must be defined
	 */
	public void translate_interface_operations( IdlObject obj, java.io.File writeInto )
	{
		java.io.PrintWriter output;

		if ( !((IdlInterface)obj).abstract_interface() )
			output = newFile(writeInto,obj.name()+"Operations");
		else
			output = newFile(writeInto,obj.name());

		String old_pkg;
		java.util.Vector list = ((IdlInterface)obj).getInheritance();

		cartouche(output,obj);

		// Interface header
		if ( !((IdlInterface)obj).abstract_interface() )
		{
			output.print("public interface "+obj.name()+"Operations");
			if ( list.size() != 0 )
				output.print(" extends ");
		}
		else
		{
			output.print("public interface "+obj.name()+" extends org.omg.CORBA.portable.IDLEntity");
			if ( list.size() != 0 )
				output.print(", ");
		}
		if ( list.size() != 0 )
		{
			for ( int i=0; i<list.size(); i++ )
			{
				if ( !((IdlInterface)list.elementAt(i)).abstract_interface() )
					output.print( fullname( ((IdlObject)list.elementAt(i)) ) + "Operations" );
				else
					output.print( fullname( ((IdlObject)list.elementAt(i)) ));
				if ( i != (list.size()-1) )
					output.print(", ");
			}
			output.println("");
		}
		else
			output.println("");

		output.println("{");

		obj.reset();
		while ( obj.end() != true )
		{
			switch ( obj.current().kind() )
			{
			case IdlType.e_operation :
				translate_operation(obj.current(),output);
				break;
			case IdlType.e_attribute :
				translate_attribute(obj.current(),output);
				break;
			case IdlType.e_const:
				// Modification by Jojakim Stahl
				if (((IdlInterface)obj).abstract_interface())
				{
					translate_constant(obj.current(), null, output);
				}
				break;
			}

			obj.next();
		}

		output.println("}");
		output.close();
	}

	/**
	 * This method an interface is empty in terms of
	 * type typedef, union, exception...
	 */
	public boolean isEmptyInterface( IdlObject obj )
	{
		obj.reset();
		while ( obj.end() != true )
		{
			switch ( obj.current().kind() )
			{
			case IdlType.e_enum :
				return false;
			case IdlType.e_struct :
				return false;
			case IdlType.e_union :
				return false;
			case IdlType.e_typedef :
				return false;
			case IdlType.e_exception :
				return false;
			case IdlType.e_native :
				return false;
			}

			obj.next();
		}

		return true;
	}

	/**
	 * This method check a value type is empty in terms of
	 * type typedef, union, exception...
	 */
	public boolean isEmptyValue( IdlObject obj )
	{
		obj.reset();
		while ( obj.end() != true )
		{
			switch ( obj.current().kind() )
			{
			case IdlType.e_enum :
				return false;
			case IdlType.e_struct :
				return false;
			case IdlType.e_union :
				return false;
			case IdlType.e_typedef :
				return false;
			case IdlType.e_exception :
				return false;
			case IdlType.e_native :
				return false;
			}

			obj.next();
		}

		return true;
	}

	/**
	 * Check a type is empty in terms of
	 * type union, struct or enum
	 */
	public boolean isEmpty( IdlObject obj )
	{
		obj.reset();
		while ( obj.end() != true )
		{
			obj.current().reset	();
			switch ( obj.current().current().kind() )
			{
			case IdlType.e_enum :
				return false;
			case IdlType.e_struct :
				return false;
			case IdlType.e_union :
				return false;
			}

			obj.next();
		}

		return true;
	}

	/**
	 * Translate an interface
	 *
	 * @param	obj	interface to translate
	 * @param	writeInto	the directory where the interface must be defined
	 */
	public void translate_interface(IdlObject obj,java.io.File writeInto)
	{
		java.io.PrintWriter output = null;

		if ( !((IdlInterface)obj).abstract_interface() )
		{
			output = newFile(writeInto,obj.name());
			cartouche(output,obj);
		}

		String old_pkg;
		java.util.Vector list = ((IdlInterface)obj).getInheritance();

		// Declare the operations interface
		translate_interface_operations(obj, writeInto);

		java.io.File intoMe = null;
		if ( isEmptyInterface( obj ) == false )
			intoMe = createDirectory(obj.name()+"Package",writeInto);
		else
			intoMe = writeInto;

		if ( !((IdlInterface)obj).abstract_interface() )
		{
			// Interface header
			output.print("public interface "+obj.name()+" extends "+obj.name()+"Operations");
			if ( list.size() != 0 )
			{
				for ( int i=0; i<list.size(); i++ )
				{
					output.print(", "+fullname( ((IdlObject)list.elementAt(i)) ) );
				}
				//output.println("");
			}
			output.println(", org.omg.CORBA.Object, org.omg.CORBA.portable.IDLEntity");
			output.println("{");
		}

		// Translate the internal definitions
		old_pkg = current_pkg;

		addToPkg(obj,obj.name()+"Package");

		obj.reset();
		while ( obj.end() != true )
		{
			switch ( obj.current().kind() )
			{
			case IdlType.e_enum :
				translate_enum(obj.current(),intoMe);
				break;
			case IdlType.e_struct :
				translate_struct(obj.current(),intoMe);
				break;
			case IdlType.e_union :
				translate_union(obj.current(),intoMe);
				break;
			case IdlType.e_typedef :
				translate_typedef(obj.current(),intoMe);
				break;
			case IdlType.e_exception :
				translate_exception(obj.current(),intoMe);
				break;
			case IdlType.e_native :
				translate_native(obj.current(),intoMe);
				break;
			case IdlType.e_const :
				// Modification by Jojakim Stahl
				if (!((IdlInterface)obj).abstract_interface())
				{
					translate_constant(obj.current(),null,output);
				}
				break;
			}

			obj.next();
		}

		if ( !((IdlInterface)obj).abstract_interface() )
		{
			output.println("}");
			output.close();
		}

		current_pkg = old_pkg;

		write_helper(obj,writeInto);
		write_holder(obj,writeInto);
	}

	/**
	 * Translate an interface for the user
	 *
	 * @param	obj	interface to translate
	 * @param	writeInto	the directory where the interface must be defined
	 */
	public void translate_user_interface( IdlObject obj, java.io.File writeInto )
	{
		// Deprecated
		// java.io.PrintStream output = ...
		java.io.PrintWriter output = newFile(initial,obj.name()+"Impl");

		String old_pkg;

		output.println("//");
		output.println("// Interface implementation definition : " +obj.name());
		output.println("//");
		output.println("// !!!! THIS CODE MUST BE COMPLETED TO BE USED !!!");
		output.println("//");
		if ( org.openorb.compiler.IdlCompiler.pidl == false )
		{
			if ( org.openorb.compiler.IdlCompiler.map_poa == true )
			{
				if ( obj.upper().upper() != null )
					output.println("public class "+obj.name()+"Impl extends "+fullname( obj.upper() )+"."+obj.name()+"POA");
				else
					output.println("public class "+obj.name()+"Impl extends "+obj.name()+"POA");
			}
			else
			{
				if ( obj.upper().upper() != null )
					output.println("public class "+obj.name()+"Impl extends "+fullname( obj.upper() )+"._"+obj.name()+"ImplBase");
				else
					output.println("public class "+obj.name()+"Impl extends _"+obj.name()+"ImplBase");
			}
		}
		else
			output.println("public class "+obj.name()+"Impl implements "+fullname(obj)+"Operations");

		output.println("{");

		// Translate the internal definitions
		old_pkg = current_pkg;

		addToPkg(obj,obj.name()+"Package");

		obj.reset();
		while ( obj.end() != true )
		{
			switch ( obj.current().kind() )
			{
			case IdlType.e_operation :
				translate_user_operation(obj.current(),output);
				break;
			case IdlType.e_attribute :
				translate_user_attribute(obj.current(),output);
				break;
			}

			obj.next();
		}

		output.println("}");
		output.close();

		current_pkg = old_pkg;
	}

	/**
	 * Check if the operation or the attribute is already in the list
	 *
	 * @param	opList operations and attributes list
	 * @param	op	the operation or attribute name
	 * @return	true if the operation or attribute is not included in the list
	 */
	public boolean isInto( java.util.Vector opList, String op )
	{
		for ( int i=0; i<opList.size(); i++ )
		{
			if ( op.equals( ((IdlObject)opList.elementAt(i)).name() ) )
				return true;
		}

		return false;
	}

	/**
	 * Return the list of operations and attributes to implement
	 * (incl. inheritance).
	 *
	 * @param	obj	interface object
	 * @param	opList already found operations and attributes list
	 * @return	operations and attributes list
	 */
	protected java.util.Vector getInheritanceOpList( IdlObject obj, java.util.Vector opList )
	{
		java.util.Vector list = ((IdlInterface)obj).getInheritance();

		obj.reset();
		while ( obj.end() != true )
		{
			switch ( obj.current().kind() )
			{
			case IdlType.e_attribute :
			case IdlType.e_operation :
				if ( !isInto(opList, obj.current().name()) )
					opList.addElement(obj.current());
				break;
			}

			obj.next();
		}

		for ( int i=0; i<list.size(); i++)
		{
			IdlInterface itf = (IdlInterface)list.elementAt(i);
			if ( itf.kind() == IdlType.e_forward_interface )
				itf = itf.getInterface();

			opList = getInheritanceOpList(itf,opList);
		}

		return opList;
	}

	/**
	 * Return the list of operations and attributes to implement
	 * (incl. the inheritance).
	 *
	 * @param	obj	the interface object
	 * @param	opList already found operations and attributes list
	 * @return	operations and attributes list
	 */
	protected java.util.Vector getInheritanceList( IdlObject obj, java.util.Vector inList )
	{
		IdlInterface itf = ( IdlInterface ) obj;

		if ( itf.isForward() )
			itf = itf.getInterface();

		java.util.Vector list = itf.getInheritance();

		boolean found = false;

		for ( int i=0; i<inList.size(); i++ )
		{
			if ( ((String)inList.elementAt(i)).equals( obj.getId() ) )
			{
				found = true;
				break;
			}
		}

		if ( found == false )
		{
			inList.addElement( obj.getId() );
		}


		for ( int i=0; i<list.size(); i++)
			inList = getInheritanceList((IdlObject)list.elementAt(i),inList);

		return inList;
	}

	/**
	 * Return the initial name.
	 */
	private String initialName( String name )
	{
		if ( name.startsWith("_") )
			return name.substring(1, name.length());

		return name;
	}

	/**
	 * Translate a syub for an interface
	 *
	 * @param	obj	the interface to translate
	 * @param	writeInto	the directory where the interface must be defined
	 */
	public void translate_interface_stub( IdlObject obj, java.io.File writeInto )
	{
		java.io.PrintWriter output = newFile(writeInto,"_"+obj.name()+"Stub");

		String old_pkg;
		java.util.Vector list = new java.util.Vector();

		list = getInheritanceList( obj, list );

		cartouche(output,obj);

		// Interface header
		output.println("public class _"+obj.name()+"Stub extends org.omg.CORBA.portable.ObjectImpl");
		output.print("\t\timplements "+obj.name());
		output.println("");
		output.println("{");

		java.util.Vector intoList = getInheritanceOpList(obj, new java.util.Vector() );

		// Construct the _ids list
		output.println("\tstatic final String[] _ids_list =");
		output.println("\t{");

		if ( list.size() != 0 )
		{
			for ( int i=0; i<list.size(); i++ )
			{
				output.print("\t\t\""+(String)list.elementAt(i)+"\"");
				if ( i+1 < list.size() )
					output.println(", ");
			}
		}

		output.println("");
		output.println("\t};");
		output.println("");

        // Translate the operation _id
		output.println("\tpublic String[] _ids()");
		output.println("\t{");
		output.println("\t	return _ids_list;");
		output.println("\t}");
		output.println("");

    // Local invocations
		if (!org.openorb.compiler.IdlCompiler.dynamic && org.openorb.compiler.IdlCompiler.local_stub)
		{
      if ( !((IdlInterface)obj).abstract_interface() )
        output.println("\tprivate final static Class _opsClass = " + fullname(obj) + "Operations.class;");
      else
        output.println("\tprivate final static Class _opsClass = " + fullname(obj) + ".class;");
			output.println("");
		}
		if(application){
		    output.println("\tprivate int retention_id=-1;");
		    output.println("\tprivate java.lang.Object synch_ret_id=new java.lang.Object();");
		    output.println("\tprivate String client_id=null;");

		    output.println("\tprivate void client_id(){");

		    output.println("\t\t\torg.openorb.CORBA.kernel.ORBLoader ol=(org.openorb.CORBA.kernel.OpenORBLoader)((org.openorb.CORBA.ORB)_orb()).getFeature(\"ORBLoader\");");
		    output.println("\t\t\tif(ol!=null){");
		    output.println("\t\t\t\torg.openorb.CORBA.kernel.Property prop;");
		    output.println("\t\t\t\tprop=ol.getProperty(\"client.id\");");
		    output.println("\t\t\t\tif(prop!=null)");
		    output.println("\t\t\t\t\tclient_id=prop.getValue();");
		    output.println("\t\t\t}");
		    output.println("if(client_id==null) client_id=Runtime.getRuntime().toString()+\"*\"+Long.toString(System.currentTimeMillis())+\"*\"+synch_ret_id.toString();");
		    output.println("\t}");
		}
		// Translate the internal definitions
		old_pkg = current_pkg;

		addToPkg(obj,obj.name()+"Package");

		for ( int i=0; i<intoList.size(); i++ )
		{
			switch ( ((IdlObject)intoList.elementAt(i)).kind() )
			{
			case IdlType.e_operation :
				if ( org.openorb.compiler.IdlCompiler.dynamic )
					translate_operation_stub(((IdlObject)intoList.elementAt(i)),output);
				else
					translate_operation_stub_stream(((IdlObject)intoList.elementAt(i)),output);
				break;
			case IdlType.e_attribute :
				if ( org.openorb.compiler.IdlCompiler.dynamic )
					translate_attribute_stub(((IdlObject)intoList.elementAt(i)),output);
				else
					translate_attribute_stub_stream(((IdlObject)intoList.elementAt(i)),output);
				break;
			}
		}

		output.println("}");
		output.close();

		current_pkg = old_pkg;
	}

	/**
	 * Translate a skeleton for an interface
	 *
	 * @param	obj	the interface to translate
	 * @param	writeInto	the directory where the interface must be defined
	 */
	public void translate_interface_skel( IdlObject obj, java.io.File writeInto )
	{
		java.io.PrintWriter output = null;

		if ( org.openorb.compiler.IdlCompiler.map_poa == false )
			output = newFile(writeInto,"_"+obj.name()+"ImplBase");
		else
			output = newFile(writeInto,obj.name()+"POA");
		String old_pkg;
		IdlObject ob;
		java.util.Vector list = new java.util.Vector();

		list = getInheritanceList(obj, list);

		cartouche(output,obj);

		// Creation du package correspond au definitions internes de l'interface
		//java.io.File intoMe = getDirectory(obj.name()+"Package",writeInto);

		// En-tete de l'interface
		if ( org.openorb.compiler.IdlCompiler.map_poa == false )
		{
			if ( org.openorb.compiler.IdlCompiler.dynamic )
				output.println("public abstract class _"+obj.name()+"ImplBase extends org.omg.CORBA.DynamicImplementation");
			else
				output.println("public abstract class _"+obj.name()+"ImplBase extends org.omg.CORBA.portable.ObjectImpl");
			output.print("\t\timplements "+obj.name());
			if ( !org.openorb.compiler.IdlCompiler.dynamic )
				output.print(", org.omg.CORBA.portable.InvokeHandler");
			output.println("");
		}
		else
		{
			if ( org.openorb.compiler.IdlCompiler.dynamic )
				output.println("public abstract class "+obj.name()+"POA extends org.omg.PortableServer.DynamicImplementation");
			else
				output.println("public abstract class "+obj.name()+"POA extends org.omg.PortableServer.Servant");
			output.print("\t\timplements "+obj.name()+"Operations");
			if ( !org.openorb.compiler.IdlCompiler.dynamic )
				output.print(", org.omg.CORBA.portable.InvokeHandler");
			output.println("");
		}

		output.println("{");

		java.util.Vector intoList = getInheritanceOpList(obj, new java.util.Vector() );

		if ( org.openorb.compiler.IdlCompiler.map_poa == false )
		{
			// Construct the _ids list
			output.println("\tstatic final String[] _ids_list = ");
			output.println("\t{");

			if ( list.size() != 0 )
			{
				for ( int i=0; i<list.size(); i++ )
				{
					output.print("\t\t\""+(String)list.elementAt(i)+"\"");
					if ( i+1 < list.size() )
						output.println(", ");
				}
			}

			output.println("");
			output.println("\t};");
			output.println("");

			// Translate the operation _id
			output.println("\tpublic String[] _ids()");
			output.println("\t{");
			output.println("\t	return _ids_list;");
			output.println("\t}");
			output.println("");
		}
		else
		{
			// Operation _this
			output.println("\tpublic "+obj.name()+" _this()");
			output.println("\t{");
			output.println("\t\treturn "+obj.name()+"Helper.narrow(_this_object());");
			output.println("\t}");
			output.println("");

			// Operation _this( ... )
			output.println("\tpublic "+obj.name()+" _this( org.omg.CORBA.ORB orb )");
			output.println("\t{");
			output.println("\t\treturn "+obj.name()+"Helper.narrow(_this_object(orb));");
			output.println("\t}");
			output.println("");

			// Operation _all_interfaces

			output.println("\tprivate static String [] _ids_list =");
			output.println("\t{");

			if ( list.size() != 0 )
			{
				for ( int i=0; i<list.size(); i++ )
				{
					output.print("\t\t\""+(String)list.elementAt(i)+"\"");
					if ( i+1 < list.size() )
						output.println(", ");
				}
			}
			output.println("");
			output.println("\t};");
			output.println("");
			output.println("\tpublic String[] _all_interfaces( org.omg.PortableServer.POA poa, byte [] objectId)");
			output.println("\t{");
			output.println("\t\treturn _ids_list;");
			output.println("\t}");
			output.println("");
		}

		boolean bUseHash = intoList.size()>8;

		// generate ops hashtable
		if(bUseHash) {
			output.println("\tprivate static java.util.Hashtable _ops_Hash = new java.util.Hashtable();");
			output.println("\tstatic {");
			for ( int i=0; i<intoList.size(); i++ )
			{
				ob = ((IdlObject)intoList.elementAt(i));

				switch ( ob.kind() )
				{
				case IdlType.e_operation :
					output.println("\t\t_ops_Hash.put(\"" + initialName(ob.name()) + "\", new Integer(" + (i<<2) + ") );");
					break;
				case IdlType.e_attribute :
					output.println("\t\t_ops_Hash.put(\"_get_" + initialName(ob.name()) + "\", new Integer(" + (i<<2) + ") );");
					if ( ((IdlAttribute)ob).readOnly() == false )
						output.println("\t\t_ops_Hash.put(\"_set_" + initialName(ob.name()) + "\", new Integer(" + ((i<<2) | 1) + ") );");
					break;
				}
			}
			output.println("\t};");
			output.println("");
		}

		// Translate the invoke operation
		if ( org.openorb.compiler.IdlCompiler.dynamic )
			output.println("\tpublic final void invoke ( org.omg.CORBA.ServerRequest request )");
		else
		{
			output.println("\tpublic org.omg.CORBA.portable.OutputStream _invoke ( String opName,");
			output.println("\t                                                     org.omg.CORBA.portable.InputStream _is,");
			output.println("\t                                                     org.omg.CORBA.portable.ResponseHandler handler)");
		}
		output.println("\t{");
		if ( org.openorb.compiler.IdlCompiler.dynamic )
		{
			output.println("\t\t\torg.omg.CORBA.ORB orb = org.omg.CORBA.ORB.init();");
			output.println("\t\tString opName = request.operation();");
		}
		else
		{
			output.println("\t\torg.omg.CORBA.portable.OutputStream _output = null;");
		}
		output.println("");

		if(bUseHash) {
			output.println("\t\tInteger opidx = (Integer)_ops_Hash.get(opName);");
			output.println("\t\tif(opidx == null)");
			output.println("\t\t\tthrow new org.omg.CORBA.BAD_OPERATION();");
			output.println("");
			output.println("\t\tswitch(opidx.intValue()) {");
		}

		for ( int i=0; i<intoList.size(); i++ )
		{
			ob = ((IdlObject)intoList.elementAt(i));

			switch ( ob.kind() )
			{
			case IdlType.e_operation :
				if(bUseHash)
					output.println("\t\tcase " + (i<<2) + ":");
				else
					output.println("\t\tif ( opName.equals(\"" + initialName(ob.name()) + "\") )");
				output.println("\t\t{");
				if ( org.openorb.compiler.IdlCompiler.dynamic )
					translate_operation_skel(ob,output);
				else
					translate_operation_skel_stream(ob,output);
				output.println("\t\t}");
				break;
			case IdlType.e_attribute :
				if(bUseHash)
					output.println("\t\tcase " + (i<<2) + ":");
				else
					output.println("\t\tif ( opName.equals(\"_get_" + initialName(ob.name()) + "\") )");
				output.println("\t\t{");
				if ( org.openorb.compiler.IdlCompiler.dynamic )
					translate_read_attribute_skel(ob,output);
				else
					translate_read_attribute_skel_stream(ob,output);
				output.println("\t\t}");
				if ( ((IdlAttribute)ob).readOnly() == false )
				{
					if(bUseHash)
						output.println("\t\tcase " + ((i<<2)|1) + ":");
					else {
						output.println("\t\telse");
						output.println("\t\tif ( opName.equals(\"_set_" + initialName(ob.name()) + "\") )");
					}
					output.println("\t\t{");
					if ( org.openorb.compiler.IdlCompiler.dynamic )
						translate_write_attribute_skel(ob,output);
					else
						translate_write_attribute_skel_stream(ob,output);
					output.println("\t\t}");
				}
				break;
			}
			if(!bUseHash)
				output.println("\t\telse");
		}
		if(bUseHash) {
			output.println("\t\t}");
			output.println("");
			output.println("\t\t// never reached");
			output.println("\t\tthrow new org.omg.CORBA.INTERNAL();");
		} else
			output.println("\t\t	throw new org.omg.CORBA.BAD_OPERATION();");
		output.println("\t}");

		output.println("}");
		output.close();
	}

	/**
	 * Translate a TIE approach for an interface
	 *
	 * @param	obj	the interface to translate
	 * @param	writeInto	the directory where the interface must be defined
	 */
	public void translate_interface_tie( IdlObject obj, java.io.File writeInto )
	{
		// Deprecated
		// java.io.PrintStream output = ...
		java.io.PrintWriter output = null;

		if ( org.openorb.compiler.IdlCompiler.map_poa == false )
			output = newFile(writeInto,obj.name()+"Tie");
		else
			output = newFile(writeInto,obj.name()+"POATie");

		cartouche(output,obj);

		// Creating package
		//java.io.File intoMe = getDirectory(obj.name()+"Package",writeInto);

		// Interface header
		if ( org.openorb.compiler.IdlCompiler.map_poa == false )
			output.println("public class "+obj.name()+"Tie extends _"+obj.name()+"ImplBase");
		else
			output.println("public class "+obj.name()+"POATie extends "+obj.name()+"POA");
		//output.println("\t\timplements "+obj.name());

		output.println("{");
		output.println("");


		// Private member for delegation
		output.println("\t//");
		output.println("\t// Private reference to implementation object");
		output.println("\t//");
		output.println("\tprivate "+obj.name()+"Operations _tie;");
		output.println("");

		if ( org.openorb.compiler.IdlCompiler.map_poa == true )
		{
			// Private member for POA
			output.println("\t//");
			output.println("\t// Private reference to POA");
			output.println("\t//");
			output.println("\tprivate org.omg.PortableServer.POA _poa;");
			output.println("");
		}

		// Constructor
		output.println("\t//");
		output.println("\t// Constructor");
		output.println("\t//");
		if ( org.openorb.compiler.IdlCompiler.map_poa == false )
			output.println("\tpublic "+obj.name()+"Tie( "+obj.name()+"Operations tieObject )");
		else
			output.println("\tpublic "+obj.name()+"POATie( "+obj.name()+"Operations tieObject )");
		output.println("\t{");
		output.println("\t\t_tie = tieObject;");
		output.println("\t}");
		output.println("");

		if ( org.openorb.compiler.IdlCompiler.map_poa == true )
		{
			output.println("\t//");
			output.println("\t// Constructor");
			output.println("\t//");
			output.println("\tpublic "+obj.name()+"POATie( "+obj.name()+"Operations tieObject, org.omg.PortableServer.POA poa )");
			output.println("\t{");
			output.println("\t\t_tie = tieObject;");
			output.println("\t\t_poa = poa;");
			output.println("\t}");
			output.println("");
		}

		output.println("\t//");
		output.println("\t// Get the delegate");
		output.println("\t//");
		output.println("\tpublic "+obj.name()+"Operations _delegate()");
		output.println("\t{");
		output.println("\t\treturn _tie;");
		output.println("\t}");
		output.println("");

		output.println("\t//");
		output.println("\t// Set the delegate");
		output.println("\t//");
		output.println("\tpublic void _delegate( "+obj.name()+"Operations delegate_)");
		output.println("\t{");
		output.println("\t\t_tie = delegate_;");
		output.println("\t}");
		output.println("");

		if ( org.openorb.compiler.IdlCompiler.map_poa == true )
		{
			output.println("\t//");
			output.println("\t//");
			output.println("\t//");
			output.println("\tpublic org.omg.PortableServer.POA _default_POA()");
			output.println("\t{");
			output.println("\t\tif ( _poa != null )");
			output.println("\t\t    return _poa;");
			output.println("\t\telse");
			output.println("\t\t    return super._default_POA();");
			output.println("\t}");
			output.println("");
		}

		java.util.Vector intoList = getInheritanceOpList(obj, new java.util.Vector() );

		for ( int i=0; i<intoList.size(); i++ )
		{
			switch ( ((IdlObject)intoList.elementAt(i)).kind() )
			{
			case IdlType.e_operation :
				translate_operation_tie(((IdlObject)intoList.elementAt(i)),output);
				break;
			case IdlType.e_attribute :
				translate_attribute_tie(((IdlObject)intoList.elementAt(i)),output);
				break;
			}
		}

		output.println("}");
		output.close();
	}

	/**
	 * Translate a module
	 *
	 * @param	obj	the module to translate
	 * @param	writeInto	the directory where the module must be defined
	 */
	public void translate_module( IdlObject obj, java.io.File writeInto, int translateType )
	{
		String old_pkg;

		java.io.File intoModule;

		if ( translateType == 0 )
			intoModule = createDirectory(obj.name(),writeInto);
		else
			intoModule = getDirectory(obj.name(),writeInto);

		old_pkg = current_pkg;

		addToPkg(obj,obj.name());

		translate_object(obj,intoModule,translateType);

		current_pkg = old_pkg;
	}

	/**
	 * Translate a user module
	 *
	 * @param	obj	the module to translate
	 * @param	writeInto	the directory where the module must be defined
	 */
	public void translate_user_module( IdlObject obj, java.io.File writeInto )
	{
		String old_pkg;

		java.io.File intoModule;

		if ( obj.getPrefix() != null )
		{
			writeInto = getPrefixDirectories(obj.getPrefix(),writeInto);
		}

		intoModule = getDirectory(obj.name(),writeInto);

		old_pkg = current_pkg;

		if ( obj.getPrefix() != null )
		{
			if ( org.openorb.compiler.IdlCompiler.reversePrefix )
				addToPkg(obj,inversedPrefix(obj.getPrefix())+"."+obj.name());
			else
				addToPkg(obj,obj.getPrefix()+"."+obj.name());
		}
		else
			addToPkg(obj,obj.name());

		translate_user_object(obj,intoModule);

		current_pkg = old_pkg;
	}

	/**
	 * Translate a value type state
	 */
	public void translate_state_member( IdlObject obj, java.io.PrintWriter output )
	{
		IdlStateMember member = ( IdlStateMember ) obj;

		if ( obj.hasComment() )
			javadoc(output,obj);
		else
		{
			output.println("\t//");
			if ( member.public_member() )
				output.println("\t// Public member : " + obj.name());
			else
				output.println("\t// Private member : " + obj.name());
			output.println("\t//");
		}

		if ( member.public_member() )
			output.print("\tpublic ");
		else
			output.print("\tprotected ");

		obj.reset();
		translate_type( obj.current(), output,true );

		output.println(" "+obj.name()+";");
		output.println("");
	}

	/**
	 * Return the list of the inherited members
	 *
	 */
	private java.util.Vector getInheritedStateMember( IdlObject obj )
	{
		java.util.Vector list = new java.util.Vector();
		java.util.Vector sub = null;

		IdlValue [] inherited = ((IdlValue)obj).getInheritance();
		if ( inherited.length != 0 )
		{
			sub = getInheritedStateMember( inherited[0] );
		}

		if ( sub != null )
		{
			for ( int i=0; i<sub.size(); i++ )
				list.addElement( sub.elementAt(i) );
		}

		obj.reset();
		while ( obj.end() != true )
		{
			if ( obj.current().kind() == IdlType.e_state_member )
				list.addElement( obj.current() );
			obj.next();
		}

		return list;
	}

	/**
	 * Translate a Value Type
	 *
	 * @param	obj	la value type to translate
	 * @param	writeInto	the directory where the module must be defined
	 */
	public void translate_value_type( IdlObject obj, java.io.File writeInto )
	{
		IdlValue value = ( IdlValue ) obj;
		boolean base_custom = false;
		String old_pkg;

		java.io.PrintWriter output = newFile(writeInto, obj.name() );

		cartouche(output,obj);

		if ( value.abstract_value() )
		{
			output.print("public interface " + obj.name()+" extends org.omg.CORBA.portable.ValueBase");

			IdlValue [] abs_inheritance = value.getInheritance();

			if ( abs_inheritance.length != 0 )
				output.print(", ");

			for ( int i=0; i<abs_inheritance.length; i++ )
			{
				output.print( fullname(abs_inheritance[i]) );

				if ( i+1 < abs_inheritance.length )
					output.print(", ");
			}

			output.println();

			if ( value.supports().size() != 0 )
			{
				output.print("\t\t, ");

				java.util.Vector list = value.supports();

				for ( int i=0; i<list.size(); i++ )
				{
					IdlInterface itf = ( IdlInterface ) list.elementAt(i);

					output.print( itf.name() );

					if ( !itf.abstract_interface() )
						output.print("Operations");

					if ( i+1 < list.size() )
						output.print(", ");
				}
			}

			output.println("{");

			// Functions from IDL description

			java.io.File intoMe = null;
			if ( isEmptyValue( obj ) == false )
				intoMe = createDirectory(obj.name()+"Package",writeInto);
			else
				intoMe = writeInto;

			// Traduit les definitions internes
			old_pkg = current_pkg;

			addToPkg(obj,obj.name()+"Package");

			obj.reset();
			while ( obj.end() != true )
			{
				switch ( obj.current().kind() )
				{
				case IdlType.e_enum :
					translate_enum(obj.current(),intoMe);
					break;
				case IdlType.e_struct :
					translate_struct(obj.current(),intoMe);
					break;
				case IdlType.e_union :
					translate_union(obj.current(),intoMe);
					break;
				case IdlType.e_typedef :
					translate_typedef(obj.current(),intoMe);
					break;
				case IdlType.e_exception :
					translate_exception(obj.current(),intoMe);
					break;
				case IdlType.e_native :
					translate_native(obj.current(),intoMe);
					break;
				case IdlType.e_const :
					translate_constant(obj.current(),null,output);
					break;
				case IdlType.e_operation :
					translate_operation(obj.current(),output);
					break;
				case IdlType.e_attribute :
					translate_attribute(obj.current(),output);
					break;
				}

				obj.next();
			}

			current_pkg = old_pkg;

			output.println("}");
			output.println("");
			output.close();
		}
		else
		{
			// -----------------------------------------
			// The class implementing the concrete value
			// -----------------------------------------

			output.print("public abstract class " + obj.name());

			// Case : Values types that do not inherit from other values or interfaces
			if ( ( value.supports().size() == 0 ) && ( value.getInheritance().length == 0 ) )
			{
				if ( value.custom_value() )
				{
					output.println(" implements org.omg.CORBA.portable.CustomValue");
				}
				else
				{
					output.println(" implements org.omg.CORBA.portable.StreamableValue");
				}
			}
			else
			{

				IdlValue [] inheritance = value.getInheritance();

				boolean extends_value = false;
				boolean streamNeed = false;

				if ( inheritance.length == 0 )
					streamNeed = true;

				if ( inheritance.length != 0 )
					extends_value = true;

				for ( int i=0; i<inheritance.length; i++ )
				{
					if ( inheritance[i].custom_value() )
						base_custom = true;
				}

				if ( ( extends_value == false ) && ( value.supports().size() == 0 ) )
				{
					output.print(" implements org.omg.CORBA.portable.StreamableValue");
				}

				boolean extend_concrete = false;
				boolean extend_abstract = false;

				for ( int i=0; i<inheritance.length; i++ )
				{
					if ( ! inheritance[i].abstract_value() )
						extend_concrete = true;
					else
						extend_abstract = true;
				}

				// case : Inheritance from other stateful values
				if ( extends_value )
				{
					if ( extend_concrete )
					{
						output.print(" extends ");
						int j = 0;
						for ( int i=0; i<inheritance.length; i++ )
						{
							if ( inheritance[i].abstract_value() == false )
							{
								j++;
								if ( j > 1 )
									output.print(", ");
								output.print( fullname(inheritance[i]) );
							}
						}
					}
					if ( extend_abstract )
					{
						output.print(" implements ");
						int j = 0;
						for ( int i=0; i<inheritance.length; i++ )
						{
							if ( inheritance[i].abstract_value() )
							{
								j++;
								if ( j > 1 )
									output.print(", ");
								output.print( fullname(inheritance[i]) );
							}
						}

						if ( !extend_concrete )
							output.print(", org.omg.CORBA.portable.StreamableValue");
					}

				}
				if (!( extend_abstract && ( value.supports().size() != 0 ) ))
					output.println("");

				// case : Inheritance from abstract values
				boolean implement_map = false;
				if ( value.custom_value() && !base_custom )
				{
					output.print("\t\timplements org.omg.CORBA.portable.CustomValue");
					implement_map = true;
				}

				// case : Supported interfaces
				if ( value.supports().size() != 0 )
				{
					if ( implement_map )
						output.print(", ");
					else
					{
						if ( !extend_abstract )
							output.print("\t\timplements ");
						else
							output.print(", ");
						if ( streamNeed )
							output.print("org.omg.CORBA.portable.StreamableValue, ");
					}

					java.util.Vector list = value.supports();

					for ( int i=0; i<list.size(); i++ )
					{
						IdlInterface itf = ( IdlInterface ) list.elementAt(i);

						output.print( itf.name() );

						if ( !itf.abstract_interface() )
							output.print("Operations");

						if ( i+1 < list.size() )
							output.print(", ");
					}
				}

				output.println("");
			}

			output.println("{");

			// Functions from IDL description

			java.io.File intoMe = null;
			if ( isEmptyValue( obj ) == false )
				intoMe = createDirectory(obj.name()+"Package",writeInto);
			else
				intoMe = writeInto;

			// Translate the internal definitions
			old_pkg = current_pkg;

			addToPkg(obj,obj.name()+"Package");

			obj.reset();
			while ( obj.end() != true )
			{
				switch ( obj.current().kind() )
				{
				case IdlType.e_enum :
					translate_enum(obj.current(),intoMe);
					break;
				case IdlType.e_struct :
					translate_struct(obj.current(),intoMe);
					break;
				case IdlType.e_union :
					translate_union(obj.current(),intoMe);
					break;
				case IdlType.e_typedef :
					translate_typedef(obj.current(),intoMe);
					break;
				case IdlType.e_exception :
					translate_exception(obj.current(),intoMe);
					break;
				case IdlType.e_native :
					translate_native(obj.current(),intoMe);
					break;
				case IdlType.e_const :
					translate_constant(obj.current(),null,output);
					break;
				case IdlType.e_operation :
					translate_operation(obj.current(),output);
					break;
				case IdlType.e_attribute :
					translate_attribute(obj.current(),output);
					break;
				case IdlType.e_state_member :
					translate_state_member(obj.current(),output);
					break;
				}

				obj.next();
			}

			current_pkg = old_pkg;

			// Functions from ValueBase

			output.println("\t//");
			output.println("\t// Return the truncatable ids");
			output.println("\t//");

			output.println("\tstatic final String[] _ids_list =");
			output.println("\t{");

			String [] list = value.truncatableList();
			int max = list.length;
			if ( !value.isTruncatable() )
				max = 1;
			for ( int i=0; i<max; i++ )
			{
				output.print("\t\t\"" + list[i] + "\"");
				if ( i+1 < list.length )
					output.print(",");
				output.println("");
			}

			output.println("\t};");
			output.println("");

			output.println("\tpublic String [] _truncatable_ids()");
			output.println("\t{");
			output.println("\t\treturn _ids_list;");
			output.println("\t}");
			output.println("");

			// Functions from Streamable
			if ( !value.custom_value() && !base_custom )
			{
				output.println("\t//");
				output.println("\t// Unmarshal the value into an InputStream");
				output.println("\t//");
				output.println("\tpublic void _read( org.omg.CORBA.portable.InputStream is )");
				output.println("\t{");

				java.util.Vector stateList = getInheritedStateMember( obj );

				for ( int i=0; i<stateList.size(); i++ )
				{
					IdlStateMember state = ( IdlStateMember ) stateList.elementAt(i);
					state.reset();
					translate_unmarshalling_member(state.current(),output,"is",state.name(),"\t\t");
				}

				output.println("\t}");
				output.println("");

				output.println("\t//");
				output.println("\t// Marshal the value into an OutputStream");
				output.println("\t//");
				output.println("\tpublic void _write( org.omg.CORBA.portable.OutputStream os )");
				output.println("\t{");

				for ( int i=0; i<stateList.size(); i++ )
				{
					IdlStateMember state = ( IdlStateMember ) stateList.elementAt(i);
					state.reset();
					translate_marshalling_member(state.current(),output,"os",state.name(),"\t\t");
				}

				output.println("\t}");
				output.println("");

				output.println("\t//");
				output.println("\t// Return the value TypeCode");
				output.println("\t//");
				output.println("\tpublic org.omg.CORBA.TypeCode _type()");
				output.println("\t{");
				output.println("\t\treturn " + fullname(obj) + "Helper.type();");
				output.println("\t}");
				output.println("");
			}

			output.println("}");
			output.println("");
			output.close();

			// ----------------------------------
			// The class implementing the factory
			// ----------------------------------
			boolean factory = false;
			obj.reset();
			while ( obj.end() != true )
			{
				if ( obj.current().kind() == IdlType.e_factory )
				{
					factory = true;
					break;
				}
				obj.next();
			}

			if ( factory )
			{
				output = newFile(writeInto, obj.name() + "ValueFactory" );

				cartouche(output,obj.current());

				output.println("public interface " + obj.name()+"ValueFactory extends org.omg.CORBA.portable.ValueFactory");
				output.println("{");

				obj.reset();
				while ( obj.end() != true )
				{
					if ( obj.current().kind() == IdlType.e_factory )
					{
						output.println("\t//");
						output.println("\t// Return the value type");
						output.println("\t//");
						output.print("\tpublic abstract "+ fullname(obj) + " " + obj.current().name() + "(");

						obj.current().reset();
						while ( obj.current().end() != true )
						{
							IdlFactoryMember member = ( IdlFactoryMember ) obj.current().current();

							member.reset();
							translate_type(member.current(), output,true);
							output.print(" "+member.name());

							obj.current().next();

							if ( obj.current().end() != true )
								output.print(", ");
						}

						output.println(");");
						output.println("");
					}

					obj.next();
				}

				output.println("}");
				output.println("");
				output.close();
			}
		}

		write_holder(obj,writeInto);
		write_helper(obj,writeInto);
	}

	/**
	 * Translate a Value Box
	 *
	 * @param	obj	the value box to translate
	 * @param	writeInto	the directory where the module must be defined
	 */
	public void translate_value_box( IdlObject obj, java.io.File writeInto )
	{
		IdlValueBox value = ( IdlValueBox ) obj;

		if ( value.simple() )
		{
			java.io.PrintWriter output = newFile(writeInto, obj.name() );

			cartouche(output,obj);

			output.println("public class " + obj.name()+" implements org.omg.CORBA.portable.ValueBase");
			output.println("{");

			output.println("\t//");
			output.println("\t// Reference to the boxed value");
			output.println("\t//");
			output.print("\tpublic ");

			obj.reset();
			translate_type(obj.current(), output,true);
			output.println(" value;");
			output.println("");

			output.println("\t//");
			output.println("\t// Constructor");
			output.println("\t//");
			output.println("\t// @param initial the initial boxed value");
			output.println("\t//");
			output.print("\tpublic "+obj.name()+"( ");
			translate_type(obj.current(), output,true);
			output.println(" initial )");
			output.println("\t{");
			output.println("\t\tvalue = initial;");
			output.println("\t}");
			output.println("");

			output.println("\t//");
			output.println("\t// Return value box id");
			output.println("\t//");
			output.println("\tprivate static String[] _ids = { "+obj.name()+"Helper.id() };");
			output.println("");

			output.println("\t//");
			output.println("\t// Return truncatable ids");
			output.println("\t//");
			output.println("\tpublic String[] _truncatable_ids()");
			output.println("\t{");
			output.println("\t\treturn _ids;");
			output.println("\t}");
			output.println("");

			output.println("}");

			output.close();

			write_holder(obj,writeInto);
			write_helper(obj,writeInto);
		}
		else
		{
			// Definition of the sub-types
			obj.reset();
			while ( obj.end() != true )
			{
				switch ( obj.current().kind() )
				{
				case IdlType.e_union :
					translate_union(obj.current(),writeInto);
					break;
				case IdlType.e_struct :
					translate_struct(obj.current(),writeInto);
					break;
				case IdlType.e_enum :
					translate_enum(obj.current(),writeInto);
					break;
				}
				obj.next();
			}

			obj.reset();
			switch ( obj.current().kind() )
			{
			case IdlType.e_string :
			case IdlType.e_wstring :
			case IdlType.e_simple :
			case IdlType.e_fixed :
			case IdlType.e_union :
			case IdlType.e_struct :
			case IdlType.e_enum :
			case IdlType.e_sequence :
			case IdlType.e_array :
			case IdlType.e_typedef :
			case IdlType.e_ident :
				write_helper(obj,writeInto);
				write_holder(obj,writeInto);
				break;
			}
		}

	}

	/**
	 * Translate a native type
	 *
	 * @param	obj	the module to translate
	 * @param	writeInto	the directory where the module must be defined
	 */
	public void translate_native( IdlObject obj, java.io.File writeInto )
	{
		write_holder(obj,writeInto);
		write_helper(obj,writeInto);
	}

	/**
	 * Translate an object content
	 */
	public void translate_object_content( IdlObject obj, java.io.File writeInto, int translateType )
	{
		switch ( obj.current().kind() )
		{
		case IdlType.e_module :
			translate_module(obj.current(),writeInto,translateType);
			break;
		case IdlType.e_const :
			if ( translateType == 0 )
				translate_constant(obj.current(),writeInto,null);
			break;
		case IdlType.e_enum :
			if ( translateType == 0 )
				translate_enum(obj.current(),writeInto);
			break;
		case IdlType.e_struct :
			if ( translateType == 0 )
				translate_struct(obj.current(),writeInto);
			break;
		case IdlType.e_union :
			if ( translateType == 0 )
				translate_union(obj.current(),writeInto);
			break;
		case IdlType.e_typedef :
			if ( translateType == 0 )
				translate_typedef(obj.current(),writeInto);
			break;
		case IdlType.e_exception :
			if ( translateType == 0 )
				translate_exception(obj.current(),writeInto);
			break;
		case IdlType.e_native :
			if ( translateType == 0 )
				translate_native(obj.current(),writeInto);
			break;
		case IdlType.e_value_box :
			if ( translateType == 0 )
				translate_value_box(obj.current(),writeInto);
			break;
		case IdlType.e_value :
			if ( translateType == 0 )
				translate_value_type(obj.current(),writeInto);
			break;
		case IdlType.e_interface :
			if ( translateType == 0 )
				translate_interface(obj.current(),writeInto);
			else
				if ( translateType == 1 )
				{
					if ( ((IdlInterface)(obj.current())).local_interface() == false )
						translate_interface_stub(obj.current(),writeInto);
				}
				else
					if ( translateType == 2 )
					{
						if ( ((IdlInterface)(obj.current())).abstract_interface() == false )
						{
							if ( ((IdlInterface)(obj.current())).local_interface() == false )
								translate_interface_skel(obj.current(),writeInto);
						}
					}
					else
					{
						if ( translateType == 3 )
						{
							if ( ((IdlInterface)(obj.current())).abstract_interface() == false )
							{
								if ( ((IdlInterface)(obj.current())).local_interface() == false )
									translate_interface_tie(obj.current(),writeInto);
							}
						}
					}
			break;
		}

	}

	/**
	 * Translate the data from a container object ( Module, Interface, Root )
	 *
	 * @param	obj	the object to translate
	 * @param	writeInto	the write access
	 * @param	translateType	the translation type ( 0=data, 1=stub, 2=skeleton )
	 */
	public void translate_object( IdlObject obj, java.io.File writeInto, int translateType )
	{
		obj.reset();
		while ( obj.end() != true )
		{
			java.io.File tmpInto = writeInto;

			if ( obj.current().included() == false )
			{
				String old_pkg = current_pkg;

				if ( org.openorb.compiler.IdlCompiler.usePrefix )
					if ( ( obj.current().getPrefix() != null ) && ( obj.kind() == IdlType.e_root ) )
					{
						if ( translateType == 0 )
						{
							tmpInto = createPrefixDirectories(obj.current().getPrefix(),writeInto);
						}
						else
						{
							tmpInto = getPrefixDirectories(obj.current().getPrefix(),writeInto);
						}

						if ( org.openorb.compiler.IdlCompiler.reversePrefix )
							addToPkg(obj,inversedPrefix(obj.current().getPrefix()));
						else
							addToPkg(obj,obj.current().getPrefix());
					}

				translate_object_content( obj, tmpInto, translateType );

				current_pkg = old_pkg;
			}

			obj.next();
		}
	}

	/**
	 * Translate the data from a container object ( Module, Interface, Root )
	 * for the user code
	 *
	 * @param	obj	the object to translate
	 * @param	writeInto	the write access
	 */
	public void translate_user_object( IdlObject obj, java.io.File writeInto )
	{
		obj.reset();

		while ( obj.end() != true )
		{
			if ( obj.included() == false )
				switch ( obj.current().kind() )
				{
				case IdlType.e_module :
					translate_user_module(obj.current(),writeInto);
					break;
				case IdlType.e_interface :
					translate_user_interface(obj.current(),writeInto);
					break;
				}

			obj.next();
		}
	}

	/**
	 * Translate the data from IDL to Java
	 *
	 * @param	obj	The compilation graph root
	 * @param	packageName	The directory where the definitions are added
	 */
	public void translateData( IdlObject obj, String packageName )
	{
		_root = obj;

		java.io.File first = null;

		if ( org.openorb.compiler.IdlCompiler.outdir != null )
			first = new java.io.File( org.openorb.compiler.IdlCompiler.outdir );

		java.io.File writeInto = createDirectory(packageName,first);

		translate_object(obj,writeInto,0);
	}

	/**
	 * Create the class for the delegation
	 *
	 * @param	obj	The compilation graph root
	 * @param	packageName	The directory where the definitions are added
	 */
	public void translateTIE( IdlObject obj, String packageName )
	{
		_root = obj;

		if ( org.openorb.compiler.IdlCompiler.pidl )
			return;

		java.io.File first = null;

		if ( org.openorb.compiler.IdlCompiler.outdir != null )
			first = new java.io.File( org.openorb.compiler.IdlCompiler.outdir );

		java.io.File writeInto = getDirectory(packageName,first);

		translate_object(obj,writeInto,3);
	}

	/**
	 * Generate the user code ( implementation class )
	 *
	 * @param	obj	The compilation graph root
	 * @param	packageName	The directory where the definitions are added
	 */
	public void translateUser( IdlObject obj, String packageName )
	{
		_root = obj;

		java.io.File first = null;

		if ( org.openorb.compiler.IdlCompiler.outdir != null )
			first = new java.io.File( org.openorb.compiler.IdlCompiler.outdir );

		java.io.File writeInto = getDirectory(packageName,first);

		initial = writeInto;

		translate_user_object(obj,writeInto);
	}

	/**
	 * Creates the stub
	 *
	 * @param	obj	The compilation graph root
	 * @param	packageName	The directory where the definitions are added
	 */
	public void translateStub(IdlObject obj, String packageName )
	{
		_root = obj;

		java.io.File first = null;

		if ( org.openorb.compiler.IdlCompiler.outdir != null )
			first = new java.io.File( org.openorb.compiler.IdlCompiler.outdir );

		java.io.File writeInto = getDirectory(packageName,first);

		translate_object(obj,writeInto,1);
	}

	/**
	 * Creates the skeleton
	 *
	 * @param	obj	The compilation graph root
	 * @param	packageName	The directory where the definitions are added
	 */
	public void translateSkeleton( IdlObject obj, String packageName )
	{
		_root = obj;

		java.io.File first = null;

		if ( org.openorb.compiler.IdlCompiler.outdir != null )
			first = new java.io.File( org.openorb.compiler.IdlCompiler.outdir );

		java.io.File writeInto = getDirectory(packageName,first);

		translate_object(obj,writeInto,2);
	}



}
