/**
 * Redistribution and use of this software and associated
 * documentation ("Software"), with or without modification, are
 * permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright statements
 *    and notices.  Redistributions must also contain a copy of this
 *    document.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Intalio Inc.  For written permission, please
 *    contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Intalio Inc. Exolab is a registered trademark of
 *    Intalio Inc.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY INTALIO AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL INTALIO OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Copyright 2000 (C) Intalio Inc. All Rights Reserved.
 *
 * $Id: AbstractServerRequest.java,v 1.12 2001/08/13 03:47:34 wood Exp $
 *
 * Date         Author  Changes
 */


/** Modified/extended by Diana Szentivanyi in June 2003
* contact diasz@ida.liu.se
*/

package org.openorb.net;

import java.util.Map;
import java.util.HashMap;

import org.omg.CORBA.CompletionStatus;

//import org.openorb.CORBA.SystemExceptionHelper;
import org.openorb.adapter.*;

import org.openorb.CORBA.MinorCodes;
import org.omg.CORBA.CompletionStatus;
import org.openorb.util.RepoIDHelper;

/**
 * Base class which implements most of the server request functionality.
 * Only protocol specific things are not implemented.
 *
 * @author <a href="mailto:wood@intalio.com">Chris Wood &lt;wood@intalio.com&gt;</a>
 * @version $Revision: 1.12 $ $Date: 2001/08/13 03:47:34 $
 */
abstract public class AbstractServerRequest
    extends org.omg.CORBA.LocalObject
    implements ServerRequest
{
    /**
     * One pass constructor for request
     */
    public AbstractServerRequest(ServerManager serverManager, ServerChannel channel,
				 Object transportAssociation, int request_id,
				 org.omg.CORBA.portable.InputStream argument_stream,
				 byte [] object_key, String operation, byte sync_scope,
				 org.omg.IOP.ServiceContext[] request_service_contexts)
    {
	init_one(serverManager, channel, transportAssociation, request_id, argument_stream);
	init(object_key, operation, sync_scope, request_service_contexts);
    }

    /**
     * Two pass constructor for locate request. this constructor must be
     * followed by a call to init.
     */
    public AbstractServerRequest(ServerManager serverManager, ServerChannel channel,
				 Object transportAssociation, int request_id,
				 org.omg.CORBA.portable.InputStream argument_stream)
    {
	init_one(serverManager, channel, transportAssociation, request_id, argument_stream);
    }

    private void init_one(ServerManager serverManager, ServerChannel channel,
			  Object transportAssociation, int request_id,
			  org.omg.CORBA.portable.InputStream argument_stream)
    {
	_serverManager = serverManager;
	_channel = channel;
	_transportAssociation = transportAssociation;
	_requestID = request_id;
	_argumentStream = argument_stream;

	_orb = _serverManager.orb();

	_piCurrent = ((org.openorb.CORBA.ORB)_orb).getPICurrent();
    }

  /**
   * second constructor pass.
   */
    public void init(byte [] object_key, String operation, byte sync_scope,
		     org.omg.IOP.ServiceContext[] request_service_contexts)
    {
	if(_state >= 0)
	    org.openorb.util.Trace.assertTrue(false);

	_objectKey = object_key;
	_operation = operation;
	_syncScope = sync_scope;
	_requestServiceContexts = request_service_contexts;

	if(_operation != null) {
	    _interceptorManager = (org.openorb.PI.ServerManager)((org.openorb.CORBA.ORB)_orb).getFeature("ServerInterceptorManager");

	    if(_interceptorManager != null) {
		_callback = new org.openorb.PI.RequestCallback() {
		    public void reply_system_exception(org.omg.CORBA.SystemException ex) {
			System.out.println("HAndle system exc. 1");
			handle_system_exception(ex);
		    }

		    public void reply_runtime_exception(java.lang.RuntimeException ex) {
			handle_runtime_exception(ex);
		    }

		    public void reply_error(java.lang.Error ex) {
			System.out.println("HAndle error 1");
			handle_error(ex);
		    }

		    public void reply_location_forward(org.omg.CORBA.Object forward, boolean permanent) {
			handle_location_forward(forward, permanent);
		    }
		};
	    }
	}

	_state = STATE_CREATED;
    }

    /**
     * One pass constructor for locate request
     */
    public AbstractServerRequest(ServerManager serverManager, ServerChannel channel,
				 Object transportAssociation, int request_id,
				 byte [] object_key)
    {
	init_one(serverManager, channel, transportAssociation, request_id, null);
	init(object_key, null, (byte)3, null);
    }

    /**
     * Two pass constructor for locate request. Must be followed by call to
     * init(byte [] object_key)
     */
    public AbstractServerRequest(ServerManager serverManager, ServerChannel channel,
				 Object transportAssociation, int request_id) {
	init_one(serverManager, channel, transportAssociation, request_id, null);
    }

    /**
     * Second locate request construction pass.
     */
    public void init(byte [] object_key) {
	init(object_key, null, (byte)3, null);
    }

    //Modified by Diana on June 21 2002 in big desperation

    org.openorb.CORBA.SystemExceptionHelper sysexchelper=new org.openorb.CORBA.SystemExceptionHelper();



    private org.omg.CORBA.ORB _orb;
    private ServerManager _serverManager;

    private ServerChannel _channel;
    private Object _transportAssociation;

    private int _requestID,_retention_id=-1;
    private long _expiration_time=0;
    private byte [] _objectKey;
    private String _operation,_client_id="";
    private byte _syncScope;
    private org.omg.IOP.ServiceContext[] _requestServiceContexts;

    private org.omg.CORBA.portable.InputStream _argumentStream;

    private org.openorb.PI.CurrentImpl _piCurrent;
    private org.omg.PortableInterceptor.CurrentOperations _rsPICurrEntry;
    private org.omg.PortableInterceptor.CurrentOperations _tsPICurrEntry;

    // self synchronized.
    private Map _serviceContexts = new HashMap();

    private org.openorb.PI.ServerManager _interceptorManager = null;
    private org.openorb.PI.RequestCallback _callback = null;

    private Object _syncState = new Object();
    private volatile int _state = -1;

    private short _replyStatus = REPLY_STATUS_UNSET;

    private org.openorb.adapter.ObjectAdapter _adapter = null;
    private org.openorb.adapter.TargetInfo _target;

    private org.omg.CORBA.portable.OutputStream _replyStream;

    private org.omg.CORBA.Object _forwardReference;
    private org.omg.IOP.IOR _forwardReferenceIOR;

    private org.omg.CORBA.SystemException _sendingSystemException;
    private org.omg.CORBA.Any _sendingSystemExceptionAny = null;
    private String _sendingSystemExceptionID;

    private static final short REPLY_STATUS_UNSET = Short.MIN_VALUE;

    /* Modified by Diana on Oct. 11 2001*/

    org.omg.CORBA.Any result=null;

    org.omg.CORBA.Any[] result_arguments=new org.omg.CORBA.Any[0];

    public void result(org.omg.CORBA.Any result){
	this.result=result;
    }

    public org.omg.CORBA.Any result(){
	return result;
    }

    public void result_arguments(org.omg.CORBA.Any result_arguments[]){
	this.result_arguments=result_arguments;
    }

    public org.omg.CORBA.Any[] result_arguments(){
	return result_arguments;
    }

    /* added by Diana on April 26 2002*/

    public void retention_id_(int retention_id){
	_retention_id=retention_id;
    }

    public void client_id_(String client_id){
	_client_id=client_id;
    }

    public void expiration_time_(long expiration_time){
	_expiration_time=expiration_time;
    }

    public int _retention_id(){
	return _retention_id;
    }

    public String _client_id(){
	return _client_id;
    }

    public long _expiration_time(){
	return _expiration_time;
    }


    /* */

    /**
     * Orb
     */
    public org.omg.CORBA.ORB orb() {
	return _orb;
    }

    /**
     * Server channel
     */
    public ServerChannel channel() {
	return _channel;
    }

    /**
     * Get the transport association.
     */
    public Object get_transport_association() {
	return _transportAssociation;
    }

    /**
     * Check the transport association, take appropriate action if association
     * is incorrect. This will be called once for each request fragment and
     * cancel request message.
     */
    public boolean checkRequestTransportAssoc(Object tpAssc) {
	// TODO: make this controlable by policy.
	return _transportAssociation.equals(tpAssc);
    }

    /**
     * Current request state.
     */
    public int state() {
	return _state;
    }

    /**
     * Returns completion status appropriate for the request state.
     */
    protected CompletionStatus state_completion_status() {
	switch(_state) {
	case -1:
	case STATE_CREATED:
	case STATE_FIND_ADAPTER:
	case STATE_QUEUED:
	    return CompletionStatus.COMPLETED_NO;
	case STATE_PROCESSING:
	    return CompletionStatus.COMPLETED_MAYBE;
	case STATE_REPLY:
	case STATE_COMPLETE:
	    return CompletionStatus.COMPLETED_YES;
	}
	throw new Error("assert failed");
    }

    /**
     * This returns true if this request is a locate request. Use replyLocate
     * to set the reply.
     *
     * Interceptors are not called for locate requests.
     *
     * ServerRequestInfo operations: arguments, exceptions, contexts,
     * operation_context, result, get_request_service_context,
     * get_reply_service_context, add_reply_service_context are not valid.
     *
     * operation returns the empty string "", response_expected returns true.
     *
     * Note the extra value for reply_status, UNKNOWN_OBJECT.
     */
    public boolean is_locate() {
	return (_operation == null);
    }

    /**
     * Object Key. This is always available.
     */
    public byte[] object_key() {
	return _objectKey;
    }


    /**
     * Request ID. This is always available.
     */
    public int request_id() {
	return _requestID;
    }

    /**
     * Equality depends on request IDs and channels.
     */
    public boolean equals(Object obj) {
	if(obj instanceof Integer)
	    return _requestID == ((Integer)obj).intValue();
	if(obj instanceof ServerRequest) {
	    ServerRequest srq2 = (ServerRequest)obj;
	    return (_requestID == srq2.request_id()) && (_channel == srq2.channel());
	}
	return false;
    }

    /**
     * Operation name. This will return null for locate requests.
     */
    public String operation() {
	return _operation;
    }

    /**
     * Response is expected to the request. If this is true then createReply or
     * createExceptionReply is expected to be called by the dispatch operation.
     */
    public boolean response_expected() {
	/* org.omg.Messaging.SYNC_WITH_TARGET.value = 3 */
	return (_syncScope == 3);
    }

    /**
     * Message sync scope.
     */
    public short sync_scope() {
	return _syncScope;
    }

    /**
     * Request Arguments. This operation is not available in java.
     */
    public org.omg.Dynamic.Parameter[] arguments() {
	throw new org.omg.CORBA.NO_RESOURCES(org.omg.CORBA.OMGVMCID.value | 1, state_completion_status());
    }

    /**
   * Request exceptions. This operation is not available in java.
   */
    public org.omg.CORBA.TypeCode[] exceptions() {
	throw new org.omg.CORBA.NO_RESOURCES(org.omg.CORBA.OMGVMCID.value | 1, state_completion_status());
    }

    /**
   * Request contexts. This operation is not available in java.
   */
    public java.lang.String[] contexts() {
	throw new org.omg.CORBA.NO_RESOURCES(org.omg.CORBA.OMGVMCID.value | 1, state_completion_status());
    }

    /**
   * Request contexts. This operation is not available in java.
   */
    public java.lang.String[] operation_context() {
	throw new org.omg.CORBA.NO_RESOURCES(org.omg.CORBA.OMGVMCID.value | 1, state_completion_status());
    }

    /**
   * Request result. This operation is not available in java.
   */
    /*	public org.omg.CORBA.Any result() {
	throw new org.omg.CORBA.NO_RESOURCES(org.omg.CORBA.OMGVMCID.value | 1, state_completion_status());
    }*/

  /**
   * Access the PI current from interceptors. From the adapter this operation
   * is also available via the PI Current.
   */
    public org.omg.CORBA.Any get_slot(int id)
	throws org.omg.PortableInterceptor.InvalidSlot
    {
	if(_rsPICurrEntry == null) {
	    synchronized(_syncState) {
		if(_rsPICurrEntry == null)
		    _rsPICurrEntry = _piCurrent.create();
	    }
	}
	synchronized(_rsPICurrEntry) {
	    return _rsPICurrEntry.get_slot(id);
	}
    }

    /**
   * Access the PI current from interceptors. From the adapter this operation
   * is also available via the PI Current.
   */
    public void set_slot(int id, org.omg.CORBA.Any data)
	throws org.omg.PortableInterceptor.InvalidSlot
    {
	if(_rsPICurrEntry == null) {
	    synchronized(_syncState) {
		if(_rsPICurrEntry == null)
		    _rsPICurrEntry = _piCurrent.create();
	    }
	}
	synchronized(_rsPICurrEntry) {
	    _rsPICurrEntry.set_slot(id, data);
	}
    }

    /**
   * Get reqest service context. This operation is always available.
   */
    public org.omg.IOP.ServiceContext get_request_service_context(int id) {
	for(int i = 0; i < _requestServiceContexts.length; ++i)
	    if(_requestServiceContexts[i].context_id == id)
		return _requestServiceContexts[i];
	throw new org.omg.CORBA.BAD_PARAM(org.omg.CORBA.OMGVMCID.value | 23, state_completion_status());
    }

    /**
   * Client intiated request cancelation. This may be called at any time to
   * indicate that a response from the request is no longer expected.
   */
    public void client_cancel() {
	synchronized(_syncState) {
	    if(_state == STATE_COMPLETE)
		return;

	    // client cancel not allowed once the request starts processing.
	    // this is needed to catch any cancel requests which occour while
	    // within the dispatch operation before dispatching the request.
	    if(_syncScope != 3 /* org.omg.Messaging.SYNC_WITH_TARGET.value */)
		switch(_state) {
		case STATE_PROCESSING:
		case STATE_REPLY:
		    return;
		}
	    System.out.println("HAndle system exc. 2");
	    handle_system_exception(new org.omg.CORBA.TRANSIENT(org.omg.CORBA.OMGVMCID.value | 2, state_completion_status()));

	    // don't send a reply, just call the interceptors and dump the request
	    if(_state != STATE_CREATED && _interceptorManager != null) {
		_interceptorManager.send_exception(this, _callback);
		_interceptorManager = null;
	    }

	    _state = STATE_COMPLETE;

	    if(_target != null)
		_adapter.cancel_dispatch(this, _target);

	    release_request();
	}
    }

    /**
   * Server intiated request cancelation. This may be called at any time by
   * the server to indicate a server initiated cancel. The request is returned
   * to the server. The completion status of the exception will be set
   * appropriatly.
   */
    public void server_cancel(Throwable ex) {
	synchronized(_syncState) {
	    if(_state == STATE_COMPLETE){
		//stem.out.println("server cancel state complete");
		return;
	    }

	    if(ex instanceof org.omg.CORBA.SystemException) {
		org.omg.CORBA.SystemException sex = (org.omg.CORBA.SystemException)ex;
		sex.completed = state_completion_status();
		System.out.println("HAndle system exc. 3");
		handle_system_exception(sex);
	    } else if(ex instanceof RuntimeException)
		handle_runtime_exception((RuntimeException)ex);
	    else if(ex instanceof Error){
		System.out.println("HAndle error 2");
		handle_error((Error)ex);
	    }
	    else{
		//System.out.println("Unknown throwable");
		handle_unknown_throwable(ex);
	    }

	    if(_target != null){
		//System.out.println("Adapter cancel dispatch");
		_adapter.cancel_dispatch(this, _target);
	    }
	    /*else
	      System.out.println("Target is null");*/
	    complete_request();
	}
    }

    /**
   * Called by the worker thread before every attempt to run the request. This
   * should call the recieve_request_service_context interception point.
   * @returns the state of the request.
   */
    public int begin_request() {
	//System.out.println("Begin request");
	synchronized(_syncState) {
	    if(_state == STATE_CREATED) {
		if(_interceptorManager != null) {
		    _interceptorManager.receive_request_service_contexts(this, _callback);

		    if(_replyStatus != REPLY_STATUS_UNSET) {
			// interceptors have forced completion
			_interceptorManager = null;
			complete_request();
			return _state;
		    }
		}
		_state = STATE_FIND_ADAPTER;
		_tsPICurrEntry = _piCurrent.copy(_rsPICurrEntry);
	    }
	    _piCurrent.set(_tsPICurrEntry);
	}

	return _state;
    }

    /**
   * Set the object adapter and transfer to the QUEUED state. This is legal
   * only in the FIND_ADAPTER or COMPLETE states. If COMPLETE this has no effect.
   *
   * @returns the state of the request. This will be either QUEUED or COMPLETE.
   */
    public int adapter(ObjectAdapter adapter) {
	synchronized(_syncState) {
	    switch(_state) {
	    case STATE_FIND_ADAPTER:
		_adapter = adapter;
		_state = STATE_QUEUED;
	    case STATE_COMPLETE:
		return _state;
	    default:
		System.out.println("Throw BAD_INV_ORDER in adapter 1");
		throw new org.omg.CORBA.BAD_INV_ORDER(org.omg.CORBA.OMGVMCID.value | 10, state_completion_status());
	    }
	}
    }

    /**
	 * Server adapter. This operation is valid in all states apart from
	 * CREATED and FIND_ADAPTER.
	 */
    public ObjectAdapter adapter() {
	synchronized(_syncState) {
	    if(_adapter == null){
		System.out.println("Throw BAD_INV_ORDER in adapter 2");
		throw new org.omg.CORBA.BAD_INV_ORDER(org.omg.CORBA.OMGVMCID.value | 10, state_completion_status());
	    }
	    return _adapter;
	}
    }

    /**
   * Find the server policy of the specified type. This operation is valid in
   * all states apart from CREATED and FIND_ADAPTER.
   */
    public org.omg.CORBA.Policy get_server_policy(int type) {
	synchronized(_syncState) {
	    if(_adapter == null){
		System.out.println("Throw BAD_INV_ORDER in get_server_policy");
		throw new org.omg.CORBA.BAD_INV_ORDER(org.omg.CORBA.OMGVMCID.value | 10, state_completion_status());
	    }
	    int [] types = new int[1];
	    types[0] = type;

	    org.omg.CORBA.Policy [] pols = _adapter.get_server_policies(types);

	    if(pols.length >= 1)
		return pols[0];

	    throw new org.omg.CORBA.INV_POLICY(org.omg.CORBA.OMGVMCID.value | 1, state_completion_status());
	}
    }


    /**
	 * Find the adapter ID.
	 */
    public byte[] adapter_id() {
	synchronized(_syncState) {
	    if(_target == null){
		System.out.println("Throw BAD_INV_ORDER in adapter_id");
		throw new org.omg.CORBA.BAD_INV_ORDER(org.omg.CORBA.OMGVMCID.value | 10, state_completion_status());

	    }
	    return _target.getAdapterID();
	}
    }

    /**
	 * Find the object ID.
	 */
    public byte[] object_id() {
	synchronized(_syncState) {
	    if(_target == null){
		System.out.println("Throw BAD_INV_ORDER in object_id");
		throw new org.omg.CORBA.BAD_INV_ORDER(org.omg.CORBA.OMGVMCID.value | 10, state_completion_status());
	    }
	    return _target.getObjectID();
	}
    }

    /**
	 * Most derrived repository ID of the target.
	 */
    public String target_most_derived_interface() {
	synchronized(_syncState) {
	    if(_target == null){
		System.out.println("Throw BAD_INV_ORDER in target_most....");
		throw new org.omg.CORBA.BAD_INV_ORDER(org.omg.CORBA.OMGVMCID.value | 10, state_completion_status());
	    }
	    return _target.getRepositoryID();
	}
    }

    /**
   * Determine if the target implements the given interface.
   */
    public boolean target_is_a(String id) {
	synchronized(_syncState) {
	    if(_target == null){
		System.out.println("Throw BAD_INV_ORDER in target_is_a");
		throw new org.omg.CORBA.BAD_INV_ORDER(org.omg.CORBA.OMGVMCID.value | 14, state_completion_status());
	    }
	    return _target.targetIsA(id);
	}
    }

    /**
   * Call predispatch to locate the target, enter the PROCESSING state,
   * call the receive_request interception points and dispatch the request
   * through the adapter. If sync scope is SYNC_WITH_SERVER an empty response
   * will be sent before the request is dispatched. When this function returns
   * all the terminating interception points will have been called, and
   * state will be COMPLETE. If the target adapter is destroyed between the
   * find_adapter and dispatch requests this function will throw an
   * AdapterDestroyedException and the CREATED state is re-entered,
   * find_adapter function will be called again once the unregister adapter
   * operation is called on the server manager with the target adapter.
   */
    public void dispatch()
	throws AdapterDestroyedException
    {
	// predispatch
	synchronized(_syncState) {
	    switch(_state) {
	    case STATE_COMPLETE:
				// request canceled
		return;
	    case STATE_QUEUED:
		break;
	    default:
		System.out.println("Throw BAD_INV_ODER in dispatch");
		throw new org.omg.CORBA.BAD_INV_ORDER(org.omg.CORBA.OMGVMCID.value | 10, state_completion_status());
	    }

	    try {
		//System.out.println("Before adapter predispatch "+_requestID);
		_target = _adapter.predispatch(this);
		//System.out.println("After adapter predispatch "+_requestID);
	    }
	    catch(AdapterDestroyedException ex) {
		//System.out.println("exc 1 "+_requestID);
		_state = STATE_FIND_ADAPTER;
		throw ex;
	    }
	    catch(org.omg.PortableInterceptor.ForwardRequest ex) {
		//System.out.println("exc 2 "+_requestID);
		handle_location_forward(ex.forward, ex.permanent);
	    }
	    catch(org.omg.CORBA.SystemException ex) {
		ex.completed = CompletionStatus.COMPLETED_NO;
		System.out.println("HAndle system exc. 4");
		handle_system_exception(ex);
	    }
	    catch(Error ex) {
		System.out.println("HAndle error 3");
		handle_error(ex);
	    }
	    catch(RuntimeException ex) {
		System.out.println("exc 3 "+_requestID);
		handle_runtime_exception(ex);
	    }

	    // exception thrown during predispatch
	    if(_replyStatus != REPLY_STATUS_UNSET) {
		//System.out.println("Before compl_re 1 "+_requestID);
		complete_request();
		//System.out.println("after compl_re 1 "+_requestID);
		return;
	    }

	    // call interceptors.
	    if(_interceptorManager != null) {
		_interceptorManager.receive_request(this, _callback);

				// handle interceptor rejection.
		if(_replyStatus != REPLY_STATUS_UNSET) {
		    // interceptors have forced completion. Cancel the dispatch
		    // and complete the request.
		    _interceptorManager = null;
		    _adapter.cancel_dispatch(this, _target);
		    complete_request();
		    return;
		}
	    }
	    /*else
	      System.out.println("Int man is null "+_requestID);*/

	    // set the request current table back to the thread scope current
	    // discarding any changes made during recieve_request
	    _rsPICurrEntry = _tsPICurrEntry;

	    // send reply if needed and handle interception points for oneway calls.
	    switch(_syncScope) {
	    case 0: /* org.omg.Messaging.SYNC_NONE.value */
				// no need to send a reply. release the request.
				// From here on the request cannot be canceled from the client end.
		//System.out.println("Before rel_req "+_requestID);
		release_request();
		//System.out.println("After rel_req "+_requestID);
		break;
	    case 1: /* org.omg.Messaging.SYNC_WITH_TRANSPORT.value */
	    case 2: /* org.omg.Messaging.SYNC_WITH_SERVER.value */
				// TODO: possibly persist the request here, then it's asynchonous
				// and reliable.

				// send the empty reply back to the client indicating an acceptance
		try {
		    // System.out.println("Before beg_m_r "+_requestID);
		    _replyStream = begin_marshal_reply();
		    //System.out.println("After beg_m_req "+_requestID);
		    complete_reply(_replyStream);
		    //System.out.println("BAfter c _r "+_requestID);
		}
		catch(org.omg.CORBA.SystemException ex) {
		    ex.completed = CompletionStatus.COMPLETED_NO;
		    System.out.println("HAndle system exc. 5");
		    handle_system_exception(ex);
		}
		catch(Error ex) {
		    System.out.println("HAndle error 4");
		    handle_error(ex);
		}
		catch(RuntimeException ex) {
		    //System.out.println("handle r e 1 "+_requestID);
		    handle_runtime_exception(ex);
		}

				// release the request. From here on in it's an unstoppable force
				// from the client end.
		//System.out.println("Release request "+_requestID);
		release_request();
		// exception while sending blank response. Treat as client cancel.
		if(_replyStatus != REPLY_STATUS_UNSET) {
		    _state = STATE_COMPLETE;
		    _adapter.cancel_dispatch(this, _target);
		    if(_interceptorManager != null)
			_interceptorManager.send_exception(this, _callback);
		    return;
		}

		break;
	    }

	    _state = STATE_PROCESSING;
	}
	// dispatching the request is performed outside of the synchronized
	// block to allow cancel requests to get processed.

	Throwable tex = null;
	int ex_type;
	try {
	    //System.out.println("Adapter dispatch "+_requestID);
	    _adapter.dispatch(this, _target);
	    //System.out.println("Adapter dispatched "+_requestID);
	    ex_type = 0;
	}
	catch(org.omg.CORBA.SystemException ex) {
	    tex = ex;
	    ex_type = 1;
	}
	catch(Error ex) {
	    tex = ex;
	    ex_type = 2;
	}
	catch(RuntimeException ex) {
	    tex = ex;
	    ex_type = 3;
	}

	synchronized(_syncState) {
	    // request has been canceled.
	    if(_state == STATE_COMPLETE)
		return;

	    switch(ex_type) {
	    case 0:
		if(_state == STATE_PROCESSING) {
		    //System.out.println("Processing");
		    if(is_locate()) {

			// successful result for a locate request.
			_replyStatus = OBJECT_HERE;
		    } else if(_syncScope == 3 /* org.omg.Messaging.SYNC_WITH_TARGET.value */) {
			// create an empty response if none has been created.
			try {
			    createReply();
			}
			catch(org.omg.CORBA.SystemException ex) {
			    ex.completed = CompletionStatus.COMPLETED_YES;
			    //System.out.println("HAndle system exc. 6");
			    handle_system_exception(ex);
			}
			catch(Error ex) {
			    //System.out.println("HAndle error 5");
			    handle_error(ex);
			}
			catch(RuntimeException ex) {
			    handle_runtime_exception(ex);
			}
		    } else {
			// a successful asynchronous request set reply status to successful
			_replyStatus = org.omg.PortableInterceptor.SUCCESSFUL.value;
		    }
		}
		//System.out.println("After test "+_state);
		break;
	    case 1:
		if(_state == STATE_REPLY)
		    ((org.omg.CORBA.SystemException)tex).completed = CompletionStatus.COMPLETED_YES;
		System.out.println("HAndle system exc. 7");
		handle_system_exception((org.omg.CORBA.SystemException)tex);
		break;
	    case 2:
		System.out.println("HAndle error 6");
		handle_error((Error)tex);
		break;
	    case 3:
		handle_runtime_exception((RuntimeException)tex);
		break;
	    }
	    //System.out.println("Before comp end req "+_requestID);
	    complete_request();
	    //System.out.println("Aftere comp end req "+_requestID);
	}
    }

    /**
   * Get the request input stream. This function can be called only once
   * while in the PROCESSING state. The returned input stream may throw a
   * system exception if some problem occours while unmarshalling the request
   * arguments. This function, and the reply functions must be called by the
   * same thread, the thread that called the dispatch function.
   */
    public org.omg.CORBA.portable.InputStream argument_stream() {
	return _argumentStream;
    }

    /* Modified by Diana on Oct. 24 2001*/

    public void argument_stream(org.omg.CORBA.portable.InputStream arg_str){
	_argumentStream=arg_str;
    }

    /**
   * Add a service context to the reply.
   */
    public void add_reply_service_context(org.omg.IOP.ServiceContext service_context, boolean replace) {
	Integer key = new Integer(service_context.context_id);
	synchronized(_serviceContexts) {
	    if(!replace && _serviceContexts.containsKey(key)){
		System.out.println("Throw BAD_INV_ORDER in add_reply_s_context");
		throw new org.omg.CORBA.BAD_INV_ORDER(org.omg.CORBA.OMGVMCID.value | 11, state_completion_status());
	    }
	    _serviceContexts.put(key, service_context);
	}
    }

    /**
   * Find a reply service context with a given ID.
   */
    public org.omg.IOP.ServiceContext get_reply_service_context(int id) {
	org.omg.IOP.ServiceContext ret = null;
	synchronized(_serviceContexts) {
	    ret = (org.omg.IOP.ServiceContext)_serviceContexts.get(new Integer(id));
	}
	if(ret == null)
	    throw new org.omg.CORBA.BAD_PARAM(org.omg.CORBA.OMGVMCID.value | 23, state_completion_status());
	return ret;
    }

    /**
   * Get a list of all the reply service contexts which been set.
   */
    public org.omg.IOP.ServiceContext [] get_reply_service_contexts() {
	synchronized(_serviceContexts) {
	    return (org.omg.IOP.ServiceContext[])_serviceContexts.values().toArray(new org.omg.IOP.ServiceContext[_serviceContexts.size()]);
	}
    }

    /**
   * Get the reply status. This operation is available in the REPLY and COMPLETE
   * states.
   */
    public short reply_status() {
	synchronized(_syncState) {
	    if(_replyStatus == REPLY_STATUS_UNSET){
		System.out.println("Throw BAD_INV_ORDER in reply_status");
		throw new org.omg.CORBA.BAD_INV_ORDER(org.omg.CORBA.OMGVMCID.value | 10, state_completion_status());
	    }
	    return _replyStatus;
	}
    }

    /**
   * The forward reference which will be sent in response to the
   * request.
   */
    public org.omg.CORBA.Object forward_reference() {
	synchronized(_syncState) {
	    switch(reply_status()) {
	    case org.omg.PortableInterceptor.LOCATION_FORWARD.value:
	    case org.omg.PortableInterceptor.LOCATION_FORWARD_PERMANENT.value:
		return _forwardReference;
	    }
	    System.out.println("Throw BAD_INV_ORDER in forwrd_Ref");
	    throw new org.omg.CORBA.BAD_INV_ORDER(org.omg.CORBA.OMGVMCID.value | 10, state_completion_status());
	}
    }

    /**
   * The IOR of the forward reference which will be sent in response to
   * the request.
   */
    public org.omg.IOP.IOR forward_reference_ior() {
	synchronized(_syncState) {
	    // reply_status throws bad inv order if not in complete state.
	    switch(reply_status()) {
	    case org.omg.PortableInterceptor.LOCATION_FORWARD.value:
	    case org.omg.PortableInterceptor.LOCATION_FORWARD_PERMANENT.value:
		if(_forwardReferenceIOR == null)
		    _forwardReferenceIOR = ((org.openorb.CORBA.Delegate)((org.omg.CORBA.portable.ObjectImpl)_forwardReference)._get_delegate()).ior();
		return _forwardReferenceIOR;
	    }
	    System.out.println("Throw BAD_INV_ORDER in forw_ref_ior");
	    throw new org.omg.CORBA.BAD_INV_ORDER(org.omg.CORBA.OMGVMCID.value | 10, state_completion_status());
	}
    }

    /**
   * Exception being sent in reply. This will only contain system exceptions.
   */
    public org.omg.CORBA.Any sending_exception() {
	synchronized(_syncState) {
	    // reply_status throws bad inv order if not in complete state.
	    switch(reply_status()) {
	    case org.omg.PortableInterceptor.SYSTEM_EXCEPTION.value:
		//System.out.println("System exception 1");
		if(_sendingSystemExceptionAny == null) {
		    _sendingSystemExceptionAny = _orb.create_any();
		    sysexchelper/*SystemExceptionHelper*/.insert(_sendingSystemExceptionAny, _sendingSystemException);
		}
		return _sendingSystemExceptionAny;
	    case org.omg.PortableInterceptor.USER_EXCEPTION.value:
		throw new org.omg.CORBA.NO_RESOURCES(org.omg.CORBA.OMGVMCID.value | 1, state_completion_status());
	    }
	    System.out.println("Throw BAD_INV_ORDER in sending_exc");
	    throw new org.omg.CORBA.BAD_INV_ORDER(org.omg.CORBA.OMGVMCID.value | 10, state_completion_status());
	}
    }

    /**
   * System exception being sent as a reply.
   */
    public org.omg.CORBA.SystemException sending_system_exception() {
	synchronized(_syncState) {
	    // reply_status throws bad inv order if not in complete state.
	    if(reply_status() == org.omg.PortableInterceptor.SYSTEM_EXCEPTION.value){
		//System.out.println("System exception 2");
		return _sendingSystemException;
	    }
	    System.out.println("Throw BAD_INV_ORDER in sending_syst_exc");
	    throw new org.omg.CORBA.BAD_INV_ORDER(org.omg.CORBA.OMGVMCID.value | 10, state_completion_status());
	}
    }

    /**
   * Repository ID of the system exception being sent as a reply.
   */
    public String sending_system_exception_id() {
	synchronized(_syncState) {
	    // reply_status throws bad inv order if not in complete state.
	    if(reply_status() == org.omg.PortableInterceptor.SYSTEM_EXCEPTION.value){
		//System.out.println("System exception 3");
		return _sendingSystemExceptionID;
	    }
	    System.out.println("Throw BAD_INV_ORDER in sending_exc_id");
	    throw new org.omg.CORBA.BAD_INV_ORDER(org.omg.CORBA.OMGVMCID.value | 10, state_completion_status());
	}
    }


    /* modified by diana on Aug. 29 2001*/
    private void handle_system_exception(org.omg.CORBA.SystemException ex) {
	_sendingSystemExceptionAny = null;
	_sendingSystemException = ex;
	_sendingSystemExceptionID = sysexchelper/*SystemExceptionHelper*/.id(ex);
	_replyStatus = org.omg.PortableInterceptor.SYSTEM_EXCEPTION.value;


	if(ex instanceof org.omg.CORBA.REPLY_ALREADY_THERE){
	    org.omg.CORBA.REPLY_ALREADY_THERE this_e=(org.omg.CORBA.REPLY_ALREADY_THERE)ex;
	    if(this_e.successful)
		_replyStatus=org.omg.PortableInterceptor.SUCCESSFUL.value;
	    else
		_replyStatus=org.omg.PortableInterceptor.USER_EXCEPTION.value;
	}
    }

    private void handle_error(Error ex) {
	if(ex instanceof ThreadDeath)
	    throw ex;
	else if(ex instanceof OutOfMemoryError){
	    //System.out.println("HAndle system exc. 20");
	    handle_system_exception(new org.omg.CORBA.NO_MEMORY(0, state_completion_status()));
	}
	else if(ex instanceof StackOverflowError){
	    //System.out.println("HAndle system exc. 21");
	    handle_system_exception(new org.omg.CORBA.NO_RESOURCES(0, state_completion_status()));
	}
	else {
	    org.openorb.util.Trace.exception( "Intercepted in AbstractServerRequest", ex );
	    org.omg.CORBA.portable.UnknownException uex = new org.omg.CORBA.portable.UnknownException(ex);
	    uex.completed = state_completion_status();
	    //System.out.println("HAndle system exc. 22");
	    handle_system_exception(uex);
	}
    }

    private void handle_runtime_exception(RuntimeException ex) {
	org.openorb.util.Trace.exception("Handle runtime exception", ex);
	org.omg.CORBA.portable.UnknownException uex = new org.omg.CORBA.portable.UnknownException(ex);
	uex.completed = state_completion_status();
	//System.out.println("HAndle system exc.23");
	handle_system_exception(uex);
    }

    private void handle_unknown_throwable(Throwable ex) {
	org.openorb.util.Trace.exception("Handle runtime exception", ex);
	org.omg.CORBA.portable.UnknownException uex = new org.omg.CORBA.portable.UnknownException(ex);
	uex.completed = state_completion_status();
	//System.out.println("HAndle system exc. 24");
	handle_system_exception(uex);
    }

    private void handle_location_forward(org.omg.CORBA.Object forward, boolean permanent) {
	_forwardReference = forward;
	_forwardReferenceIOR = null;
	_replyStatus = permanent?org.omg.PortableInterceptor.LOCATION_FORWARD_PERMANENT.value:org.omg.PortableInterceptor.LOCATION_FORWARD.value;
    }

    public org.omg.CORBA.portable.OutputStream createReply() {
	synchronized(_syncState) {
	    if(is_locate()){
		System.out.println("Throw BAD_INV_ORDER in createReply");
		throw new org.omg.CORBA.BAD_INV_ORDER(MinorCodes.BAD_INV_ORDER_SERVER, state_completion_status());
	    }
	    // throw cancelation exception
	    if(_replyStatus == org.omg.PortableInterceptor.SYSTEM_EXCEPTION.value){
		//System.out.println("System exception 5");
		throw _sendingSystemException;
	    }
	    // If if we have a oneway request return a dummy buffer for
	    // marshaling the response.
	    if(_syncScope != 3 /* org.omg.Messaging.SYNC_WITH_TARGET.value */){
		/* Modified by diana on Jan. 28 2002*/
		if(_interceptorManager!=null){
		    _interceptorManager.send_reply(this, _callback);
		    _interceptorManager=null;
		}
		return _orb.create_output_stream();
	    }

	    _replyStatus = org.omg.PortableInterceptor.SUCCESSFUL.value;

	    if(_interceptorManager != null) {
		//System.out.println("Send reply form here");
		_interceptorManager.send_reply(this, _callback);
		_interceptorManager = null;

		if(_replyStatus == org.omg.PortableInterceptor.SYSTEM_EXCEPTION.value) {
		    // set the interceptor manager to null, the request sequence
		    // is complete.
		    //System.out.println("System exception 6");
		    throw _sendingSystemException;
		}
	    }

	    _state = STATE_REPLY;
	    //System.out.println("before beg. marshal");
	    _replyStream = begin_marshal_reply();
	    //System.out.println("After beg. marshal");
	    return _replyStream;
	}
    }


    // Modified by Diana on Aug. 29 2001

    public org.omg.CORBA.portable.OutputStream createNoInvokeReply() {
	if(_syncScope != 3 /* org.omg.Messaging.SYNC_WITH_TARGET.value */)
	    return _orb.create_output_stream();
	_replyStatus = org.omg.PortableInterceptor.SUCCESSFUL.value;
	_state = STATE_REPLY;
	_replyStream = begin_marshal_reply();
	return _replyStream;
    }



    public org.omg.CORBA.portable.OutputStream createExceptionReply() {
	synchronized(_syncState) {
	    if(is_locate()){
		System.out.println("Throw BAD_INV_ORDER in createExcRepl");
		throw new org.omg.CORBA.BAD_INV_ORDER(MinorCodes.BAD_INV_ORDER_SERVER, state_completion_status());}

	    // If in the complete state due to a cancelation or we have a oneway
	    // request we return a dummy buffer for marshaling the response.
	    if(_state == STATE_COMPLETE || _syncScope != 3 /* org.omg.Messaging.SYNC_WITH_TARGET.value */) {
		if(_replyStatus == org.omg.PortableInterceptor.SYSTEM_EXCEPTION.value){
		    //System.out.println("System exception 7");
		    throw _sendingSystemException;
		}
		//System.out.println("Before create output_stream");
		return _orb.create_output_stream();
	    }

	    _replyStatus = org.omg.PortableInterceptor.USER_EXCEPTION.value;

	    if(_interceptorManager != null) {
		_interceptorManager.send_exception(this, _callback);
		_interceptorManager = null;

		if(_replyStatus == org.omg.PortableInterceptor.SYSTEM_EXCEPTION.value) {
		    //System.out.println("System exception 8");
		    // set the interceptor manager to null, the request sequence
		    // is complete.
		    throw _sendingSystemException;
		}
	    }

	    _state = STATE_REPLY;
	    _replyStream = begin_marshal_user_exception();
	    //System.out.println("before ret replyStream");
	    return _replyStream;
	}
    }

    /**
   * call interceptors and send response. syncState must be owned.
   */
    private void complete_request() {
	org.openorb.util.Trace.assertTrue(_replyStatus != REPLY_STATUS_UNSET, "Completing request without setting reply status.");

	//System.out.println("The synchScope is "+_syncScope);

	if(_state != STATE_CREATED && _interceptorManager != null)
	    switch(_replyStatus) {
	    case org.omg.PortableInterceptor.SUCCESSFUL.value:
		//System.out.println("Complete request state successfull");
		_interceptorManager.send_reply(this, _callback);
		break;
	    case org.omg.PortableInterceptor.SYSTEM_EXCEPTION.value:
		//System.out.println("System exception 9");
		_interceptorManager.send_exception(this, _callback);
		break;
	    case org.omg.PortableInterceptor.LOCATION_FORWARD.value:
	    case org.omg.PortableInterceptor.LOCATION_FORWARD_PERMANENT.value:
		_interceptorManager.send_other(this, _callback);
		break;
	    }

	// send a reply if needed.
	switch(_syncScope) {
	case 1: /* org.omg.Messaging.SYNC_WITH_TRANSPORT.value */
	case 2: /* org.omg.Messaging.SYNC_WITH_SERVER.value */
	    //System.out.println("Synchsocpe is "+_syncScope);
	    if(_state < STATE_PROCESSING)
		break;
	    //System.out.println("With reply");
	    // fallthrough
	case 0: /* org.omg.Messaging.SYNC_NONE.value */
	    // release_request called allready, take no action.
	    //System.out.println("No reply");
	    _state = STATE_COMPLETE;
	    //System.out.println("return synchScope");
	    return;
	}

	// construct or complete a reply.

	// NOTE: the REPLY state is only entered if a reply has begun to be
	// constructed. At this stage any exception will either be a problem
	// with marshaling, or a comm channel loss. In both cases we just send
	// a new reply containing the exception. This may cause problems if the
	// reply is fragmented, and the client does not understand that an
	// exception following a fragment message results in an exception.

	try {
	    // send reply message.
	    switch(_replyStatus) {
	    case OBJECT_HERE:
		marshal_locate_reply(true);
		break;
	    case UNKNOWN_OBJECT:
		marshal_locate_reply(false);
		break;
	    case org.omg.PortableInterceptor.SUCCESSFUL.value:
	    case org.omg.PortableInterceptor.USER_EXCEPTION.value:
		//System.out.println("Complete reply for user exc or succ "+_replyStatus+" "+org.omg.PortableInterceptor.SUCCESSFUL.value);
		if(_replyStream!=null)
		    complete_reply(_replyStream);
		break;
	    case org.omg.PortableInterceptor.SYSTEM_EXCEPTION.value:
		//System.out.println("System exception 10");
		marshal_system_exception(_sendingSystemExceptionID, _sendingSystemException);
		break;
	    case org.omg.PortableInterceptor.LOCATION_FORWARD.value:
		marshal_forward_request(_forwardReference, false);
		break;
	    case org.omg.PortableInterceptor.LOCATION_FORWARD_PERMANENT.value:
		marshal_forward_request(_forwardReference, true);
		break;
	    }
	}
	catch(org.omg.CORBA.SystemException ex) {
	    org.openorb.util.Trace.high(this, "Transport failure while replying to request: " + ex.toString());
	}
	catch(Error ex) {
	    org.openorb.util.Trace.high(this, "Java error while replying to request" + ex.toString());
	}
	catch(RuntimeException ex) {
	    org.openorb.util.Trace.high(this, "Java runtime exception while replying to request" + ex.toString());
	}
	_state = STATE_COMPLETE;
	//System.out.println("RElease_Request");
	release_request();
    }

    /**
   * Send a system exception result. Note that a failed locate request will
   * always result in an OBJECT_NOT_FOUND system exception, which this
   * function is free to convert into a failed locate reply. This may
   * throw a system exception indicating a transport problem.
   */
    abstract protected void marshal_system_exception(String repo_id, org.omg.CORBA.SystemException ex);

    /**
   * Send a forward request result. This may throw a system
   * exception indicating a transport problem.
   */
    abstract protected void marshal_forward_request(org.omg.CORBA.Object target, boolean perminent);

    /**
   * Reply to a locate request. This argument to this function will be true
   * when called from this class, however marshal_system_exception may convert
   * a system exception response into a locate failure. This may throw a system
   * exception indicating a transport problem.
   */
    abstract protected void marshal_locate_reply(boolean object_is_here);

    /**
   * Create a stream for marshaling a successful response. This is paired
   * with a call to complete_marshal. The returned stream may throw a system
   * exception at any time to indicate transport problems.
   */
    abstract protected org.omg.CORBA.portable.OutputStream begin_marshal_reply();

    /**
   * Create a stream for marshaling a user exception response. This is paired
   * with a call to complete_marshal. The returned stream may throw a system
   * exception at any time to indicate transport problems.
   */
    abstract protected org.omg.CORBA.portable.OutputStream begin_marshal_user_exception();

    /**
   * Complete the marshaling process. Paired with a call to begin_marshal_* .
   * This may throw a system exception to indicate transport problems.
   */
    abstract protected void complete_reply(org.omg.CORBA.portable.OutputStream os);

    /**
   * Release any resources associated with the request. Thisis called when the
   * complete state is entered.
   */
    abstract protected void release_request();
}
